// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "ckd.h"

#include "io.h"
#include "nc_object.h"
#include "spex_settings.h"

#include <numeric>
#include <spdlog/spdlog.h>

namespace spex {

// Retrieve values from a NetCDF variable into a buffer and reshape
// for destination.
static auto getAndReshape(const netCDF::NcVar& nc_var,
                          std::vector<SharedArray<double>>& dest) -> void
{
    std::vector<double> buffer(dest.size() * dest.front().size());
    nc_var.getVar(buffer.data());
    for (int i {}; i < static_cast<int>(dest.size()); ++i) {
        for (int j {}; j < static_cast<int>(dest.at(i).size()); ++j) {
            dest[i][j] = buffer[i * dest.front().size() + j];
        }
    }
}

// Similar but for a 3-dimensional array
static auto getAndReshape(
  const netCDF::NcVar& nc_var,
  std::array<std::vector<SharedArray<double>>, dims::pol>& dest) -> void
{
    const std::array<int, 3> dims {
        static_cast<int>(dest.size()),
        static_cast<int>(dest.front().size()),
        static_cast<int>(dest.front().front().size()),
    };
    // Sometimes the first dimension is dims::pol in the destination
    // array but second dimension in the NetCDF variable.
    const bool flip_dim { nc_var.getDim(0).getSize() != dims[0] };
    std::vector<double> buffer(
      std::accumulate(dims.cbegin(), dims.cend(), 1, std::multiplies<>()));
    nc_var.getVar(buffer.data());
    // Later idx + k is the full index of the source array
    int idx {};
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int j {}; j < dims[1]; ++j) {
            if (flip_dim) {
                idx = (j * dims::pol + i_pol) * dims[2];
            } else {
                idx = (i_pol * dims[1] + j) * dims[2];
            }
            for (int k {}; k < dims[2]; ++k) {
                dest[i_pol][j][k] = buffer[idx + k];
            }
        }
    }
}

auto CKD::readCKD(
  const std::string& ckd_file_name,
  const ProcLevel target_level,
  const std::array<bool, static_cast<int>(ProcLevel::n_levels)>& debug,
  std::array<bool, static_cast<int>(ProcLevel::n_levels)>& skip) -> void
{
    NcObject nc { netCDF::NcFile::read, ckd_file_name };
    dim_detector_spat =
      static_cast<int>(nc.getDim("spatial_detector_pixels").getSize());
    dim_detector_spec =
      static_cast<int>(nc.getDim("spectral_detector_pixels").getSize());
    dim_vp = static_cast<int>(nc.getDim("viewports").getSize());
    // Number of pixels is the product of length and width. This is
    // unbinned CKD.
    npix = dim_detector_spec * dim_detector_spat;
    // Print the basic dimensions which are fixed from now on
    spdlog::info("Using an input CKD: {}", ckd_file_name);
    spdlog::info("Detector dimension  (spatial): {}", dim_detector_spat);
    spdlog::info("Detector dimension (spectral): {}", dim_detector_spec);
    spdlog::info("Number of viewports: {}", dim_vp);
    // Read the pixel mask which is always there, possibly not yet in
    // its final state. It may be updated by one or more detector
    // calibration steps. When reading into shared arrays make sure to
    // call the root process only.
    std::vector<uint8_t> pixel_mask_u8(npix);
    nc.getVar("mask").getVar(pixel_mask_u8.data());
    pixel_mask.resize(npix);
    std::transform(pixel_mask_u8.cbegin(),
                   pixel_mask_u8.cend(),
                   pixel_mask.begin(),
                   [](const auto v) { return v == 1; });
    // Only read data if the target level is equal or higher than the
    // process. Do not read data for the process that the CKD is set
    // to be mature for.
    if (const netCDF::NcGroup grp { nc.getGroup("DARK") };
        target_level >= ProcLevel::dark && !grp.isNull()) {
        level = ProcLevel::dark;
        skip.at(static_cast<int>(level)) = false;
        dim_dark_order =
          static_cast<int>(grp.getDim("dark_number_of_coefficients").getSize());
        dark_offset.resize(npix * dim_dark_order);
        dark_current.resize(npix * dim_dark_order);
        grp.getVar("dark_offset").getVar(dark_offset.data());
        grp.getVar("dark_current").getVar(dark_current.data());
        grp.getVar("dark_nominal_temperature")
          .getVar(&dark_nominal_temperature);
        spdlog::info("Reading the dark CKD");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("NOISE") };
        target_level >= ProcLevel::noise && !grp.isNull()) {
        level = ProcLevel::noise;
        skip.at(static_cast<int>(level)) = false;
        noise.g = { true, mpi, npix };
        noise.n = { true, mpi, npix };
        if (debug.at(static_cast<int>(ProcLevel::noise))) {
            noise.std_dev_g = { true, mpi, npix };
            noise.std_dev_n = { true, mpi, npix };
        }
        if (mpi.rank == 0) {
            grp.getVar("g").getVar(noise.g.data());
            grp.getVar("n").getVar(noise.n.data());
            if (debug.at(static_cast<int>(ProcLevel::noise))) {
                const auto grp { nc.getGroup("DEBUG").getGroup("NOISE") };
                grp.getVar("std_dev_g").getVar(noise.std_dev_g.data());
                grp.getVar("std_dev_n").getVar(noise.std_dev_n.data());
            }
        }
        spdlog::info("Reading the noise CKD");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("NON_LINEARITY") };
        target_level >= ProcLevel::nonlin && !grp.isNull()) {
        level = ProcLevel::nonlin;
        skip.at(static_cast<int>(level)) = false;
        dim_nonlin_exptime = static_cast<int>(
          grp.getDim("nonlin_number_of_exposure_times").getSize());
        dim_nonlin_spline = static_cast<int>(
          grp.getDim("nonlin_number_of_coefficients").getSize());
        dim_nonlin_knot =
          static_cast<int>(grp.getDim("nonlin_number_of_knots").getSize());
        nonlin_knots.resize(dim_nonlin_knot);
        nonlin_fit.resize(npix * dim_nonlin_exptime * dim_nonlin_spline);
        // Exposure times for which nonlinearity is calculated
        nonlin_exptimes.resize(dim_nonlin_exptime);
        // Scaling signal for the abscissa of the nonlinearity fit
        nonlin_signal_scale.resize(npix * dim_nonlin_exptime);
        int intorder {}; // For integer type conversion
        grp.getVar("nonlin_order").getVar(&intorder);
        nonlin_order = intorder;
        grp.getVar("nonlin_knots").getVar(nonlin_knots.data());
        grp.getVar("nonlin_exptimes").getVar(nonlin_exptimes.data());
        grp.getVar("nonlin_signal_scale").getVar(nonlin_signal_scale.data());
        grp.getVar("nonlin_fit").getVar(nonlin_fit.data());
        spdlog::info("Reading the nonlinearity CKD");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("PRNU") };
        target_level >= ProcLevel::prnu && !grp.isNull()) {
        level = ProcLevel::prnu;
        skip.at(static_cast<int>(level)) = false;
        prnu.prnu = { true, mpi, npix };
        if (debug.at(static_cast<int>(ProcLevel::prnu))) {
            prnu.sensitivity_fit = { true, mpi, npix };
        }
        if (mpi.rank == 0) {
            grp.getVar("prnu").getVar(prnu.prnu.data());
            if (debug.at(static_cast<int>(ProcLevel::prnu))) {
                const auto grp { nc.getGroup("DEBUG").getGroup("PRNU") };
                grp.getVar("sensitivity").getVar(prnu.sensitivity_fit.data());
            }
        }
        spdlog::info("Reading the PRNU CKD");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("STRAYLIGHT") };
        target_level >= ProcLevel::stray && !grp.isNull()) {
        level = ProcLevel::stray;
        skip.at(static_cast<int>(level)) = false;
        stray.n_kernels = static_cast<int>(grp.getDim("kernels").getSize());
        uint8_t reduced_kernels_u8 {};
        grp.getVar("reduced_kernels").getVar(&reduced_kernels_u8);
        stray.reduced_kernels = static_cast<bool>(reduced_kernels_u8);
        stray.n_spatial = static_cast<int>(grp.getDim("spatial").getSize());
        stray.n_spectral = static_cast<int>(grp.getDim("spectral").getSize());
        stray.kernel_rows.resize(stray.n_kernels);
        grp.getVar("kernel_rows").getVar(stray.kernel_rows.data());
        stray.kernel_cols.resize(stray.n_kernels);
        grp.getVar("kernel_cols").getVar(stray.kernel_cols.data());
        stray.kernel_fft_sizes.resize(stray.n_kernels);
        grp.getVar("kernel_fft_sizes").getVar(stray.kernel_fft_sizes.data());
        // Generate the FFTW descriptors
        stray.fftw_data.resize(stray.n_kernels);
        for (int i_kernel {}; i_kernel < stray.n_kernels; ++i_kernel) {
            stray.fftw_data[i_kernel] = std::make_unique<FFTWData>(
              stray.kernel_rows[i_kernel], stray.kernel_cols[i_kernel]);
        }
        stray.kernels_fft.resize(stray.n_kernels);
        for (int i {}; i < stray.n_kernels; ++i) {
            stray.kernels_fft[i] = { true, mpi, stray.kernel_fft_sizes[i] };
        }
        if (mpi.rank == 0) {
            std::vector<size_t> start { 0 };
            for (int i_kernel {}; i_kernel < stray.n_kernels; ++i_kernel) {
                // Need to multiply by 2 because of complex numbers
                std::vector<size_t> count {
                    2 * static_cast<size_t>(stray.kernel_fft_sizes.at(i_kernel))
                };
                grp.getVar("kernels_fft")
                  .getVar(start, count, stray.kernels_fft.at(i_kernel).data());
                start.front() += count.front();
            }
        }
        stray.eta =
          SharedArray<double> { true, mpi, stray.n_spatial * stray.n_spectral };
        if (mpi.rank == 0) {
            grp.getVar("eta").getVar(stray.eta.data());
        }
        resize(stray.weights,
               mpi,
               stray.n_kernels,
               stray.n_spatial * stray.n_spectral);
        if (mpi.rank == 0) {
            getAndReshape(grp.getVar("weights"), stray.weights);
        }
        stray.edges.resize(4 * stray.n_kernels);
        grp.getVar("edges").getVar(stray.edges.data());
        if (debug.at(static_cast<int>(ProcLevel::stray))) {
            const auto grp { nc.getGroup("DEBUG").getGroup("STRAYLIGHT") };
            stray.kernel_n_rows_full =
              static_cast<int>(grp.getDim("kernel_rows").getSize());
            stray.kernel_n_cols_full =
              static_cast<int>(grp.getDim("kernel_cols").getSize());
            resize(stray.kernels_real,
                   mpi,
                   stray.n_kernels,
                   stray.kernel_n_rows_full * stray.kernel_n_cols_full);
            if (mpi.rank == 0) {
                getAndReshape(grp.getVar("kernels_real"), stray.kernels_real);
            }
        }
        spdlog::info("Reading stray light CKD");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("FIELD_OF_VIEW") };
        target_level >= ProcLevel::fov && !grp.isNull()) {
        level = ProcLevel::fov;
        skip.at(static_cast<int>(level)) = false;
        fov.n_act =
          static_cast<int>(nc.getDim("spatial_samples_per_image").getSize());
        fov.n_act_vp.resize(dim_vp);
        fov.start_vp.resize(dim_vp);
        fov.act_angles.resize(fov.n_act);
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            resize(fov.spat_indices[i_pol], mpi, fov.n_act, dim_detector_spec);
            resize(fov.pix1[i_pol], mpi, fov.n_act, dim_detector_spec);
            resize(fov.pix2[i_pol], mpi, fov.n_act, dim_detector_spec);
            resize(fov.weight[i_pol], mpi, fov.n_act, dim_detector_spec);
        }
        fov.dims_spec.resize(fov.n_act, dim_detector_spec);
        grp.getVar("fov_nfov_vp").getVar(fov.n_act_vp.data());
        grp.getVar("fov_ifov_start_vp").getVar(fov.start_vp.data());
        grp.getVar("fov_act_angles").getVar(fov.act_angles.data());
        if (mpi.rank == 0) {
            getAndReshape(grp.getVar("fov_ispat"), fov.spat_indices);
        }
        // Convert fractional spatial detector coordinates into pixels
        // and a weight factor.
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            for (int i_fov {}; i_fov < fov.n_act; ++i_fov) {
                for (int i {}; i < dim_detector_spec; ++i) {
                    const int spat_idx { static_cast<int>(
                      fov.spat_indices[i_pol][i_fov][i]) };
                    fov.pix1[i_pol][i_fov][i] =
                      spat_idx * dim_detector_spec + i;
                    fov.pix2[i_pol][i_fov][i] =
                      fov.pix1[i_pol][i_fov][i] + dim_detector_spec;
                    fov.weight[i_pol][i_fov][i] =
                      spat_idx + 1.0 - fov.spat_indices[i_pol][i_fov][i];
                }
            }
        }
        if (debug.at(static_cast<int>(ProcLevel::fov))) {
            const auto grp { nc.getGroup("DEBUG").getGroup("FOV") };
            fov.n_l1a_total = grp.getDim("measurements").getSize();
            fov.n_columns = grp.getDim("peak_positions").getSize();
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                resize(
                  fov.peak_pos.at(i_pol), mpi, fov.n_l1a_total, fov.n_columns);
                resize(fov.peak_pos_b.at(i_pol),
                       mpi,
                       fov.n_l1a_total,
                       dim_detector_spec);
            }
            if (mpi.rank == 0) {
                getAndReshape(grp.getVar("peak_positions"), fov.peak_pos);
                getAndReshape(grp.getVar("peak_positions_b"), fov.peak_pos_b);
            }
        }
        spdlog::info("Reading the FOV CKD");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("SWATH") };
        target_level >= ProcLevel::swath && !grp.isNull()) {
        level = ProcLevel::swath;
        skip.at(static_cast<int>(level)) = false;
        swath_swathvectors.resize(fov.n_act * dims::vec);
        grp.getVar("swath_swathvectors").getVar(swath_swathvectors.data());
        swath_vectorplane_normals.resize(dim_vp * dims::vec);
        grp.getVar("swath_vectorplane_normals")
          .getVar(swath_vectorplane_normals.data());
        spdlog::info("Reading swath vectors");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("WAVELENGTH") };
        target_level >= ProcLevel::wave && !grp.isNull()) {
        level = ProcLevel::wave;
        skip.at(static_cast<int>(level)) = false;
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            resize(wave_full.at(i_pol), mpi, fov.n_act, dim_detector_spec);
        }
        resize(wave_common, mpi, fov.n_act, dim_detector_spec);
        if (mpi.rank == 0) {
            std::vector<double> buffer(fov.n_act * dims::pol
                                       * dim_detector_spec);
            grp.getVar("wave_full").getVar(buffer.data());
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                for (int i_act {}; i_act < fov.n_act; ++i_act) {
                    for (int i {}; i < dim_detector_spec; ++i) {
                        wave_full[i_pol][i_act][i] =
                          buffer[(i_pol * fov.n_act + i_act) * dim_detector_spec
                                 + i];
                    }
                }
            }
            getAndReshape(grp.getVar("wave_common"), wave_common);
        }
        spdlog::info("Reading the wavelength CKD");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("RADIOMETRIC") };
        target_level >= ProcLevel::rad && !grp.isNull()) {
        level = ProcLevel::rad;
        skip.at(static_cast<int>(level)) = false;
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            resize(rad_spectra.at(i_pol), mpi, fov.n_act, dim_detector_spec);
        }
        if (mpi.rank == 0) {
            std::vector<double> buffer(fov.n_act * dims::pol
                                       * dim_detector_spec);
            grp.getVar("rad_spectra").getVar(buffer.data());
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                for (int i_act {}; i_act < fov.n_act; ++i_act) {
                    for (int i {}; i < dim_detector_spec; ++i) {
                        rad_spectra[i_pol][i_act][i] =
                          buffer[(i_act * dims::pol + i_pol) * dim_detector_spec
                                 + i];
                    }
                }
            }
        }
        spdlog::info("Reading the radiometric CKD");
    }
    if (const netCDF::NcGroup grp { nc.getGroup("POLARIMETRIC") };
        target_level >= ProcLevel::pol && !grp.isNull()) {
        level = ProcLevel::pol;
        skip.at(static_cast<int>(level)) = false;
        resize(pol.m_q, mpi, fov.n_act, dim_detector_spec);
        resize(pol.m_u, mpi, fov.n_act, dim_detector_spec);
        resize(pol.m_t, mpi, fov.n_act, dim_detector_spec);
        if (debug.at(static_cast<int>(ProcLevel::pol))) {
            resize(pol.eps, mpi, fov.n_act, dim_detector_spec);
            resize(pol.delta, mpi, fov.n_act, dim_detector_spec);
        }
        if (mpi.rank == 0) {
            getAndReshape(grp.getVar("pol_m_q"), pol.m_q);
            getAndReshape(grp.getVar("pol_m_u"), pol.m_u);
            getAndReshape(grp.getVar("pol_m_t"), pol.m_t);
            if (debug.at(static_cast<int>(ProcLevel::pol))) {
                const auto grp { nc.getGroup("DEBUG").getGroup(
                  "POLARIMETRIC") };
                getAndReshape(grp.getVar("eps"), pol.eps);
                getAndReshape(grp.getVar("delta"), pol.delta);
            }
        }
        spdlog::info("Reading the polarimetric CKD");
    }
}

CKD::CKD(const gadfit::MPIVars& mpi, const std::string ckd_file_name)
  : mpi { mpi }
{
    // Ignore all debug information present in the CKD file here
    std::array<bool, static_cast<int>(ProcLevel::n_levels)> debug {};
    std::ranges::fill(debug, false);
    // Dummy variable, not used for anything later
    std::array<bool, static_cast<int>(ProcLevel::n_levels)> skip {};
    readCKD(ckd_file_name, ProcLevel::n_levels, debug, skip);
}

CKD::CKD(const gadfit::MPIVars& mpi,
         const ProcLevel target_level,
         SPEXSettings& settings)
  : mpi { mpi }
  , dim_detector_spec { settings.main.detector_spec }
  , dim_detector_spat { settings.main.detector_spat }
  , dim_detector_spec_full { settings.main.detector_spec_full.value() }
  , dim_detector_spat_full { settings.main.detector_spat_full.value() }
  , npix { settings.main.detector_spec * settings.main.detector_spat }
  , dim_vp { settings.main.n_viewports }
{
    if (settings.main.processes.front() < ProcLevel::l1c) {
        printHeading("CKD initialization");
    }
    // Star with an empty pixel mask (all pixels are live)
    pixel_mask.resize(npix, false);
    if (settings.main.ckd_file_in.empty()) {
        if (settings.main.processes.front() < ProcLevel::l1b) {
            spdlog::info("Starting with an empty CKD: {}",
                         settings.main.ckd_file_out);
            spdlog::info("Detector dimension  (spatial): {}",
                         dim_detector_spat);
            spdlog::info("Detector dimension (spectral): {}",
                         dim_detector_spec);
            spdlog::info("Number of viewports: {}", dim_vp);
        }
        return;
    }
    std::array<bool, static_cast<int>(ProcLevel::n_levels)> debug {};
    for (int i {}; i < static_cast<int>(ProcLevel::n_levels); ++i) {
        debug.at(i) = settings.debug.at(i).value();
    }
    readCKD(settings.main.ckd_file_in, target_level, debug, settings.skip);
    // If the user defined detector dimensions to not match those from
    // the CKD it is undefined behavior. Better to stop here.
    if (settings.main.detector_spat != 0
        && settings.main.detector_spat != dim_detector_spat) {
        throw std::runtime_error { "[main][detector_spat] does not match the "
                                   "detector dimension from the CKD" };
    }
    if (settings.main.detector_spec != 0
        && settings.main.detector_spec != dim_detector_spec) {
        throw std::runtime_error { "[main][detector_spec] does not match the "
                                   "detector dimension from the CKD" };
    }
    // Since more functions use the settings than the CKD copy the
    // dimensions to the settings so that they are easily available
    // (they might already be set).
    settings.main.detector_spec = dim_detector_spec;
    settings.main.detector_spat = dim_detector_spat;
}

// Like getAndReshape but in reverse
static auto reshapeAndPut(const std::vector<SharedArray<double>>& source,
                          netCDF::NcVar& nc_var) -> void
{
    std::vector<double> buffer(source.size() * source.front().size());
    for (int i {}; i < static_cast<int>(source.size()); ++i) {
        for (int j {}; j < static_cast<int>(source.at(i).size()); ++j) {
            buffer[i * source.front().size() + j] = source[i][j];
        }
    }
    nc_var.putVar(buffer.data());
}

// Similar but for a 3-dimensional array
static auto reshapeAndPut(
  const std::array<std::vector<SharedArray<double>>, dims::pol>& source,
  netCDF::NcVar& nc_var) -> void
{
    const std::array<int, 3> dims {
        static_cast<int>(source.size()),
        static_cast<int>(source.front().size()),
        static_cast<int>(source.front().front().size()),
    };
    // Same comments as above in getAndReshape
    const bool flip_dim { nc_var.getDim(0).getSize() != dims[0] };
    std::vector<double> buffer(
      std::accumulate(dims.cbegin(), dims.cend(), 1, std::multiplies<>()));
    int idx {};
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int j {}; j < dims[1]; ++j) {
            if (flip_dim) {
                idx = (j * dims::pol + i_pol) * dims[2];
            } else {
                idx = (i_pol * dims[1] + j) * dims[2];
            }
            for (int k {}; k < dims[2]; ++k) {
                buffer[idx + k] = source[i_pol][j][k];
            }
        }
    }
    nc_var.putVar(buffer.data());
}

static auto getOrCreateDebugGroup(const std::string& group_name, NcObject& nc)
  -> netCDF::NcGroup
{
    if (const netCDF::NcGroup grp { nc.getGroup("DEBUG") }; grp.isNull()) {
        return nc.addGroup("DEBUG").addGroup(group_name);
    } else {
        return grp.addGroup(group_name);
    }
}

auto CKD::write(const SPEXSettings& settings) -> void
{
    // Open an output CKD file and write some global attributes and a
    // reduced YAML configuration into it.
    NcObject nc { netCDF::NcFile::replace, settings.main.ckd_file_out };
    nc.writeReducedConfig(settings);
    nc.putAtt("title", "PACE SPEXone calibration key data (CKD)");
    // Define basic dimensions and pixel mask
    const netCDF::NcDim polarization_states {
        nc.addDim("polarization_directions", dims::pol),
    };
    const netCDF::NcVar polarization_states_str {
        nc.addVar(
          "polarization_directions", netCDF::ncString, polarization_states),
    };
    const char* polarization_states_map[dims::pol] { "-", "+" };
    polarization_states_str.putVar(&polarization_states_map);
    const netCDF::NcDim vector_elements {
        nc.addDim("vector_elements", dims::vec),
    };
    const netCDF::NcDim spatial_detector_pixels {
        nc.addDim("spatial_detector_pixels", dim_detector_spat),
    };
    const netCDF::NcDim spectral_detector_pixels {
        nc.addDim("spectral_detector_pixels", dim_detector_spec),
    };
    const netCDF::NcDim viewports {
        nc.addDim("viewports", dim_vp),
    };
    netCDF::NcVar viewports_str {
        nc.addVar("viewports", netCDF::ncString, viewports),
    };
    viewports_str.putAtt("units", "degrees");
    const char* viewports_map[dim_vp] { "-50", "-20", "0", "20", "50" };
    viewports_str.putVar(&viewports_map);
    // Mask is the only global variable
    const std::vector<netCDF::NcDim> dims_mask { spatial_detector_pixels,
                                                 spectral_detector_pixels };
    netCDF::NcVar mask_nc { nc.addVar("mask", netCDF::ncUbyte, dims_mask) };
    mask_nc.putAtt("long_name", "detector pixel mask");
    constexpr std::array<uint8_t, 2> mask_values { 0, 1 };
    mask_nc.putAtt(
      "flag_values", netCDF::ncUbyte, mask_values.size(), mask_values.data());
    mask_nc.putAtt("flag_meanings", "good bad");
    std::vector<uint8_t> pixel_mask_u8(npix);
    std::transform(pixel_mask.cbegin(),
                   pixel_mask.cend(),
                   pixel_mask_u8.begin(),
                   [](const auto v) { return v ? 1 : 0; });
    mask_nc.putVar(pixel_mask_u8.data());
    // Variable that is directly filled
    netCDF::NcVar var {};
    if (!settings.skip.at(static_cast<int>(ProcLevel::dark))) {
        netCDF::NcGroup grp { nc.addGroup("DARK") };
        // Create order dimension for temperature
        const netCDF::NcDim dimid_dark_order { grp.addDim(
          "dark_number_of_coefficients", dim_dark_order) };
        var = grp.addVar("dark_offset",
                         netCDF::ncDouble,
                         { spatial_detector_pixels,
                           spectral_detector_pixels,
                           dimid_dark_order });
        var.putAtt("long_name", "dark offset");
        const std::vector<double> knots { dark_nominal_temperature,
                                          dark_nominal_temperature + 1.0 };
        var.putAtt("knots", netCDF::ncDouble, knots.size(), knots.data());
        var.putAtt("units", "counts");
        var.putVar(dark_offset.data());
        var = grp.addVar("dark_current",
                         netCDF::ncDouble,
                         { spatial_detector_pixels,
                           spectral_detector_pixels,
                           dimid_dark_order });
        var.putAtt("long_name", "dark current");
        var.putAtt("knots", netCDF::ncDouble, knots.size(), knots.data());
        var.putAtt("comment",
                   "basis set continues as polynomials outside knot domain");
        var.putAtt("units", "counts/s");
        var.putVar(dark_current.data());
        // Dark nominal temperature
        var = grp.addVar("dark_nominal_temperature", netCDF::ncDouble);
        var.putAtt("long_name", "detector temperature at nominal conditions");
        var.putAtt("units", "K");
        var.putVar(&dark_nominal_temperature);
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::noise))) {
        netCDF::NcGroup grp { nc.addGroup("NOISE") };
        var = grp.addVar("g",
                         netCDF::ncDouble,
                         { spatial_detector_pixels, spectral_detector_pixels });
        var.putAtt("long_name", "noise conversion gain");
        var.putAtt("comment", "noise model: sigma = sqrt(g*signal + n)");
        var.putAtt("units", "counts/e");
        var.putVar(noise.g.data());
        var = grp.addVar("n",
                         netCDF::ncDouble,
                         { spatial_detector_pixels, spectral_detector_pixels });
        var.putAtt("long_name", "scaled read noise");
        var.putAtt("comment", "noise model: sigma = sqrt(g*signal + n)");
        var.putAtt("units", "counts^2/e");
        var.putVar(noise.n.data());
        if (settings.debug.at(static_cast<int>(ProcLevel::noise)).value()) {
            auto grp { getOrCreateDebugGroup("NOISE", nc) };
            var =
              grp.addVar("std_dev_g",
                         netCDF::ncDouble,
                         { spatial_detector_pixels, spectral_detector_pixels });
            var.putAtt("long_name",
                       "standard deviation of noise conversion gain");
            var.putAtt("units", "counts");
            var.putVar(noise.std_dev_g.data());
            var =
              grp.addVar("std_dev_n",
                         netCDF::ncDouble,
                         { spatial_detector_pixels, spectral_detector_pixels });
            var.putAtt("long_name", "standard deviation scaled dark noise");
            var.putAtt("units", "counts^2");
            var.putVar(noise.std_dev_n.data());
        }
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::nonlin))) {
        netCDF::NcGroup grp { nc.addGroup("NON_LINEARITY") };
        // These dimensions identifiers are only needed to write
        // this output.  The series dimension is only for
        // diagnostic CKD, so its dimension size is even a
        // non-owned pointer.
        const netCDF::NcDim dimid_nonlin_exptime { grp.addDim(
          "nonlin_number_of_exposure_times", dim_nonlin_exptime) };
        const netCDF::NcDim dimid_nonlin_spline { grp.addDim(
          "nonlin_number_of_coefficients", dim_nonlin_spline) };
        const netCDF::NcDim dimid_nonlin_knot { grp.addDim(
          "nonlin_number_of_knots", dim_nonlin_knot) };
        // Order (scalar)
        var = grp.addVar("nonlin_order", netCDF::ncInt);
        var.putAtt("long_name", "b-spline order of non-linearity fit");
        var.putAtt("definition",
                   "parameter n, so polynomial order is one lower");
        var.putVar(&nonlin_order);
        // Knots
        var = grp.addVar("nonlin_knots", netCDF::ncDouble, dimid_nonlin_knot);
        var.putAtt("long_name", "scaled b-spline knots of non-linearity fit");
        var.putVar(nonlin_knots.data());
        // Exposure times
        var =
          grp.addVar("nonlin_exptimes", netCDF::ncDouble, dimid_nonlin_exptime);
        var.putAtt(
          "long_name",
          "exposure times for which non-linearity correction is derived");
        var.putAtt("units", "s");
        var.putVar(nonlin_exptimes.data());
        // Signal scales
        var = grp.addVar("nonlin_signal_scale",
                         netCDF::ncDouble,
                         { spatial_detector_pixels,
                           spectral_detector_pixels,
                           dimid_nonlin_exptime });
        var.putAtt("long_name", "scaling factor of the b-spline knots");
        var.putAtt("units", "counts");
        var.putVar(nonlin_signal_scale.data());
        // Fit results
        var = grp.addVar("nonlin_fit",
                         netCDF::ncDouble,
                         { spatial_detector_pixels,
                           spectral_detector_pixels,
                           dimid_nonlin_exptime,
                           dimid_nonlin_spline });
        var.putAtt("long_name",
                   "b-spline coefficients for non-linearity curve");
        var.putAtt("units", "counts");
        var.putVar(nonlin_fit.data());
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::prnu))) {
        netCDF::NcGroup grp { nc.addGroup("PRNU") };
        var = grp.addVar("prnu",
                         netCDF::ncDouble,
                         { spatial_detector_pixels, spectral_detector_pixels });
        var.putAtt("long_name", "pixel response non-uniformity");
        var.putAtt("units", "1");
        var.putVar(prnu.prnu.data());
        if (settings.debug.at(static_cast<int>(ProcLevel::prnu)).value()) {
            auto grp { getOrCreateDebugGroup("PRNU", nc) };
            var =
              grp.addVar("sensitivity",
                         netCDF::ncDouble,
                         { spatial_detector_pixels, spectral_detector_pixels });
            var.putAtt("long_name", "polynomial fit of the calibration data");
            var.putAtt("units", "1");
            var.putVar(prnu.sensitivity_fit.data());
        }
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::stray))) {
        netCDF::NcGroup grp { nc.addGroup("STRAYLIGHT") };
        // Dimensions
        const netCDF::NcDim n_kernels {
            grp.addDim("kernels", stray.n_kernels),
        };
        const auto reduced_kernels_u8 { static_cast<uint8_t>(
          stray.reduced_kernels) };
        grp.addVar("reduced_kernels", netCDF::ncUbyte)
          .putVar(&reduced_kernels_u8);
        const netCDF::NcDim n_spatial { grp.addDim("spatial",
                                                   stray.n_spatial) };
        const netCDF::NcDim n_spectral { grp.addDim("spectral",
                                                    stray.n_spectral) };
        const netCDF::NcDim kernels_fft_size {
            // Need to multiply by 2 because of complex numbers
            grp.addDim("kernels_fft_size",
                       2
                         * std::accumulate(stray.kernel_fft_sizes.cbegin(),
                                           stray.kernel_fft_sizes.cend(),
                                           0)),
        };
        var = grp.addVar("kernel_rows", netCDF::ncInt, n_kernels);
        var.putAtt("long_name", "number of rows in each kernel");
        var.putAtt("units", "1");
        var.putVar(stray.kernel_rows.data());
        var = grp.addVar("kernel_cols", netCDF::ncInt, n_kernels);
        var.putAtt("long_name", "number of columns in each kernel");
        var.putAtt("units", "1");
        var.putVar(stray.kernel_cols.data());
        var = grp.addVar("kernel_fft_sizes", netCDF::ncInt, n_kernels);
        var.putAtt("long_name",
                   "sizes of kernel FFTs in units of complex double");
        var.putAtt("units", "1");
        var.putVar(stray.kernel_fft_sizes.data());
        // Variables
        var = grp.addVar("kernels_fft", netCDF::ncDouble, kernels_fft_size);
        var.putAtt("long_name", "Fourier transforms of the kernels");
        var.putAtt("units", "1");
        std::vector<size_t> start { 0 };
        for (int i_kernel {}; i_kernel < stray.n_kernels; ++i_kernel) {
            std::vector<size_t> count {
                2 * static_cast<size_t>(stray.kernel_fft_sizes.at(i_kernel))
            };
            var.putVar(start, count, stray.kernels_fft.at(i_kernel).data());
            start.front() += count.front();
        }
        var = grp.addVar("eta", netCDF::ncDouble, { n_spatial, n_spectral });
        var.putAtt("long_name", "internal scattering factor");
        var.putAtt("units", "1");
        var.putVar(stray.eta.data());
        var = grp.addVar(
          "weights", netCDF::ncDouble, { n_kernels, n_spatial, n_spectral });
        var.putAtt("long_name", "kernel weights");
        var.putAtt("units", "1");
        reshapeAndPut(stray.weights, var);
        const netCDF::NcDim edges_dim { grp.addDim("edges",
                                                   stray.edges.size()) };
        var = grp.addVar("edges", netCDF::ncInt, edges_dim);
        var.putAtt("units", "1");
        var.putVar(stray.edges.data());
        if (settings.debug.at(static_cast<int>(ProcLevel::stray)).value()) {
            auto grp { getOrCreateDebugGroup("STRAYLIGHT", nc) };
            const netCDF::NcDim n_kernels {
                grp.addDim("kernels", stray.n_kernels),
            };
            const netCDF::NcDim kernel_n_rows {
                grp.addDim("kernel_rows", stray.kernel_n_rows_full),
            };
            const netCDF::NcDim kernel_n_cols {
                grp.addDim("kernel_cols", stray.kernel_n_cols_full),
            };
            const netCDF::NcDim n_spatial { grp.addDim("spatial",
                                                       stray.n_spatial) };
            const netCDF::NcDim n_spectral { grp.addDim("spectral",
                                                        stray.n_spectral) };
            var = grp.addVar("kernels_real",
                             netCDF::ncDouble,
                             { n_kernels, kernel_n_rows, kernel_n_cols });
            var.putAtt("long_name", "real space kernels");
            var.putAtt("units", "1");
            reshapeAndPut(stray.kernels_real, var);
            var = grp.addVar(
              "etas", netCDF::ncDouble, { n_kernels, n_spatial, n_spectral });
            reshapeAndPut(stray.etas, var);
            var.putAtt("long_name", "individual internal scattering factors");
            var.putAtt("units", "1");
        }
    }
    // Define this dimension outside the FOV scope because it is also
    // used by other steps.
    netCDF::NcDim spatial_samples_per_image {};
    if (fov.n_act > 0) {
        spatial_samples_per_image =
          nc.addDim("spatial_samples_per_image", fov.n_act);
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::fov))) {
        netCDF::NcGroup grp { nc.addGroup("FIELD_OF_VIEW") };
        var = grp.addVar("fov_nfov_vp", netCDF::ncInt, viewports);
        var.putAtt("long_name", "number of spatial samples per viewport");
        var.putVar(fov.n_act_vp.data());
        var = grp.addVar("fov_ifov_start_vp", netCDF::ncInt, viewports);
        var.putAtt("long_name", "first spatial sample index in each viewport");
        var.putVar(fov.start_vp.data());
        var = grp.addVar(
          "fov_act_angles", netCDF::ncDouble, spatial_samples_per_image);
        var.putAtt("long_name",
                   "across track rotation stage position per spatial sample");
        var.putAtt("units", "radians");
        var.putVar(fov.act_angles.data());
        var = grp.addVar("fov_ispat",
                         netCDF::ncDouble,
                         { spatial_samples_per_image,
                           polarization_states,
                           spectral_detector_pixels });
        var.putAtt(
          "long_name",
          "floating point spatial detector pixel indices of the spectra");
        reshapeAndPut(fov.spat_indices, var);
        if (settings.debug.at(static_cast<int>(ProcLevel::fov)).value()) {
            auto grp { getOrCreateDebugGroup("FOV", nc) };
            const auto n_l1a { grp.addDim("measurements", fov.n_l1a_total) };
            const auto n_columns { grp.addDim("peak_positions",
                                              fov.n_columns) };
            var = grp.addVar("peak_positions",
                             netCDF::ncDouble,
                             { n_l1a, polarization_states, n_columns });
            var.putAtt(
              "long_name",
              "peak positions determined from Gaussian fits across columns");
            var.putAtt("units", "1");
            reshapeAndPut(fov.peak_pos, var);
            var = grp.addVar(
              "peak_positions_b",
              netCDF::ncDouble,
              { n_l1a, polarization_states, spectral_detector_pixels });
            var.putAtt("long_name", "B-spline fits of peak positions");
            var.putAtt("units", "1");
            reshapeAndPut(fov.peak_pos_b, var);
        }
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::swath))) {
        netCDF::NcGroup grp { nc.addGroup("SWATH") };
        // Swath vectors
        const std::vector<netCDF::NcDim> dims_swathvectors {
            spatial_samples_per_image, vector_elements
        };
        var =
          grp.addVar("swath_swathvectors", netCDF::ncDouble, dims_swathvectors);
        var.putAtt("long_name",
                   "normalized pointing vectors in spacecraft coordinates");
        var.putAtt("axes", "x = flying direction, y = to the right, z = down");
        var.putVar(swath_swathvectors.data());
        const std::vector<netCDF::NcDim> dims_vectorplane_normals {
            viewports, vector_elements
        };
        var = grp.addVar("swath_vectorplane_normals",
                         netCDF::ncDouble,
                         dims_vectorplane_normals);
        var.putAtt(
          "long_name",
          "normal vectors to plane through swath vectors of viewport.");
        var.putAtt("axes", "x = flying direction, y = to the right, z = down");
        var.putVar(swath_vectorplane_normals.data());
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::wave))) {
        netCDF::NcGroup grp { nc.addGroup("WAVELENGTH") };
        // Wavelengths per spectrum
        const std::vector<netCDF::NcDim> dims_wave_full {
            polarization_states,
            spatial_samples_per_image,
            spectral_detector_pixels
        };
        var = grp.addVar("wave_full", netCDF::ncDouble, dims_wave_full);
        var.putAtt("long_name", "wavelengths of S and P spectra");
        var.putAtt("units", "nm");
        reshapeAndPut(wave_full, var);
        const std::vector<netCDF::NcDim> dims_wave_common {
            spatial_samples_per_image, spectral_detector_pixels
        };
        var = grp.addVar("wave_common", netCDF::ncDouble, dims_wave_common);
        var.putAtt("long_name", "wavelengths of modulation spectra");
        var.putAtt("units", "nm");
        reshapeAndPut(wave_common, var);
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::rad))) {
        netCDF::NcGroup grp { nc.addGroup("RADIOMETRIC") };
        // Radiometric calibration data on spectrum perspective
        const std::vector<netCDF::NcDim> dims_rad_spectra {
            spatial_samples_per_image,
            polarization_states,
            spectral_detector_pixels
        };
        var = grp.addVar("rad_spectra", netCDF::ncDouble, dims_rad_spectra);
        var.putAtt("long_name", "radiometric calibration constants");
        var.putAtt("units", "J m-2 sr-1 um-1 counts-1");
        reshapeAndPut(rad_spectra, var);
    }
    if (!settings.skip.at(static_cast<int>(ProcLevel::pol))) {
        netCDF::NcGroup grp { nc.addGroup("POLARIMETRIC") };
        // Polarimetric calibration sampled on spectral pixels
        const std::vector<netCDF::NcDim> dims_pol { spatial_samples_per_image,
                                                    spectral_detector_pixels };
        // Mueller matrix ratio for Q
        var = grp.addVar("pol_m_q", netCDF::ncDouble, dims_pol);
        var.putAtt("long_name", "modulation sensitivy to q-polarization");
        var.putAtt("units", "1");
        reshapeAndPut(pol.m_q, var);
        // Mueller matrix ratio for U
        var = grp.addVar("pol_m_u", netCDF::ncDouble, dims_pol);
        var.putAtt("long_name", "modulation sensitivy to u-polarization");
        var.putAtt("units", "1");
        reshapeAndPut(pol.m_u, var);
        // Mueller element for telescope polarization
        var = grp.addVar("pol_m_t", netCDF::ncDouble, dims_pol);
        var.putAtt("long_name", "telescope polarization");
        var.putAtt("units", "1");
        reshapeAndPut(pol.m_t, var);
        if (settings.debug.at(static_cast<int>(ProcLevel::pol)).value()) {
            auto grp { getOrCreateDebugGroup("POLARIMETRIC", nc) };
            var = grp.addVar("eps", netCDF::ncDouble, dims_pol);
            var.putAtt("long_name", "polarimetric efficiency");
            var.putAtt("units", "1");
            reshapeAndPut(pol.eps, var);
            var = grp.addVar("delta", netCDF::ncDouble, dims_pol);
            var.putAtt("long_name", "polarimetric retardance");
            var.putAtt("units", "nm");
            reshapeAndPut(pol.delta, var);
        }
    }
}

} // namespace spex
