// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "darkcal.h"

#include "array.h"
#include "bspline.h"
#include "ckd.h"
#include "functions.h"
#include "l1a.h"
#include "lininv.h"
#include "matrix.h"

#include <cassert>
#include <spdlog/spdlog.h>
#include <string>

namespace spex {

// Dark correction calibration software
// 1. Extract necessary metadata from input
// 2. Fit temperature-dependent dark signal
// 3. Add a noise estimate
auto Darkcal::processBatch(const int /*i_batch*/) -> void
{
    const auto& l1a_products { batches.front().l1a_products };
    const int n_l1a { static_cast<int>(l1a_products.size()) };
    std::vector<double> time(n_l1a, 0.0); // Should end up with size ntime
    std::vector<double> temp(n_l1a, 0.0); // Should end up with size ntemp

    for (size_t il1a=0 ; il1a<n_l1a ; il1a++) {
        time[il1a] = l1a_products[il1a].get().exposure_time;
        temp[il1a] = l1a_products[il1a].get().temp_detector;
    }

    // Everything is independent of pixels what so ever
    // Same idea, such that first term is at nominal temperature
    std::vector<double> knots_temp = {settings.dark.nominal_temperature,settings.dark.nominal_temperature+1};
    Bspline b_temp(settings.dark.order,2,knots_temp.data());
    ckd.dim_dark_order = b_temp.nspline;
    size_t nstate = 2*b_temp.nspline;

    // Construct full Jacobian. The state will be just the temperature B-splines followed
    // by the temperature B-splines multiplied with the exposure time. So the
    // first nspline components are for the dark offset and the last nspline
    // components are for the dark current.
    std::vector<double> jac(nstate*n_l1a, 0.0);
    b_temp.jaccalc(n_l1a,temp.data(),jac.data());
    // That is the first half of the Jacobian. Now, the second.
    for (size_t ispline=0 ; ispline<b_temp.nspline ; ispline++) {
        for (size_t il1a=0 ; il1a<n_l1a ; il1a++) {
            jac[(ispline+b_temp.nspline)*n_l1a+il1a] = jac[ispline*n_l1a+il1a] * time[il1a];
        }
    }

    // Saturation protection: Figure out what percentage of well filling each
    // measurement is. That will be based on the medians of the full images.
    // Next, a fit is performed as normal, using these medians and based on the
    // chi squared, it is figured out which measurements to include.
    // We could use the same measurements for each pixel, or figure out the
    // subset again for each pixel.
    medians.resize(n_l1a); // This is a member variable, because if detailed output is selected, this is included in the output.
    spdlog::info("Acquire image medians");
    for (size_t il1a=0 ; il1a<n_l1a ; il1a++) {
        median(ckd.npix, l1a_products[il1a].get().image.data(), medians[il1a]);
    }
    // Sort this thing
    std::vector<size_t> indexarray(n_l1a);
    std::vector<double> medians_sorted(n_l1a);
    Array<double>::sort(n_l1a,medians.data(),medians_sorted.data(),indexarray.data());

    spdlog::info("Construct saturation filter based on fit through medians");
    size_t n_l1a_include_start = 0;
    // Include all measurements with less well filling than threshold from setting
    double mn = medians_sorted[0];
    double mx = medians_sorted[n_l1a-1];
    for (size_t il1a=0 ; il1a<n_l1a ; il1a++) {
        double& med = medians_sorted[il1a];
        if ((med-mn) / (mx-mn) < settings.dark.saturation_fraction_start) n_l1a_include_start++;
    }

    // The fit will fail if there are not enough included L1A images, but
    // if there are just enough points, the fit will be perfect and the
    // chi squared will be undefined, so the measurement selection protocol
    // will fail.
    assert(!(n_l1a_include_start <= nstate) && "Insufficient measurements in initial reference fraction for protocol. Increase setting 'saturation_fraction_start' or include more low measurements.");

    // Create mask
    std::vector<bool> mask(n_l1a,false);

    // Remove excluded measurements from the Jacobian
    for (size_t irem=n_l1a_include_start ; irem<n_l1a ; irem++) {
        mask[indexarray[irem]] = true;
    }
    // Perform a reference retrieval for the chosen start threshold
    std::vector<double> res(nstate);
    linear_invert(nstate,n_l1a,jac.data(),OPT_NONE,NULL,&mask,medians.data(),res.data());
    std::vector<double> mod(n_l1a, 0.0);
    Matrix::matmul_rl_fold_slow(nstate,n_l1a,jac.data(),res.data(),mod.data());
    double chi2_ref = 0.0;
    for (size_t iinc=0 ; iinc<n_l1a_include_start ; iinc++) {
        size_t& idx = indexarray[iinc];
        chi2_ref += pow(medians[idx]-mod[idx],2.0) / (n_l1a_include_start-nstate);
    }

    // If the loop is completed without break, all measurements should be
    // included. Then, nl1a_include will not be overwritten.
    size_t nl1a_include = n_l1a;
    for (size_t iinc_add=n_l1a_include_start ; iinc_add<n_l1a ; iinc_add++) {
        // Restore measurement number iinc_add
        mask[indexarray[iinc_add]] = false;
        // Perform the retrieval
        std::vector<double> res(nstate, 0.0);
        linear_invert(nstate,n_l1a,jac.data(),OPT_NONE,NULL,&mask,medians.data(),res.data());
        std::vector<double> mod(n_l1a, 0.0);
        Matrix::matmul_rl_fold_slow(nstate,n_l1a,jac.data(),res.data(),mod.data());
        double chi2_test = 0.0;
        // Index iinc_add is included now
        for (size_t iinc=0 ; iinc<=iinc_add ; iinc++) {
            size_t& idx = indexarray[iinc];
            chi2_test += pow(medians[idx]-mod[idx],2.0) / (n_l1a_include_start-nstate);
        }
        if (chi2_test > settings.dark.chi2_increment_threshold*chi2_ref) {
            // Remove last measurement again
            mask[indexarray[iinc_add]] = true;
            // Now, index iinc_add is not included anymore, so total number is equal to iinc_add
            nl1a_include = iinc_add; // Overwrite this thing
            break;
        }
    }
    spdlog::debug("Including {} measurements in the fit", nl1a_include);
    // Regardless of whether the loop is completed or broken. nl1a_include is
    // now set to the right value.
    // More importantly, mask is now correct.

    // We make a gain matrix that is complete, it will have zeros for the
    // excluded measurements. That has to do with the misery that would arise
    // because they are not sorted.
    std::vector<double> gain(nstate*n_l1a, 0.0);
    linear_invert_gain(nstate,n_l1a,jac.data(),OPT_NONE,NULL,&mask,gain.data());

    // Construct output
    ckd.dark_offset.resize(ckd.dim_dark_order*ckd.npix);
    ckd.dark_current.resize(ckd.dim_dark_order*ckd.npix);
    dark_chi2.resize(ckd.npix);

    // Moving pointers
    double* dark_offset_cur = ckd.dark_offset.data();
    double* dark_current_cur = ckd.dark_current.data();
    double* dark_chi2_cur = dark_chi2.data();

    spdlog::info("Fitting temperature-dependent dark signal per pixel");
    for (size_t ipix=0 ; ipix<ckd.npix ; ipix++) {
        std::vector<double> meas(n_l1a);
        for (size_t il1a=0 ; il1a<n_l1a ; il1a++) {
            meas[il1a] = l1a_products[il1a].get().image[ipix];
        }
        std::vector<double> res(nstate);

        Matrix::matmul_rl_fold_quick(nstate,n_l1a,gain.data(),meas.data(),res.data());

        // Split results into dark offset and dark current
        memcpy(dark_offset_cur,res.data(),b_temp.nspline*sizeof(double));
        memcpy(dark_current_cur,&res[b_temp.nspline],b_temp.nspline*sizeof(double));

        // Calculate chi squared
        *dark_chi2_cur = 0.0;
        std::vector<double> mod(n_l1a);
        Matrix::matmul_rl_fold_slow(nstate,n_l1a,jac.data(),res.data(),mod.data());
        for (size_t iinc=0 ; iinc<nl1a_include ; iinc++) {
            size_t& idx = indexarray[iinc];
            *dark_chi2_cur += pow(meas[idx]-mod[idx],2.0) / (nl1a_include - nstate);
        }

        // Apply pixel mask
        if (settings.dark.mask_chi2_max && *dark_chi2_cur > settings.dark.mask_chi2_max) ckd.pixel_mask[ipix] = true;
        // The offset and current is checked for nominal temperature
        // That is the first element
        if (settings.dark.mask_offset_min && dark_offset_cur[0] < settings.dark.mask_offset_min) ckd.pixel_mask[ipix] = true;
        if (settings.dark.mask_offset_max && dark_offset_cur[0] > settings.dark.mask_offset_max) ckd.pixel_mask[ipix] = true;
        if (settings.dark.mask_current_min && dark_current_cur[0] < settings.dark.mask_current_min) ckd.pixel_mask[ipix] = true;
        if (settings.dark.mask_current_max && dark_current_cur[0] > settings.dark.mask_current_max) ckd.pixel_mask[ipix] = true;

        // Progress pointers
        dark_offset_cur += ckd.dim_dark_order;
        dark_current_cur += ckd.dim_dark_order;
        dark_chi2_cur++;

    }
    // Do not forget to communicate the nominal temperature to the CKD
    ckd.dark_nominal_temperature = settings.dark.nominal_temperature;
}

} // namespace spex
