// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "driver.h"

#include "ckd.h"
#include "constants.h"
#include "darkcal.h"
#include "date.h"
#include "fov_gen.h"
#include "io.h"
#include "l1b.h"
#include "l1c.h"
#include "noise_gen.h"
#include "nonlincal.h"
#include "pol_gen.h"
#include "prnu_gen.h"
#include "processor.h"
#include "rad_gen.h"
#include "settings.h"
#include "stray_gen.h"
#include "swathcal.h"
#include "wave_gen.h"

#include <ranges>
#include <spdlog/spdlog.h>

namespace spex {

// NOLINTNEXTLINE(misc-misplaced-const)
auto driver(const MPI_Comm mpi_comm, SPEXSettings& settings) -> void
{
    // Initialize MPI variables
    const gadfit::MPIVars mpi { mpi_comm };
    // Set up loggers
    initLogging(mpi);
    // General description of the executable
    printHeading("SPEXone processor", false);
    spdlog::get("plain")->info("Version                 : {}",
                               SPEXONE_PROJECT_VERSION);
    spdlog::get("plain")->info("Commit hash             : {}",
                               SPEXONE_GIT_COMMIT_ABBREV);
    spdlog::get("plain")->info("Date and timezone       : {}", getDate());
    spdlog::get("plain")->info(
      "Contacts                : raullaasner@gmail.com\n"
      "                          spexone.slack.com (request permission)\n"
      "                          bitbucket.org/sron_earth/spexone_cal/issues "
      "(request permission)");
    spdlog::get("plain")->info("Host system             : {}",
                               SPEXONE_CMAKE_HOST_SYSTEM);
    spdlog::get("plain")->info("Executable location     : {}",
                               SPEXONE_EXECUTABLE);
    spdlog::get("plain")->info("C++ compiler            : {}",
                               SPEXONE_CXX_COMPILER);
    spdlog::get("plain")->info("C++ compiler flags      : {}",
                               SPEXONE_CXX_COMPILER_FLAGS);
#ifdef USE_MPI
    spdlog::get("plain")->info("Number of MPI processes : {}", mpi.n_procs);
#endif
#ifndef __clang__
    for (bool first_line { true };
         const auto lib : SPEXONE_LIBRARIES | std::views::split(' ')
                            | std::views::transform([](auto&& str) {
                                  return std::string_view(
                                    &*str.begin(), std::ranges::distance(str));
                              })) {
        if (first_line) {
            spdlog::get("plain")->info("Linking against         : {}", lib);
            first_line = false;
        } else {
            spdlog::get("plain")->info("                          {}", lib);
        }
    }
#endif
    // At first the CKD is mature for the lowest process specified in
    // the input file. The maturity can increase in the CKD
    // constructor depending on processes in the input CKD.
    const ProcLevel target_ckd_level { settings.main.processes.front() };
    // Define the CKD which will be partially or fully constructed
    // from an input CKD. If there is no input CKD, this constructor
    // only sets the basic dimensions from the configuration.
    CKD ckd { mpi, target_ckd_level, settings };
    // Execute each process from settings.main.processes. These are in
    // correct order.
    for (const auto& proc_level : settings.main.processes) {
        std::unique_ptr<Processor> processor {};
        switch (proc_level) {
        case ProcLevel::dark:
            processor = std::make_unique<Darkcal>(mpi, settings, ckd);
            break;
        case ProcLevel::noise:
            processor = std::make_unique<NoiseGen>(mpi, settings, ckd);
            break;
        case ProcLevel::nonlin:
            processor = std::make_unique<Nonlincal>(mpi, settings, ckd);
            break;
        case ProcLevel::prnu:
            processor = std::make_unique<PrnuGen>(mpi, settings, ckd);
            break;
        case ProcLevel::stray:
            processor = std::make_unique<StrayGen>(mpi, settings, ckd);
            break;
        case ProcLevel::fov:
            processor = std::make_unique<FovGen>(mpi, settings, ckd);
            break;
        case ProcLevel::swath:
            processor = std::make_unique<Swathcal>(mpi, settings, ckd);
            break;
        case ProcLevel::wave:
            processor = std::make_unique<WaveGen>(mpi, settings, ckd);
            break;
        case ProcLevel::rad:
            processor = std::make_unique<RadGen>(mpi, settings, ckd);
            break;
        case ProcLevel::pol:
            processor = std::make_unique<PolGen>(mpi, settings, ckd);
            break;
        case ProcLevel::l1b:
            processor = std::make_unique<L1B>(mpi, settings, ckd);
            break;
        case ProcLevel::l1c:
            processor = std::make_unique<L1C>(mpi, settings, ckd);
            break;
        default:
            throw std::runtime_error { "unknown process" };
        }
        processor->execute();
        // If generating the CKD, write to output if the current
        // process is the last of the CKD steps.
        if (const auto next_proc { std::next(
              std::ranges::find(settings.main.processes, proc_level)) };
            proc_level <= ProcLevel::pol
            && (next_proc == settings.main.processes.cend()
                || *next_proc > ProcLevel::pol)
            && mpi.rank == 0) {
            ckd.write(settings);
        }
    }
}

} // namespace spex
