// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "fourier.h"

#include "fftw_data.h"

#include <cassert>

namespace spex {

auto fft_r2c(const int n_rows,
             const int n_cols,
             std::vector<double>& image,
             SharedArray<std::complex<double>>& image_fft) -> void
{
    fftw_plan fft_plan =
      fftw_plan_dft_r2c_2d(n_rows,
                           n_cols,
                           image.data(),
                           reinterpret_cast<fftw_complex*>(image_fft.data()),
                           FFTW_ESTIMATE);
    fftw_execute(fft_plan);
    fftw_destroy_plan(fft_plan);
}

auto convolve(const int image_n_rows,
              const int image_n_cols,
              const std::vector<double>& image_in,
              const int kernel_n_rows,
              const int kernel_n_cols,
              const spex::SharedArray<double>& kernel_in,
              const spex::SharedArray<std::complex<double>>& kernel_fft,
              std::vector<double>& image_out) -> void
{
    const int kernel_size { kernel_n_rows * kernel_n_cols };
    assert(image_n_rows <= kernel_n_rows
           && "Image dimension must be less than the kernel dimension");
    assert(image_n_cols <= kernel_n_cols
           && "Image dimension must be less than the kernel dimension");
    FFTWData fftw_data { kernel_n_rows, kernel_n_cols };
    for (int i {}; i < image_n_rows; ++i) {
        for (int j {}; j < image_n_cols; ++j) {
            fftw_data.image[i * kernel_n_cols + j] =
              image_in[i * image_n_cols + j];
        }
    }
    for (int i {}; i < kernel_n_rows; ++i) {
        for (int j = image_n_cols; j < kernel_n_cols; ++j) {
            fftw_data.image[i * kernel_n_cols + j] = 0.0;
        }
    }
    for (int i = image_n_rows; i < kernel_n_rows; ++i) {
        for (int j {}; j < image_n_cols; ++j) {
            fftw_data.image[i * kernel_n_cols + j] = 0.0;
        }
    }

    fftw_data.execute(spex::FFTWData::FORWARD);

    for (int i {}; i < static_cast<int>(kernel_fft.size()); ++i) {
        const auto image_r { fftw_data.image_fft[i][0] };
        const auto image_i { fftw_data.image_fft[i][1] };
        const auto kernel_r { kernel_fft[i].real() };
        const auto kernel_i { kernel_fft[i].imag() };
        fftw_data.image_fft[i][0] = image_r * kernel_r - image_i * kernel_i;
        fftw_data.image_fft[i][1] = image_r * kernel_i + image_i * kernel_r;
    }

    fftw_data.execute(spex::FFTWData::REVERSE);

    image_out.resize(image_n_rows * image_n_cols);
    for (int i {}; i < image_n_rows; ++i) {
        for (int j {}; j < image_n_cols; ++j) {
            image_out[i * image_n_cols + j] =
              fftw_data.image[i * kernel_n_cols + j] / kernel_size;
        }
    }
}

} // namespace spex
