// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "fov_gen.h"

#include "bspline.h"
#include "ckd.h"
#include "io.h"
#include "l1a.h"
#include "l1a_metadata.h"
#include "linalg.h"
#include "spex_math.h"

#include <gadfit/lm_solver.h>
#include <numeric>
#include <spdlog/spdlog.h>

namespace spex {

auto FovGen::processInit() -> void
{
    ckd.fov.n_columns = ckd.dim_detector_spec
                        - settings.fov.exclude_columns_right
                        - settings.fov.exclude_columns_left;
    ckd.fov.start_vp.resize(ckd.dim_vp, 0);
    ckd.fov.n_act_vp.resize(ckd.dim_vp, 0);
    ckd.fov.dims_spec.resize(ckd.dim_detector_spat, ckd.dim_detector_spec);
    // Allocate maximum size for the CKD and shrink in the end when we know the
    // number of FOVs (ACT angles used)
    ckd.fov.act_angles.resize(ckd.dim_detector_spat);
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        resize(ckd.fov.spat_indices[i_pol],
               mpi,
               ckd.dim_detector_spat,
               ckd.dim_detector_spec);
        resize(ckd.fov.pix1[i_pol],
               mpi,
               ckd.dim_detector_spat,
               ckd.dim_detector_spec);
        resize(ckd.fov.pix2[i_pol],
               mpi,
               ckd.dim_detector_spat,
               ckd.dim_detector_spec);
        resize(ckd.fov.weight[i_pol],
               mpi,
               ckd.dim_detector_spat,
               ckd.dim_detector_spec);
        resize(ckd.fov.peak_pos[i_pol],
               mpi,
               ckd.dim_detector_spat,
               ckd.fov.n_columns);
        resize(ckd.fov.peak_pos_b[i_pol],
               mpi,
               ckd.dim_detector_spat,
               ckd.dim_detector_spec);
    }
}

auto FovGen::prepareBatches() -> void
{
    batchViewport();
}

// Remove those calibration measurements from consideration that are
// not visible on the detector. This is determined by comparing light
// sums (sum over all pixel values) with the highest light sum of all
// measurements of the current batch.
static auto filterByLightSums(
  std::vector<std::reference_wrapper<L1A>>& l1a_refs) -> void
{
    const int n_l1a { static_cast<int>(l1a_refs.size()) };
    const int npix { static_cast<int>(l1a_refs.front().get().image.size()) };
    std::vector<double> light_sums(n_l1a, 0.0);
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        for (int i {}; i < npix; ++i) {
            if (!l1a_refs[i_l1a].get().pixel_mask[i]) {
                light_sums[i_l1a] += l1a_refs[i_l1a].get().image[i];
            }
        }
    }
    const double max_light_sum { *std::ranges::max_element(light_sums) };
    constexpr double light_sum_thres { 0.1 };
    auto l1a_first { l1a_refs.begin() };
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        if (light_sums.at(i_l1a) > light_sum_thres * max_light_sum) {
            l1a_first = l1a_refs.begin() + i_l1a;
            break;
        } else {
            spdlog::debug(
              "Discarding {}",
              shortenFilename(l1a_refs.at(i_l1a).get().meta->filename));
        }
    }
    auto l1a_last { l1a_refs.end() };
    for (int i_l1a { n_l1a - 1 }; i_l1a >= 0; --i_l1a) {
        if (light_sums.at(i_l1a) > light_sum_thres * max_light_sum) {
            l1a_last = l1a_refs.begin() + i_l1a + 1;
            break;
        } else {
            spdlog::debug(
              "Discarding {}",
              shortenFilename(l1a_refs.at(i_l1a).get().meta->filename));
        }
    }
    l1a_refs = { l1a_first, l1a_last };
}

auto FovGen::processBatch(const int i_vp) -> void
{
    spdlog::info("Generating field of view CKD for viewport: {} deg",
                 viewport_idx_to_angle.at(
                   batches.at(i_vp).l1a_products.front().get().meta->viewport));
    // Sort all L1A instances of this batch by ACT angle
    std::vector<std::reference_wrapper<L1A>> l1a_refs {
        batches[i_vp].l1a_products.begin(), batches[i_vp].l1a_products.end()
    };
    std::ranges::sort(l1a_refs, [](const auto& a, const auto& b) {
        return a.get().meta->act_angle < b.get().meta->act_angle;
    });
    // Remove measurements that are not visible in this viewport
    filterByLightSums(l1a_refs);
    spdlog::info("  Smallest ACT angle used: {} deg",
                 l1a_refs.front().get().meta->act_angle / math::deg_to_rad);
    spdlog::info("  Largest  ACT angle used:  {} deg",
                 l1a_refs.back().get().meta->act_angle / math::deg_to_rad);
    // Total number of measurements used from this batch after filtering
    const int n_l1a { static_cast<int>(l1a_refs.size()) };
    // Prepare B-spline
    std::vector<double> knots(settings.fov.n_knots, 0.0);
    for (int i {}; i < settings.fov.n_knots; ++i) {
        knots[i] = static_cast<double>(i) * ckd.dim_detector_spec
                   / (settings.fov.n_knots - 1);
    }
    // The knots should be slightly outside the data range
    knots.front() -= 1e-10;
    knots.back() += 1e-10;
    BSpline b_spline { settings.fov.bspline_order, knots };
    const int n_states { b_spline.nStates() };
    // x-values for constructing the B-spline matrix
    std::vector<double> x_data(ckd.fov.n_columns);
    std::iota(x_data.begin(),
              x_data.end(),
              static_cast<double>(settings.fov.exclude_columns_left));
    // B-spline basis matrix
    std::vector<double> B_mat {};
    b_spline.evalBasis(x_data, B_mat);
    // Parallelize over columns (spectral dimension)
    std::vector<int> mpi_ranges {};
    genWorkRanges(ckd.fov.n_columns, mpi.n_procs, mpi_ranges);
    // Standard deviations of the peak positions. This array store one
    // value per column for a given detector row.
    std::array<SharedArray<double>, dims::pol> std_devs {};
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        std_devs.at(i_pol) = { true, mpi, ckd.fov.n_columns };
    }
    // Polarization states are handled separately
    const int dim_detector_spat_half { ckd.dim_detector_spat / 2 };
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        spdlog::info("    Fitting peaks along detector columns: {:5.1f}%",
                     100.0 * i_l1a / n_l1a);
        for (int i_spec { mpi_ranges.at(mpi.rank) };
             i_spec < mpi_ranges.at(mpi.rank + 1);
             ++i_spec) {
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                // For the Gaussian fit, x-data are row (spatial)
                // indices and y-data are the pixel intensities.
                std::vector<double> x_data {};
                std::vector<double> y_data {};
                std::vector<double> noise {};
                for (int i_spat { i_pol * dim_detector_spat_half };
                     i_spat < (i_pol + 1) * dim_detector_spat_half;
                     ++i_spat) {
                    const int idx { i_spat * ckd.dim_detector_spec + i_spec
                                    + settings.fov.exclude_columns_left };
                    if (!l1a_refs[i_l1a].get().pixel_mask[idx]) {
                        x_data.push_back(i_spat);
                        y_data.push_back(l1a_refs[i_l1a].get().image[idx]);
                        noise.push_back(l1a_refs[i_l1a].get().noise[idx]);
                    }
                }
                gadfit::LMsolver solver {
                    [](const std::vector<gadfit::AdVar>& parameters,
                       const double x) {
                        // a * exp(-(x0-x)^2/sigma) + bgr
                        gadfit::AdVar diff { parameters[0] - x };
                        return parameters[2] * exp(-diff * diff / parameters[1])
                               + parameters[3];
                    }
                };
                solver.addDataset(x_data, y_data, noise);
                // Maximum pixel intensity across the column is set as
                // the initial guess for the Levenberg-Marquardt
                // solver.
                const auto max_loc { std::ranges::distance(
                  y_data.cbegin(), std::ranges::max_element(y_data)) };
                solver.setPar(0, x_data[max_loc], true); // x0
                solver.setPar(1, 1.0, true);             // sigma
                solver.setPar(2, y_data[max_loc], true); // a
                solver.setPar(3, 0.0, true);             // bgr
                solver.settings.iteration_limit = 30;
                loggerOnOff();
                solver.fit();
                loggerOnOff();
                ckd.fov.peak_pos[i_pol][ckd.fov.n_l1a_total + i_l1a][i_spec] =
                  solver.getParValue(0);
                std_devs[i_pol][i_spec] = solver.chi2()
                                          / solver.degreesOfFreedom()
                                          * solver.getInvJTJ().front();
                // Filter bad values. For instance, some columns close
                // to the edge of the detector produce unusually low
                // standard deviations.
                if (std::isnan(std_devs[i_pol][i_spec])
                    || std_devs[i_pol][i_spec] < 1e-10) {
                    std_devs[i_pol][i_spec] = 1e30;
                }
            }
        }
        // Wait before the root process starts working on some of the
        // shared arrays.
#ifdef USE_MPI
        MPI_Barrier(mpi.comm);
#endif
        if (mpi.rank == 0) {
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                std::vector<double> B_mat_copy { B_mat };
                std::vector<double> state_vector(n_states, 0.0);
                dppsv(ckd.fov.peak_pos[i_pol][ckd.fov.n_l1a_total + i_l1a],
                      std_devs[i_pol],
                      B_mat_copy,
                      state_vector);
                for (int i {}; i < ckd.dim_detector_spec; ++i) {
                    ckd.fov.peak_pos_b[i_pol][ckd.fov.n_l1a_total + i_l1a][i] =
                      b_spline.deBoor(state_vector, static_cast<double>(i));
                }
            }
        }
#ifdef USE_MPI
        MPI_Barrier(mpi.comm);
#endif
    }
    spdlog::info("  Fitting peaks along detector columns: 100.0%");
    // Number of negative L1B ACT angles determined by the lowest ACT
    // angle in the measurement set (after filtering by light sums
    // above) and the user supplied ACT sampling.
    const int n_act_neg { static_cast<int>(
      -l1a_refs.front().get().meta->act_angle
      / settings.fov.act_sampling.at(i_vp)) };
    // Number of positive L1B ACT angles
    const int n_act_pos { static_cast<int>(
      l1a_refs.back().get().meta->act_angle
      / settings.fov.act_sampling.at(i_vp)) };
    // Total number is the sum plus one in the middle corresponding to ACT = 0
    ckd.fov.n_act_vp.at(i_vp) = n_act_neg + n_act_pos + 1;
    if (i_vp > 0) {
        ckd.fov.start_vp.at(i_vp) =
          ckd.fov.start_vp.at(i_vp - 1) + ckd.fov.n_act_vp.at(i_vp - 1);
    }
    spdlog::info("Interpolating to the target L1B ACT angles ({})",
                 ckd.fov.n_act_vp.at(i_vp));
    // Generate a list of ACT angle for this viewport starting with
    // the most negative one.
    ckd.fov.act_angles.at(ckd.fov.start_vp.at(i_vp)) =
      -n_act_neg * settings.fov.act_sampling.at(i_vp);
    for (int i_act { 1 }; i_act < ckd.fov.n_act_vp.at(i_vp); ++i_act) {
        ckd.fov.act_angles.at(ckd.fov.start_vp.at(i_vp) + i_act) =
          ckd.fov.act_angles.at(ckd.fov.start_vp.at(i_vp))
          + i_act * settings.fov.act_sampling.at(i_vp);
    }
    // For the cubic spline fit, x-data are the ACT angles and y-data
    // are the peak positions along a given detector column.
    x_data.resize(n_l1a);
    for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
        x_data[i_l1a] = l1a_refs[i_l1a].get().meta->act_angle;
    }
    if (mpi.rank == 0) {
        std::vector<double> y_data(n_l1a);
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            for (int i_spec {}; i_spec < ckd.dim_detector_spec; ++i_spec) {
                for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
                    y_data[i_l1a] =
                      ckd.fov
                        .peak_pos_b[i_pol][ckd.fov.n_l1a_total + i_l1a][i_spec];
                }
                CubicSpline spline { x_data, y_data };
                for (int i_act {}; i_act < ckd.fov.n_act_vp[i_vp]; ++i_act) {
                    const int fov_idx { ckd.fov.start_vp[i_vp] + i_act };
                    // This is where the primary FOV CKD quantity is constructed
                    ckd.fov.spat_indices[i_pol][fov_idx][i_spec] =
                      spline.eval(ckd.fov.act_angles[fov_idx]);
                }
            }
            // Derived quantities
            for (int i_act {}; i_act < ckd.fov.n_act_vp.at(i_vp); ++i_act) {
                const int act_idx { ckd.fov.start_vp[i_vp] + i_act };
                for (int i_spec {}; i_spec < ckd.dim_detector_spec; ++i_spec) {
                    const int spat_idx { static_cast<int>(
                      ckd.fov.spat_indices[i_pol][act_idx][i_spec]) };
                    ckd.fov.pix1[i_pol][act_idx][i_spec] =
                      spat_idx * ckd.dim_detector_spec + i_spec;
                    ckd.fov.pix2[i_pol][act_idx][i_spec] =
                      ckd.fov.pix1[i_pol][act_idx][i_spec]
                      + ckd.dim_detector_spec;
                    ckd.fov.weight[i_pol][act_idx][i_spec] =
                      spat_idx + 1.0
                      - ckd.fov.spat_indices[i_pol][act_idx][i_spec];
                }
            }
        }
    }
    ckd.fov.n_l1a_total += n_l1a;
}

auto FovGen::processFinalize() -> void
{
    // Now that we know the number of ACT angles of all viewports,
    // shrink all relevant arrays to their optimal sizes.
    ckd.fov.n_act =
      std::accumulate(ckd.fov.n_act_vp.cbegin(), ckd.fov.n_act_vp.cend(), 0);
    ckd.fov.dims_spec.resize(ckd.fov.n_act);
    ckd.fov.act_angles.resize(ckd.fov.n_act);
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        ckd.fov.spat_indices.at(i_pol).resize(ckd.fov.n_act);
        ckd.fov.pix1.at(i_pol).resize(ckd.fov.n_act);
        ckd.fov.pix2.at(i_pol).resize(ckd.fov.n_act);
        ckd.fov.weight.at(i_pol).resize(ckd.fov.n_act);
        if (!ckd.fov.peak_pos[i_pol].empty()) {
            ckd.fov.peak_pos[i_pol].resize(ckd.fov.n_l1a_total);
            ckd.fov.peak_pos_b[i_pol].resize(ckd.fov.n_l1a_total);
        }
    }
}

} // namespace spex
