// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "io.h"

#include <filesystem>
#include <fstream>
#include <gadfit/mpi.h>
#include <spdlog/sinks/stdout_color_sinks.h>
#include <spdlog/spdlog.h>
#include <unistd.h>

namespace spex {

auto spex_formatter_flag::format(const spdlog::details::log_msg& log_msg,
                                 const std::tm& /*tm_time*/,
                                 spdlog::memory_buf_t& dest) -> void
{
    std::string text {};
    switch (log_msg.level) {
    case spdlog::level::info:
        break;
    case spdlog::level::warn:
        text = " [warning]";
        break;
    case spdlog::level::err:
        text = " [error]";
        break;
    case spdlog::level::debug:
        text = " [debug]";
        break;
    default:
        text = " [unknown]";
    }
    // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-pointer-arithmetic)
    dest.append(text.data(), text.data() + text.size());
}

auto spex_formatter_flag::clone() const
  -> std::unique_ptr<custom_flag_formatter>
{
    return spdlog::details::make_unique<spex_formatter_flag>();
}

auto initLogging(const gadfit::MPIVars& mpi) -> void
{
    auto formatter { std::make_unique<spdlog::pattern_formatter>() };
    formatter->add_flag<spex_formatter_flag>('*').set_pattern(
      "[%H:%M:%S]%* %v");
    spdlog::set_formatter(std::move(formatter));
    spdlog::stdout_color_mt("plain");
    spdlog::get("plain")->set_pattern("%v");
    setLoggerLevel(mpi, false);
}

auto setLoggerLevel(const gadfit::MPIVars& mpi, const bool use_debug_level)
  -> void
{
    if (mpi.rank == 0) {
        if (use_debug_level) {
            spdlog::set_level(spdlog::level::debug);
        } else {
            spdlog::set_level(spdlog::level::info);
        }
    } else {
        spdlog::set_level(spdlog::level::off);
    }
}

auto loggerOnOff() -> void
{
    static auto saved_logger_level { spdlog::level::off };
    auto logger_level { spdlog::get_level() };
    std::swap(logger_level, saved_logger_level);
    spdlog::set_level(logger_level);
}

auto printHeading(const std::string& heading, const bool incl_empty_line)
  -> void
{
    if (incl_empty_line) {
        spdlog::get("plain")->info("");
    }
    std::string hash_line(heading.size() + 4, '#');
    spdlog::get("plain")->info(hash_line);
    spdlog::get("plain")->info("# " + heading + " #");
    spdlog::get("plain")->info(hash_line);
}

auto shortenFilename(const std::string& filename) -> std::string
{
    constexpr int substring_length { 70 };
    if (static_cast<int>(filename.size()) <= substring_length) {
        return filename;
    }
    return ".."
           + filename.substr(filename.size() - substring_length,
                             filename.size() - 1);
}

auto checkPresenceOfFile(const std::string& filename, const bool required)
  -> void
{
    if (!required && filename.empty()) {
        return;
    }
    try {
        std::ifstream file { filename };
        file.exceptions(std::ifstream::failbit);
    } catch (const std::ifstream::failure& e) {
        const std::string msg { "\nCould not open file: " + filename };
        throw std::ifstream::failure { e.what() + msg, e.code() };
    }
}

auto checkFileWritable(const std::string& filename) -> void
{
    if (!filename.empty()) {
        std::string dir { std::filesystem::path { filename }.parent_path() };
        if (!dir.empty() && static_cast<bool>(access(dir.c_str(), W_OK))) {
            throw std::runtime_error { filename + " is not writable" };
        }
    }
}

auto extractFilesFromList(const std::string& filename,
                          std::vector<std::string>& list,
                          const bool required) -> void
{
    checkPresenceOfFile(filename, required);
    std::ifstream file { filename };
    std::string line {};
    while (std::getline(file, line)) {
        list.push_back(line);
    }
}

} // namespace spex
