// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a.h"

#include "bspline.h"
#include "ckd.h"
#include "cubic_spline.h"
#include "fourier.h"
#include "functions.h"
#include "io.h"
#include "l1a_metadata.h"
#include "l1x.h"
#include "linear_spline.h"
#include "lininv.h"
#include "matrix.h"
#include "spex_math.h"
#include "spex_settings.h"

#include <cassert>
#include <chrono>
#include <numeric>
#include <spdlog/spdlog.h>

namespace spex {

L1A::L1A(const gadfit::MPIVars& mpi, const CKD& ckd, const bool is_background)
  : mpi { mpi }, ckd { ckd }, is_background { is_background }
{
    l1x.resize(static_cast<int>(ProcLevel::n_levels));
}

auto L1A::readImageAttributes() -> void
{
    meta->nc->reOpen();
    const auto image_attributes { meta->nc->getGroup("image_attributes") };
    // In calibration mode, image_offset is 0 and number_of_images is
    // 1 or more. In flight mode, image_offset is 0 or higher and
    // number_of_images is 1. It does not mean the number of images in
    // the full L1A product but the number of frames that make up one
    // flight image.
    const std::vector<size_t> start { static_cast<size_t>(image_offset) };
    const std::vector<size_t> count {
        static_cast<size_t>(meta->number_of_images),
    };
    std::vector<double> image_time_values(meta->number_of_images);
    image_attributes.getVar("image_time")
      .getVar(start, count, image_time_values.data());
    std::vector<uint32_t> seconds_values(meta->number_of_images);
    image_attributes.getVar("image_CCSDS_sec")
      .getVar(start, count, seconds_values.data());
    std::vector<uint16_t> image_CCSDS_subsec_values(meta->number_of_images);
    image_attributes.getVar("image_CCSDS_subsec")
      .getVar(start, count, image_CCSDS_subsec_values.data());
    std::vector<uint8_t> binning_table_values(meta->number_of_images);
    image_attributes.getVar("binning_table")
      .getVar(start, count, binning_table_values.data());
    std::vector<int16_t> digital_offset_values(meta->number_of_images);
    image_attributes.getVar("digital_offset")
      .getVar(start, count, digital_offset_values.data());
    std::vector<uint16_t> nr_coadditions_values(meta->number_of_images);
    image_attributes.getVar("nr_coadditions")
      .getVar(start, count, nr_coadditions_values.data());
    std::vector<double> exposure_time_values(meta->number_of_images);
    image_attributes.getVar("exposure_time")
      .getVar(start, count, exposure_time_values.data());
    meta->nc->close();
    // Temperature is interpolated from engineering data
    std::vector<double> temp_detector_values(meta->number_of_images);
    if (meta->l1x_level == ProcLevel::l1a && meta->HK_tlm_time.size() > 1) {
        const LinearSpline spline { meta->HK_tlm_time, meta->temp_detector };
        for (int i {}; i < meta->number_of_images; ++i) {
            temp_detector_values.at(i) = spline.eval(image_time_values.at(i));
        }
    } else {
        // Take over the already interpolated temperature
        temp_detector_values.front() = meta->temp_detector.at(image_offset);
    }
    // The binning table should be consistent within an image
    binning_table_id = binning_table_values.front();
    // Coadd image attributes
    const auto coadd { [&nr_coadditions_values](const auto& arr) {
        return std::transform_reduce(arr.begin(),
                                     arr.end(),
                                     nr_coadditions_values.begin(),
                                     0.0,
                                     std::plus<>(),
                                     std::multiplies<>());
    } };
    exposure_time = coadd(exposure_time_values);
    digital_offset = static_cast<int>(coadd(digital_offset_values));
    temp_detector = coadd(temp_detector_values);
    nr_coadditions = std::accumulate(
      nr_coadditions_values.begin(), nr_coadditions_values.end(), 0);
    // Some of the coadded quantities should be averaged
    exposure_time /= nr_coadditions;
    temp_detector /= nr_coadditions;
    // Time is defined as the beginning of the first exposure
    image_time = image_time_values.front();
    seconds = (static_cast<long int>(seconds_values.front())
               + static_cast<long int>(seconds_values.back()))
              / 2;
    second_fraction = (static_cast<double>(image_CCSDS_subsec_values.front())
                       + static_cast<double>(image_CCSDS_subsec_values.back()))
                      / 2 / 65535;
}

auto L1A::processNavigationData() -> void
{
    if (meta->l1x_level > ProcLevel::l1a) {
        attitude_quaternion = meta->nav_att_quat[image_offset];
        satellite_position = meta->nav_orb_pos[image_offset];
        satellite_velocity = meta->nav_orb_vel[image_offset];
        return;
    }
    if (image_time < meta->nav_att_time.front()
        || image_time > meta->nav_att_time.back()) {
        throw std::runtime_error {
            "Image time is not within the range of the attitude time axis"
        };
    }
    if (meta->sc_records == 1) {
        // One navigation data time slot, possibly through
        // simulated image with one image with geolocation
        attitude_quaternion = meta->nav_att_quat.front();
    } else {
        // We interpolate the index array, because quaternions
        // are interpolated differently from a regular linear
        // interpolation
        std::vector<double> indexarray(meta->sc_records);
        for (int i {}; i < meta->sc_records; ++i) {
            indexarray[i] = static_cast<double>(i);
        }
        double idx {};
        linear_interpol(meta->sc_records,
                        1,
                        meta->nav_att_time.data(),
                        &image_time,
                        indexarray.data(),
                        &idx);
        size_t intidx { (size_t)idx };
        // Prevent out-of-bounds problems in the case the time
        // is exactly the last attitude time
        if (intidx == meta->sc_records - 1) {
            --intidx;
        }
        double weightright { idx - intidx };
        // Prevent numeric errors in L1X loop
        attitude_quaternion = meta->nav_att_quat[intidx].interpolate(
          meta->nav_att_quat[intidx + 1], weightright);
        // Normalize the quaternion. It comes from independent
        // input, possibly polluted by ASCII roundings before
        // it ended up in the NetCDF.
    }
    attitude_quaternion.normalize();
    attitude_quaternion.intuitive_representation();
    // Orbit
    // Same guard for single time instance. Separate clause,
    // because it is more logical that orb_time and att_time
    // can have different sizes, even though the can't in the
    // current L1A format.
    if (meta->sc_records == 1) {
        // One navigation data time slot, possibly through
        // simulated image with one image with geolocation
        satellite_position = meta->nav_orb_pos.front();
        // Only needed for L1X
        satellite_velocity = meta->nav_orb_vel.front();
        return;
    }
    // Same interpolation issue. Now, it is even worse, we
    // can only interpolate once we have seen the planet.
    // means: geostationary is zero speed.
    std::vector<double> indexarray(meta->sc_records);
    for (int i {}; i < meta->sc_records; ++i) {
        indexarray[i] = (double)i;
    }
    double idx {};
    // Recycles idx
    linear_interpol(meta->sc_records,
                    1,
                    meta->nav_orb_time.data(),
                    &image_time,
                    indexarray.data(),
                    &idx);
    // Recycles intidx
    size_t intidx = (size_t)idx;
    // Prevent out-of-bounds problems in the case the time
    // is exactly the last orbit time
    if (intidx == meta->sc_records - 1) {
        --intidx;
    }
    Vector pos_left = meta->nav_orb_pos[intidx];
    Vector pos_right = meta->nav_orb_pos[intidx + 1];
    Vector vel_left = meta->nav_orb_vel[intidx];
    Vector vel_right = meta->nav_orb_vel[intidx + 1];
    double time_left = meta->nav_orb_time[intidx];
    double time_right = meta->nav_orb_time[intidx + 1];
    // Interpolate satellite position and velocity.  These
    // are already defined in an inertial system, so they
    // can be interpolated right away. The interpolation
    // is done with a cubic function where the velocities
    // are used as derivatives of the
    // position. Afterwards, we do not need the velocity
    // anymore.
    // Minimize numeric errors by going to small number
    // for the interpolation axes. Then, the numeric
    // errors are made only once.
    double time_diff { time_right - time_left };
    double time_center { image_time - time_left };
    // Construct B-spline
    std::vector<double> knots { 0.0, time_diff };
    const size_t nmeas { 4 };
    Bspline b(4, 2, knots.data());
    // Retrieve gain matrix (state as function of input)
    std::vector<double> x { 0.0, 0.0, time_diff, time_diff };
    std::vector<int> ideriv { 0, 1, 0, 1 };
    // 4 is the number of measurements, the lengths of x and ideriv.
    std::vector<double> mat(b.nspline * nmeas);
    // In principle, b.nspline is also 4, so it is exactly
    // determined.  We make a gain matrix, because our
    // measurements are vectors, not scalars, so we have
    // to repeat the same retrieval three times.  And
    // linear_invert does not know what a vector is.
    b.jaccalc(nmeas, x.data(), ideriv.data(), mat.data());
    std::vector<double> gain(b.nspline * nmeas);
    linear_invert_gain(
      b.nspline, nmeas, mat.data(), OPT_NONE, NULL, NULL, gain.data());
    // Evaluate the position at target
    // Evaluation from b.nspline (4) spline parameters to
    // the position at the target time.
    std::vector<double> mat_eval(b.nspline);
    b.jaccalc(1, &time_center, mat_eval.data());
    // Now, get the matrix from measurement to target,
    // that a matrix multiplication of the gain with this
    // evaluation Jacobian. Because we have only one
    // evalution, the latter is not a matrix. We have to
    // fold the slow dimension of the gain.
    std::vector<double> trans(nmeas);
    Matrix::matmul_rl_fold_slow(
      b.nspline, nmeas, gain.data(), mat_eval.data(), trans.data());
    // Execute evaluation
    satellite_position = trans.front() * pos_left + trans[1] * vel_left
                         + trans[2] * pos_right + trans[3] * vel_right;
    // This is the interpolated satellite position in the
    // J2000 frame
    // For velocity (only needed for L1X)
    int one { 1 }; // For first derivative
    // We recycle mat_eval and trans
    b.jaccalc(1, &time_center, &one, mat_eval.data());
    Matrix::matmul_rl_fold_slow(
      b.nspline, nmeas, gain.data(), mat_eval.data(), trans.data());
    satellite_velocity = trans.front() * pos_left + trans[1] * vel_left
                         + trans[2] * pos_right + trans[3] * vel_right;
}

// Bin raw data according to the ratio [main][detector_spec] /
// [main][detector_spec_reduced] in the spectral dimension and
// similarly in the spatial dimension. Here the data are almost raw -
// only the digital offset is applied after reading in images.
static auto binImage(const CKD& ckd,
                     const std::span<uint16_t>& image,
                     std::vector<double>& image_binned) -> void
{
    const int bin_spec { ckd.dim_detector_spec_full / ckd.dim_detector_spec };
    const int bin_spat { ckd.dim_detector_spat_full / ckd.dim_detector_spat };
    std::ranges::fill(image_binned, 0.0);
    for (int i {}; i < ckd.dim_detector_spat; ++i) {
        for (int j {}; j < ckd.dim_detector_spec; ++j) {
            for (int k {}; k < bin_spat; ++k) {
                for (int l {}; l < bin_spec; ++l) {
                    image_binned[i * ckd.dim_detector_spec + j] +=
                      static_cast<double>(image[(i * bin_spat + k) * bin_spec
                                                  * ckd.dim_detector_spec
                                                + j * bin_spec + l]);
                }
            }
        }
    }
}

auto L1A::readImage(const int proc_for_shared) -> void
{
    spdlog::debug("Reading image from {}",
                  shortenFilename(meta->filename.c_str()));
    meta->nc->reOpen();
    // Allocate the main work variables in either shared or serial
    // mode. If shared then variables such as the image are created as
    // shared arrays using the MPI RMA model. In calibration mode the
    // arrays are typically allocated as shared. In L1B they are
    // serial because parallelization is over the total number of
    // images.
    const bool shared { proc_for_shared != math::idx_fill_value };
    image = { shared, mpi, binning_table->nBins() };
    noise = { shared, mpi, binning_table->nBins() };
    pixel_mask = { shared, mpi, binning_table->nBins() };
    image_buf = { shared, mpi, binning_table->nBins() };
    for (int i {}; i < binning_table->nBins(); ++i) {
        pixel_mask[i] = ckd.pixel_mask[i];
    }
    // Whether images should be immediately binned after reading. This
    // is done only if the user defined one of both of the reduced
    // dimensions in the [main] section.
    const bool bin_images {
        ckd.dim_detector_spec_full > 0
          && ckd.dim_detector_spec != ckd.dim_detector_spec_full
        || ckd.dim_detector_spat_full > 0
             && ckd.dim_detector_spat != ckd.dim_detector_spat_full
    };
    // This function only reads from the science_data group
    const netCDF::NcGroup science_data { meta->nc->getGroup("science_data") };
    if (meta->l1x_level == ProcLevel::l1a) {
        if (shared && proc_for_shared != mpi.rank) {
            meta->nc->close();
            return;
        }
        const std::vector<size_t> start { static_cast<size_t>(image_offset),
                                          0 };
        const std::vector<size_t> count {
            static_cast<size_t>(meta->number_of_images),
            static_cast<size_t>(meta->npix),
        };
        std::vector<uint16_t> frames_raw(meta->number_of_images * meta->npix);
        science_data.getVar("detector_images")
          .getVar(start, count, frames_raw.data());
        std::vector<std::span<uint16_t>> frames(meta->number_of_images);
        for (int i {}; i < meta->number_of_images; ++i) {
            frames[i] = { frames_raw.begin() + i * meta->npix,
                          frames_raw.begin() + (i + 1) * meta->npix };
        }
        // Start with minus the total digital offset which is already
        // summed over images but not yet over binned pixels. Note
        // that the digital offset can vary between different images
        // (usually not).
        if (!meta->keep_frames) {
            for (int i {}; i < binning_table->nBins(); ++i) {
                image[i] = -static_cast<double>(digital_offset
                                                * binning_table->binSize(i));
            }
            for (int i_img {}; i_img < meta->number_of_images; ++i_img) {
                if (bin_images) {
                    std::vector<double> image_binned(ckd.npix);
                    binImage(ckd, frames[i_img], image_binned);
                    for (int i {}; i < ckd.npix; ++i) {
                        image[i] += image_binned[i];
                    }
                } else {
                    for (int i {}; i < binning_table->nBins(); ++i) {
                        image[i] += frames[i_img][i];
                    }
                }
            }
            for (int i {}; i < binning_table->nBins(); ++i) {
                image[i] /= nr_coadditions * binning_table->binSize(i);
            }
        } else {
            this->frames.resize(meta->number_of_images,
                                std::vector<double>(binning_table->nBins()));
            // Digital offset per single frame
            const double offset_single { static_cast<double>(
              digital_offset / meta->number_of_images) };
            for (int i_img {}; i_img < meta->number_of_images; ++i_img) {
                if (bin_images) {
                    std::vector<double> frame_binned(ckd.npix);
                    binImage(ckd, frames[i_img], frame_binned);
                    for (int i {}; i < binning_table->nBins(); ++i) {
                        this->frames[i_img][i] =
                          frame_binned[i] - offset_single;
                    }
                } else {
                    for (int i {}; i < ckd.npix; ++i) {
                        this->frames[i_img][i] =
                          static_cast<double>(frames[i_img][i]) - offset_single;
                    }
                }
            }
        }
    } else if (meta->l1x_level <= ProcLevel::stray) {
        const std::vector<size_t> start { static_cast<size_t>(image_offset),
                                          0 };
        const std::vector<size_t> count { 1, image.size() };
        science_data.getVar("detector_images")
          .getVar(start, count, image.data());
        science_data.getVar("detector_noise")
          .getVar(start, count, noise.data());
        std::vector<uint8_t> mask_u8(binning_table->nBins());
        science_data.getVar("pixel_mask").getVar(start, count, mask_u8.data());
        std::transform(mask_u8.cbegin(),
                       mask_u8.cend(),
                       pixel_mask.begin(),
                       [](const auto x) { return x == 1; });
        std::ranges::fill(noise, 1.0);
        std::ranges::fill(pixel_mask, false);
    } else if (meta->l1x_level < ProcLevel::l1b) {
        l1x_spectra.resize(ckd.fov.n_act);
        std::vector<int> spectra_sizes(ckd.fov.n_act);
        science_data.getVar("spectra_sizes").getVar(spectra_sizes.data());
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                const std::vector<size_t> start { l1x_image_offset,
                                                  static_cast<size_t>(i_fov),
                                                  static_cast<size_t>(i_pol),
                                                  0 };
                const std::vector<size_t> count {
                    1, 1, 1, static_cast<size_t>(spectra_sizes[i_fov])
                };
                l1x_spectra[i_fov].signal[i_pol].resize(spectra_sizes[i_fov]);
                science_data.getVar("spectra").getVar(
                  start, count, l1x_spectra[i_fov].signal[i_pol].data());
                l1x_spectra[i_fov].noise[i_pol].resize(spectra_sizes[i_fov]);
                science_data.getVar("spectra_noise")
                  .getVar(start, count, l1x_spectra[i_fov].noise[i_pol].data());
                l1x_spectra[i_fov].mask[i_pol].resize(spectra_sizes[i_fov]);
                std::vector<uint8_t> mask_u8(spectra_sizes[i_fov]);
                science_data.getVar("spectra_mask")
                  .getVar(start, count, mask_u8.data());
                std::transform(mask_u8.cbegin(),
                               mask_u8.cend(),
                               l1x_spectra[i_fov].mask[i_pol].begin(),
                               [](const auto x) { return x == 1; });
            }
        }
    } else {
        throw std::runtime_error { "attempting to read an L1A file with an "
                                   "unknown L1X level" };
    }
    meta->nc->close();
}

auto L1A::rowDependentOffset(const SPEXSettings& settings,
                             const ProcLevel proc_level) -> void
{
    const int& dark_cols_left { settings.dark_cols_left.at(
      static_cast<int>(proc_level)) };
    const int& dark_cols_right { settings.dark_cols_right.at(
      static_cast<int>(proc_level)) };
    if (dark_cols_left + dark_cols_right == 0) {
        return;
    }
    for (int i {}; i < static_cast<int>(image_buf.size()); ++i) {
        image_buf[i] = 0.0;
    }
    std::vector<double> edge_pixels(dark_cols_left + dark_cols_right);
    for (int i_row {}; i_row < ckd.dim_detector_spat; ++i_row) {
        if (binning_table->skipRow(i_row)) {
            continue;
        }
        const int row_idx { i_row * ckd.dim_detector_spec };
        int edge_pixel_idx {};
        for (int i_col {}; i_col < dark_cols_left; ++i_col) {
            const int idx { binning_table->binIndex(row_idx + i_col) };
            if (!pixel_mask[idx]) {
                edge_pixels[edge_pixel_idx] = image[idx];
                ++edge_pixel_idx;
            }
        }
        for (int i_col {}; i_col < dark_cols_right; ++i_col) {
            const int idx { binning_table->binIndex(
              row_idx + ckd.dim_detector_spec - 1 - i_col) };
            if (!pixel_mask[idx]) {
                edge_pixels[edge_pixel_idx] = image[idx];
                ++edge_pixel_idx;
            }
        }
        const double med { median(edge_pixels) };
        for (int i_col {}; i_col < ckd.dim_detector_spec; ++i_col) {
            const int idx { binning_table->binIndex(row_idx + i_col) };
            image_buf[idx] += (image[idx] - med) / binning_table->binSize(idx);
        }
    }
    for (int i {}; i < static_cast<int>(image.size()); ++i) {
        image[i] = image_buf[i];
    }
}

auto L1A::calibrateStrayLight() -> void
{
    // CKD ckd { mpi, "data/ckd_stray.nc" };
    // Sub image size in pixels
    const int sub_npix { ckd.stray.n_spatial * ckd.stray.n_spectral };
    std::vector<double> image_conv(sub_npix);
    binning_table_stray->unbin(image, image_conv);
    std::vector<double> conv_result(sub_npix);
    std::vector<double> image_ideal { image_conv };
    for (int i_vc {}; i_vc < 3; ++i_vc) {
        std::vector<double> image_tmp(sub_npix, 0.0);
        for (int i_kernel {}; i_kernel < ckd.stray.n_kernels; ++i_kernel) {
            std::vector<double> image_weighted { image_ideal };
            for (int i {}; i < sub_npix; ++i) {
                image_weighted[i] *= ckd.stray.weights[i_kernel][i];
            }
            // Number of rows and column in this subimage
            const int image_n_rows {
                ckd.stray.edges[i_kernel * box::n + box::t]
                - ckd.stray.edges[i_kernel * box::n + box::b]
            };
            const int image_n_cols {
                ckd.stray.edges[i_kernel * box::n + box::r]
                - ckd.stray.edges[i_kernel * box::n + box::l]
            };
            std::vector<double> image_shifted(image_n_rows * image_n_cols);
            for (int i {}; i < image_n_rows; ++i) {
                for (int j {}; j < image_n_cols; ++j) {
                    image_shifted[i * image_n_cols + j] = image_weighted
                      [(i + ckd.stray.edges[i_kernel * box::n + box::b])
                         * ckd.stray.n_spectral
                       + j + ckd.stray.edges[i_kernel * box::n + box::l]];
                }
            }
            std::vector<double> conv_shifted {};
            convolve(image_n_rows,
                     image_n_cols,
                     image_shifted,
                     ckd.stray.kernel_rows[i_kernel],
                     ckd.stray.kernel_cols[i_kernel],
                     ckd.stray.kernels_real[i_kernel],
                     ckd.stray.kernels_fft[i_kernel],
                     conv_shifted);
            std::ranges::fill(conv_result, 0.0);
            for (int i {}; i < image_n_rows; ++i) {
                for (int j {}; j < image_n_cols; ++j) {
                    conv_result[(i
                                 + ckd.stray.edges[i_kernel * box::n + box::b])
                                  * ckd.stray.n_spectral
                                + j
                                + ckd.stray.edges[i_kernel * box::n + box::l]] =
                      conv_shifted[i * image_n_cols + j];
                }
            }
            for (int i {}; i < sub_npix; ++i) {
                image_tmp[i] += conv_result[i];
            }
        }
        for (int i {}; i < sub_npix; ++i) {
            image_ideal[i] =
              (image_conv[i] - image_tmp[i]) / (1 - ckd.stray.eta[i]);
        }
    }
    binning_table_stray->binImage(image_ideal, image);
}

auto L1A::calibrateImage(const SPEXSettings& settings,
                         const ProcLevel proc_level) -> void
{
    // A calibration step is run if two conditions are satisfied:
    // * The L1X level is less than the corresponding process. For
    //   instance, if the L1X level is ProcLevel::prnu then the PRNU
    //   correction is not applied because the data is already
    //   corrected for PRNU.
    // * The CKD level is equal or higher than the given process. If
    //   the CKD level is not high enough then the calibration step
    //   cannot be applied because that section is missing from the
    //   CKD.
    // * If the process is not skipped

    // If requested, write the data without any calibration. Still,
    // this differs from the original L1A contents in that the data
    // has been coadded, binning table counts have been applied, and
    // the data is saved in a floating point format.
    if (meta->l1x_level < ProcLevel::raw
        && l1x[static_cast<int>(ProcLevel::raw)]) {
        l1x[static_cast<int>(ProcLevel::raw)]->writeImage(*this);
    }

    // Dark calibration: dark offset
    if (meta->l1x_level < ProcLevel::dark && ckd.level >= ProcLevel::dark
        && !settings.skip.at(static_cast<int>(ProcLevel::dark))) {
        timers[static_cast<int>(ProcLevel::dark)]->start();
        // Evaluate the fit. The B-splines are fixed to a polynomial
        // where the first term is the value at nominal temperature.
        std::vector<double> knots { ckd.dark_nominal_temperature,
                                    ckd.dark_nominal_temperature + 1.0 };
        Bspline b { ckd.dim_dark_order, 2, knots.data() };
        std::vector<double> terms(ckd.dim_dark_order);
        b.jaccalc(1, &temp_detector, terms.data());
        for (int i {}; i < binning_table->nBins(); ++i) {
            if (!pixel_mask[i]) {
                for (int i_order {}; i_order < ckd.dim_dark_order; ++i_order) {
                    image[i] -=
                      ckd.dark_offset[i * ckd.dim_dark_order + i_order]
                      * terms[i_order];
                }
            }
        }
        if (meta->keep_frames) {
            for (int i_img {}; i_img < meta->number_of_images; ++i_img) {
                for (int i {}; i < binning_table->nBins(); ++i) {
                    for (int i_order {}; i_order < ckd.dim_dark_order;
                         ++i_order) {
                        frames[i_img][i] -=
                          ckd.dark_offset[i * ckd.dim_dark_order + i_order]
                          * terms[i_order];
                    }
                }
            }
        }
        timers[static_cast<int>(ProcLevel::dark)]->stop();
    }

    // Noise estimation
    if (meta->l1x_level < ProcLevel::noise && ckd.level >= ProcLevel::noise) {
        if (!settings.skip.at(static_cast<int>(ProcLevel::noise))) {
            timers[static_cast<int>(ProcLevel::noise)]->start();
            for (int i {}; i < binning_table->nBins(); ++i) {
                if (pixel_mask[i]) {
                    noise[i] = NC_FILL_DOUBLE;
                } else {
                    const double var { ckd.noise.g[i] * image[i]
                                       + ckd.noise.n[i] };
                    if (var <= 0.0) {
                        image[i] = NC_FILL_DOUBLE;
                        noise[i] = NC_FILL_DOUBLE;
                        pixel_mask[i] = true;
                    } else {
                        noise[i] = std::sqrt(
                          var / (nr_coadditions * binning_table->binSize(i)));
                    }
                }
            }
            timers[static_cast<int>(ProcLevel::noise)]->stop();
        } else {
            for (int i {}; i < binning_table->nBins(); ++i) {
                if (pixel_mask[i]) {
                    noise[i] = NC_FILL_DOUBLE;
                } else {
                    noise[i] = 1.0;
                }
            }
        }
    }

    // Dark calibration: dark current
    if (meta->l1x_level < ProcLevel::dark && ckd.level >= ProcLevel::dark
        && !settings.skip.at(static_cast<int>(ProcLevel::dark))) {
        timers[static_cast<int>(ProcLevel::dark)]->start();
        // Evaluate the fit. The B-splines are fixed to a polynomial
        // where the first term is the value at nominal temperature.
        std::vector<double> knots { ckd.dark_nominal_temperature,
                                    ckd.dark_nominal_temperature + 1.0 };
        Bspline b { ckd.dim_dark_order, 2, knots.data() };
        std::vector<double> terms(ckd.dim_dark_order);
        b.jaccalc(1, &temp_detector, terms.data());
        for (int i {}; i < binning_table->nBins(); ++i) {
            if (!pixel_mask[i]) {
                for (int i_order {}; i_order < ckd.dim_dark_order; ++i_order) {
                    image[i] -=
                      ckd.dark_current[i * ckd.dim_dark_order + i_order]
                      * terms[i_order] * exposure_time;
                }
            }
        }
        timers[static_cast<int>(ProcLevel::dark)]->stop();
        if (l1x[static_cast<int>(ProcLevel::dark)]) {
            l1x[static_cast<int>(ProcLevel::dark)]->writeImage(*this);
        }
    }

    // If row_dep_bgr then apply row-dependent offset to both the
    // calibration image and background image.
    if (settings.main.row_dep_bgr) {
        rowDependentOffset(settings, proc_level);
    }
    // Subtract background image
    if (bgr_image) {
        for (int ipix {}; ipix < binning_table->nBins(); ++ipix) {
            if (!pixel_mask[ipix]) {
                image[ipix] -= (*bgr_image)[ipix];
            }
        }
    }
    // If !row_dep_bgr then apply row-dependent offset to the
    // corrected image but not the calibration image and
    // background image separately.
    if (!settings.main.row_dep_bgr && !is_background) {
        rowDependentOffset(settings, proc_level);
    }

    // Nonlinearity correction
    if (meta->l1x_level < ProcLevel::nonlin && ckd.level >= ProcLevel::nonlin
        && !settings.skip.at(static_cast<int>(ProcLevel::nonlin))) {
        timers[static_cast<int>(ProcLevel::nonlin)]->start();
        // Choose the relevant exposure time
        double mindiff { abs(ckd.nonlin_exptimes.front() - exposure_time) };
        int iexptime_best {};
        for (int iexptime { 1 }; iexptime < ckd.dim_nonlin_exptime;
             ++iexptime) {
            if (abs(ckd.nonlin_exptimes[iexptime] - exposure_time) < mindiff) {
                mindiff = abs(ckd.nonlin_exptimes[iexptime] - exposure_time);
                iexptime_best = iexptime;
            }
        }
        // Use the B-spline to evaluate the fit on the desired input
        // (the signal)
        Bspline b(
          ckd.nonlin_order, ckd.dim_nonlin_knot, ckd.nonlin_knots.data());
        std::vector<int> ideriv { 0, 1 };
        const double* nonlin_fit_cur {
            &ckd.nonlin_fit[iexptime_best * ckd.dim_nonlin_spline]
        };
        const double* nonlin_signal_scale_cur {
            &ckd.nonlin_signal_scale[iexptime_best]
        };
        for (int i_pix {}; i_pix < binning_table->nBins(); ++i_pix) {
            if (!pixel_mask[i_pix]) {
                double signal_cur = image[i_pix];
                bool converged = false;
                for (size_t iter {}; iter < 100; ++iter) {
                    std::vector<double> model(2);
                    std::vector<double> abscissa(2);
                    abscissa[0] = signal_cur / (*nonlin_signal_scale_cur);
                    abscissa[1] = abscissa[0];
                    b.jacapply(2,
                               abscissa.data(),
                               ideriv.data(),
                               nonlin_fit_cur,
                               model.data());
                    double gain = *nonlin_signal_scale_cur / model[1];
                    double addition = (image[i_pix] - model[0]) * gain;
                    signal_cur += addition;
                    if (abs(addition) < 1e-3) {
                        image[i_pix] = signal_cur;
                        noise[i_pix] *= abs(gain);
                        converged = true;
                        break;
                    }
                }
                // Flag pixel if non-linearity inversion did not converge
                if (!converged) {
                    image[i_pix] = NC_FILL_DOUBLE;
                    noise[i_pix] = NC_FILL_DOUBLE;
                    pixel_mask[i_pix] = true;
                }
            }
            nonlin_fit_cur += ckd.dim_nonlin_spline * ckd.dim_nonlin_exptime;
            nonlin_signal_scale_cur += ckd.dim_nonlin_exptime;
        }
        timers[static_cast<int>(ProcLevel::nonlin)]->stop();
        if (l1x[static_cast<int>(ProcLevel::nonlin)]) {
            l1x[static_cast<int>(ProcLevel::nonlin)]->writeImage(*this);
        }
    }

    // Pixel response nonuniformity (PRNU)
    if (meta->l1x_level < ProcLevel::prnu && ckd.level >= ProcLevel::prnu
        && !settings.skip.at(static_cast<int>(ProcLevel::prnu))) {
        timers[static_cast<int>(ProcLevel::prnu)]->start();
        for (int i {}; i < binning_table->nBins(); ++i) {
            if (!pixel_mask[i]) {
                image[i] /= ckd.prnu.prnu[i];
                noise[i] /= ckd.prnu.prnu[i];
            }
        }
        timers[static_cast<int>(ProcLevel::prnu)]->stop();
        if (l1x[static_cast<int>(ProcLevel::prnu)]) {
            l1x[static_cast<int>(ProcLevel::prnu)]->writeImage(*this);
        }
    }

    // Stray light
    if (ckd.level >= ProcLevel::stray && meta->l1x_level < ProcLevel::stray
        && proc_level == ProcLevel::l1b
        && !settings.skip.at(static_cast<int>(ProcLevel::stray))) {
        timers[static_cast<int>(ProcLevel::stray)]->start();
        calibrateStrayLight();
        timers[static_cast<int>(ProcLevel::stray)]->stop();
        if (l1x[static_cast<int>(ProcLevel::stray)]) {
            l1x[static_cast<int>(ProcLevel::stray)]->writeImage(*this);
        }
    }
}

// Spectrally bin the signal, noise, and mask of both polarization
// channels of FOV i_fov. If the spectral binning table does not exist
// this function has no affect. That check is done inside binSpectral.
static auto binSpectra(const BinningTable& binning_table,
                       const int i_fov,
                       Spectra& spectra) -> void
{
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        binning_table.binSpectral(i_fov, spectra.signal[i_pol]);
        binning_table.binSpectral(i_fov, spectra.noise[i_pol]);
        binning_table.binSpectral(i_fov, spectra.mask[i_pol]);
    }
}

auto L1A::extractSpectra(const int i_fov, Spectra& spectra) const -> void
{
    if (ckd.level >= ProcLevel::fov) {
        if (meta->l1x_level < ProcLevel::fov) {
            timers[static_cast<int>(ProcLevel::fov)]->start();
            spectra.extractFOV(ckd, image, noise, pixel_mask, i_fov);
            binSpectra(*binning_table, i_fov, spectra);
            timers[static_cast<int>(ProcLevel::fov)]->stop();
            if (l1x[static_cast<int>(ProcLevel::fov)]) {
                l1x[static_cast<int>(ProcLevel::fov)]->writeSpectrum(
                  spectra, l1x_image_offset, i_fov);
            }
        } else if (meta->l1x_level == ProcLevel::fov) {
            spectra = l1x_spectra[i_fov];
        }
    }
    if (ckd.level >= ProcLevel::rad) {
        if (meta->l1x_level < ProcLevel::rad) {
            timers[static_cast<int>(ProcLevel::rad)]->start();
            if (!ckd.rad_spectra.front().empty()) {
                spectra.calibrate(ckd, exposure_time, i_fov);
            }
            spectra.interpolateToCommon(ckd, i_fov);
            timers[static_cast<int>(ProcLevel::rad)]->stop();
            if (l1x[static_cast<int>(ProcLevel::rad)]) {
                l1x[static_cast<int>(ProcLevel::rad)]->writeSpectrum(
                  spectra, l1x_image_offset, i_fov);
            }
        } else if (meta->l1x_level == ProcLevel::rad) {
            spectra = l1x_spectra[i_fov];
        }
    }
}

auto L1A::destroy() -> void
{
    spdlog::debug("Deallocating image {}",
                  shortenFilename(meta->filename.c_str()));
    image = {};
    noise = {};
    pixel_mask = {};
    image_buf = {};
    frames = {};
    l1x_spectra = {};
}

} // namespace spex
