// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for reading and storing the contents of one L1A product and
// for performing operations on it such as calibrating.

#pragma once

#include "binning_table.h"
#include "constants.h"
#include "egse.h"
#include "nc_object.h"
#include "shared_array.h"
#include "spectra.h"
#include "vector.h"

#include <array>
#include <cstddef>
#include <gadfit/timer.h>
#include <vector>

namespace spex {

class L1AMetadata;
class L1X;

class L1A
{
private:
    const gadfit::MPIVars& mpi;
    const CKD& ckd;
    // Whether this is a background image
    bool is_background {};
    // Apply row-dependent offset to an image
    auto rowDependentOffset(const SPEXSettings& settings,
                            const ProcLevel proc_level) -> void;
    auto calibrateStrayLight() -> void;

public:
    L1A(const gadfit::MPIVars& mpi, const CKD& ckd, const bool is_background);
    ~L1A() = default;

    // Main data

    // This variable starts with a raw detector image and is updated
    // at each calibration step. After the spectra have been extracted
    // we don't need the detector image anymore (steps FOV and
    // higher).
    SharedArray<double> image {};
    // Same thing but for the detector image noise
    SharedArray<double> noise {};
    // Background image associated with this L1A product (default is
    // no background image)
    SharedArray<double>* bgr_image {};
    // Individual frames that make up an image. Typically these are
    // discarded right after reading in and coadding. However, for
    // some operations such as noise CKD generation these need to be
    // saved for later use.
    std::vector<std::vector<double>> frames {};
    // Some operations require the original image to be saved and
    // overwritten later.
    SharedArray<double> image_buf {};
    // Pixel mask is initially copied from the CKD but may be updated
    // during one or more calibrations steps. Only used for detector
    // steps, hence called pixel mask. From FOV onward we use the
    // spectral mask (part of Spectra).
    SharedArray<bool> pixel_mask {};

    // Meta data

    // Some of the meta data is in an L1AMetadata instance and the
    // rest is held by the L1A class directly.
    std::shared_ptr<L1AMetadata> meta {};
    // Image number where the current observation starts. In
    // calibration mode it is typically 0 unless some images need to
    // be skipped. In flight mode, where a single L1A file contains
    // all the images which are not coadded, image_offset is the index
    // of the current image to be processed.
    int image_offset {};
    // Seconds since the start of the day
    double image_time {};
    // Absolute time for the solar model
    long int seconds {};
    // Additional fraction of a second to be added
    double second_fraction {};
    // Binning table
    uint8_t binning_table_id {};
    std::shared_ptr<BinningTable> binning_table {};
    // Stray light might require special treatment for unbinning so
    // keep a separate binning_table for that.
    std::shared_ptr<BinningTable> binning_table_stray {};
    // Coadded digital offset
    int digital_offset {};
    // Number of coadditions. In calibration mode, this is summed over
    // images in the L1A product.
    int nr_coadditions {};
    // Coadded and averaged exposure time
    double exposure_time {};
    // Detector temperature taken from engineering_data and
    // interpolated to image_time
    double temp_detector {};
    // Equals either pol_minus or pol_plus
    int polarization_state {};

    // Naviagion data

    // Rotation quaternion for attitude
    Quaternion attitude_quaternion {};
    // Interpolated satellite position
    Vector satellite_position {};
    // Both are in J2000 system
    Vector satellite_velocity {}; // Only for L1X

    // L1X related

    // L1X instances for each step (null if not requested)
    std::vector<std::shared_ptr<L1X>> l1x {};
    // Which image in the L1X is owned by this L1A instance
    size_t l1x_image_offset {};
    // When spectra are read they are stashed into this array. This is
    // needed because the L1A usually has no spectra as members. When
    // a spectrum is requested it is retrieved from this stash rather
    // than acquired from the image which does not exist for data
    // ProcLevel::fov or higher.
    std::vector<Spectra> l1x_spectra {};

    // Timers for most steps in the process chain. All L1A instances
    // within a process (e.g. fov, rad, l1b) share the same timers so
    // that cumulative times can be reported in the end. The timers
    // are created and associated with L1A instances at the beginning
    // of each process after the L1A manager has been initialized.
    std::array<std::shared_ptr<gadfit::Timer>,
               static_cast<int>(ProcLevel::n_levels)>
      timers {};

    // Read and process data from the image_attributes NetCDF group
    auto readImageAttributes() -> void;
    // Process data that was already read from the navigation_data NetCDF group
    auto processNavigationData() -> void;
    // A high-data-load function to read the image. This includes
    // coadding of the images. If proc_for_shared is given then the
    // main variables such as the image are allocated as shared
    // arrays. The reading is done by process proc_for_shared after
    // which the data are accessible to all processes. This function
    // recognizes both L1A and L1X data levels.
    auto readImage(const int proc_for_shared = math::idx_fill_value) -> void;
    // Calibrate an image up to and including stray light
    auto calibrateImage(const SPEXSettings& settings,
                        const ProcLevel proc_level) -> void;
    // Extract a pair of radiometrically calibrated S+/S- spectra from
    // a detector image
    auto extractSpectra(const int i_fov, Spectra& spectra) const -> void;
    // Deallocate the largest arrays
    auto destroy() -> void;
};

} // namespace spex
