// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_batch.h"

#include "l1a.h"
#include "settings.h"
#include "spex_math.h"

#include <gadfit/mpi.h>
#include <spdlog/spdlog.h>

namespace spex {

auto L1ABatch::genWorkRanges(const gadfit::MPIVars& mpi) -> void
{
    spex::genWorkRanges(
      l1a_products.size(), mpi.n_procs, mpi_ranges, work_proc_map);
    my_mpi_range = { mpi_ranges.at(mpi.rank), mpi_ranges.at(mpi.rank + 1) };
}

auto L1ABatch::genWorkRanges() -> void
{
    my_mpi_range = { 0, static_cast<int>(l1a_products.size()) };
    work_proc_map = { math::idx_fill_value };
}

auto L1ABatch::calibrate(const SPEXSettings& settings,
                         const ProcLevel proc_level,
                         const bool print_progress) -> void
{
    // Reading of the images runs across all L1A products because it
    // involves creating shared arrays that need the participation of
    // all processes.
    const bool print { print_progress && l1a_products.size() > 1 };
    if (print) {
        spdlog::info("Reading images");
    }
    for (int i {}; i < static_cast<int>(l1a_products.size()); ++i) {
        l1a_products.at(i).get().readImage(work_proc_map.at(i));
    }
    // Determine the range of L1A products from the current batch to
    // be processed by this MPI process.
    for (int i { my_mpi_range.front() }; i < my_mpi_range.back(); ++i) {
        if (print) {
            spdlog::info("Calibrating {:5.1f}%",
                         100.0 * i / my_mpi_range.back());
        }
        l1a_products.at(i).get().calibrateImage(settings, proc_level);
    }
    if (print) {
        spdlog::info("Calibrating 100.0%");
    }
}

auto L1ABatch::destroy() -> void
{
    for (auto& l1a : l1a_products) {
        l1a.get().destroy();
    }
}

} // namespace spex
