// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// In general L1A products (images) are processed in batches. A batch
// may contain a single image (L1B), all images, or a subset of
// images. For instance, during CKD generation, depending on the
// process, it makes sense to partition L1A images into batches by
// viewport if the viewports can be processed independently. This
// class holds the L1A products and functions for reading and
// calibrating them.

#pragma once

#include "constants.h"

#include <array>
#include <vector>

namespace gadfit {

class MPIVars;

} // namespace gadfit

namespace spex {

class L1A;
class SPEXSettings;

class L1ABatch
{
private:
    // Range of images that each MPI process should process. The
    // starting index for a given process is mpi_ranges[mpi.rank] and
    // the end index is mpi_ranges[mpi.rank+1]-1. Example: for 7
    // images and 3 MPI processes we would have mpi_ranges { 0, 3, 5,
    // 7 } with the workloads 3,2,2. Each MPI process holds the full
    // vector, i.e. information about what the other processes are
    // doing. This is needed later in the synchronization step.
    std::vector<int> mpi_ranges {};
    // Contains either the segment (mpi_ranges[my_rank],
    // mpi_ranges[my_rank+1]) of [0, N] where N is the total amount of
    // work depending on whether the batch needs to be processed in
    // parallel or not.
    std::array<int, 2> my_mpi_range {};
    std::vector<int> work_proc_map { 0 };

public:
    // All L1A products associated with this batch
    std::vector<std::reference_wrapper<L1A>> l1a_products {};
    // Generate mpi_ranges and work_proc_map such that work is
    // distributed evenly among MPI processes.
    auto genWorkRanges(const gadfit::MPIVars& mpi) -> void;
    // In this version, the batch is meant to be processed in
    // serial. my_mpi_range is set accordingly and work_proc_map is
    // filled with a fill value.
    auto genWorkRanges() -> void;
    // Read and calibrate all L1A images in this batch. If running in
    // parallel, explicit synchronization might be necessary after
    // this call.
    auto calibrate(const SPEXSettings& settings,
                   const ProcLevel proc_level,
                   const bool print_progress = false) -> void;
    // Deallocate the largest arrays in L1A products
    auto destroy() -> void;
};

} // namespace spex
