// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_manager.h"

#include "l1a_metadata.h"
#include "l1x.h"
#include "spex_settings.h"

#include <spdlog/spdlog.h>

namespace spex {

// Must use an initialization function because it calls some pure
// virtual functions.
auto L1AManager::init(const SPEXSettings& settings,
                      const ProcLevel proc_level,
                      const CKD& ckd) -> void
{
    const auto& l1a_files {
        is_background ? settings.l1a_bgr_files.at(static_cast<int>(proc_level))
                      : settings.l1a_files.at(static_cast<int>(proc_level))
    };
    // First read all the L1A meta data (low data load). That
    // determines how many L1A products should be allocated and how
    // they are treated.
    std::vector<std::shared_ptr<L1AMetadata>> l1a_meta(l1a_files.size());
    if (!is_background) {
        spdlog::info("Reading L1A metadata ..");
    }
    for (int i_file {}; i_file < static_cast<int>(l1a_files.size()); ++i_file) {
        l1a_meta.at(i_file) = std::make_shared<L1AMetadata>(
          l1a_files.at(i_file), settings, proc_level);
    }
    if (!is_background) {
        spdlog::info("Reading L1A metadata done");
    }
    l1a_products = std::vector<L1A>(
      getNumberOfImages(
        settings, l1a_meta.front(), static_cast<int>(l1a_files.size())),
      { mpi, ckd, is_background });
    organize(settings, proc_level, l1a_meta);
    for (auto& l1a : l1a_products) {
        l1a.readImageAttributes();
        if (l1a.meta->sc_records > 0) {
            l1a.processNavigationData();
        }
    }
}

auto L1AManager::initL1X(const SPEXSettings& settings,
                         const ProcLevel proc_level,
                         const CKD& ckd) -> void
{
    std::vector<std::shared_ptr<L1X>> l1x(static_cast<int>(ProcLevel::n_levels),
                                          nullptr);
    for (int i_l1x {}; i_l1x < static_cast<int>(ProcLevel::n_levels); ++i_l1x) {
        const ProcLevel l1x_level { static_cast<ProcLevel>(i_l1x) };
        std::string l1x_filename {};
        if (l1x_level == ProcLevel::raw) {
            l1x_filename = settings.l1x_raw.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::dark) {
            l1x_filename = settings.l1x_dark.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::nonlin) {
            l1x_filename = settings.l1x_nonlin.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::prnu) {
            l1x_filename = settings.l1x_prnu.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::stray) {
            l1x_filename = settings.l1x_stray.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::fov) {
            l1x_filename = settings.l1x_fov.at(static_cast<int>(proc_level));
        } else if (l1x_level == ProcLevel::rad) {
            l1x_filename = settings.l1x_rad.at(static_cast<int>(proc_level));
        }
        if (!l1x_filename.empty()) {
            l1x.at(i_l1x) = std::make_shared<L1X>(
              mpi, ckd, l1a_products, l1x_level, l1x_filename);
        }
    }
    for (int i {}; i < static_cast<int>(l1a_products.size()); ++i) {
        l1a_products.at(i).l1x = l1x;
        l1a_products.at(i).l1x_image_offset = i;
    }
}

} // namespace spex
