// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_metadata.h"

#include "spex_settings.h"

namespace spex {

L1AMetadata::L1AMetadata(const std::string& filename,
                         const SPEXSettings& settings,
                         const ProcLevel proc_level)
  : filename { filename }, keep_frames { proc_level == ProcLevel::noise }
{
    nc = std::make_shared<NcObject>(netCDF::NcFile::read);
    nc->open(filename);
    for (const auto& [name, att] : nc->getAtts()) {
        if (name == "l1x_level") {
            att.getValues(&l1x_level);
        }
    }
    number_of_images = nc->getDim("number_of_images").getSize();
    if (l1x_level < ProcLevel::fov) {
        npix = nc->getDim("samples_per_image").getSize();
    }

    const auto engineering_data { nc->getGroup("engineering_data") };
    hk_packets = nc->getDim("hk_packets").getSize();
    HK_tlm_time.resize(hk_packets);
    temp_detector.resize(hk_packets);
    engineering_data.getVar("HK_tlm_time").getVar(HK_tlm_time.data());
    engineering_data.getVar("temp_detector").getVar(temp_detector.data());
    if (const auto navigation_data { nc->getGroup("navigation_data") };
        !navigation_data.isNull()) {
        sc_records = nc->getDim("SC_records").getSize();
        nav_att_time.resize(sc_records);
        nav_att_quat.resize(sc_records);
        std::vector<double> nav_att_quat_raw(sc_records * dims::quat);
        nav_orb_time.resize(sc_records);
        nav_orb_pos.resize(sc_records);
        std::vector<double> nav_orb_pos_raw(sc_records * dims::vec);
        nav_orb_vel.resize(sc_records);
        std::vector<double> nav_orb_vel_raw(sc_records * dims::vec);
        navigation_data.getVar("att_time").getVar(nav_att_time.data());
        navigation_data.getVar("att_quat").getVar(nav_att_quat_raw.data());
        for (int i {}; i < sc_records; ++i) {
            nav_att_quat[i] = Quaternion(&nav_att_quat_raw[i * dims::quat]);
        }
        navigation_data.getVar("orb_time").getVar(nav_orb_time.data());
        navigation_data.getVar("orb_pos").getVar(nav_orb_pos_raw.data());
        for (int i {}; i < sc_records; ++i) {
            nav_orb_pos[i] = Vector(&nav_orb_pos_raw[i * dims::vec]);
        }
        navigation_data.getVar("orb_vel").getVar(nav_orb_vel_raw.data());
        for (int i {}; i < sc_records; ++i) {
            nav_orb_vel[i] = Vector(&nav_orb_vel_raw[i * dims::vec]);
        }
    }
    if (const auto gse_data { nc->getGroup("gse_data") }; !gse_data.isNull()) {
        for (const auto& [name, dim] : gse_data.getDims()) {
            if (name == "wavelength") {
                dim_refspec = dim.getSize();
            }
        }
        refspec_wavelength.resize(dim_refspec, NC_FILL_DOUBLE);
        refspec_radiance.resize(dim_refspec, NC_FILL_DOUBLE);
        for (const auto& [name, var] : gse_data.getVars()) {
            if (name == "wavelength") {
                var.getVar(refspec_wavelength.data());
            } else if (name == "signal") {
                var.getVar(refspec_radiance.data());
            } else if (name == "viewport") {
                var.getVar(&viewport);
            }
        }
        if (const auto wave_monitor { gse_data.getGroup("WaveMonitor") };
            !wave_monitor.isNull()) {
            dim_refspec = wave_monitor.getDim("wavelength").getSize();
            refspec_wavelength.resize(dim_refspec);
            wave_monitor.getVar("wavelength").getVar(refspec_wavelength.data());
            // Reference spectra will be taken as the sum over different times
            const auto dim_time { wave_monitor.getDim("time").getSize() };
            std::vector<double> refspec_radiance_raw(dim_refspec * dim_time);
            wave_monitor.getVar("spectrum").getVar(refspec_radiance_raw.data());
            refspec_radiance.resize(dim_refspec);
            std::ranges::fill(refspec_radiance, 0.0);
            int start_idx {};
            for (int i_time {}; i_time < dim_time; ++i_time) {
                for (int i {}; i < dim_refspec; ++i) {
                    refspec_radiance[i] += refspec_radiance_raw[i + start_idx];
                }
                start_idx += dim_refspec;
            }
        }
        if (const auto egse_var { gse_data.getVar("egse") };
            !egse_var.isNull()) {
            egse.resize(egse_var.getDim(0).getSize());
            egse_var.getVar(egse.data());
        }
        // Degree of linear polarization constant over the spectrum
        // (might or might not be needed)
        double dolp_value {};
        for (const auto& [name, att] : gse_data.getAtts()) {
            if (name == "Illumination_level") {
                att.getValues(&illumination_level);
            } else if (name == "DoLP") {
                att.getValues(&dolp_value);
            } else if (name == "ACT_rotationAngle") {
                att.getValues(&act_angle);
                act_angle *= math::deg_to_rad;
            } else if (name == "ALT_rotationAngle") {
                // If ALT angle is present override the previously set
                // viewport value.
                att.getValues(&alt_angle);
                constexpr double angle_comp_margin { 10.0 };
                for (const auto& [vp_idx, angle] : viewport_idx_to_angle) {
                    if (std::abs(angle - alt_angle) < angle_comp_margin) {
                        viewport = vp_idx;
                    }
                }
                alt_angle *= math::deg_to_rad;
            }
        }
        if (const auto group { gse_data.getGroup("SpectralDolP") };
            !group.isNull()) {
            const auto n_wavelengths { group.getDim("wavelength").getSize() };
            std::vector<double> wavelengths(n_wavelengths);
            group.getVar("wavelength").getVar(wavelengths.data());
            std::vector<double> dolp(n_wavelengths);
            group.getVar("DoLP").getVar(dolp.data());
            dolp_spline = std::make_unique<CubicSpline>(wavelengths, dolp);
        } else {
            dolp_spline = std::make_unique<CubicSpline>(
              std::vector<double> { 0.0, NC_FILL_FLOAT },
              std::vector<double> { dolp_value, dolp_value });
        }
    }
    // Close the file because there may be an onslaught of L1A files
    nc->close();
}

} // namespace spex
