// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1b.h"

#include "binning_table.h"
#include "bspline.h"
#include "ckd.h"
#include "functions.h"
#include "l1a.h"
#include "l1a_metadata.h"
#include "linalg.h"
#include "linear_spline.h"
#include "lininv.h"
#include "matrix.h"
#include "spex_math.h"

#include <cassert>
#include <cstring>
#include <fstream>
#include <gadfit/mpi.h>
#include <iomanip>
#include <netcdf.h>
#include <numbers>
#include <spdlog/spdlog.h>

namespace spex {

// Wavelength range for defining the valid min and max attributes of
// some NetCDF variables. The official spectral coverage of SPEXone is
// 385–770 nm but here we have taken the values by inspecting a CKD.
constexpr float wavelengths_valid_min { 340.0f };
constexpr float wavelengths_valid_max { 800.0f };

L1B::L1B(const gadfit::MPIVars& mpi, const SPEXSettings& settings, CKD& ckd)
  : Processor { mpi, settings, ProcLevel::l1b, ckd }
{
    if (settings.l1b.last_calibration_step
        && settings.l1b.last_calibration_step < ckd.level) {
        ckd.level = static_cast<ProcLevel>(
          static_cast<int>(settings.l1b.last_calibration_step.value())
          + 1);
    }
}

// Copy the navigation data over from an L1A file
static auto copyNavigationData(NcObject& nc_l1a, NcObject& nc_l1b) -> void
{
    nc_l1a.reOpen();
    const auto SC_records { nc_l1a.getDim("SC_records").getSize() };
    const auto nav_l1a { nc_l1a.getGroup("navigation_data") };
    auto nav_l1b { nc_l1b.addGroup("NAVIGATION_DATA") };
    // Attitude times
    const auto SC_records_dim { nc_l1b.addDim("SC_records", SC_records) };
    auto att_time_nc { nav_l1b.addVar(
          "att_time", netCDF::ncDouble, SC_records_dim) };
    att_time_nc.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    att_time_nc.putAtt("long_name", "attitude sample time (seconds of day)");
    att_time_nc.putAtt("valid_min", netCDF::ncDouble, 0.0);
    att_time_nc.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
    att_time_nc.putAtt("units", "seconds");
    std::vector<double> att_time(SC_records);
    nav_l1a.getVar("att_time").getVar(att_time.data());
    att_time_nc.putVar(att_time.data());
    // Attitude quaternions
    const auto quaternion_elements_dim { nc_l1b.addDim(
          "quaternion_elements", dims::quat) };
    auto att_quat_nc { nav_l1b.addVar(
          "att_quat",
          netCDF::ncFloat,
          { SC_records_dim, quaternion_elements_dim })
    };
    att_quat_nc.putAtt("_FillValue", netCDF::ncFloat, -999.9f);
    att_quat_nc.putAtt("long_name",
                       "attitude quaternions (J2000 to spacecraft)");
    att_quat_nc.putAtt("valid_min", netCDF::ncFloat, -1.0f);
    att_quat_nc.putAtt("valid_max", netCDF::ncFloat, 1.0f);
    att_quat_nc.putAtt("units", "seconds");
    std::vector<float> att_quat(SC_records * dims::quat);
    nav_l1a.getVar("att_quat").getVar(att_quat.data());
    att_quat_nc.putVar(att_quat.data());
    // Orbit times
    auto orb_time_nc { nav_l1b.addVar(
          "orb_time", netCDF::ncDouble, SC_records_dim) };
    orb_time_nc.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    orb_time_nc.putAtt("long_name", "Orbit vector time (seconds of day)");
    orb_time_nc.putAtt("valid_min", netCDF::ncDouble, 0.0);
    orb_time_nc.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
    orb_time_nc.putAtt("units", "seconds");
    nav_l1a.getVar("orb_time").getVar(att_time.data());
    orb_time_nc.putVar(att_time.data());
    // Orbit positions
    const auto vector_elements_dim { nc_l1b.addDim("vector_elements",
                                                   dims::vec) };
    auto orb_pos_nc { nav_l1b.addVar("orb_pos",
                                     netCDF::ncFloat,
                                     { SC_records_dim, vector_elements_dim }) };
    orb_pos_nc.putAtt("_FillValue", netCDF::ncFloat, -9999999.f);
    orb_pos_nc.putAtt("long_name", "Orbit positions vectors (ECR)");
    orb_pos_nc.putAtt("valid_min", netCDF::ncFloat, -7200000.f);
    orb_pos_nc.putAtt("valid_max", netCDF::ncFloat, 7200000.f);
    orb_pos_nc.putAtt("units", "m");
    std::vector<float> orb_pos(SC_records * dims::vec);
    nav_l1a.getVar("orb_pos").getVar(orb_pos.data());
    orb_pos_nc.putVar(orb_pos.data());
    // Orbit velocities
    auto orb_vel_nc { nav_l1b.addVar("orb_vel",
                                     netCDF::ncFloat,
                                     { SC_records_dim, vector_elements_dim }) };
    orb_vel_nc.putAtt("_FillValue", netCDF::ncFloat, -9999999.f);
    orb_vel_nc.putAtt("long_name", "Orbit velocity vectors (ECR)");
    orb_vel_nc.putAtt("valid_min", netCDF::ncFloat, -7600.f);
    orb_vel_nc.putAtt("valid_max", netCDF::ncFloat, 7600.f);
    orb_vel_nc.putAtt("units", "meters/seconds");
    std::vector<float> orb_vel(SC_records * dims::vec);
    nav_l1a.getVar("orb_vel").getVar(orb_vel.data());
    orb_vel_nc.putVar(orb_vel.data());
    // Global attributes of orbit data
    for (const auto& [name, att] : nc_l1a.getAtts()) {
        std::string buffer {};
        if (name == "startDirection") {
            att.getValues(buffer);
            nc_l1b.putAtt("startDirection", buffer);
        } else if (name == "endDirection") {
            att.getValues(buffer);
            nc_l1b.putAtt("endDirection", buffer);
        } else if (name == "time_coverage_start") {
            att.getValues(buffer);
            nc_l1b.putAtt("time_coverage_start", buffer);
        } else if (name == "time_coverage_stop") {
            att.getValues(buffer);
            nc_l1b.putAtt("time_coverage_stop", buffer);
        } else if (name == "reference") {
            // Reference on which image time is based
            att.getValues(buffer);
            nc_l1b.putAtt("reference", buffer);
        } else if (name == "orbit_number") {
            long int orbit_number {};
            att.getValues(&orbit_number);
            nc_l1b.putAtt("orbit_number", netCDF::ncInt64, orbit_number);
        }
    }
    nc_l1a.close();
}

auto L1B::processInit() -> void
{
    // Some administration to begin with
    timer_Sx_BT.reset();
    timer_Squ.reset();
    timer_SQU.reset();
    timer_Smt.reset();
    timer_SXT.reset();
    timer_S_QUI_conv.reset();
    do_geolocation = settings.l1b.geolocation;
    if (settings.l1b.geolocation
        && (settings.skip.at(static_cast<int>(ProcLevel::swath))
            || ckd.level <= ProcLevel::swath)) {
        spdlog::warn("Geolocation will not be performed because the CKD does "
                     "not contain swath vectors");
        do_geolocation = false;
    }

    if (mpi.rank == 0) {
        nc_l1b.open(settings.l1b.outputfile);
        nc_l1b.writeReducedConfig(settings);
    }

    // Determine how many images each MPI process needs to handle
    std::vector<int> workloads(mpi.n_procs, 0);
    // In a serial run the relative load on process 0 cannot be variable
    double first_proc_rel_workload { settings.l1b.first_proc_rel_workload };
    if (mpi.n_procs == 1) {
        first_proc_rel_workload = 1.0;
    }
    // The first one is assigned a fixed number determined by an input setting
    workloads.front() =
      std::max(1,
               static_cast<int>(std::round(first_proc_rel_workload
                                           * nL1A() / mpi.n_procs)));
    // Distribute the remaining images equally among the remaining MPI
    // processes.
    const int nl1a_remaining { nL1A() - workloads.front() };
    for (int i_rank { 1 }; i_rank < static_cast<int>(workloads.size());
         ++i_rank) {
        workloads.at(i_rank) =
          static_cast<int>(nl1a_remaining) / (mpi.n_procs - 1);
    }
    // Some processes might be assigned one more image than others
    if (mpi.n_procs > 1) {
        for (int i { 1 };
             i < static_cast<int>(nl1a_remaining) % (mpi.n_procs - 1) + 1;
             ++i) {
            ++workloads.at(i);
        }
    }
    mpi_ranges.resize(mpi.n_procs + 1);
    mpi_ranges.front() = 0;
    for (int i_rank { 1 }; i_rank < mpi.n_procs + 1; ++i_rank) {
        mpi_ranges.at(i_rank) =
          mpi_ranges.at(i_rank - 1) + workloads.at(i_rank - 1);
    }
    if (mpi.n_procs > 1) {
        if (mpi.n_procs > nL1A()) {
            throw std::runtime_error { "Error: number of MPI tasks should not "
                                       "be less than the number of images." };
        }
        spdlog::info(
          "Images are distributed among the MPI processes as follows:");
        std::stringstream workloads_line {};
        for (const auto& workload : workloads) {
            workloads_line << "  " << workload;
        }
        spdlog::info( "{}", workloads_line.str().c_str());
    }

    // Define L1B-specific global attributes
    if (mpi.rank == 0) {
        nc_l1b.putAtt("cdm_data_type", "swath");
        nc_l1b.putAtt("processing_level", "L1B");
        nc_l1b.putAtt("title", "PACE SPEX Level-1b data");
    }

    // Define groups
    netCDF::NcGroup grp_bands {};
    netCDF::NcGroup grp_bin {};
    netCDF::NcGroup grp_geolocation {};
    netCDF::NcGroup grp_observation {};
    if (mpi.rank == 0) {
        grp_bands = nc_l1b.addGroup("SENSOR_VIEW_BANDS");
        grp_bin = nc_l1b.addGroup("BIN_ATTRIBUTES");
        grp_geolocation = nc_l1b.addGroup("GEOLOCATION_DATA");
        grp_observation = nc_l1b.addGroup("OBSERVATION_DATA");
    }

    // Define dimensions
    netCDF::NcDim dimid_obs {}; // Number of images
    netCDF::NcDim dimid_vp {}; // Number of viewports
    netCDF::NcDim dimid_fov {}; // Number of swath positions per viewport
    netCDF::NcDim dimid_int_wave {}; // Number of intensity (raw) wavelengths
    netCDF::NcDim dimid_pol_wave {}; // Number of L1B output wavelengths
    if (mpi.rank == 0) {
        dimid_obs = nc_l1b.addDim("bins_along_track", nL1A());
    }

    // Define variables
    if (mpi.rank == 0) {
        var_time = grp_bin.addVar("image_time", netCDF::ncDouble, dimid_obs);
        var_time.putAtt("long_name", "image time (seconds of day)");
        var_time.putAtt("_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        var_time.putAtt("valid_min", netCDF::ncDouble, 0.0);
        var_time.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
        var_time.putAtt("units", "seconds");
    }

    // If anything is calculated related to a spectrum. This can be
    // radiance, demodulation or geolocation.
    if ((ckd.level >= ProcLevel::rad || do_geolocation) && mpi.rank == 0) {
        dimid_vp = nc_l1b.addDim("number_of_views",ckd.dim_vp);
        dimid_fov = nc_l1b.addDim("spatial_samples_per_image",ckd.fov.n_act);
        // Map of viewport indices per FOV
        std::vector<uint8_t> ivp_fov(ckd.fov.n_act);
        size_t ivp = 0;
        for (size_t ifov=0 ; ifov<ckd.fov.n_act ; ifov++) {
            while (ivp != ckd.dim_vp-1 && ifov == ckd.fov.start_vp[ivp+1]) ivp++;
            ivp_fov[ifov] = (uint8_t) ivp;
        }
        netCDF::NcVar viewport_index { grp_bands.addVar(
              "viewport_index", netCDF::ncUbyte, dimid_fov) };
        viewport_index.putAtt("long_name",
                   "zero-based viewport index for each spatial sample");
        viewport_index.putAtt("_FillValue", netCDF::ncUbyte, NC_FILL_UBYTE);
        viewport_index.putAtt("valid_min", netCDF::ncUbyte, 0);
        viewport_index.putAtt(
          "valid_max", netCDF::ncUbyte, static_cast<uint8_t>(ckd.dim_vp-1));
        std::vector<double> vp_angles(ckd.dim_vp, 0);
        std::transform(viewport_idx_to_angle.cbegin(),
                       viewport_idx_to_angle.cend(),
                       vp_angles.begin(),
                       [](const auto& pair) { return pair.second; });
        viewport_index.putAtt(
          "mapping", netCDF::ncDouble, ckd.dim_vp, vp_angles.data());
        viewport_index.putVar(ivp_fov.data());
    }

    // Define L1B wavelengths for intensity grid. This will be updated
    // to the raw grid with binning. TODO: Do this.
    // Now, we read intensity wavelengths from the settings.
    dim_int_wave = settings.l1b.intensity_wavelength.size();

    // Everything for radiance
    if (ckd.level >= ProcLevel::rad && mpi.rank == 0) {
        // The dimension for intensity wavelengths
        dimid_int_wave =
          nc_l1b.addDim("intensity_bands_per_view", dim_int_wave);
        std::vector<netCDF::NcDim> dims_int = {dimid_vp,dimid_int_wave};
        // Wavelengths intensity grid
        netCDF::NcVar intensity_wavelengths { grp_bands.addVar(
              "intensity_wavelengths", netCDF::ncFloat, dims_int) };
        intensity_wavelengths.putAtt("long_name",
                                     "wavelength at intensity samples");
        intensity_wavelengths.putAtt(
          "_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        intensity_wavelengths.putAtt(
          "valid_min", netCDF::ncFloat, wavelengths_valid_min);
        intensity_wavelengths.putAtt(
          "valid_max", netCDF::ncFloat, wavelengths_valid_max);
        intensity_wavelengths.putAtt("units", "nm");
        // The wavelengths now come from the settings. In the future, they will come from
        // the CKD, but will still be the same for all viewports and FOVs. To conform the
        // L1B format, we will copy the same array dim_vp times. TODO: Change this comment
        // when we no longer acquire these wavelengths from the settings.
        for (size_t ivp=0 ; ivp<ckd.dim_vp ; ivp++) {
            const std::vector<size_t> strt { ivp, 0 };
            const std::vector<size_t> cnt { 1,
                                            static_cast<size_t>(dim_int_wave) };
            intensity_wavelengths.putVar(
              strt, cnt, settings.l1b.intensity_wavelength.data());
        }

        // Intensity bandpasses. This should be an indication of the ISRF. We do not have
        // an ISRF yet, nor do we know for certain how it depends on viewport or FOV. We
        // now implement it as if it is per view port, but we do not fill in anything, because
        // we do not know anything.
        netCDF::NcVar intensity_bandpasses { grp_bands.addVar(
              "intensity_bandpasses", netCDF::ncFloat, dims_int) };
        intensity_bandpasses.putAtt("long_name", "FWHM of intensity bands");
        intensity_bandpasses.putAtt(
          "_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        intensity_bandpasses.putAtt("units", "nm");
        std::vector<double> fake_intensity_bandpasses(
          ckd.dim_vp * dim_int_wave, NC_FILL_DOUBLE);
        intensity_bandpasses.putVar(fake_intensity_bandpasses.data());

        // Solar spectrum on intensity grid
        netCDF::NcVar intensity_f0 { grp_bands.addVar(
              "intensity_f0", netCDF::ncFloat, dims_int) };
        intensity_f0.putAtt("long_name",
                   "solar irradiance at intensity wavelength grid");
        intensity_f0.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        intensity_f0.putAtt("units", "W.m-2");
        // TODO: Make a real solar spectrum
        std::vector<double> fake_solar_spectrum(
          ckd.dim_vp * dim_int_wave, NC_FILL_DOUBLE);
        intensity_f0.putVar(fake_solar_spectrum.data());

        // The wavelengths and the intensity-only spectra
        std::vector<netCDF::NcDim> dims_spectra_int = {dimid_obs,dimid_fov,dimid_int_wave};

        // Intensity
        var_radiance =
          grp_observation.addVar("I", netCDF::ncFloat, dims_spectra_int);
        var_radiance.putAtt("long_name", "I stokes vector");
        var_radiance.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_radiance.putAtt("units", "W.m-2.sr-1.um-1");
        var_radiance.putAtt("ancillary_variables", "I_noise");
        std::vector<std::size_t> chunksize_intensity {
            1,
            static_cast<size_t>(ckd.fov.n_act),
            settings.l1b.intensity_wavelength.size()
        };
        var_radiance.setChunking(netCDF::NcVar::nc_CHUNKED,
                                 chunksize_intensity);
        var_radiance.setCompression(true, true, 1);

        // Noise on intensity
        var_radiance_noise =
          grp_observation.addVar("I_noise", netCDF::ncFloat, dims_spectra_int);
        var_radiance_noise.putAtt("long_name",
                                  "noise (standard deviation) of I");
        var_radiance_noise.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_radiance_noise.putAtt("units", "W.m-2.sr-1.um-1");
        var_radiance_noise.setChunking(netCDF::NcVar::nc_CHUNKED,
                                       chunksize_intensity);
        var_radiance_noise.setCompression(true, true, 1);
    }
    // L1B output wavelengths. They consist of central wavelengths and
    // full-width half-maxima.
    dim_pol_wave = settings.l1b.l1b_wavelength.size();
    // Save the FWHMs for in the output
    std::vector<double> fwhm(dim_pol_wave, 0.0);
    // Pre-calculations for radiometric calibration
    if (ckd.level >= ProcLevel::rad) {
        // Find out where the retrieval windows are and how big they
        // are.
        fov_start_bin.resize(ckd.fov.n_act);
        fov_nmeas.resize(ckd.fov.n_act);
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            int ibin {};
            if (settings.l1b.wave_min) {
                while (ckd.wave_common[i_fov][ibin] < settings.l1b.wave_min) {
                    ++ibin;
                }
            }
            fov_start_bin[i_fov] = ibin;
            ibin = ckd.fov.dims_spec[i_fov] - 1;
            if (settings.l1b.wave_max) {
                while (ckd.wave_common[i_fov][ibin] > settings.l1b.wave_max) {
                    --ibin;
                }
            }
            fov_nmeas[i_fov] = ibin + 1 - fov_start_bin[i_fov];
        }
    }
    // Pre-calculations for demodulation
    if (ckd.level >= ProcLevel::pol) {
        // We need to add (order-1) because the last order knots and
        // the endpoint are not included.
        n_demod_states = 2
                         * (static_cast<int>(settings.l1b.knots.size())
                            + settings.l1b.order - 1);
        n_demod_states_half = n_demod_states / 2;
        demod_size_full = n_demod_states * n_demod_states;
        demod_size_half = n_demod_states_half * n_demod_states_half;
        BSpline demod_b_spline { settings.l1b.order, settings.l1b.knots };
        // Initialize demodulation splines and the Gaussian kernel
        demod_spline_matrix.resize(ckd.fov.n_act);
        gauss_kernel.resize(ckd.fov.n_act);
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            // This is the pure spline matrix used in demodulation
            // with the dimension n_demod_states_half because q and u
            // are not included yet.
            demod_spline_matrix[i_fov] = std::make_unique<PackedMatrix>(
              fov_nmeas[i_fov], n_demod_states_half);
            for (int i {}; i < fov_nmeas[i_fov]; ++i) {
                demod_b_spline.evalBasis(
                  ckd.wave_common[i_fov][fov_start_bin[i_fov] + i],
                  i,
                  *demod_spline_matrix[i_fov]);
            }
            // Gauss conv
            gauss_kernel[i_fov] = std::make_unique<PackedMatrix>(
              dim_pol_wave, fov_nmeas[i_fov]);
            for (int i {}; i < dim_pol_wave; ++i) {
                gauss_kernel[i_fov]->row_locations[i] =
                  static_cast<int>(gauss_kernel[i_fov]->data.size());
                const double fwhm { settings.l1b.l1b_wavelength[i]
                                    / settings.l1b.resolving_power };
                const double sigma { fwhm / (2.0 * sqrt(2.0 * log(2.0))) };
                const double norm_inv {
                    1.0 / (sigma * std::sqrt(2 * std::numbers::pi))
                };
                for (int j {}; j < fov_nmeas[i_fov]; ++j) {
                    const int j0 { fov_start_bin[i_fov] + j };
                    double res { ckd.wave_common[i_fov][j0]
                                 - settings.l1b.l1b_wavelength[i] };
                    res *= res;
                    res /= 2 * sigma * sigma;
                    res = norm_inv * std::exp(-res);
                    double delta_lambda {};
                    if (j == 0) {
                        delta_lambda = ckd.wave_common[i_fov][j0 + 1]
                                       - ckd.wave_common[i_fov][j0];
                    } else if (j == fov_nmeas[i_fov] - 1) {
                        delta_lambda = ckd.wave_common[i_fov][j0]
                                       - ckd.wave_common[i_fov][j0 - 1];
                    } else {
                        delta_lambda = 0.5 * (ckd.wave_common[i_fov][j0 + 1]
                                              - ckd.wave_common[i_fov][j0 - 1]);
                    }
                    const double result { res * delta_lambda };
                    if (result > settings.l1b.gauss_range_tol) {
                        gauss_kernel[i_fov]->data.push_back(result);
                        ++gauss_kernel[i_fov]->row_sizes[i];
                        if (gauss_kernel[i_fov]->global_col_indices[i]
                            == math::idx_fill_value) {
                            gauss_kernel[i_fov]->global_col_indices[i] = j;
                        }
                    }
                }
            }
        }
    }
    // Now do the NetCDF things
    if (ckd.level >= ProcLevel::pol && mpi.rank == 0) {
        // Dimension: Wavelengths on polarization output
        dimid_pol_wave = nc_l1b.addDim("polarization_bands_per_view",dim_pol_wave);
        std::vector<netCDF::NcDim> dims_pol = {dimid_vp,dimid_pol_wave};
        // Polarization wavelengths
        netCDF::NcVar polarization_wavelengths { grp_bands.addVar(
              "polarization_wavelengths", netCDF::ncFloat, dims_pol) };
        polarization_wavelengths.putAtt(
          "long_name", "wavelength at center of polarization bands");
        polarization_wavelengths.putAtt(
          "_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        polarization_wavelengths.putAtt(
          "valid_min", netCDF::ncFloat, wavelengths_valid_min);
        polarization_wavelengths.putAtt(
          "valid_max", netCDF::ncFloat, wavelengths_valid_max);
        polarization_wavelengths.putAtt("units", "nm");
        // Repeat wavelengths for each viewport. See comments on intensity section.
        // TODO: Update comments when there had been a discussion.
        for (size_t ivp=0 ; ivp<ckd.dim_vp ; ivp++) {
            const std::vector<size_t> strt { ivp, 0 };
            const std::vector<size_t> cnt { 1,
                                            static_cast<size_t>(dim_pol_wave) };
            polarization_wavelengths.putVar(
              strt, cnt, settings.l1b.l1b_wavelength.data());
        }

        // Polarization bandpasses
        netCDF::NcVar polarization_bandpasses { grp_bands.addVar(
              "polarization_bandpasses", netCDF::ncFloat, dims_pol) };
        polarization_bandpasses.putAtt("long_name",
                                       "FWHM of polarization bands");
        polarization_bandpasses.putAtt(
          "_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        polarization_bandpasses.putAtt("units", "nm");
        // Repeat same data set for all viewports
        for (size_t ivp=0 ; ivp<ckd.dim_vp ; ivp++) {
            const std::vector<size_t> strt { ivp, 0 };
            const std::vector<size_t> cnt { 1,
                                            static_cast<size_t>(dim_pol_wave) };
            polarization_bandpasses.putVar(strt,cnt, fwhm.data());
        }

        // Solar spectrum at polarization wavelengths
        netCDF::NcVar polarization_f0 { grp_bands.addVar(
              "polarization_f0", netCDF::ncFloat, dims_pol) };
        polarization_f0.putAtt("long_name",
                   "solar irradiance at polarization wavelength grid");
        polarization_f0.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        polarization_f0.putAtt("units", "W.m-2");
        // TODO: Make a real solar spectrum
        std::vector<double> fake_solar_spectrum(ckd.dim_vp*dim_pol_wave,
                                                NC_FILL_DOUBLE);
        polarization_f0.putVar(fake_solar_spectrum.data());

        const std::vector<netCDF::NcDim> dims_spectra_pol { dimid_obs,
                                                            dimid_fov,
                                                            dimid_pol_wave };
        // Intensity at polarimeter sampling
        var_radiance_pol =  grp_observation.addVar(
          "I_polsample", netCDF::ncFloat, dims_spectra_pol);
        var_radiance_pol.putAtt("long_name", "I at polarimeter sampling");
        var_radiance_pol.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_radiance_pol.putAtt("units", "W.m-2.sr-1.um-1");
        var_radiance_pol.putAtt("ancillary_variables", "I_pol_noise");
        std::vector<size_t> chunksize_pol { 1,
                                            static_cast<size_t>(ckd.fov.n_act),
                                            static_cast<size_t>(dim_pol_wave) };
        var_radiance_pol.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_radiance_pol.setCompression(true, true, 1);

        // Noise on intensity at polarimeter sampling
        var_radiance_noise_pol = grp_observation.addVar(
          "I_polsample_noise", netCDF::ncFloat, dims_spectra_pol);
        var_radiance_noise_pol.putAtt(
          "long_name",
          "noise (standard deviation) of I at polarimeter sampling");
        var_radiance_noise_pol.putAtt("_FillValue", netCDF::ncFloat,
                                      NC_FILL_FLOAT);
        var_radiance_noise_pol.putAtt("units", "W.m-2.sr-1.um-1");
        var_radiance_noise_pol.setChunking(netCDF::NcVar::nc_CHUNKED,
                                           chunksize_pol);
        var_radiance_noise_pol.setCompression(true, true, 1);

        // Small q
        var_small_q =
          grp_observation.addVar("q", netCDF::ncFloat, dims_spectra_pol);
        var_small_q.putAtt("long_name", "q");
        var_small_q.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_small_q.putAtt("valid_min", netCDF::ncFloat, 0.0f);
        var_small_q.putAtt("valid_max", netCDF::ncFloat, 1.0f);
        var_small_q.putAtt("units", "1");
        var_small_q.putAtt("ancillary_variables", "q_noise");
        var_small_q.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_small_q.setCompression(true, true, 1);

        // Noise on q
        var_small_q_noise =
          grp_observation.addVar("q_noise", netCDF::ncFloat, dims_spectra_pol);
        var_small_q_noise.putAtt("long_name", "noise (standard deviation) of q");
        var_small_q_noise.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_small_q_noise.putAtt("units", "1");
        var_small_q_noise.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_small_q_noise.setCompression(true, true, 1);

        // Small u
        var_small_u =
          grp_observation.addVar("u", netCDF::ncFloat, dims_spectra_pol);
        var_small_u.putAtt("long_name", "u");
        var_small_u.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_small_u.putAtt("valid_min", netCDF::ncFloat, 0.0f);
        var_small_u.putAtt("valid_max", netCDF::ncFloat, 1.0f);
        var_small_u.putAtt("units", "1");
        var_small_u.putAtt("ancillary_variables", "u_noise");
        var_small_u.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_small_u.setCompression(true, true, 1);

        // Noise on u
        var_small_u_noise =
          grp_observation.addVar("u_noise", netCDF::ncFloat, dims_spectra_pol);
        var_small_u_noise.putAtt("long_name",
                                 "noise (standard deviation) of u");
        var_small_u_noise.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_small_u_noise.putAtt("units", "1");
        var_small_u_noise.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_small_u_noise.setCompression(true, true, 1);

        // DoLP
        var_dolp =
          grp_observation.addVar("DoLP", netCDF::ncFloat, dims_spectra_pol);
        var_dolp.putAtt("long_name", "degree of linear polarization");
        var_dolp.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_dolp.putAtt("valid_min", netCDF::ncFloat, 0.0f);
        var_dolp.putAtt("valid_max", netCDF::ncFloat, 1.0f);
        var_dolp.putAtt("units", "1");
        var_dolp.putAtt("ancillary_variables", "DoLP_noise");
        var_dolp.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_dolp.setCompression(true, true, 1);

        // Noise on DoLP
        var_dolp_noise = grp_observation.addVar(
          "DoLP_noise", netCDF::ncFloat, dims_spectra_pol);
        var_dolp_noise.putAtt("long_name", "noise (standard deviation) of DoLP");
        var_dolp_noise.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_dolp_noise.putAtt("units", "1");
        var_dolp_noise.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_dolp_noise.setCompression(true, true, 1);

        // AoLP
        var_aolp =
          grp_observation.addVar("AoLP", netCDF::ncFloat, dims_spectra_pol);
        var_aolp.putAtt("long_name", "angle of linear polarization");
        var_aolp.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_aolp.putAtt("valid_min", netCDF::ncFloat, -90.0f);
        var_aolp.putAtt("valid_max", netCDF::ncFloat, 90.0f);
        var_aolp.putAtt("units", "degrees");
        var_aolp.putAtt("ancillary_variables", "AoLP_noise");
        var_aolp.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_aolp.setCompression(true, true, 1);

        // Noise on AoLP
        var_aolp_noise = grp_observation.addVar("AoLP_noise",netCDF::ncFloat,dims_spectra_pol);
        var_aolp_noise.putAtt("long_name", "noise (standard deviation) of AoLP");
        var_aolp_noise.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_aolp_noise.putAtt("units", "degrees");
        var_aolp_noise.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_pol);
        var_aolp_noise.setCompression(true, true, 1);
    }

    // NetCDF variables for geolocation
    if (do_geolocation) {
        // Read UTC file
        NcObject nc_utc { netCDF::NcFile::read };
        nc_utc.open(settings.l1b.utcfile);
        dim_utc = nc_utc.getDim("sz").getSize();
        mjd_utc.resize(dim_utc);
        tdiff_utc.resize(dim_utc);
        nc_utc.getVar("mjd").getVar(mjd_utc.data());
        nc_utc.getVar("timediff").getVar(tdiff_utc.data());

        // Construct the planet
        earth = std::make_unique<Planet>(settings.l1b.semi_major_axis,settings.l1b.semi_minor_axis,settings.l1b.latitude_tol);
        if (settings.l1b.demfile.compare("") != 0) {
            earth->include_dem(settings.l1b.demfile);
        }
    }
    // Create geolocation output
    if (do_geolocation && mpi.rank == 0) {
        const std::vector<netCDF::NcDim> dims_geo { dimid_obs, dimid_fov };
        // Latitude
        var_lat = grp_geolocation.addVar("latitude", netCDF::ncFloat, dims_geo);
        var_lat.putAtt("standard_name", "latitude");
        var_lat.putAtt("long_name", "latitude");
        var_lat.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_lat.putAtt("valid_min", netCDF::ncFloat, -90.0f);
        var_lat.putAtt("valid_max", netCDF::ncFloat, 90.0f);
        var_lat.putAtt("units", "degrees_north");

        // Longitude
        var_lon =
          grp_geolocation.addVar("longitude", netCDF::ncFloat, dims_geo);
        var_lon.putAtt("standard_name", "longitude");
        var_lon.putAtt("long_name", "longitude");
        var_lon.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_lon.putAtt("valid_min", netCDF::ncFloat, -180.0f);
        var_lon.putAtt("valid_max", netCDF::ncFloat, 180.0f);
        var_lon.putAtt("units", "degrees_east");

        // Altitude
        var_alt = grp_geolocation.addVar("altitude", netCDF::ncFloat, dims_geo);
        var_alt.putAtt("standard_name", "altitude");
        var_alt.putAtt("long_name", "height above mean sea level");
        var_alt.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_alt.putAtt("valid_min", netCDF::ncFloat, -1000.0f);
        var_alt.putAtt("valid_max", netCDF::ncFloat, 10000.0f);
        var_alt.putAtt("units", "m");
        var_alt.putAtt("positive", "up");
        var_alt.putAtt("axis", "Z");

        // Viewing zenith angle
        var_vza = grp_geolocation.addVar("sensor_zenith",netCDF::ncFloat,dims_geo);
        var_vza.putAtt("long_name", "sensor zenith angle");
        var_vza.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_vza.putAtt("valid_min",netCDF::ncFloat, -90.0f);
        var_vza.putAtt("valid_max",netCDF::ncFloat, 90.0f);
        var_vza.putAtt("units", "degrees");

        // Viewing azimuth angle
        var_vaa = grp_geolocation.addVar("sensor_azimuth",netCDF::ncFloat,dims_geo);
        var_vaa.putAtt("long_name", "sensor azimuth angle");
        var_vaa.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_vaa.putAtt("valid_min",netCDF::ncFloat, -180.0f);
        var_vaa.putAtt("valid_max",netCDF::ncFloat, 180.0f);
        var_vaa.putAtt("units", "degrees");

        // Solar zenith angle
        var_sza = grp_geolocation.addVar("solar_zenith",netCDF::ncFloat,dims_geo);
        var_sza.putAtt("long_name", "solar zenith angle");
        var_sza.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_sza.putAtt("valid_min",netCDF::ncFloat, -90.0f);
        var_sza.putAtt("valid_max",netCDF::ncFloat, 90.0f);
        var_sza.putAtt("units", "degrees");

        // Solar azimuth angle
        var_saa = grp_geolocation.addVar("solar_azimuth", netCDF::ncFloat, dims_geo);
        var_saa.putAtt("long_name", "solar azimuth angle");
        var_saa.putAtt("_FillValue", netCDF::ncFloat, NC_FILL_FLOAT);
        var_saa.putAtt("valid_min",netCDF::ncFloat, -180.0f);
        var_saa.putAtt("valid_max",netCDF::ncFloat, 180.0f);
        var_saa.putAtt("units", "degrees");

        // Copy navigation data from L1A to L1B
        copyNavigationData(*l1a_manager->l1a_products.front().meta->nc, nc_l1b);
    }

#ifdef USE_MPI
    // These variables are targeted for MPI parallelization. Edit this
    // list if necessary.
    mpi_nc_vars = { { MpiNcVarId::time, var_time } };
    if (ckd.level >= ProcLevel::rad) {
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::radiance, var_radiance });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::radiance_noise, var_radiance_noise });
    }
    if (ckd.level >= ProcLevel::pol) {
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::radiance_pol, var_radiance_pol });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::radiance_noise_pol, var_radiance_noise_pol });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::small_q, var_small_q });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::small_q_noise, var_small_q_noise });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::small_u, var_small_u });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::small_u_noise, var_small_u_noise });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::dolp, var_dolp });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::dolp_noise, var_dolp_noise });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::aolp, var_aolp });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::aolp_noise, var_aolp_noise });
    }
    if (do_geolocation) {
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::lat, var_lat });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::lon, var_lon });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::alt, var_alt });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::vza, var_vza });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::vaa, var_vaa });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::sza, var_sza });
        mpi_nc_vars.emplace_back(std::pair{ MpiNcVarId::saa, var_saa });
    }
    mpi_buffer_starts.resize(mpi_nc_vars.size() + 1, 0);
    mpi_nc_starts.resize(mpi_nc_vars.size());
    mpi_nc_counts.resize(mpi_nc_vars.size());
    mpi_buffers.resize(nL1A());
    mpi_buffer_markers.resize(mpi.n_procs);
    mpi_requests.resize(nL1A(), MPI_REQUEST_NULL);
    mpi_buf_multiplicities.resize(nL1A(), 0);
    mpi_io_timer.reset();
    mpi_final_io_timer_total.reset();
    mpi_final_io_timer_active.reset();
    // Get the dimensions and size of each netCDF variable in mpi_nc_vars
    for (int i_var {}; i_var < static_cast<int>(mpi_nc_vars.size()); ++i_var) {
        int n_dims {};
        if (mpi.rank == 0) {
            n_dims = mpi_nc_vars.at(i_var).second.getDimCount();
        }
        MPI_Bcast(&n_dims, 1, MPI_INT, 0, mpi.comm);
        mpi_nc_starts.at(i_var).resize(n_dims, 0);
        mpi_nc_counts.at(i_var).resize(n_dims, 1);
        int var_size { 1 };
        for (int i_dim { 1 }; i_dim < n_dims; ++i_dim) {
            int dim {};
            if (mpi.rank == 0) {
                dim = mpi_nc_vars.at(i_var).second.getDim(i_dim).getSize();
            }
            MPI_Bcast(&dim, 1, MPI_INT, 0, mpi.comm);
            mpi_nc_counts.at(i_var).at(i_dim) = dim;
            var_size *= mpi_nc_counts.at(i_var).at(i_dim);
        }
        // Double-precision variables take twice as much space in the buffer
        int size_ratio {};
        if (mpi.rank == 0) {
            size_ratio = mpi_nc_vars.at(i_var).second.getType().getSize() / sizeof(float);
        }
        MPI_Bcast(&size_ratio, 1, MPI_INT, 0, mpi.comm);
        var_size *= size_ratio;
        mpi_buffer_starts.at(i_var + 1) =
          mpi_buffer_starts.at(i_var) + var_size;
    }
    mpi_buffer_size = mpi_buffer_starts.back();
    mpi_send_size = settings.l1b.mpi_send_size;
    for (int i_rank { 1 }; i_rank < mpi.n_procs; ++i_rank) {
        for (int i_img { mpi_ranges.at(i_rank) };
             i_img < mpi_ranges.at(i_rank + 1);
             i_img += mpi_send_size) {
            // Set every Mth value equal to M where
            // M=mpi_send_size. Be careful at process boundaries (see
            // the comment at mpi_buf_multiplicities).
            mpi_buf_multiplicities.at(i_img) =
              std::min(mpi_send_size, mpi_ranges.at(i_rank + 1) - i_img);
        }
    }
    // The user might have entered a large value for this
    // variable. Reduced it to the true maximum value here (this only
    // matters for output).
    mpi_send_size = std::min(mpi_send_size,
                             *std::max_element(mpi_buf_multiplicities.cbegin(),
                                               mpi_buf_multiplicities.cend()));
    // Initialize mpi_buffer_markers and send out the first
    // asynchronous MPI calls. Most MPI_Irecv calls are initiated at
    // the end of process_batch which is called multiple times.
    if (mpi.rank == 0) {
        for (int i_rank { 1 }; i_rank < mpi.n_procs; ++i_rank) {
            mpi_buffer_markers.at(i_rank) = mpi_ranges.at(i_rank);
        }
        static constexpr int n_buffer_rounds { 2 };
        for (int i {}; i < n_buffer_rounds; ++i) {
            for (int i_rank { 1 }; i_rank < mpi.n_procs; ++i_rank) {
                if (mpi_buffer_markers[i_rank] >= mpi_ranges[i_rank + 1]) {
                    break;
                }
                const int first_idx { mpi_buffer_markers.at(i_rank) };
                mpi_buffer_markers.at(i_rank) =
                  first_idx + mpi_buf_multiplicities.at(first_idx);
                mpi_buffers.at(first_idx).resize(
                  mpi_buf_multiplicities.at(first_idx) * mpi_buffer_size);
                if (!settings.l1b.disable_mpi_comm) {
                    MPI_Irecv(
                      mpi_buffers[first_idx].data(),
                      mpi_buf_multiplicities[first_idx] * mpi_buffer_size,
                      MPI_FLOAT,
                      i_rank,
                      first_idx,
                      MPI_COMM_WORLD,
                      &mpi_requests[first_idx]);
                }
            }
        }
    }
    // Buffer size per image in GB
    const double buf_in_GB { static_cast<double>(mpi_buffer_size)
                             * sizeof(float) / 1024 / 1024 / 1024 };
    // Typically, the MPI buffer size on each process is proportional
    // to two times mpi_send_size. Hence the word "Estimated" and the
    // factor of 2.
    spdlog::info("Estimated MPI buffer size  per process: {:.3f} GB",
                 2 * mpi_send_size * buf_in_GB);
    // Again we say "Estimated" here because the first process will
    // probably hold a slightly higher number of MPI buffers in memory.
    spdlog::info("Estimated MPI buffer size on process 0: {:.3f} GB ({:.1f}% of "
                 "the orbit)",
                 (mpi.n_procs - 1) * mpi_send_size * buf_in_GB,
                 1e2 * (mpi.n_procs - 1) * mpi_send_size / nL1A());
    if (mpi.rank == 0) {
        tmp_buf_dbl.resize(mpi_send_size * mpi_buffer_size);
        tmp_buf_float.resize(mpi_send_size * mpi_buffer_size);
    }
#endif
}

auto L1B::prepareBatches() -> void
{
    batchOne();
}

// Process one image
auto L1B::processBatch(const int i_img) -> void
{
    spdlog::info("Processing image {}/{}", i_img + 1, mpi_ranges.at(1));
    const auto& l1a { batches.at(i_img).l1a_products.front().get() };
    // All output to be written in the NetCDF
    std::vector<double> lat {};
    std::vector<double> lon {};
    std::vector<double> alt {};
    std::vector<double> vza {};
    std::vector<double> vaa {};
    std::vector<double> sza {};
    std::vector<double> saa {};
    // Detailed output
    std::vector<double> satpos_j2000 {}; // L1A content, but then in double array
    std::vector<double> rotation_quaternion {}; // L1A content, but then in double array
    double t {}; // To keep it in different if-geolocation scopes
    double greenwich_hour_angle {}; // This is a scalar, a new variable has to be made for unit conversion
    std::vector<double> satpos_ecr {};
    std::vector<double> sunvec_ecr {};
    std::vector<double> pointings_ecr {};
    std::vector<double> geoloc_ecr {};

    // Help array from geolocation that is required for the demodulation,
    // for transferring polarization to local meridian plane.
    std::vector<double> aolp_shift(ckd.fov.n_act);

    // 2. Construct the member arrays for output.
    // S-min and S-plus are used inside the L1A module, so are eventually
    // used for intermediate L1X output, so they are allocated when they
    // start making sense (after FOV calibration.
    if (do_geolocation) {
        lat.resize(ckd.fov.n_act,NC_FILL_DOUBLE);
        lon.resize(ckd.fov.n_act,NC_FILL_DOUBLE);
        alt.resize(ckd.fov.n_act,NC_FILL_DOUBLE);
        vza.resize(ckd.fov.n_act,NC_FILL_DOUBLE);
        vaa.resize(ckd.fov.n_act,NC_FILL_DOUBLE);
        sza.resize(ckd.fov.n_act,NC_FILL_DOUBLE);
        saa.resize(ckd.fov.n_act,NC_FILL_DOUBLE);
        // Detailed output
        if (settings.l1b.l1b_output_level >= 1) {
            satpos_j2000.resize(dims::vec,NC_FILL_DOUBLE);
            rotation_quaternion.resize(dims::quat,NC_FILL_DOUBLE);
            // No need to resize greenwich_hour_angle to size one. It is a double, not a std::vector<double>.
            satpos_ecr.resize(dims::vec,NC_FILL_DOUBLE);
            sunvec_ecr.resize(dims::vec,NC_FILL_DOUBLE);
            pointings_ecr.resize(ckd.fov.n_act*dims::vec,NC_FILL_DOUBLE);
            geoloc_ecr.resize(ckd.fov.n_act*dims::vec,NC_FILL_DOUBLE);
        }
        // Geolocation instructions that do not depend on the viewport or FOV

        // Execute the solar model.
        // Much of which comes from the Astronomical Almanac (1983 or 1984).

        // Calculate the sun position and the Greenwich hour angle. Both
        // need a complicated astronomical model from the Astronomical Almanac.
        // To keep the perspective as consistent as possible, we will calculate
        // the sun position in J2000 inside the routine, because our satellite
        // position is also still in J2000.

        // The solar model works for the Earth. If you have a
        // simulation where the planet is adapted, the sun still thinks it
        // is the Earth. Just imagine that the sun is orbiting the Earth,
        // not the other way around.

        // Get absolute date and time.
        // We have a number of seconds (int) and a fraction (double) after
        // 1970 midnight.
        int day = l1a.seconds/86400;
        double dayfract = (l1a.seconds % 86400 + l1a.second_fraction) / 86400.0;

        // If day is one, it is 2 January 1970
        int j = juliandate(1, 1, 1970) + day;
        // This is the day, an integer. Day fractions are not considered yet.

        // Piece of code from NASA

        // Find time difference at this day
        int mjd = j - 2400000;
        double tdiff = NC_FILL_DOUBLE;
        // Search this day
        for (size_t iutc=0 ; iutc<dim_utc ; iutc++) {
            if (mjd == mjd_utc[iutc]) {
                tdiff = tdiff_utc[iutc];
                break;
            }
        }
        if (tdiff == NC_FILL_DOUBLE) {
            // spdlog::warn("No UTC time difference found in file for this day. "
            //              "Continuing with zero time difference.");
            tdiff = 0.0;
        }
        dayfract += tdiff / 86400.0;

        // Compute the Sun vector in geocentric inertial
        // (equatorial) coodinates. It uses the model referenced in The
        // Astronomical Almanac for 1984, Section S (Supplement) and documented
        // for the SeaWiFS Project in "Constants and Parameters for SeaWiFS
        // Mission Operations", in TBD. The accuracy of the Sun vector is
        // approximately 0.1 arcminute.

        // Compute floating point days since Jan 1st, 2000 at noon in Greenwich.
        // The day fraction is from zero to one, so half at noon. That is
        // fixed with the point five in the constant number.
        t = j - 2451545.5 + dayfract;

        // Compute solar ephemeris parameters. The formulas are in degrees and
        // the modulo over 360 is calculated in advance to stay consistent with
        // common implementation. Changing the order of modulo and unit
        // conversion, or just execute sine or cosine without modulo over full
        // circles leads to different numeric errors.

        // Compute ephemeris parameters used by other Mission Operations
        // routines: the solar mean longitude and mean anomaly, and
        // the lunar mean longitude and mean ascending node. It uses the model
        // referenced in The Astronomical Almanac for 1984, Section S
        // (Supplement) and documented for the SeaWiFS Project in "Constants
        // and Parameters for SeaWiFS Mission Operations", in TBD. These
        // parameters are used to compute the solar longitude and the nutation
        // in longitude and obliquity.
        double xls = fmod(280.46592 + 0.9856473516*t , 360.0) * math::deg_to_rad; // Sun mean longitude
        double gs = fmod(357.52772 + 0.9856002831*t , 360.0) * math::deg_to_rad; // Sun mean anomaly
        double xlm = fmod(218.31643 + 13.17639648*t , 360.0) * math::deg_to_rad; // Moon mean longitude
        double omega = fmod(125.04452 - 0.0529537648*t , 360.0) * math::deg_to_rad; // Ascending node of Moon's mean orbit

        // Compute the nutation in longitude and the obliquity
        // of the ecliptic corrected for nutation. It uses the model referenced
        // in The Astronomical Almanac for 1984, Section S (Supplement) and
        // documented for the SeaWiFS Project in "Constants and Parameters for
        // SeaWiFS Mission Operations", in TBD. These parameters are used to
        // compute the apparent time correction to the Greenwich Hour Angle and
        // for the calculation of the geocentric Sun vector. The input
        // ephemeris parameters are computed using subroutine ephparms. Terms
        // are included to 0.1 arcsecond.

        // Nutation in longitude
        // Conversion from arcseconds to radians at the end
        double dpsi = (-17.1996*sin(omega) + 0.2062*sin(2.0*omega) - 1.3187*sin(2.0*xls) + 0.1426*sin(gs) - 0.2274*sin(2.0*xlm)) * math::deg_to_rad / 3600.0;

        // Obliquity of the ecliptic
        double epsm = (23.439291 - 3.560e-7*t) * math::deg_to_rad;
        double eps = epsm + (9.2025*cos(omega) + 0.5736*cos(2.0*xls)) * math::deg_to_rad / 3600.0;

        // Compute planet mean anomalies
        double g2 = fmod(50.40828 + 1.60213022*t , 360.0) * math::deg_to_rad; // Venus mean anomaly
        double g4 = fmod(19.38816 + 0.52402078*t , 360.0) * math::deg_to_rad; // Mars mean anomaly
        double g5 = fmod(20.35116 + 0.08309121*t , 360.0) * math::deg_to_rad; // Jupiter mean anomaly
        double rs = 1.00014 - 0.01671*cos(gs) - 0.00014*cos(2.0*gs); // Solar distance in astronomical units (AU)

        // Compute geometric solar longitude
        double xlsg = xls + (
            (6893.0 - 4.6543463e-4*t)*sin(gs)
            + 72.0*sin(2.0*gs)
            - 7.0*cos(gs - g5)
            + 6.0*sin(xlm - xls)
            + 5.0*sin(4.0*gs - 8.0*g4 + 3.0*g5)
            - 5.0*cos(2.0*gs - 2.0*g2)
            - 4.0*sin(gs - g2)
            + 4.0*cos(4.0*gs - 8.0*g4 + 3.0*g5)
            + 3.0*sin(2.0*gs - 2.0*g2)
            - 3.0*sin(g5)
            - 3.0*sin(2.0*gs - 2.0*g5)
        ) * math::deg_to_rad / 3600.0;

        // Compute apparent solar longitude includes corrections for nutation
        // in longitude and velocity aberration.
        double xk = 0.0056932 * math::deg_to_rad; // Constant of aberration
        double xlsa = xlsg + dpsi - xk/rs;

        // Compute unit sun vector in inertial frame
        Vector sun(cos(xlsa),sin(xlsa)*cos(eps),sin(xlsa)*sin(eps));

        // Compute Greenwich Mean Sidereal Time (degrees)
        double gmst = fmod(100.4606184 + 0.9856473663 * t + 2.908e-13 * pow(t,2.0) , 360.0)*math::deg_to_rad;

        // Include apparent time correction and time-of-day
        double gha = gmst + dpsi*cos(eps) + dayfract*360.0*math::deg_to_rad;
        // Transform Greenwich hour angle to rotation Quaternion
        Quaternion gha_rot(0.0,0.0,-sin(gha/2.0),cos(gha/2.0));
        gha_rot.intuitive_representation();

        // Transform the sun and the satellite into ECR. The nutation and
        // such is already done. Only the Greenwich hour angle has to be applied.
        sun = gha_rot.quaternion_rotate(sun);

        // Get full J2000 to ECR rotation. Unfortunately, the gha rotation
        // is only part of it.

        // First rotation: J2000 to mean-of-date
        double tcent = t / 36525.0; // J2000 time parameter in centuries
        double tcent2 = pow(tcent,2.0);
        double tcent3 = pow(tcent,3.0);
        double zeta0 = (2306.2181*tcent + 0.302*tcent2 + 0.018*tcent3)*math::deg_to_rad / 3600.0;
        double thetap = (2004.3109*tcent - 0.4266*tcent2 - 0.04160*tcent3)*math::deg_to_rad / 3600.0;
        double xip = (2306.2181*tcent + 1.095*tcent2 + 0.018*tcent3)*math::deg_to_rad / 3600.0;
        Quaternion qz(0.0,0.0,sin(zeta0/2.0),cos(zeta0/2.0)); // First
        Quaternion qx(0.0,0.0,sin(xip/2.0),cos(xip/2.0)); // Last
        Quaternion qt(0.0,-sin(thetap/2.0),0.0,cos(thetap/2.0)); // Middle
        Quaternion j2000_to_mod = qx*qt*qz;

        // Second rotation the nutation
        Quaternion qe(-sin(eps/2.0),0.0,0.0,cos(eps/2.0)); // First
        Quaternion qm(sin(epsm/2.0),0.0,0.0,cos(epsm/2.0)); // Last
        Quaternion qd(0.0,0.0,-sin(dpsi/2.0),cos(dpsi/2.0)); // Middle
        Quaternion nut = qm*qd*qe; // Must be reversed (conjugate)

        // The third rotation is the GHA angle
        Quaternion ecr = gha_rot*nut.conjug()*j2000_to_mod;

        Vector sat = l1a.satellite_position;

        // Start with having the satellite position in both lat/lon/alt and
        // in XYZ coordinates.
        double satlat;
        double satlon;
        double satalt;
        earth->lla(sat,&satlat,&satlon,&satalt);

        // Initialize the running pointers over FOV
        double* lat_cur = lat.data();
        double* lon_cur = lon.data();
        double* alt_cur = alt.data();
        double* vza_cur = vza.data();
        double* vaa_cur = vaa.data();
        double* sza_cur = sza.data();
        double* saa_cur = saa.data();
        double* swathvector_cur = ckd.swath_swathvectors.data();
        double* vectorplane_normal_cur = ckd.swath_vectorplane_normals.data();
        // Detailed output
        double* pointings_ecr_cur = pointings_ecr.data();
        double* geoloc_ecr_cur = geoloc_ecr.data();
        // Intermediate reults to be transferred to the demodulation
        double* aolp_shift_cur = aolp_shift.data();

        size_t ivp = 0;
        for (size_t i_fov=0 ; i_fov<ckd.fov.n_act ; i_fov++) {
            while (ivp != ckd.dim_vp-1 && i_fov == ckd.fov.start_vp[ivp+1]) {
                ivp++;
                vectorplane_normal_cur += dims::vec; // This is a vector per viewport, not per FOV
            }

            // Breakable do-once loop
            for (bool ctd=true ; ctd ; ctd=false) {
                // Get the swath vector into ECR coordinates. That is applying
                // the rotation from CKD to J2000 and from J2000 to ECR.
                Quaternion q = ecr*l1a.attitude_quaternion;
                Vector pointing = q.quaternion_rotate(Vector(swathvector_cur));
                // Also put the intended X-axis of the calibration reference frame
                // into ECR coordinartes. See 'local meridian plane'
                // conversion code.
                Vector crp_desired_y = q.quaternion_rotate(Vector(vectorplane_normal_cur));
                // Copy the satellite in both ECR and lat/lon/alt systems
                Vector pos_track(sat);
                double lat_track = satlat;
                double lon_track = satlon;
                double alt_track = satalt;

                // Altitude differences, for iterative conversion. In the beginning, the only
                // thing is that it should be positive.
                std::vector<double> altdiffs {satalt,satalt}; // This is now still altitude of LOS. Dem will be subtracted later on

                bool fail = false; // Turns true when the LOS misses earth

                // Step 1. Approach the earth to the DEM's reach
                while (alt_track > settings.l1b.mountain) {
                    double currentmove = std::max(alt_track - settings.l1b.mountain, static_cast<double>(settings.l1b.movedistance));
                    pos_track += currentmove*pointing;
                    earth->lla(pos_track,&lat_track,&lon_track,&alt_track);
                    if (alt_track > satalt) {
                        // Certainly misses earth: Unable to approach earth at altitude as low as the highest mountain
                        fail = true;
                        break;
                    }
                    altdiffs[0] = altdiffs[1];
                    altdiffs[1] = alt_track;
                }
                if (fail) break; // Go to next pixel

                // Now, the altitude differences make sense
                double dem_height;
                earth->dem(lat_track,lon_track,dem_height);
                altdiffs[0] -= dem_height;
                altdiffs[1] -= dem_height;

                // Step 2. Approach the elevated surface

                while (dem_height < alt_track) {
                    pos_track += settings.l1b.movedistance * pointing;
                    earth->lla(pos_track,&lat_track,&lon_track,&alt_track);
                    if (alt_track > settings.l1b.mountain) {
                        // Misses earth: Though a part of the line of sight is in the orography's domain, the surface is not visible.
                        fail = true;
                        break;
                    }
                    earth->dem(lat_track,lon_track,dem_height);
                    altdiffs[0] = altdiffs[1];
                    altdiffs[1] = alt_track-dem_height;
                }
                if (fail) break; // Go to next pixel

                // Step 3. Converge

                double mv = settings.l1b.movedistance;
                bool cur_right = true;
                while (mv > settings.l1b.geolocation_tol) {
                    double weightright = std::max(static_cast<double>(settings.l1b.extremeweight),std::min(1.0-settings.l1b.extremeweight,altdiffs[0]/(altdiffs[0]-altdiffs[1])));
                    if (cur_right) {
                        pos_track -= (1.0-weightright)*mv*pointing;
                    } else {
                        pos_track += weightright*mv*pointing;
                    }
                    earth->lla(pos_track,&lat_track,&lon_track,&alt_track);
                    earth->dem(lat_track,lon_track,dem_height);
                    double newaltdiff = alt_track-dem_height;
                    cur_right = newaltdiff < 0.0;
                    if (cur_right) {
                        altdiffs[1] = newaltdiff;
                        mv *= weightright;
                    } else {
                        altdiffs[0] = newaltdiff;
                        mv *= 1.0-weightright;
                    }
                }

                // Go to the middle of the two
                if (cur_right) {
                    pos_track -= 0.5*mv*pointing;
                } else {
                    pos_track += 0.5*mv*pointing;
                }

                // Detailed output: The geolocation in ECR
                if (settings.l1b.l1b_output_level >= 1) {
                    pos_track.get(geoloc_ecr_cur);
                }

                // Write the latitude, longitude and altitude to the output now
                earth->lla(pos_track,lat_cur,lon_cur,alt_cur);

                // Reference frame for zenith and azimuth angles
                Vector zen(cos(*lat_cur)*cos(*lon_cur),cos(*lat_cur)*sin(*lon_cur),sin(*lat_cur));
                Vector north(-sin(*lat_cur)*cos(*lon_cur),-sin(*lat_cur)*sin(*lon_cur),cos(*lat_cur));
                Vector east = north.cross(zen);

                // The reference plane for polarization in local meridian
                // plane is defined as:
                // Z = Light travel direction (+pointing)
                // X = Local zenith with Z-contributions taken out.
                // Y = Right handed coordinate system.
                Vector lmp_y = pointing.cross(zen); // y = z cross x, with x possibly contaminated with z
                lmp_y.normalize();
                Vector lmp_x = lmp_y.cross(pointing); // x = y cross z without contamination

                // In the highly exceptional case of exactly nadir,
                // the nornalized lmp_y will still be zero and hence,
                // lmp_x will be zero. The AoLP shift does not matter,
                // because there is no reference plane, so L2 people will
                // not model any polarization whatsoever.

                // For the calibration reference frame, an YZ plane
                // is defined, so that is X as normal, but the Z may
                // be slightly off the YZ-plane, because of imperfections,
                // So, the X needs to be set to the exact perpendicular part
                // of the normal.
                Vector crp_x = crp_desired_y.cross(pointing); // x = y cross z, with y possibly contaminated with z.
                crp_x.normalize();

                // The case of crp_desired_y to be in line with pointing
                // should be practically impossible, since the crp_desired_y
                // is normal on the plane fitted through all pointing vectors
                // in one viewport and they should be close to in a plane.
                // So, one should not be perpendicular to it, and surely not
                // exactly. Though, we will try to avoid any zero divisions even by this.
                double sin_aolp_shift = pointing.dot(lmp_x.cross(crp_x));
                double cos_aolp_shift = lmp_x.dot(crp_x);
                *aolp_shift_cur = atan2(sin_aolp_shift,cos_aolp_shift);

                // Viewing geometry
                *vza_cur = acos(zen.dot(-pointing));
                *vaa_cur = atan2(east.dot(-pointing),north.dot(-pointing));

                // Solar geometry
                *sza_cur = acos(zen.dot(sun));
                *saa_cur = atan2(east.dot(sun),north.dot(sun));

                // Convert units
                *lat_cur /= math::deg_to_rad;
                *lon_cur /= math::deg_to_rad;
                *vza_cur /= math::deg_to_rad;
                *vaa_cur /= math::deg_to_rad;
                *sza_cur /= math::deg_to_rad;
                *saa_cur /= math::deg_to_rad;

            }

            // Move pointers
            swathvector_cur += dims::vec;
            lat_cur++;
            lon_cur++;
            alt_cur++;
            vza_cur++;
            vaa_cur++;
            sza_cur++;
            saa_cur++;
            // Detailed output
            if (settings.l1b.l1b_output_level >= 1) {
                pointings_ecr_cur += dims::vec;
                geoloc_ecr_cur += dims::vec;
            }
            aolp_shift_cur++;
        }
    }
    // Perform demodulation
    l1a.timers[static_cast<int>(ProcLevel::pol)]->start();
    const int fov_dim_max { ckd.level >= ProcLevel::fov
                            ? *std::max_element(ckd.fov.dims_spec.cbegin(),
                                                ckd.fov.dims_spec.cend())
                            : 0 };
    const int nmeas_max { ckd.level >= ProcLevel::rad
                          ? *std::max_element(fov_nmeas.cbegin(),
                                              fov_nmeas.cend())
                          : 0 };
    // Radiance on the fine wavelength grid
    std::vector<double> radiance_fine {};
    std::vector<double> radiance_noise_fine {};
    std::vector<double> radiance {};
    std::vector<double> radiance_noise {};
    if (ckd.level >= ProcLevel::rad) {
        radiance_fine.resize(fov_dim_max, NC_FILL_DOUBLE);
        radiance_noise_fine.resize(fov_dim_max, NC_FILL_DOUBLE);
        radiance.resize(ckd.fov.n_act * dim_int_wave, NC_FILL_DOUBLE);
        radiance_noise.resize(ckd.fov.n_act * dim_int_wave, NC_FILL_DOUBLE);
    }
    std::vector<double> radiance_pol {};
    std::vector<double> small_q {};
    std::vector<double> small_u {};
    std::vector<double> dolp {};
    std::vector<double> aolp {};
    std::vector<double> radiance_noise_pol {};
    std::vector<double> small_q_noise {};
    std::vector<double> small_u_noise {};
    std::vector<double> dolp_noise {};
    std::vector<double> aolp_noise {};
    if (ckd.level >= ProcLevel::pol) {
        radiance_pol.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        small_q.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        small_u.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        dolp.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        aolp.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        radiance_noise_pol.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        small_q_noise.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        small_u_noise.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        dolp_noise.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
        aolp_noise.resize(ckd.fov.n_act * dim_pol_wave, NC_FILL_DOUBLE);
    }
    // Linear inversion covariance matrix
    std::vector<double> Sx(demod_size_full, NC_FILL_DOUBLE);
    // The q-quarter of Sx
    std::vector<double> Sxq(
      n_demod_states_half * n_demod_states_half, NC_FILL_DOUBLE);
    // The u-quarter of Sx
    std::vector<double> Sxu(
      n_demod_states_half * n_demod_states_half, NC_FILL_DOUBLE);
    // The qu-quarter of Sx
    std::vector<double> Sxqu(
      n_demod_states_half * n_demod_states_half, NC_FILL_DOUBLE);
    // (Sxq*B^T)^T where B is the B-spline matrix
    std::vector<double> Sxq_BT_T(
      nmeas_max * n_demod_states_half, NC_FILL_DOUBLE);
    // (Sxu*B^T)^T
    std::vector<double> Sxu_BT_T(
      nmeas_max * n_demod_states_half, NC_FILL_DOUBLE);
    // (Sxu*B^T)^T
    std::vector<double> Sxqu_BT_T(
      nmeas_max * n_demod_states_half, NC_FILL_DOUBLE);
    // Covariance matrices of q, u, and their cross-term on the fine
    // wavelength grid
    std::vector<double> Sq(nmeas_max * nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> Su(nmeas_max * nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> Squ(nmeas_max * nmeas_max, NC_FILL_DOUBLE);
    // Covariance matrices of Q, U, and their cross-terms, including
    // with I, on the fine wavelength grid
    std::vector<double> SQ(nmeas_max * nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> SU(nmeas_max * nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> SQU(nmeas_max * nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> SIQ(nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> SIU(nmeas_max, NC_FILL_DOUBLE);
    // Covariance matrices related to application of telescope
    std::vector<double> SQ_1(nmeas_max * nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> I_var_1(nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> SIQ_1(nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> SIU_1(nmeas_max, NC_FILL_DOUBLE);
    std::vector<double> SQU_1(nmeas_max * nmeas_max, NC_FILL_DOUBLE);
    // Modulation G values
    std::vector<double> modulation(nmeas_max);
    // G variances
    std::vector<double> modulation_var(nmeas_max);
    // Intensity variances
    std::vector<double> I_var(nmeas_max);
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        Spectra spectra {};
        l1a.extractSpectra(i_fov, spectra);
        if (ckd.level >= ProcLevel::rad) {
            for (int i {}; i < ckd.fov.dims_spec[i_fov]; ++i) {
                radiance_fine[i] = spectra.signal[pol_plus][i]
                                   + spectra.signal[pol_minus][i];
                radiance_noise_fine[i] = sqrt(spectra.noise[pol_minus][i]
                                              * spectra.noise[pol_minus][i]
                                              + spectra.noise[pol_plus][i]
                                              * spectra.noise[pol_plus][i]);
            }
        }
        if (ckd.level >= ProcLevel::pol) {
            const int nmeas { fov_nmeas[i_fov] };
            for (int imeas {}; imeas < nmeas; ++imeas) {
                const int i { fov_start_bin[i_fov] + imeas };
                const double m { spectra.signal[pol_minus][i] };
                const double p { spectra.signal[pol_plus][i] };
                modulation[imeas] = (p - m) / (p + m);
                const double ns_m { spectra.noise[pol_minus][i] };
                const double ns_p { spectra.noise[pol_plus][i] };
                const double p_m { p + m };
                const double p_m2 { p_m * p_m };
                modulation_var[imeas] = 4 * (m * m * ns_p * ns_p + p * p
                                             * ns_m * ns_m) / (p_m2 * p_m2);
                I_var[imeas] = ns_p * ns_p + ns_m * ns_m;
            }
            PackedMatrix spline_matrix_q { *demod_spline_matrix[i_fov] };
            PackedMatrix spline_matrix_u { *demod_spline_matrix[i_fov] };
            for (int i {}; i < nmeas; ++i) {
                for (int k {}; k < spline_matrix_q.row_sizes[i]; ++k) {
                    const int idx { spline_matrix_q.row_locations[i] + k };
                    spline_matrix_q.data[idx] *=
                      ckd.pol.m_q[i_fov][fov_start_bin[i_fov] + i]
                      / std::sqrt(modulation_var[i]);
                    spline_matrix_u.data[idx] *=
                      ckd.pol.m_u[i_fov][fov_start_bin[i_fov] + i]
                      / std::sqrt(modulation_var[i]);
                }
            }
            PackedMatrix spline_matrix_q_T { n_demod_states_half, nmeas };
            PackedMatrix spline_matrix_u_T { n_demod_states_half, nmeas };
            spline_matrix_q_T.transpose(spline_matrix_q);
            spline_matrix_u_T.transpose(spline_matrix_u);
            // BT*B is the linear inversion matrix. It is constructed
            // from four different segments.
            std::vector<double> BTB_11(demod_size_half);
            spline_matrix_q_T.multRightTransposed(spline_matrix_q_T, BTB_11);
            std::vector<double> BTB_12(demod_size_half);
            spline_matrix_q_T.multRightTransposed(spline_matrix_u_T, BTB_12);
            std::vector<double> BTB_22(demod_size_half);
            spline_matrix_u_T.multRightTransposed(spline_matrix_u_T, BTB_22);
            std::vector<double> BTB(demod_size_full);
            for (int i {}; i < n_demod_states_half; ++i) {
                for (int j {}; j < n_demod_states_half; ++j) {
                    BTB[i * n_demod_states + j] =
                      BTB_11[i * n_demod_states_half + j];
                }
            }
            for (int i {}; i < n_demod_states_half; ++i) {
                for (int j {}; j < n_demod_states_half; ++j) {
                    BTB[i * n_demod_states + j + n_demod_states_half] =
                      BTB_12[i * n_demod_states_half + j];
                }
            }
            for (int i {}; i < n_demod_states_half; ++i) {
                for (int j {}; j < n_demod_states_half; ++j) {
                    BTB[(i + n_demod_states_half) * n_demod_states + j] =
                      BTB_12[i * n_demod_states_half + j];
                }
            }
            for (int i {}; i < n_demod_states_half; ++i) {
                for (int j {}; j < n_demod_states_half; ++j) {
                    BTB[(i + n_demod_states_half) * n_demod_states + j
                        + n_demod_states_half] =
                      BTB_22[i * n_demod_states_half + j];
                }
            }
            // BT*Y is the B-spline matrix multiplied with the modulation data
            std::vector<double> BTY(n_demod_states, 0.0);
            for (int i {}; i < n_demod_states_half; ++i) {
                for (int k {}; k < spline_matrix_q_T.row_sizes[i]; ++k) {
                    const int packed_idx { spline_matrix_q_T.row_locations[i]
                                           + k };
                    const int col_idx { spline_matrix_q_T.global_col_indices[i]
                                        + k };
                    BTY[i] += spline_matrix_q_T.data[packed_idx]
                              * modulation[col_idx]
                              / std::sqrt(modulation_var[col_idx]);
                    BTY[i + n_demod_states_half] +=
                      spline_matrix_u_T.data[packed_idx]
                      * modulation[col_idx]
                      / std::sqrt(modulation_var[col_idx]);
                }
            }
            // Convert BTB into an upper triangular form
            std::vector<double> BTB_tri(n_demod_states * (n_demod_states + 1)
                                        / 2);
            for (int j {}; j < n_demod_states; ++j) {
                for (int i {}; i <= j; ++i) {
                    BTB_tri[i + (j+1)*j/2] = BTB[i * n_demod_states + j];
                }
            }
            // Do the linear inversion. x is the resulting q/u state
            // vector and Sx the corresponding covariance matrix.
            std::ranges::fill(Sx, 0.0);
            for (int i {}; i < n_demod_states; ++i) {
                Sx[i * n_demod_states + i] = 1.0;
            }
            std::vector<int> IPIV(n_demod_states);
            const char uplo { 'u' };
            int info {};
            dppsv_(&uplo,
                   &n_demod_states,
                   &n_demod_states,
                   BTB_tri.data(),
                   Sx.data(),
                   &n_demod_states,
                   &info);
            std::vector<double> x_result(n_demod_states, 0.0);
            for (int i {}; i < n_demod_states; ++i) {
                for (int k {}; k < n_demod_states; ++k) {
                    x_result[i] += Sx[i * n_demod_states + k] * BTY[k];
                }
            }
            // Divide Sx into sectors
            for (int i {}; i < n_demod_states_half; ++i) {
                for (int j {}; j < n_demod_states_half; ++j) {
                    Sxq[i * n_demod_states_half + j] =
                      Sx[i * n_demod_states + j];
                    Sxu[i * n_demod_states_half + j] =
                      Sx[(i + n_demod_states_half) * n_demod_states + j
                         + n_demod_states_half];
                    Sxqu[i * n_demod_states_half + j] =
                      Sx[i * n_demod_states + j + n_demod_states_half];
                }
            }
            // q and u values on the fine grid
            std::vector<double> q_fine(nmeas, 0.0);
            std::vector<double> u_fine(nmeas, 0.0);
            for (int i {}; i < nmeas; ++i) {
                for (int k {}; k < demod_spline_matrix[i_fov]->row_sizes[i];
                     ++k) {
                    const int spl_idx {
                        demod_spline_matrix[i_fov]->row_locations[i] + k
                    };
                    const int x_idx {
                        demod_spline_matrix[i_fov]->global_col_indices[i] + k
                    };
                    q_fine[i] += demod_spline_matrix[i_fov]->data[spl_idx]
                                 * x_result[x_idx];
                    u_fine[i] += demod_spline_matrix[i_fov]->data[spl_idx]
                                 * x_result[n_demod_states_half + x_idx];
                }
            }
            // q and u variances on the fine grid
            timer_Sx_BT.start();
            if (settings.l1b.noise_propagation) {
                demod_spline_matrix[i_fov]->multLeftTransposedT(
                  n_demod_states_half, Sxq, Sxq_BT_T);
                demod_spline_matrix[i_fov]->multLeftTransposedT(
                  n_demod_states_half, Sxu, Sxu_BT_T);
                demod_spline_matrix[i_fov]->multLeftTransposedT(
                  n_demod_states_half, Sxqu, Sxqu_BT_T);
            }
            timer_Sx_BT.stop();
            timer_Squ.start();
            if (settings.l1b.noise_propagation) {
                demod_spline_matrix[i_fov]->multRightTransposed(
                  nmeas, Sxq_BT_T, Sq);
                demod_spline_matrix[i_fov]->multRightTransposed(
                  nmeas, Sxu_BT_T, Su);
                demod_spline_matrix[i_fov]->multRightTransposed(
                  nmeas, Sxqu_BT_T, Squ);
            }
            timer_Squ.stop();
            // Q and U
            std::vector<double> Q_fine { q_fine };
            std::vector<double> U_fine { u_fine };
            for (int i {}; i < nmeas; ++i) {
                Q_fine[i] *= radiance_fine[fov_start_bin[i_fov] + i];
                U_fine[i] *= radiance_fine[fov_start_bin[i_fov] + i];
            }
            timer_SQU.start();
            if (settings.l1b.noise_propagation) {
                for (int i {}; i < nmeas; ++i) {
                    for (int j {}; j < nmeas; ++j) {
                        const double rad_prod {
                            radiance_fine[fov_start_bin[i_fov] + i]
                            * radiance_fine[fov_start_bin[i_fov] + j]
                        };
                        const int idx { i * nmeas + j };
                        SQ[idx] = Sq[idx] * rad_prod;
                        SU[idx] = Su[idx] * rad_prod;
                        SQU[idx] = Squ[idx] * rad_prod;
                    }
                    SQ[i * nmeas + i] += q_fine[i] * I_var[i] * q_fine[i];
                    SU[i * nmeas + i] += u_fine[i] * I_var[i] * u_fine[i];
                    SIQ[i] = q_fine[i] * I_var[i];
                    SIU[i] = u_fine[i] * I_var[i];
                    SQU[i * nmeas + i] += q_fine[i] * I_var[i] * u_fine[i];
                }
            }
            timer_SQU.stop();
            // Apply telescope polarization (Mueller element mt) to I and Q
            std::vector<double> mt2(nmeas);
            std::vector<double> mt22(nmeas);
            std::vector<double> mt22_inv(nmeas);
            std::vector<double> mt21_1_inv(nmeas);
            for (int i {}; i < nmeas; ++i) {
                const double mt {
                    ckd.pol.m_t[i_fov][fov_start_bin[i_fov] + i]
                };
                mt2[i] = mt * mt;
                mt22[i] = (1 - mt2[i]) * (1 - mt2[i]);
                mt22_inv[i] = 1.0 / mt22[i];
                mt21_1_inv[i] = 1.0 / (1.0 - mt2[i]);
            }
            std::vector<double> I_1(nmeas);
            for (int i {}; i < nmeas; ++i) {
                const double mt {
                    ckd.pol.m_t[i_fov][fov_start_bin[i_fov] + i]
                };
                const double I_1 {
                    1 / (1 - mt2[i])
                    * radiance_fine[fov_start_bin[i_fov] + i]
                    - ckd.pol.m_t[i_fov][fov_start_bin[i_fov] + i]
                    / (1 - mt2[i]) * Q_fine[i]
                };
                const double Q_fine_1 {
                    1 / (1 - mt2[i]) * Q_fine[i] - mt / (1 - mt2[i])
                    * radiance_fine[fov_start_bin[i_fov] + i]
                };
                radiance_fine[fov_start_bin[i_fov] + i] = I_1;
                Q_fine[i] = Q_fine_1;
            }
            if (settings.l1b.noise_propagation) {
                timer_Smt.start();
                for (int i {}; i < nmeas; ++i) {
                    const int row_start { i * nmeas };
                    for (int j {}; j < nmeas; ++j) {
                        const int idx { row_start + j };
                        SQ_1[idx] = SQ[idx] * mt22_inv[i];
                        SQU_1[idx] = SQU[idx] * mt21_1_inv[i];
                    }
                }
                for (int i {}; i < nmeas; ++i) {
                    const double mt { ckd.pol.m_t[i_fov][fov_start_bin[i_fov]
                                                        + i] };
                    SIQ_1[i] = -mt * SQ[i * nmeas + i] / mt22[i];
                    SIU_1[i] =
                      (SIU[i] - SQU[i * nmeas + i] * mt) / (1 - mt2[i]);
                    SQ_1[i * nmeas + i] += mt2[i] / mt22[i] * I_var[i];
                    SQ_1[i * nmeas + i] += -2 * mt * SIQ[i] / mt22[i];
                    I_var_1[i] = (I_var[i]+ SQ[i * nmeas + i] * mt2[i]
                                  - 2 * SIQ[i] * mt) / mt22[i];
                    SIQ_1[i] +=
                      (SIQ[i] - mt * I_var[i] + mt2[i] * SIQ[i]) / mt22[i];
                    SQU_1[i * nmeas + i] -= SIU[i] * mt * mt21_1_inv[i];
                }
                std::swap(SQ, SQ_1);
                std::swap(I_var, I_var_1);
                std::swap(SIQ, SIQ_1);
                std::swap(SIU, SIU_1);
                std::swap(SQU, SQU_1);
                timer_Smt.stop();
            }
            // I Q U convolved with Gaussian
            std::vector<double> Q_conv(dim_pol_wave, 0.0);
            std::vector<double> U_conv(dim_pol_wave, 0.0);
            std::vector<double> I_conv(dim_pol_wave, 0.0);
            std::vector<double> SI_conv(dim_pol_wave, 0.0);
            for (int i {}; i < dim_pol_wave; ++i) {
                for (int k {}; k < gauss_kernel[i_fov]->row_sizes[i]; ++k) {
                    const int spl_idx { gauss_kernel[i_fov]->row_locations[i]
                                        + k };
                    const int x_idx { gauss_kernel[i_fov]->global_col_indices[i]
                                      + k };
                    Q_conv[i] +=
                      gauss_kernel[i_fov]->data[spl_idx] * Q_fine[x_idx];
                    U_conv[i] +=
                      gauss_kernel[i_fov]->data[spl_idx] * U_fine[x_idx];
                    I_conv[i] += gauss_kernel[i_fov]->data[spl_idx]
                                 * radiance_fine[fov_start_bin[i_fov] + x_idx];
                    SI_conv[i] += gauss_kernel[i_fov]->data[spl_idx]
                                  * gauss_kernel[i_fov]->data[spl_idx]
                                  * I_var[x_idx];
                }
            }
            timer_SXT.start();
            std::vector<double> SQ_KT(nmeas * dim_pol_wave);
            std::vector<double> SU_KT(nmeas * dim_pol_wave);
            std::vector<double> SIQ_KT(nmeas * dim_pol_wave);
            std::vector<double> SIU_KT(nmeas * dim_pol_wave);
            std::vector<double> SQU_KT(nmeas * dim_pol_wave);
            if (settings.l1b.noise_propagation) {
                gauss_kernel[i_fov]->multLeftTransposed(nmeas, SQ, SQ_KT);
                gauss_kernel[i_fov]->multLeftTransposed(nmeas, SU, SU_KT);
                gauss_kernel[i_fov]->multLeftTransposed(nmeas, SQU, SQU_KT);
            }
            timer_SXT.stop();
            timer_S_QUI_conv.start();
            std::vector<double> SQ_conv(dim_pol_wave, 0.0);
            std::vector<double> SU_conv(dim_pol_wave, 0.0);
            std::vector<double> SIQ_conv(dim_pol_wave, 0.0);
            std::vector<double> SIU_conv(dim_pol_wave, 0.0);
            std::vector<double> SQU_conv(dim_pol_wave, 0.0);
            for (int i {}; i < dim_pol_wave; ++i) {
                for (int k {}; k < gauss_kernel[i_fov]->row_sizes[i]; ++k) {
                    const int spl_idx { gauss_kernel[i_fov]->row_locations[i]
                                        + k };
                    const int x_idx { gauss_kernel[i_fov]->global_col_indices[i]
                                      + k };
                    SQ_conv[i] += gauss_kernel[i_fov]->data[spl_idx]
                                  * SQ_KT[x_idx * dim_pol_wave + i];
                    SU_conv[i] += gauss_kernel[i_fov]->data[spl_idx]
                                  * SU_KT[x_idx * dim_pol_wave + i];
                    SQU_conv[i] += gauss_kernel[i_fov]->data[spl_idx]
                                   * SQU_KT[x_idx * dim_pol_wave + i];
                }
            }
            for (int i {}; i < dim_pol_wave; ++i) {
                for (int k {}; k < gauss_kernel[i_fov]->row_sizes[i]; ++k) {
                    const double g { gauss_kernel[i_fov]->data[
                          gauss_kernel[i_fov]->row_locations[i] + k] };
                    const int idx { gauss_kernel[i_fov]->global_col_indices[i]
                                    + k };
                    SIQ_conv[i] += g * g * SIQ[idx];
                    SIU_conv[i] += g * g * SIU[idx];
                }
            }
            timer_S_QUI_conv.stop();
            std::vector<double> q_conv(dim_pol_wave);
            std::vector<double> u_conv(dim_pol_wave);
            for (int i {}; i < dim_pol_wave; ++i) {
                q_conv[i] = Q_conv[i] / I_conv[i];
                u_conv[i] = U_conv[i] / I_conv[i];
            }
            std::vector<double> Sq_conv(dim_pol_wave);
            std::vector<double> Su_conv(dim_pol_wave);
            std::vector<double> SIq_conv(dim_pol_wave);
            std::vector<double> SIu_conv(dim_pol_wave);
            std::vector<double> Squ_conv(dim_pol_wave);
            for (int i {}; i < dim_pol_wave; ++i) {
                const double I_conv2 { I_conv[i] * I_conv[i] };
                const double q_conv2 { q_conv[i] * q_conv[i] };
                const double u_conv2 { u_conv[i] * u_conv[i] };
                Sq_conv[i] = (SQ_conv[i]
                              + SI_conv[i] * q_conv2
                              - 2 * SIQ_conv[i] * q_conv[i]) / I_conv2;
                Su_conv[i] = (SU_conv[i]
                              + SI_conv[i] * u_conv2
                              - 2 * SIU_conv[i] * u_conv[i]) / I_conv2;
                SIq_conv[i] = (SIQ_conv[i] - SI_conv[i] * q_conv[i])
                              / I_conv[i];
                SIu_conv[i] = (SIU_conv[i] - SI_conv[i] * u_conv[i])
                              / I_conv[i];
                Squ_conv[i] = (SQU_conv[i]
                               + SI_conv[i] * q_conv[i] * u_conv[i]
                               - SIU_conv[i] * q_conv[i]
                               - SIQ_conv[i] * u_conv[i])
                              / I_conv2;
            }
            std::vector<double> Sdolp(dim_pol_wave);
            std::vector<double> Saolp(dim_pol_wave);
            std::vector<double> Saolp_dolp(dim_pol_wave);
            const int pol_wave_idx_start { i_fov * dim_pol_wave };
            for (int i {}; i < dim_pol_wave; ++i) {
                dolp[pol_wave_idx_start + i] =
                  std::sqrt(q_conv[i] * q_conv[i] + u_conv[i] * u_conv[i]);
                aolp[pol_wave_idx_start + i] =
                  0.5 * std::atan2(u_conv[i], q_conv[i]);
                const double p2 { dolp[pol_wave_idx_start + i]
                                  * dolp[pol_wave_idx_start + i] };
                const double q_conv2 { q_conv[i] * q_conv[i] };
                const double u_conv2 { u_conv[i] * u_conv[i] };
                Sdolp[i] = (Sq_conv[i] * q_conv2 + Su_conv[i] * u_conv2
                            + 2 * Squ_conv[i] * q_conv[i] * u_conv[i]) / p2;
                Saolp[i] = (Su_conv[i] * q_conv2 + Sq_conv[i] * u_conv2
                            - 2 * Squ_conv[i] * q_conv[i] * u_conv[i])
                           / (4 * p2 * p2);
                Saolp_dolp[i] =
                  ((Su_conv[i] - Sq_conv[i]) * q_conv[i] * u_conv[i]
                   + Squ_conv[i] * (q_conv2 - u_conv2))
                  / (2 * p2 * dolp[pol_wave_idx_start + i]);
            }
            // Rotate q and u to the local meridian
            std::vector<double> q_conv_rot(dim_pol_wave);
            std::vector<double> u_conv_rot(dim_pol_wave);
            std::vector<double> Sq_conv_rot(dim_pol_wave);
            std::vector<double> Su_conv_rot(dim_pol_wave);
            std::vector<double> Squ_conv_rot(dim_pol_wave);
            for (int i {}; i < dim_pol_wave; ++i) {
                aolp[pol_wave_idx_start + i] += aolp_shift[i_fov];
                aolp[pol_wave_idx_start + i] -=
                  180.0 * math::deg_to_rad
                  * floor((aolp[pol_wave_idx_start + i] + 90.0
                           * math::deg_to_rad) / (180.0 * math::deg_to_rad));
                q_conv_rot[i] = dolp[pol_wave_idx_start + i]
                                * std::cos(2.0 * aolp[pol_wave_idx_start + i]);
                u_conv_rot[i] = dolp[pol_wave_idx_start + i]
                                * std::sin(2.0 * aolp[pol_wave_idx_start + i]);
                const double p2 { dolp[pol_wave_idx_start + i]
                                  * dolp[pol_wave_idx_start + i] };
                const double q_conv_rot2 { q_conv_rot[i] * q_conv_rot[i] };
                const double u_conv_rot2 { u_conv_rot[i] * u_conv_rot[i] };
                Sq_conv_rot[i] =
                  4 * Saolp[i] * u_conv_rot2 + Sdolp[i] * q_conv_rot2 / p2
                  - 4 * Saolp_dolp[i] * q_conv_rot[i] * u_conv_rot[i]
                  / dolp[pol_wave_idx_start + i];
                Su_conv_rot[i] =
                  4 * Saolp[i] * q_conv_rot2 + Sdolp[i] * u_conv_rot2 / p2
                  + 4 * Saolp_dolp[i] * q_conv_rot[i] * u_conv_rot[i]
                  / dolp[pol_wave_idx_start + i];
                Squ_conv_rot[i] =
                  -4 * Saolp[i] * q_conv_rot[i] * u_conv_rot[i]
                  + Sdolp[i] * q_conv_rot[i] * u_conv_rot[i] / p2
                  + 2 * Saolp_dolp[i] * (q_conv_rot2 - u_conv_rot2)
                  / dolp[pol_wave_idx_start + i];
            }
            for (int i {}; i < dim_pol_wave; ++i) {
                aolp[pol_wave_idx_start + i] /= math::deg_to_rad;
                Saolp[i] /= math::deg_to_rad * math::deg_to_rad;
            }
            for (int iwave {}; iwave < dim_pol_wave; ++iwave) {
                radiance_pol[pol_wave_idx_start + iwave] = I_conv[iwave];
                radiance_noise_pol[pol_wave_idx_start + iwave] = SI_conv[iwave];
                small_q[pol_wave_idx_start + iwave] = q_conv_rot[iwave];
                small_q_noise[pol_wave_idx_start + iwave] =
                  std::sqrt(Sq_conv_rot[iwave]);
                small_u[pol_wave_idx_start + iwave] = u_conv_rot[iwave];
                small_u_noise[pol_wave_idx_start + iwave] =
                  std::sqrt(Su_conv_rot[iwave]);
                dolp_noise[pol_wave_idx_start + iwave] =
                  std::sqrt(Sdolp[iwave]);
                aolp_noise[pol_wave_idx_start + iwave] =
                  std::sqrt(Saolp[iwave]);
            }
        }
        if (ckd.level >= ProcLevel::rad) {
            LinearSpline radiance_spline { ckd.wave_common[i_fov],
                                           radiance_fine };
            LinearSpline radiance_noise_spline { ckd.wave_common[i_fov],
                                                 radiance_noise_fine };
            for (int i {}; i < dim_int_wave; ++i) {
                radiance[i_fov * dim_int_wave + i] =
                  radiance_spline.eval(settings.l1b.intensity_wavelength[i]);
                radiance_noise[i_fov * dim_int_wave + i] =
                  radiance_noise_spline.eval(
                    settings.l1b.intensity_wavelength[i]);
            }
        }
    }
    l1a.timers[static_cast<int>(ProcLevel::pol)]->stop();
    // 5. Write it into the output
    if (mpi.rank == 0 && !settings.l1b.disable_io) {
        const std::vector<size_t> strt_img { static_cast<size_t>(i_img) };
        const std::vector<size_t> cnt_img { 1 };
        var_time.putVar(strt_img,cnt_img,&l1a.image_time);
        // Start indices for some NetCDF variables when calling putVar
        const std::vector<size_t> startp { static_cast<size_t>(i_img), 0, 0 };
        if (ckd.level >= ProcLevel::rad) {
            const std::vector<size_t> counts {
                1,
                static_cast<size_t>(ckd.fov.n_act),
                static_cast<size_t>(dim_int_wave)
            };
            var_radiance.putVar(startp, counts, radiance.data());
            var_radiance_noise.putVar(startp, counts, radiance_noise.data());
        }
        if (ckd.level >= ProcLevel::pol) {
            const std::vector<size_t> counts {
                1,
                static_cast<size_t>(ckd.fov.n_act),
                static_cast<size_t>(dim_pol_wave)
            };
            var_radiance_pol.putVar(startp, counts, radiance_pol.data());
            var_small_q.putVar(startp, counts, small_q.data());
            var_small_u.putVar(startp, counts, small_u.data());
            var_dolp.putVar(startp, counts, dolp.data());
            var_aolp.putVar(startp, counts, aolp.data());
            var_radiance_noise_pol.putVar(
              startp, counts, radiance_noise_pol.data());
            var_small_q_noise.putVar(startp, counts, small_q_noise.data());
            var_small_u_noise.putVar(startp, counts, small_u_noise.data());
            var_dolp_noise.putVar(startp, counts, dolp_noise.data());
            var_aolp_noise.putVar(startp, counts, aolp_noise.data());
        }
        if (do_geolocation) {
            const std::vector<size_t> counts {
                1, static_cast<size_t>(ckd.fov.n_act)
            };
            const std::vector<size_t> startp { static_cast<size_t>(i_img), 0 };
            var_lat.putVar(startp,counts,lat.data());
            var_lon.putVar(startp,counts,lon.data());
            var_alt.putVar(startp,counts,alt.data());
            var_vza.putVar(startp,counts,vza.data());
            var_vaa.putVar(startp,counts,vaa.data());
            var_sza.putVar(startp,counts,sza.data());
            var_saa.putVar(startp,counts,saa.data());
            // Write detailed output
            if (settings.l1b.l1b_output_level >= 1) {
                const std::vector<size_t> cnt_geo_vec { 1, dims::vec };
                const std::vector<size_t> cnt_geo_quat { 1, dims::quat };
                const std::vector<size_t> cnt_geo_point {
                    1, static_cast<size_t>(ckd.fov.n_act), dims::vec
                };
                const std::vector<size_t> cnt_obs_only {
                    static_cast<size_t>(1)
                };
                const std::vector<size_t> strt_geo_vec {
                    static_cast<size_t>(i_img), 0
                };
                const std::vector<size_t> strt_geo_quat {
                    static_cast<size_t>(i_img), 0
                };
                const std::vector<size_t> strt_geo_point {
                    static_cast<size_t>(i_img), 0, 0
                };
                const std::vector<size_t> strt_obs_only {
                    static_cast<size_t>(i_img)
                };
                var_satpos_j2000.putVar(strt_geo_vec,cnt_geo_vec,satpos_j2000.data());
                var_rotation_quaternion.putVar(strt_geo_quat,cnt_geo_quat,rotation_quaternion.data());
                var_t.putVar(strt_obs_only,cnt_obs_only,&t);
                var_greenwich_hour_angle.putVar(strt_obs_only,cnt_obs_only,&greenwich_hour_angle);
                var_satpos_ecr.putVar(strt_geo_vec,cnt_geo_vec,satpos_ecr.data());
                var_sunvec_ecr.putVar(strt_geo_vec,cnt_geo_vec,sunvec_ecr.data());
                var_pointings_ecr.putVar(strt_geo_point,cnt_geo_point,pointings_ecr.data());
                var_geoloc_ecr.putVar(strt_geo_point,cnt_geo_point,geoloc_ecr.data());
            }
        }
        nc_l1b.sync();
    }
#ifdef USE_MPI
    if (mpi.rank > 0) {
        // Start the send process to process 0. Note that
        // mpi_buf_multiplicities is often zero so a buffer is
        // actually only allocated at selected iterations.
        mpi_buffers[i_img].resize(mpi_buf_multiplicities[i_img]
                                 * mpi_buffer_size);
        // The current and maximum allocation size is recorded only
        // for output purposes.
        cur_buf_size += mpi_buffers[i_img].size();
        max_buf_size = std::max(max_buf_size, cur_buf_size);
        // Distance between this iteration and the last iteration when
        // a buffer was allocated. If mpi_buf_multiplicities is zero
        // for this and the previous round but was non-zero two
        // iterations ago then i_shift=2.
        const int i_shift { (static_cast<int>(i_img) - mpi_ranges[mpi.rank])
                            % mpi_send_size };
        // Index of the last iteration when a buffer was allocated. If
        // a buffer was allocated on this round,
        // i.e. mpi_buf_multiplicities is non-zero for this round then
        // i_active_idx=i_img.
        const int i_active_idx { static_cast<int>(i_img) - i_shift };
        // Where to start writing targeted variables (DoLP, AoLP, ...)
        // in the MPI buffer. There might already be data from
        // previous iterations so need to shift the starting index
        // accordingly.
        const int buf_start_idx { i_shift * mpi_buffer_size };
        auto buf_start { mpi_buffers[i_active_idx].begin() + buf_start_idx };
        // Main loop where we copy each netCDF variable into the MPI
        // buffer one by one.
        for (int i_var {}; i_var < static_cast<int>(mpi_nc_vars.size());
             ++i_var) {
            const auto copy { [this, &buf_start, i_var](const auto& source) {
                std::transform(source.cbegin(),
                               source.cend(),
                               buf_start + mpi_buffer_starts[i_var],
                               [](const double x) {
                                   return static_cast<float>(x);
                               });
            } };
            switch (mpi_nc_vars[i_var].first) {
            case MpiNcVarId::time:
                // We are copying a double precision variable into a
                // single precision buffer.
                std::memcpy(&mpi_buffers[i_active_idx][
                              buf_start_idx + mpi_buffer_starts[i_var]],
                            &l1a.image_time,
                            sizeof(double));
                break;
            case MpiNcVarId::radiance:
                copy(radiance);
                break;
            case MpiNcVarId::radiance_noise:
                copy(radiance_noise);
                break;
            case MpiNcVarId::radiance_pol:
                copy(radiance_pol);
                break;
            case MpiNcVarId::radiance_noise_pol:
                copy(radiance_noise_pol);
                break;
            case MpiNcVarId::small_q:
                copy(small_q);
                break;
            case MpiNcVarId::small_q_noise:
                copy(small_q_noise);
                break;
            case MpiNcVarId::small_u:
                copy(small_u);
                break;
            case MpiNcVarId::small_u_noise:
                copy(small_u_noise);
                break;
            case MpiNcVarId::dolp:
                copy(dolp);
                break;
            case MpiNcVarId::dolp_noise:
                copy(dolp_noise);
                break;
            case MpiNcVarId::aolp:
                copy(aolp);
                break;
            case MpiNcVarId::aolp_noise:
                copy(aolp_noise);
                break;
            case MpiNcVarId::lat:
                copy(lat);
                break;
            case MpiNcVarId::lon:
                copy(lon);
                break;
            case MpiNcVarId::alt:
                copy(alt);
                break;
            case MpiNcVarId::vza:
                copy(vza);
                break;
            case MpiNcVarId::vaa:
                copy(vaa);
                break;
            case MpiNcVarId::sza:
                copy(sza);
                break;
            case MpiNcVarId::saa:
                copy(saa);
                break;
            }
        }
        // If this is the final iteration for the given MPI process or
        // for the next iteration mpi_buf_multiplicities is non-zero
        // again, it's time to send the buffer contents to process 0.
        if ((i_img == mpi_ranges[mpi.rank + 1] - 1
             || mpi_buf_multiplicities[i_img + 1] > 0)
            && !settings.l1b.disable_mpi_comm) {
            MPI_Isend(mpi_buffers[i_active_idx].data(),
                      mpi_buf_multiplicities[i_active_idx] * mpi_buffer_size,
                      MPI_FLOAT,
                      0,
                      i_active_idx,
                      MPI_COMM_WORLD,
                      &mpi_requests[i_active_idx]);
        }
        // Check if process 0 has recieved any of the previous
        // requests. If so, deallocate those MPI buffers.
        for (int i_img { mpi_ranges[mpi.rank] };
             i_img < mpi_ranges[mpi.rank + 1];
             i_img += mpi_send_size) {
            if (mpi_requests[i_img] != MPI_REQUEST_NULL) {
                MPI_Status status {};
                int flag {};
                if (!settings.l1b.disable_mpi_comm) {
                    MPI_Test(&mpi_requests[i_img], &flag, &status);
                }
                if (flag || settings.l1b.disable_mpi_comm) {
                    cur_buf_size -= mpi_buffers[i_img].size();
                    mpi_buffers[i_img] = std::vector<float>{};
                }
            }
        }
    } else {
        mpi_io_timer.start();
        const bool last_round { i_img == mpi_ranges[1] - 1 };
        if (last_round) {
            mpi_final_io_timer_total.start();
        }
        // Tests if any data was received in the main double loop
        bool active_run {};
        do {
            active_run = false;
            for (int i_rank { 1 }; i_rank < mpi.n_procs; ++i_rank) {
                for (int i_img { mpi_ranges[i_rank] };
                     i_img < mpi_ranges[i_rank + 1];
                     i_img += mpi_send_size) {
                    // MPI_REQUEST_NULL signals that data has already
                    // been received from that buffer. Finish this
                    // iteration.
                    if (mpi_requests[i_img] == MPI_REQUEST_NULL) {
                        continue;
                    }
                    active_run = true;
                    MPI_Status status;
                    int flag {};
                    // Test if the MPI_Isend operation for image i_img
                    // has been completed.
                    if (!settings.l1b.disable_mpi_comm) {
                        MPI_Test(&mpi_requests[i_img], &flag, &status);
                    }
                    // If not, try again later but finish this iteration
                    if (!flag || settings.l1b.disable_mpi_comm) {
                        continue;
                    }
                    if (last_round) {
                        mpi_final_io_timer_active.start();
                    }
                    // MPI buffer i_img contains data. Write that data
                    // to output file. First treat image_time which is
                    // the only double-precision variable.
                    for (int i_count {};
                         i_count < mpi_buf_multiplicities[i_img];
                         ++i_count) {
                        // i_count iterates over the number of images
                        // received in this batch.
                        const int start_buf { i_count * mpi_buffer_size
                                              + mpi_buffer_starts.front() };
                        std::memcpy(&tmp_buf_dbl[i_count],
                                    &mpi_buffers[i_img][start_buf],
                                    sizeof(double));
                    }
                    mpi_nc_starts.front().front() = i_img;
                    mpi_nc_counts.front().front() =
                      mpi_buf_multiplicities[i_img];
                    if (!settings.l1b.disable_io) {
                        mpi_nc_vars.front().second.putVar(
                          mpi_nc_starts.front(),
                          mpi_nc_counts.front(),
                          tmp_buf_dbl.data());
                    }
                    // Next treat all the other variables, which are
                    // single-precision.
                    for (int i_var { 1 };
                         i_var < static_cast<int>(mpi_nc_vars.size());
                         ++i_var) {
                        for (int i_count {};
                             i_count < mpi_buf_multiplicities[i_img];
                             ++i_count) {
                            const int var_size { mpi_buffer_starts[i_var + 1]
                                                 - mpi_buffer_starts[i_var] };
                            const int start_tmp { i_count * var_size };
                            const int start_buf { i_count * mpi_buffer_size
                                                  + mpi_buffer_starts[i_var] };
                            for (int i {}; i < var_size; ++i) {
                                tmp_buf_float[i + start_tmp] =
                                  mpi_buffers[i_img][i + start_buf];
                            }
                        }
                        mpi_nc_starts[i_var].front() =
                          i_img;
                        mpi_nc_counts[i_var].front() =
                          mpi_buf_multiplicities[i_img];
                        if (!settings.l1b.disable_io) {
                            mpi_nc_vars[i_var].second.putVar(
                              mpi_nc_starts[i_var],
                              mpi_nc_counts[i_var],
                              tmp_buf_float.data());
                        }
                    }
                    // Mark operation as complete (see the conditional
                    // at the beginning of this loop).
                    mpi_requests[i_img] = MPI_REQUEST_NULL;
                    // Free the used up MPI buffer and continue to
                    // allocate a new buffer and send out a
                    // corresponding asynchronous receive request.
                    mpi_buffers[i_img] = std::vector<float>{};
                    // If the buffer marker has passed the image range
                    // assigned to a given process then all buffers
                    // have already been allocated.
                    if (mpi_buffer_markers[i_rank] >= mpi_ranges[i_rank + 1]) {
                        if (last_round) {
                            mpi_final_io_timer_active.stop();
                        }
                        continue;
                    }
                    const int next_img { mpi_buffer_markers[i_rank] };
                    mpi_buffers[next_img].resize(mpi_buf_multiplicities[i_img]
                                                 * mpi_buffer_size);
                    MPI_Irecv(mpi_buffers[next_img].data(),
                              mpi_buf_multiplicities[i_img] * mpi_buffer_size,
                              MPI_FLOAT,
                              i_rank,
                              next_img,
                              MPI_COMM_WORLD,
                              &mpi_requests[next_img]);
                    mpi_buffer_markers[i_rank] += mpi_buf_multiplicities[i_img];
                    if (last_round) {
                        mpi_final_io_timer_active.stop();
                    }
                }
            }
            // The double loop runs as many times as necessary until
            // there are no more active MPI requests and all data has
            // been received by process 0.
        } while (last_round && active_run && !settings.l1b.disable_mpi_comm);
        if (last_round) {
            mpi_final_io_timer_total.stop();
        }
        mpi_io_timer.stop();
    }
#endif
}

// Finalize (write the creation time)
auto L1B::processFinalize() -> void
{
#ifdef USE_MPI
    const int n_peak_images { static_cast<int>(max_buf_size
                                               / mpi_buffer_size) };
    std::vector<int> n_peak_images_all(mpi.n_procs, 0);
    gadfit::gatherScalar(mpi, n_peak_images, n_peak_images_all);
    for (int i_rank { 1 }; i_rank < mpi.n_procs; ++i_rank) {
        spdlog::info("Maximum number of images held in the MPI buffer "
                     "on process {:3}: {}",
                     i_rank,
                     n_peak_images_all.at(i_rank));
    }
    spdlog::info("MPI communication time on process 0: {:8.3f} s",
                 mpi_io_timer.totalWallTime());
    spdlog::info("      MPI waiting time on process 0: {:8.3f} s",
                 mpi_final_io_timer_total.totalWallTime()
                 - mpi_final_io_timer_active.totalWallTime());
#endif
    spdlog::debug("Demodulation error propagation timings:");
    spdlog::debug("           Sx_BT: {:8.3f} s ({} calls)",
                  timer_Sx_BT.totalWallTime(),
                  timer_Sx_BT.totalNumberOfCalls());
    spdlog::debug("             Squ: {:8.3f} s ({} calls)",
                  timer_Squ.totalWallTime(),
                  timer_Squ.totalNumberOfCalls());
    spdlog::debug("             SQU: {:8.3f} s ({} calls)",
                  timer_SQU.totalWallTime(),
                  timer_SQU.totalNumberOfCalls());
    spdlog::debug("             Smt: {:8.3f} s ({} calls)",
                  timer_Smt.totalWallTime(),
                  timer_Smt.totalNumberOfCalls());
    spdlog::debug("             SXT: {:8.3f} s ({} calls)",
                  timer_SXT.totalWallTime(),
                  timer_SXT.totalNumberOfCalls());
    spdlog::debug("      S_QUI_conv: {:8.3f} s ({} calls)",
                  timer_S_QUI_conv.totalWallTime(),
                  timer_S_QUI_conv.totalNumberOfCalls());
}

} // namespace spex
