// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include "packed_matrix.h"
#include "planet.h"
#include "processor.h"

#include <cstdint>

#ifdef USE_MPI
#include <mpi.h>
#endif

#include <vector>

namespace spex {

class L1B : public Processor
{
public:
    L1B(const gadfit::MPIVars& mpi, const SPEXSettings& settings, CKD& ckd);
    ~L1B() = default;

protected:
    auto processInit() -> void override;
    // Each image is in its own batch
    auto prepareBatches() -> void override;
    auto processBatch(const int i_img) -> void override;
    auto processFinalize() -> void override;

private:
    // If geolocation was requested ([l1b][geolocation]) but the CKD
    // does not contain swath vectors it cannot be performed. We leave
    // the value of [l1b][geolocation] but use a custom flag to signal
    // that geolocation should be skipped.
    bool do_geolocation { true };
    // Binned spectral index marking the beginning of the retrieval
    // window per FOV (determined by [l1b][l1b_wavelength])
    std::vector<int> fov_start_bin {};
    // Number of measurements in the retrieval window per FOV
    std::vector<int> fov_nmeas {};
    // Number of wavelenghts in (polarization) L1B output
    int dim_pol_wave {};
    // Number of wavelengths in intensity only L1B output
    int dim_int_wave {};
    // Number of linear inversion states when doing a B-spline fit is
    // close to twice the number of knots because both q and u are
    // included.
    int n_demod_states {};
    int n_demod_states_half {};
    // Size of matrices with dimensions (n_demod_states, n_demod_states)
    int demod_size_full {};
    int demod_size_half {};
    std::vector<std::unique_ptr<PackedMatrix>> demod_spline_matrix {};
    // Demodulation Gaussian kernel matrix
    std::vector<std::unique_ptr<PackedMatrix>> gauss_kernel {};
    // Output
    NcObject nc_l1b { netCDF::NcFile::replace };
    netCDF::NcVar var_time {};
    // Raw Radiance
    netCDF::NcVar var_radiance {};
    netCDF::NcVar var_radiance_noise {};
    // Demodulation
    netCDF::NcVar var_radiance_pol {};
    netCDF::NcVar var_radiance_noise_pol {};
    netCDF::NcVar var_small_q {};
    netCDF::NcVar var_small_q_noise {};
    netCDF::NcVar var_small_u {};
    netCDF::NcVar var_small_u_noise {};
    netCDF::NcVar var_dolp {};
    netCDF::NcVar var_dolp_noise {};
    netCDF::NcVar var_aolp {};
    netCDF::NcVar var_aolp_noise {};
    // Geolocation
    netCDF::NcVar var_lat {};
    netCDF::NcVar var_lon {};
    netCDF::NcVar var_alt {};
    // Geometry
    netCDF::NcVar var_vza {};
    netCDF::NcVar var_vaa {};
    // Solar geometry
    netCDF::NcVar var_sza {};
    netCDF::NcVar var_saa {};
    // Detailed output
    netCDF::NcGroup grp_detailed_output_geolocation {};
    netCDF::NcVar var_satpos_j2000 {};
    netCDF::NcVar var_rotation_quaternion {};
    netCDF::NcVar var_t {};
    netCDF::NcVar var_greenwich_hour_angle {};
    netCDF::NcVar var_satpos_ecr {};
    netCDF::NcVar var_sunvec_ecr {};
    netCDF::NcVar var_pointings_ecr {};
    netCDF::NcVar var_geoloc_ecr {};
    // UTC file contents (for geolocation)
    size_t dim_utc {};
    std::vector<int> mjd_utc {};
    std::vector<double> tdiff_utc {};

    // The planet
    std::unique_ptr<Planet> earth {};

    // We use these enums to connect netCDF variables with the
    // corresponding data, e.g. var_dolp and dolp.
    enum class MpiNcVarId
    {
        time,
        radiance,
        radiance_noise,
        radiance_pol,
        radiance_noise_pol,
        small_q,
        small_q_noise,
        small_u,
        small_u_noise,
        dolp,
        dolp_noise,
        aolp,
        aolp_noise,
        lat,
        lon,
        alt,
        vza,
        vaa,
        sza,
        saa,
    };
#ifdef USE_MPI
    // These netCDF variables are targeted for MPI parallelization
    std::vector<std::pair<MpiNcVarId, netCDF::NcVar>> mpi_nc_vars {};
    // Each MPI buffer in mpi_buffers corresponds to one image,
    // although not all nl1a_total images are allocated at the same
    // time. Each MPI buffer contains the data defined by mpi_nc_vars,
    // i.e. variables such as DoLP, AoLP, etc. are serialized into a
    // single buffer per iteration (image).
    std::vector<std::vector<float>> mpi_buffers {};
    // The starting index of each quantity (DoLP, AoLP, ...) in an MPI buffer
    std::vector<int> mpi_buffer_starts {};
    // When writing data to file, defines the starting indices of each
    // netCDF variable (the startp argument of netCDF::NcVar::putVar)
    std::vector<std::vector<size_t>> mpi_nc_starts {};
    // When writing data to file, defines the number of indices along
    // each dimension for the netCDF variables (the countp argument of
    // netCDF::NcVar::putVar)
    std::vector<std::vector<size_t>> mpi_nc_counts {};
    // Size of an MPI buffer per image. Note that each member of
    // mpi_buffer is resized to an integer multiple (greater than 1)
    // of this variable because it's more efficient to send multiple
    // images in a batch, i.e. not initiate an asynchronous MPI
    // command after each iteration.
    int mpi_buffer_size {};
    // For process 0 (my_rank=0), keeps track of images for which the
    // data has already been received. The total number of those
    // images is nl1a_total minus those assigned to process 0. For all
    // other processes keeps track of MPI_Isend operations that have
    // successfully completed and thus signals that the corresponding
    // memory may be released.
    std::vector<MPI_Request> mpi_requests {};
    // Keeps track of the highest image index for which a buffer has
    // been allocated for the given MPI process. For instance, with
    // 9000 images and 20 MPI processes, let's say process 15 has been
    // assigned images 6750...7200 but only the MPI buffers for images
    // 6750...7014 have been allocated because it's too expensive to
    // keep the whole orbit in memory. Then
    // mpi_buffer_markers[14]=7014. This value is increased as the
    // calculation progresses until all buffers have been allocated
    // and all data transmitted to process 0.
    std::vector<int> mpi_buffer_markers {};
    int mpi_send_size { 1 };
    // Determines the size of the MPI buffer allocated on a given
    // iteration inside the main loop over images. Its values are
    // typically either mpi_send_size or 0. In the following example
    // the columns are R - MPI rank, N - image index, M -
    // mpi_buf_multiplicities[N].
    //
    // R  N  M
    // ...
    // 1 34  0
    // 2 35  3 # Allocate MPI_buffer with size 3 * mpi_buffer_size
    // 2 36  0 # Do not allocate new buffer; write data into previously
    // 2 37  0 #    allocated buffer
    // 2 38  2 # Allocate MPI_buffer with size 2 * mpi_buffer_size and not
    // 2 39  0 #    2 * mpi_buffer_size because there aren't enough images left.
    // 3 40  3
    // ...
    //
    // For image 35 an MPI buffer is allocated with size 3 *
    // mpi_buffer_size (meaning mpi_send_size=3). In subsequent
    // iterations a new buffer is not allocated and instead data is
    // written into the previously allocated buffer. At the end of
    // iteration 37 data is sent to process 0 and soon after the
    // buffer is released. At iteration 38 a new buffer is allocated
    // but with the size 2 * mpi_buffer_size because there are only
    // two images left (image #40 is assigned to the next MPI
    // process).
    std::vector<int> mpi_buf_multiplicities {};
    // Keeps track of the time spent by process 0 collecting data from
    // other processes.
    gadfit::Timer mpi_io_timer {};
    // Keeps track of the time spent by process 0 in the final step
    // collecting data from other processes.
    gadfit::Timer mpi_final_io_timer_total {};
    // Same but avoids dead time and only records time spent on MPI
    // communication. The difference of these two timers yields the
    // time spent by process 0 on waiting for the other processes to
    // finish. If it is large then l1b.first_proc_rel_workload should
    // be increased.
    gadfit::Timer mpi_final_io_timer_active {};
    // Current number of MPI buffers allocated on this MPI process
    size_t cur_buf_size {};
    // Maximum number of MPI buffers allocated on this MPI process
    size_t max_buf_size {};
    // In mpi_ranges, data corresponding to different images is stored
    // sequentially. This means that all data corresponding to a
    // single variable such as dolp is stored non-contiguously. We use
    // tmp_buf_dbl/float for rearranging the data layout before
    // writing it to file.
    std::vector<double> tmp_buf_dbl {};
    std::vector<float> tmp_buf_float {};
#endif

    // Timers for the most expensive demodulation steps
    gadfit::Timer timer_Sx_BT {};
    gadfit::Timer timer_Squ {};
    gadfit::Timer timer_SQU {};
    gadfit::Timer timer_Smt {};
    gadfit::Timer timer_SXT {};
    gadfit::Timer timer_S_QUI_conv {};
};

} // namespace spex
