// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1c.h"

#include "planet.h"
#include "vector.h"

#include <cassert>
#include <cmath>
#include <spdlog/spdlog.h>

namespace spex {

static auto getSide(Vector& line1, Vector& line2, Vector& tar, Vector& zen)
  -> double
{
    Vector leg1 = tar-line1;
    Vector leg2 = line2-line1;
    return zen.dot(leg1.cross(leg2));
}

auto L1C::processInit() -> void
{
    size_t dim_l1c_obs = settings.l1c.lat.size();

    // This process has no loop over images or something like that. So
    // everything is in images. The batch is absent, because there are no
    // L1A files.

    // This process init function is the whole algorithm.

    // Create target grid, using the Earth and the selected points.
    // TODO: Make a more convenient grid generator or selector.

    // Read L1B along-track dimension
    NcObject nc_l1b { netCDF::NcFile::read };
    nc_l1b.open(settings.l1c.l1bfile);

    // Orbit metadata to copy
    std::string startDirection {};
    std::string endDirection {};
    long long orbit_number {};
    std::string time_coverage_start {};
    std::string time_coverage_stop {};

    { // Search global attributes for orbit data
        std::multimap<std::string,netCDF::NcGroupAtt> mp {};
        mp = nc_l1b.getAtts();
        std::multimap<std::string,netCDF::NcGroupAtt>::iterator search {};
        search = mp.find("startDirection");
        if (search == mp.end()) startDirection = "Unknown";
        else {
            search->second.getValues(startDirection);
        }
        search = mp.find("endDirection");
        if (search == mp.end()) endDirection = "Unknown";
        else {
            search->second.getValues(endDirection);
        }
        search = mp.find("orbit_number");
        if (search == mp.end()) orbit_number = NC_FILL_INT64;
        else {
            search->second.getValues(&orbit_number);
        }
        search = mp.find("time_coverage_start");
        if (search == mp.end()) time_coverage_start = "Unknown";
        else {
            search->second.getValues(time_coverage_start);
        }
        search = mp.find("time_coverage_stop");
        if (search == mp.end()) time_coverage_stop = "Unknown";
        else {
            search->second.getValues(time_coverage_stop);
        }
    }

    size_t dim_along {};
    dim_along = nc_l1b.getDim("bins_along_track").getSize();
    assert(!(dim_along < 2) && "Cannot collocate if along-track dimension is less than two");
    size_t dim_vp {};;
    dim_vp = nc_l1b.getDim("number_of_views").getSize();

    // Number of across-track positions is variable per viewport
    size_t dim_fov {}; // All fields of view, of all viewports combined
    dim_fov = nc_l1b.getDim("spatial_samples_per_image").getSize();

    // Number of wavelengths
    size_t dim_pol_wave {};
    dim_pol_wave = nc_l1b.getDim("polarization_bands_per_view").getSize();
    size_t dim_int_wave {};
    dim_int_wave = nc_l1b.getDim("intensity_bands_per_view").getSize();

    // Viewport map, to be turned into the old (CKD) way of representing it.
    // The philosophy is still that we do not need the CKD in this L1C processor.
    netCDF::NcGroup grp {};
    grp = nc_l1b.getGroup("SENSOR_VIEW_BANDS");
    std::vector<uint8_t> viewportmap(dim_fov, 0);
    grp.getVar("viewport_index").getVar(viewportmap.data());
    std::vector<size_t> dims_across(dim_vp, 0);
    std::vector<size_t> ifov_start(dim_vp, 0);
    for (size_t ifov=0 ; ifov<dim_fov ; ifov++) dims_across[viewportmap[ifov]]++;
    ifov_start[0] = 0;
    for (size_t ivp=1 ; ivp<dim_vp ; ivp++) ifov_start[ivp] = ifov_start[ivp-1] + dims_across[ivp-1];
    // Read the wavelengths
    std::vector<double> pol_wavelength(dim_vp*dim_pol_wave, 0.0);
    grp.getVar("polarization_wavelengths").getVar(pol_wavelength.data());
    std::vector<double> int_wavelength(dim_vp*dim_int_wave, 0.0);
    grp.getVar("intensity_wavelengths").getVar(int_wavelength.data());
    // TODO: Read other variables in SENSOR_VIEW_BANDS to copy and/or combine to the L1C product.

    // Open the observation data
    grp = nc_l1b.getGroup("OBSERVATION_DATA");
    netCDF::NcVar var_l1b_i {};
    var_l1b_i = grp.getVar("I");
    netCDF::NcVar var_l1b_i_noise {};
    var_l1b_i_noise = grp.getVar("I_noise");
    netCDF::NcVar var_l1b_i_polsample {};
    var_l1b_i_polsample = grp.getVar("I_polsample");
    netCDF::NcVar var_l1b_i_polsample_noise {};
    var_l1b_i_polsample_noise = grp.getVar("I_polsample_noise");
    netCDF::NcVar var_l1b_q {};
    var_l1b_q = grp.getVar("q");
    netCDF::NcVar var_l1b_q_noise {};
    var_l1b_q_noise = grp.getVar("q_noise");
    netCDF::NcVar var_l1b_u {};
    var_l1b_u = grp.getVar("u");
    netCDF::NcVar var_l1b_u_noise {};
    var_l1b_u_noise = grp.getVar("u_noise");
    // We need the noise on the DoLP (or AoLP) so we can reconstruct the
    // noise covariance between Q and U. This coveriance term in not in
    // the L1B, because no one undestands it.
    netCDF::NcVar var_l1b_dolp_noise {};
    var_l1b_dolp_noise = grp.getVar("DoLP_noise");

    // Geolocation
    grp = nc_l1b.getGroup("GEOLOCATION_DATA");
    netCDF::NcVar var_l1b_lat {};
    var_l1b_lat = grp.getVar("latitude");
    netCDF::NcVar var_l1b_lon {};
    var_l1b_lon = grp.getVar("longitude");
    netCDF::NcVar var_l1b_alt {};
    var_l1b_alt = grp.getVar("altitude");
    netCDF::NcVar var_l1b_sza {};
    var_l1b_sza = grp.getVar("solar_zenith");
    netCDF::NcVar var_l1b_saa {};
    var_l1b_saa = grp.getVar("solar_azimuth");
    netCDF::NcVar var_l1b_vza {};
    var_l1b_vza = grp.getVar("sensor_zenith");
    netCDF::NcVar var_l1b_vaa {};
    var_l1b_vaa = grp.getVar("sensor_azimuth");

    // Initialize L1C output. {{{

    NcObject nc_l1c { netCDF::NcFile::replace };
    nc_l1c.open(settings.l1c.l1cfile);
    nc_l1c.writeReducedConfig(settings);
    // Write orbit metadata
    nc_l1c.putAtt("processing_level", "L1C");
    nc_l1c.putAtt("startDirection",startDirection);
    nc_l1c.putAtt("endDirection",endDirection);
    nc_l1c.putAtt("orbit_number",netCDF::ncInt64,orbit_number);
    nc_l1c.putAtt("time_coverage_start",time_coverage_start);
    nc_l1c.putAtt("time_coverage_stop",time_coverage_stop);

    // TODO (again). How to derive the target grid. The IODS proposes
    // a swath-type L1C output file. We like the nadir grid, but maybe
    // there will be a common PACE grid.
    // Now, we make a one by N grid.
    size_t dim_l1c_along = dim_l1c_obs;
    size_t dim_l1c_across = 1;
    netCDF::NcDim dimid_along {};
    dimid_along = nc_l1c.addDim("bins_along_track",dim_l1c_along);
    netCDF::NcDim dimid_across {};
    dimid_across = nc_l1c.addDim("bins_across_track",dim_l1c_across);
    netCDF::NcDim dimid_vp {};
    dimid_vp = nc_l1c.addDim("number_of_views",dim_vp);
    netCDF::NcDim dimid_pol_wave {};
    dimid_pol_wave = nc_l1c.addDim("polarization_bands_per_view",dim_pol_wave);
    netCDF::NcDim dimid_int_wave {};
    dimid_int_wave = nc_l1c.addDim("intensity_bands_per_view",dim_int_wave);

    // Group SENSOR_VIEW_BANDS
    grp = nc_l1c.addGroup("SENSOR_VIEW_BANDS");
    grp.addVar("polarization_wavelengths",netCDF::ncDouble,{dimid_vp,dimid_pol_wave}).putVar(pol_wavelength.data());
    grp.addVar("intensity_wavelengths",netCDF::ncDouble,{dimid_vp,dimid_int_wave}).putVar(int_wavelength.data());
    // TODO: More stuff to be added

    // Group OBSERVATION_DATA
    grp = nc_l1c.addGroup("OBSERVATION_DATA");
    std::vector<netCDF::NcDim> dims_pol = {dimid_along,dimid_across,dimid_vp,dimid_pol_wave};
    std::vector<netCDF::NcDim> dims_int = {dimid_along,dimid_across,dimid_vp,dimid_int_wave};
    netCDF::NcVar var_l1c_i {};
    var_l1c_i = grp.addVar("I",netCDF::ncDouble,dims_int);
    netCDF::NcVar var_l1c_i_noise {};
    var_l1c_i_noise = grp.addVar("I_noise",netCDF::ncDouble,dims_int);
    netCDF::NcVar var_l1c_i_polsample {};
    var_l1c_i_polsample = grp.addVar("I_polsample",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_i_polsample_noise {};
    var_l1c_i_polsample_noise = grp.addVar("I_polsample_noise",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_q {};
    var_l1c_q = grp.addVar("q",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_q_noise {};
    var_l1c_q_noise = grp.addVar("q_noise",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_u {};
    var_l1c_u = grp.addVar("u",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_u_noise {};
    var_l1c_u_noise = grp.addVar("u_noise",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_dolp {};
    var_l1c_dolp = grp.addVar("DoLP",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_dolp_noise {};
    var_l1c_dolp_noise = grp.addVar("DoLP_noise",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_aolp {};
    var_l1c_aolp = grp.addVar("AoLP",netCDF::ncDouble,dims_pol);
    netCDF::NcVar var_l1c_aolp_noise {};
    var_l1c_aolp_noise = grp.addVar("AoLP_noise",netCDF::ncDouble,dims_pol);

    // Group GEOLOCATION_DATA
    grp = nc_l1c.addGroup("GEOLOCATION_DATA");
    std::vector<netCDF::NcDim> dims_geo = {dimid_along,dimid_across};

    auto latitude_nc { grp.addVar("latitude",netCDF::ncDouble,dims_geo) };
    latitude_nc.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    latitude_nc.putAtt("long_name", "Orbit latitude (degrees North)");
    latitude_nc.putAtt("valid_min", netCDF::ncDouble, -90.0);
    latitude_nc.putAtt("valid_max", netCDF::ncDouble, 90.0);
    latitude_nc.putAtt("units", "degrees");
    latitude_nc.putVar(settings.l1c.lat.data());

    auto longitude_nc { grp.addVar("longitude",netCDF::ncDouble,dims_geo) };
    longitude_nc.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    longitude_nc.putAtt("long_name", "Orbit longitude (degrees East)");
    longitude_nc.putAtt("valid_min", netCDF::ncDouble, -180.0);
    longitude_nc.putAtt("valid_max", netCDF::ncDouble, 180.0);
    longitude_nc.putAtt("units", "degrees");
    longitude_nc.putVar(settings.l1c.lon.data());

    netCDF::NcVar var_l1c_alt {};
    var_l1c_alt = grp.addVar("altitude",netCDF::ncDouble,dims_geo);
    var_l1c_alt.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    var_l1c_alt.putAtt("long_name", "Orbit altitude");
    var_l1c_alt.putAtt("valid_min", netCDF::ncDouble, -1000.0);
    var_l1c_alt.putAtt("valid_max", netCDF::ncDouble, 10000.0);
    var_l1c_alt.putAtt("units", "m");
    // I think that the solar angles also depend on viewport, because
    // of time differences.
    std::vector<netCDF::NcDim> dims_ang = {dimid_along,dimid_across,dimid_vp};
    netCDF::NcVar var_l1c_sza {};
    var_l1c_sza = grp.addVar("solar_zenith",netCDF::ncDouble,dims_ang);
    var_l1c_sza.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    var_l1c_sza.putAtt("long_name", "Solar zenith angle");
    var_l1c_sza.putAtt("valid_min", netCDF::ncDouble, 0.0);
    var_l1c_sza.putAtt("valid_max", netCDF::ncDouble, 180.0);
    var_l1c_sza.putAtt("units", "degrees");
    netCDF::NcVar var_l1c_saa {};
    var_l1c_saa = grp.addVar("solar_azimuth",netCDF::ncDouble,dims_ang);
    var_l1c_saa.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    var_l1c_saa.putAtt("long_name", "Solar azimuth angle");
    var_l1c_saa.putAtt("valid_min", netCDF::ncDouble, -180.0);
    var_l1c_saa.putAtt("valid_max", netCDF::ncDouble, 180.0);
    var_l1c_saa.putAtt("units", "degrees");
    netCDF::NcVar var_l1c_vza {};
    var_l1c_vza = grp.addVar("sensor_zenith",netCDF::ncDouble,dims_ang);
    var_l1c_vza.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    var_l1c_vza.putAtt("long_name", "Sensor zenith angle");
    var_l1c_vza.putAtt("valid_min", netCDF::ncDouble, 0.0);
    var_l1c_vza.putAtt("valid_max", netCDF::ncDouble, 180.0);
    var_l1c_vza.putAtt("units", "degrees");
    netCDF::NcVar var_l1c_vaa {};
    var_l1c_vaa = grp.addVar("sensor_azimuth",netCDF::ncDouble,dims_ang);
    var_l1c_vaa.putAtt("_FillValue", netCDF::ncDouble, -999.9);
    var_l1c_vaa.putAtt("long_name", "Sensor azimuth angle");
    var_l1c_vaa.putAtt("valid_min", netCDF::ncDouble, -180.0);
    var_l1c_vaa.putAtt("valid_max", netCDF::ncDouble, 180.0);
    var_l1c_vaa.putAtt("units", "degrees");

    // Construct the Earth. We will have latitudes and longitudes, and we need
    // cartesian coordinates.
    Planet earth(settings.l1c.semi_major_axis,settings.l1c.semi_minor_axis,settings.l1c.latitude_tol);
    if (settings.l1c.demfile.compare("") != 0) {
        earth.include_dem(settings.l1c.demfile);
    }

    // Construct array of target positions. For target positions, altitudes are
    // not relevant, because the final projection is along the zenith.
    std::vector<Vector> target(dim_l1c_obs);
    std::vector<Vector> tarzen(dim_l1c_obs); // The zenith of the target
    std::vector<double> alt(dim_l1c_obs, 0.0);
    for (size_t iobs=0 ; iobs<dim_l1c_obs ; iobs++) {
        double lat = settings.l1c.lat[iobs] * math::deg_to_rad;
        double lon = settings.l1c.lon[iobs] * math::deg_to_rad;
        earth.dem(lat,lon,alt[iobs]);
        earth.xyz(lat,lon,alt[iobs],&target[iobs]);
        tarzen[iobs][0] = std::cos(lat)*std::cos(lon);
        tarzen[iobs][1] = std::cos(lat)*std::sin(lon);
        tarzen[iobs][2] = std::sin(lat);
    }
    var_l1c_alt.putVar(alt.data());

    // From now on, we will have an along-track dimensions, which is fixed
    // for all viewports, and an across-track dimension, which may differ
    // for different viewport. Optimal would be to keep all the along
    // track things outside the for loop, but that is an optimization step
    // that may be done in the future and only if it saves significant time.
    // Keeping all this inside the loop keeps the related code together.

    // Loop over viewports (source grids)
    for (size_t ivp=0 ; ivp<dim_vp ; ivp++) {
        spdlog::info("Constructing grid for viewport {}", ivp);

        size_t& dim_across = dims_across[ivp];
        // Interpret some dimension sizes
        // Go to 'int' type, because we may virtually move to indices like -1 and that
        // should be recognized properly.
        int ncell_along = (int)dim_along-1;
        int ncell_across = (int)dim_across-1;
        int ncell = ncell_along*ncell_across;

        // Construct 2D B-splines over an index grid. The index grid is the same for
        // all the viewports. TODO: No. On the contrary, no.
        std::vector<double> gain_along((dim_along+2)*(dim_along+2));
        std::vector<double> gain_across((dim_across+2)*(dim_across+2));
        for (size_t al_ac=0 ; al_ac<2 ; al_ac++) {
            size_t dim = al_ac==0?dim_along:dim_across;
            double* gain = al_ac==0?gain_along.data():gain_across.data();
            // Start with the inverse of the corresponding Toeplitz matrix.
            // Initialize the Chebyshev polynomials, without recycling between along-track
            // and across track. The x-value of these is always two.
            // The Chebyshev polynomials become huge, so we save factors
            // between adjacent orders. chebfacts[i] = cheb[i+1] / cheb[i],
            // usually close to four.
            std::vector<double> chebfacts(dim+2); // Negative definition to get rid of the sign flip action
            chebfacts[0] = -4.0;
            for (size_t icheb=1 ; icheb<dim+2 ; icheb++) {
                chebfacts[icheb] = -4.0 - 1.0/chebfacts[icheb-1];
            }
            std::vector<double> els_cur((dim+3)/2);
            els_cur[0] = -6.0/chebfacts[dim+1];
            gain[0] = els_cur[0];
            gain[(dim+1)*(dim+2)+dim+1] = els_cur[0];
            bool firsthalf = true;
            for (size_t idim1=1 ; idim1<dim+2 ; idim1++) {
                size_t sz2 = firsthalf?idim1:dim+2-idim1;
                for (size_t idim2=0 ; idim2<sz2 ; idim2++) els_cur[idim2] /= chebfacts[dim+1-idim1];
                if (firsthalf) {
                    sz2++;
                    els_cur[sz2-1] = els_cur[sz2-2] * chebfacts[sz2-2];
                }
                for (size_t idim2=0 ; idim2<sz2 ; idim2++) {
                    gain[idim1*(dim+2)+idim2] = els_cur[idim2];
                    // Possibly writing the same index again, but ifs
                    // may be more expensive than additional write actions.
                    gain[idim2*(dim+2)+idim1] = els_cur[idim2];
                    gain[(dim+1-idim1)*(dim+2)+dim+1-idim2] = els_cur[idim2];
                    gain[(dim+1-idim2)*(dim+2)+dim+1-idim1] = els_cur[idim2];
                }
                if (2*sz2 > dim+1) firsthalf = false;
            }
            // Apply perturbations
            std::vector<size_t> ifirst = {0,0,0,dim+1,dim+1,dim+1};
            std::vector<size_t> isecond = {0,1,2,dim-1,dim,dim+1};
            std::vector<double> pert = {1.0/3.0,-13.0/6.0,1.0,1.0,-13.0/6.0,1.0/3.0};
            for (size_t icase=0 ; icase<6 ; icase++) {
                std::vector<double> one(dim+2);
                std::vector<double> two(dim+2);
                for (size_t idim=0 ; idim<dim+2 ; idim++) {
                    one[idim] = gain[idim*(dim+2)+ifirst[icase]];
                    two[idim] = gain[isecond[icase]*(dim+2)+idim];
                }
                double cross = gain[isecond[icase]*(dim+2)+ifirst[icase]];
                for (size_t idim_first=0 ; idim_first<dim+2 ; idim_first++) {
                    for (size_t idim_second=0 ; idim_second<dim+2 ; idim_second++) {
                        gain[idim_first*(dim+2)+idim_second] -= one[idim_first]*two[idim_second]*pert[icase] / (1.0 + cross*pert[icase]);
                    }
                }
            }
        }

        // Read latitudes and longitudes for this view port
        std::vector<double> l1b_lat(dim_along*dim_across);
        std::vector<double> l1b_lon(dim_along*dim_across);
        std::vector<double> l1b_alt(dim_along*dim_across);
        const std::vector<size_t> l1b_start = {0,ifov_start[ivp]};
        const std::vector<size_t> l1b_cnt = {dim_along,dim_across};
        var_l1b_lat.getVar(l1b_start,l1b_cnt,l1b_lat.data());
        var_l1b_lon.getVar(l1b_start,l1b_cnt,l1b_lon.data());
        var_l1b_alt.getVar(l1b_start,l1b_cnt,l1b_alt.data());

        // Construct the grid in cartesian coordinates
        std::vector<Vector> grid(dim_along*dim_across);

        for (size_t igrid=0 ; igrid<dim_along*dim_across ; igrid++) {
            double lat = l1b_lat[igrid] * math::deg_to_rad;
            double lon = l1b_lon[igrid] * math::deg_to_rad;
            earth.xyz(lat,lon,l1b_alt[igrid],&grid[igrid]);
        }

        // Cell zeniths (for mesh investigation). They need not be precise, only such that
        // the cells do not become complex after projection.
        std::vector<Vector> zens(ncell);
        for (int icell_along=0 ; icell_along<ncell_along ; icell_along++) {
            for (int icell_across=0 ; icell_across<ncell_across ; icell_across++) {
                Vector& ll = grid[icell_along*dim_across+icell_across];
                Vector& lr = grid[icell_along*dim_across+icell_across+1];
                Vector& ul = grid[(icell_along+1)*dim_across+icell_across];
                Vector& ur = grid[(icell_along+1)*dim_across+icell_across+1];
                Vector avg = 0.25*(ll+lr+ul+ur);
                double lat;
                double lon;
                double alt;
                earth.lla(avg,&lat,&lon,&alt);
                zens[icell_along*ncell_across+icell_across] = Vector(cos(lat)*cos(lon),cos(lat)*sin(lon),sin(lat));
            }
        }
        // Alternatively, use corner grid and center cells for zeniths.

        // Calibration for what is left or right.
        // We assume that if along-flight is forward and the zenith is
        // upward, across-track is to the right. This must be verified.
        // We use the zeniths of the cells. We may have used the zeniths
        // of the points that are the target points here (ATBD), but the cell
        // should also suffice.
        double cal = 0.0; // Neutral
        for (int icell_along=0 ; icell_along<ncell_along ; icell_along++) {
            for (int icell_across=0 ; icell_across<ncell_across ; icell_across++) {
                Vector* base = &grid[icell_along*dim_across+icell_across];
                Vector& z = zens[icell_along*ncell_across+icell_across];
                size_t nright = 0;
                if (getSide(base[0],base[dim_across],base[dim_across+1],z) > 0.0) nright++;
                if (getSide(base[dim_across],base[dim_across+1],base[1],z) > 0.0) nright++;
                if (getSide(base[dim_across+1],base[1],base[0],z) > 0.0) nright++;
                if (getSide(base[1],base[0],base[dim_across],z) > 0.0) nright++;
                assert(nright != 2 && "Complex quadrilateral in grid");
                if (cal == 0.0) cal = nright>2?1.0:-1.0;
                assert(!((cal > 0.0) != (nright>2)) && "Clockwise and counter-clockwise cells co-exist");
                if (nright == 1 || nright == 3) {
                    spdlog::warn("Concave quadrilateral in grid. This is rare but still supported.");
                }
            }
        }

        spdlog::info("Collocating observations into grid for viewport {}", ivp);

        // The observation loop calculates a factor matrix of interpolation factors
        // from the L1B along-across grid to the observation index on L1C. If it
        // fails, the failure flag is turned on.
        std::vector<double> factormat(dim_l1c_obs*dim_along*dim_across,0.0);
        double* factormat_cur = factormat.data();
        std::vector<bool> fail(dim_l1c_obs,false);

        for (size_t iobs=0 ; iobs<dim_l1c_obs ; iobs++) {

            // Target point and its zenith, used many times
            Vector& tar = target[iobs];
            Vector& zen = tarzen[iobs];

            // Set a-priori cell where the dot product between cell zenith
            // and target zenith is maximum. We could have taken the middle
            // of the cell and the target, but we forgot the middle of the
            // cells and we still have the zeniths for verifying the mesh.
            // The a-priori need not be precise, so it does not matter.

            int icell = 0;
            double zenith_dot_product = zen.dot(zens[0]);
            for (int icell_search=1 ; icell_search<ncell ; icell_search++) {
                double newdot = zen.dot(zens[icell_search]);
                if (newdot > zenith_dot_product) {
                    zenith_dot_product = newdot;
                    icell = icell_search;
                }
            }
            // Split up a-priori cell to along and across track coordinates
            int icell_along = icell / ncell_across; // Integer division
            int icell_across = icell - icell_along*ncell_across; // Or icell modulo ncell_across

            // getSide constructions
            double construction_left {};
            double construction_right {};
            double construction_down {};
            double construction_up {};

            // Execute search algorithm
            // Rounding errors cannot cause the search algorithm to go
            // back and forth, in which case we will compare the badnesses,
            // which should approach zero. This can occur if the target
            // point is on a grid line. In such a case, both cells are
            // correct, but the lowest zero badness will be taken.

            // Concave quadrilaterals are handled such that, if a corner
            // is also outside the line, it is a concave quadrilateral
            // and that direction is flagged as a concave direction. Only
            // if there are desires to get out in two concave directions,
            // there is a wish to get out, in the second-worst direction.
            // If there is a concave quadrilateral and we are on a rim or
            // a vertex, we just save the badness and end at the least bad
            // cell, which should have a badness equal to a rounding error.

            // We have to save the concave (overruled) movement desires
            std::vector<double> badness(ncell,0.0);
            std::vector<bool> visited(ncell,false);
            std::vector<bool> concave_left(ncell,false);
            std::vector<bool> concave_right(ncell,false);
            std::vector<bool> concave_down(ncell,false);
            std::vector<bool> concave_up(ncell,false);
            bool continue_investigation = false; // Flag for cycle due to rounding errors
            for (int iter=0 ; iter<ncell+1 ; iter++) { // Maximum number of loop iterations is smaller than this
                // Protection against infinite loop
                assert(iter != ncell && "Too many iterations in search loop. Infinite loop in search algorithm or zero-sized grid.");
                icell = icell_along*ncell_across + icell_across;
                if (visited[icell]) {
                    // Loop detected
                    // For the final step, these must be rounding errors and
                    // the candidates will be compared.
                    continue_investigation = true;
                    break;
                }
                visited[icell] = true;

                // Calculate movement desires
                // Index offset of the cell in the grid
                Vector* base = &grid[icell_along*dim_across+icell_across];
                // Left
                construction_left = getSide(base[0],base[dim_across],tar,zen)*cal;
                bool wish_left = construction_left < 0.0;
                // Right
                construction_right = getSide(base[dim_across+1],base[1],tar,zen)*cal;
                bool wish_right = construction_right < 0.0;
                // Down
                construction_down = getSide(base[1],base[0],tar,zen)*cal;
                bool wish_down = construction_down < 0.0;
                // Up
                construction_up = getSide(base[dim_across],base[dim_across+1],tar,zen)*cal;
                bool wish_up = construction_up < 0.0;

                // Check for concave angles. In such a case, it is possible
                // that a desire to move out is calculated while the
                // target point is in the cell. More specifically, if the
                // reflex-angle condition is fulfilled and this is the
                // only desire, the point is in the cell. We will not
                // remove the wish flag but just add a concave flag under
                // the condition. Flags are saved in arrays, because
                // for numeric rounding-error loops, we may finally
                // converge to a cell that is not the last one visited.
                if (wish_left) {
                    double construction_corner1 = getSide(base[0],base[dim_across],base[1],zen)*cal;
                    double construction_corner2 = getSide(base[0],base[dim_across],base[dim_across+1],zen)*cal;
                    if (construction_corner1 < 0.0 || construction_corner2 < 0.0) concave_left[icell] = true;
                }
                if (wish_right) {
                    double construction_corner1 = getSide(base[dim_across+1],base[1],base[dim_across],zen)*cal;
                    double construction_corner2 = getSide(base[dim_across+1],base[1],base[0],zen)*cal;
                    if (construction_corner1 < 0.0 || construction_corner2 < 0.0) concave_right[icell] = true;
                }
                if (wish_down) {
                    double construction_corner1 = getSide(base[1],base[0],base[dim_across+1],zen)*cal;
                    double construction_corner2 = getSide(base[1],base[0],base[dim_across],zen)*cal;
                    if (construction_corner1 < 0.0 || construction_corner2 < 0.0) concave_down[icell] = true;
                }
                if (wish_up) {
                    double construction_corner1 = getSide(base[dim_across],base[dim_across+1],base[0],zen)*cal;
                    double construction_corner2 = getSide(base[dim_across],base[dim_across+1],base[1],zen)*cal;
                    if (construction_corner1 < 0.0 || construction_corner2 < 0.0) concave_up[icell] = true;
                }

                size_t nwish = 0;
                if (wish_left) nwish++;
                if (wish_right) nwish++;
                if (wish_down) nwish++;
                if (wish_up) nwish++;
                if (nwish == 0) {
                    // Perfect conversion
                    break;
                }
                if (nwish == 1) {
                    // Wish to move in one side. If this movement is under discussion
                    // because of a reflex angle, we know we are in the cell and we are
                    // ready. Otherwise, move in that direction.
                    if (wish_left && concave_left[icell]) break; // Ready
                    if (wish_right && concave_right[icell]) break; // Ready
                    if (wish_down && concave_down[icell]) break; // Ready
                    if (wish_up && concave_up[icell]) break; // Ready
                    // Not converged, thus badness is written for if it is a rounding error
                    if (wish_left) badness[icell] = construction_left;
                    if (wish_right) badness[icell] = construction_right;
                    if (wish_down) badness[icell] = construction_down;
                    if (wish_up) badness[icell] = construction_up;

                    // Move in that direction
                    if (wish_left) icell_across--;
                    if (wish_right) icell_across++;
                    if (wish_down) icell_along--;
                    if (wish_up) icell_along++;
                }
                if (nwish == 2) {
                    // If these are opposite, it is an error unless one side is concave. Then,
                    // go to the other side.
                    if (wish_up && wish_down) {
                        if (concave_down[icell] == concave_up[icell]) {
                            spdlog::error("Complex quadrilateral encountered during search");
                            fail[iobs] = true;
                            break;
                        }
                        if (concave_down[icell]) icell_along++;
                        else icell_along--;
                        continue;
                    }
                    if (wish_left && wish_right) {
                        if (concave_left[icell] == concave_right[icell]) {
                            spdlog::error("Complex quadrilateral encountered during search");
                            fail[iobs] = true;
                            break;
                        }
                        if (concave_left[icell]) icell_across++;
                        else icell_across--;
                        continue;
                    }
                    // Now, we assume there is a desire to move in two perpendicular directions.
                    // Objection flags.
                    bool object_across = false;
                    bool object_along = false;
                    // In theory, it is impossible to have a desire to move in two
                    // perpendicular and one is concave. In such a case, the other
                    // concave border should also trigger. Here, however, it is
                    // most important to save badness for eventual rounding-error
                    // driven infinite loop and stay resillient against numeric
                    // errors anyway.
                    // Concave wish always has lowest priority.
                    object_across = (concave_left[icell] || concave_right[icell]);
                    object_along = (concave_down[icell] || concave_up[icell]);
                    // Save badness. If there is one objection, save the other.
                    // If there are two objections, save the least bad one.
                    // If there are no objections, save the worst one.
                    double badness_across = wish_left?construction_left:construction_right; // Flag wish_left equals !wish_right here.
                    double badness_along = wish_down?construction_down:construction_up; // Flag wish_down equals !wish_up here
                    if (object_across) {
                        if (object_along) {
                            // Save least bad one. Low (strongly negative) is bad.
                            // Also, overrule concaveness for the least bad badness. Only the
                            // worst one can be concave. This is important if this ends up
                            // being the correct cell and that the one of the two movement
                            // desires is a rounding error and the other one in concave.
                            // Then, only the true concave direction should be flagged as concave
                            // for the interpolation part.
                            if (badness_across < badness_along) {
                                // Across are concave, along are not
                                badness[icell] = badness_along;
                                concave_down[icell] = false;
                                concave_up[icell] = false;
                            } else {
                                // Along are concave, across are not
                                badness[icell] = badness_across;
                                concave_left[icell] = false;
                                concave_right[icell] = false;
                            }
                        } else badness[icell] = badness_along;
                    } else {
                        if (object_along) badness[icell] = badness_across;
                        else {
                            // Save worst one. Low (strongly negative) is bad.
                            if (badness_across < badness_along) badness[icell] = badness_across;
                            else badness[icell] = badness_along;
                        }
                    }
                    // Second rule
                    if (object_across == object_along) {
                        // Prefer moving in along-track direction, except when this involves wrapping or
                        // revisiting.
                        // If across-track movement also involves wrapping or revisiting, execute the wrap or
                        // revisit in along-track direction.

                        // Move virtually
                        int icell_along_desire = icell_along;
                        int icell_across_desire = icell_across;
                        if (wish_left) icell_across_desire--;
                        if (wish_right) icell_across_desire++;
                        if (wish_down) icell_along_desire--;
                        if (wish_up) icell_along_desire++;

                        // Object if you move out of the grid or to a visited cell
                        object_across = icell_across_desire == -1 || icell_across_desire == ncell_across;
                        // Prevent any chance of out-of-bound errors
                        // in reading visited
                        if (!object_across) {
                            object_across = visited[icell_along*ncell_across+icell_across_desire];
                        }
                        object_along = icell_along_desire == -1 || icell_along_desire == ncell_along;
                        // Prevent any chance of out-of-bound errors in reading visited
                        if (!object_along) {
                            object_along = visited[icell_along_desire*ncell_across+icell_across];
                        }
                    }
                    // These are the final objection flags. If it is still a draw, take
                    // the along-track direction.
                    if (object_along && !object_across) {
                        if (wish_left) icell_across--;
                        if (wish_right) icell_across++;
                    } else {
                        if (wish_down) icell_along--;
                        if (wish_up) icell_along++;
                    }
                }
                if (nwish == 3) {
                    // Do opposite the one that is not there, even if this involves wrapping
                    if (!wish_left) icell_across++;
                    if (!wish_right) icell_across--;
                    if (!wish_down) icell_along++;
                    if (!wish_up) icell_along--;
                    // Do not write badness, it will have a badness of zero, which, in the
                    // interpretation will be flagged as infinitely bad. It is not possible
                    // to have three rounding errors unless the grid is ill.
                }
                if (nwish == 4) {
                    spdlog::warn(
                      "Current cell is on the opposite side as target. This "
                      "should be impossible or nearly impossible, although it "
                      "should not harm the algorithm.");
                    // This is rare. You are exactly on the opposite of the globe. Do somehting,
                    // it does not matter if you wrap, but do not die if not needed.
                    // We can assume that you did not wrap to this position, so moving in
                    // any direction along track should be safe.
                    icell_along++;
                    // Do not write badness, it will have a badness of zero, which, in the
                    // interpretation will be flagged as infinitely bad. It is not possible
                    // to have four rounding errors unless the grid is ill.
                }

                // Detect moving out of the orbit
                if (icell_across == -1 || icell_across == ncell_across || icell_along == -1 || icell_along == ncell_along) {
                    fail[iobs] = true;
                    break;
                }

            }

            // If the continue investigation flag is turned on, it is
            // a loop by rounding erors and the badness array is consulted.
            // Terribly bad cells will not have a badness written, hence
            // their array value is at zero. These are indeed interpreted as
            // very bad and will not be considered. For the rest, the best
            // cell is the one with the smallest negative badness (highest
            // numeric value of badness).
            if (continue_investigation) {
                double best_badness = 0.0;
                int icell_best = -1; // Nothing found
                for (int icell_loop=0 ; icell_loop<ncell ; icell_loop++) {
                    if (badness[icell_loop] == 0.0) continue; // Not visited or flagged as chanceless (badness is zero)
                    if (best_badness == 0.0 || badness[icell_loop] > best_badness) {
                        icell_best = icell_loop;
                        best_badness = badness[icell_loop];
                    }
                }
                // Set 2D and 1D indices accordingly
                icell = icell_best;
                icell_along = icell/ncell_across;
                icell_across = icell - icell_along*ncell_across;
            }

            // Failed collocations are ready. No floating indices need to be calculated.
            if (!fail[iobs]) {

                assert(!(icell < 0 || icell >= ncell) && "Program error: Variable icell not in boundaries after search. It is a miracle that no worse error occured before.");

                // Find the weight factors so that the floating part of the index
                // can be retrieved.

                // Start iterative interpolation
                double weightleft = 0.5;
                double weightright = 0.5;
                double weightdown = 0.5;
                double weightup = 0.5;
                double mv = 0.25;
                Vector& lowleft = grid[icell_along*dim_across+icell_across];
                Vector& lowright = grid[icell_along*dim_across+icell_across+1];
                Vector& upleft = grid[(icell_along+1)*dim_across+icell_across];
                Vector& upright = grid[(icell_along+1)*dim_across+icell_across+1];
                for (size_t i=0 ; i<200 ; i++) { // Iterative loop. TODO: Set loop limit
                    Vector left = weightdown*lowleft + weightup*upleft;
                    Vector right = weightdown*lowright + weightup*upright;
                    Vector down = weightleft*lowleft + weightright*lowright;
                    Vector up = weightleft*upleft + weightright*upright;
                    bool force_left = false;
                    bool force_right = false;
                    if (concave_left[icell]) {
                        double construction_corner1 = getSide(down,up,lowright,zen)*cal;
                        double construction_corner2 = getSide(down,up,upright,zen)*cal;
                        if (construction_corner1 < 0.0 || construction_corner2 < 0.0) force_right = true;
                    }
                    if (concave_right[icell]) {
                        double construction_corner1 = getSide(down,up,lowleft,zen)*cal;
                        double construction_corner2 = getSide(down,up,upleft,zen)*cal;
                        if (construction_corner1 > 0.0 || construction_corner2 > 0.0) force_left = true;
                    }
                    // Only if nothing is forced here, do normal
                    bool go_left {};
                    if (force_left) go_left = true;
                    else if (force_right) go_left = false;
                    else {
                        double construction = getSide(down,up,tar,zen)*cal;
                        go_left = construction < 0.0;
                    }
                    // Repeat joke for down and up
                    bool force_down = false;
                    bool force_up = false;
                    if (concave_down[icell]) {
                        double construction_corner1 = getSide(left,right,upleft,zen)*cal;
                        double construction_corner2 = getSide(left,right,upright,zen)*cal;
                        if (construction_corner1 > 0.0 || construction_corner2 > 0.0) force_up = true;
                    }
                    if (concave_up[icell]) {
                        double construction_corner1 = getSide(left,right,lowleft,zen)*cal;
                        double construction_corner2 = getSide(left,right,lowright,zen)*cal;
                        if (construction_corner1 < 0.0 || construction_corner2 < 0.0) force_down = true;
                    }
                    // Only if nothing is forced here, do normal
                    bool go_up {};
                    if (force_down) go_up = false;
                    else if (force_up) go_up = true;
                    else {
                        double construction = getSide(left,right,tar,zen)*cal;
                        go_up = construction < 0.0;
                    }
                    // Execute movement
                    if (go_left) {
                        weightleft += mv;
                        weightright -= mv;
                    } else {
                        weightleft -= mv;
                        weightright += mv;
                    }
                    if (go_up) {
                        weightdown -= mv;
                        weightup += mv;
                    } else {
                        weightdown += mv;
                        weightup -= mv;
                    }
                    mv /= 2.0;
                }

                std::vector<double> splinecoefs_along(4);
                std::vector<double> splinecoefs_across(4);
                // Zeroth element: decreasing part of the low section
                splinecoefs_along[0] = 1.0/6.0 * pow(weightdown,3.0);
                splinecoefs_across[0] = 1.0/6.0 * pow(weightleft,3.0);
                // First element: decreasing part of the high section
                splinecoefs_along[1] = -1.0/2.0 * pow(weightdown,3.0) + 1.0/2.0 * pow(weightdown,2.0) + 1.0/2.0 * weightdown + 1.0/6.0;
                splinecoefs_across[1] = -1.0/2.0 * pow(weightleft,3.0) + 1.0/2.0 * pow(weightleft,2.0) + 1.0/2.0 * weightleft + 1.0/6.0;
                // Second element: increasing part of high section
                splinecoefs_along[2] = -1.0/2.0 * pow(weightup,3.0) + 1.0/2.0 * pow(weightup,2.0) + 1.0/2.0 * weightup + 1.0/6.0;
                splinecoefs_across[2] = -1.0/2.0 * pow(weightright,3.0) + 1.0/2.0 * pow(weightright,2.0) + 1.0/2.0 * weightright + 1.0/6.0;
                // Third element: increasing part of low section
                splinecoefs_along[3] = 1.0/6.0 * pow(weightup,3.0);
                splinecoefs_across[3] = 1.0/6.0 * pow(weightright,3.0);

                for (size_t isect_along=0 ; isect_along<4 ; isect_along++) {
                    for (size_t isect_across=0 ; isect_across<4 ; isect_across++) {
                        for (size_t ialong=0 ; ialong<dim_along ; ialong++) {
                            for (size_t iacross=0 ; iacross<dim_across ; iacross++) {
                                factormat_cur[ialong*dim_across+iacross] += splinecoefs_along[isect_along]*splinecoefs_across[isect_across] * gain_along[(icell_along+isect_along)*(dim_along+2)+ialong+1] * gain_across[(icell_across+isect_across)*(dim_across+2)+iacross+1];
                            }
                        }
                    }
                }
            }

            // Move running pointer
            factormat_cur += dim_along*dim_across;

        } // Loop over observations

        // Apply the factor matrix to everything that is being collocated

        spdlog::info("Interpolating intensity product");
        // Radiance
        {
            std::vector<double> l1b_radiance(dim_along*dim_across*dim_int_wave);
            std::vector<double> l1b_radiance_noise(dim_along*dim_across*dim_int_wave);
            // Slice out one viewport
            std::vector<size_t> l1b_start = {0,ifov_start[ivp],0};
            std::vector<size_t> l1b_cnt = {dim_along,dim_across,dim_int_wave};
            var_l1b_i.getVar(l1b_start,l1b_cnt,l1b_radiance.data());
            var_l1b_i_noise.getVar(l1b_start,l1b_cnt,l1b_radiance_noise.data());
            factormat_cur = factormat.data();
            for (size_t iobs=0 ; iobs<dim_l1c_obs ; iobs++) {

                // Where to write
                std::vector<size_t> l1c_start = {iobs,0,ivp,0};
                std::vector<size_t> l1c_cnt = {1,1,1,dim_int_wave};

                if (fail[iobs]) {
                    // Write fill values
                    std::vector<double> fillvalues(dim_int_wave,NC_FILL_DOUBLE);
                    var_l1c_i.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_i_noise.putVar(l1c_start,l1c_cnt,fillvalues.data());
                } else {
                    std::vector<double> l1c_radiance(dim_int_wave,0.0);
                    std::vector<double> l1c_radiance_noise(dim_int_wave,0.0);
                    double* l1b_radiance_cur = l1b_radiance.data();
                    double* l1b_radiance_noise_cur = l1b_radiance_noise.data();
                    for (size_t ialong=0 ; ialong<dim_along ; ialong++) {
                        for (size_t iacross=0 ; iacross<dim_across ; iacross++) {
                            for (size_t iwave=0 ; iwave<dim_int_wave ; iwave++) {
                                l1c_radiance[iwave] += factormat_cur[ialong*dim_across+iacross] * l1b_radiance_cur[iwave];
                                // L1C radiance noise is here L1C radiance variance
                                l1c_radiance_noise[iwave] += pow(factormat_cur[ialong*dim_across+iacross] * l1b_radiance_noise_cur[iwave],2.0);
                            }
                            l1b_radiance_cur += dim_int_wave;
                            l1b_radiance_noise_cur += dim_int_wave;
                        }
                    }
                    // Convert variance to noise
                    for (size_t iwave=0 ; iwave<dim_int_wave ; iwave++) l1c_radiance_noise[iwave] = sqrt(l1c_radiance_noise[iwave]);

                    // Write the results
                    var_l1c_i.putVar(l1c_start,l1c_cnt,l1c_radiance.data());
                    var_l1c_i_noise.putVar(l1c_start,l1c_cnt,l1c_radiance_noise.data());
                }
                factormat_cur += dim_along*dim_across;
            }
        }

        spdlog::info("Interpolating polarization product");
        // Polarization. TODO: Discuss on what to collocate, small or large Q/U?
        // TODO: Discuss averaging terhnique with or without averaging out noise.
        {
            std::vector<double> l1b_i_polsample(dim_along*dim_across*dim_pol_wave);
            std::vector<double> l1b_i_polsample_noise(dim_along*dim_across*dim_pol_wave);
            std::vector<double> l1b_q(dim_along*dim_across*dim_pol_wave);
            std::vector<double> l1b_q_noise(dim_along*dim_across*dim_pol_wave);
            std::vector<double> l1b_u(dim_along*dim_across*dim_pol_wave);
            std::vector<double> l1b_u_noise(dim_along*dim_across*dim_pol_wave);
            std::vector<double> l1b_dolp_noise(dim_along*dim_across*dim_pol_wave);
            std::vector<double> l1b_qu_noise_covar(dim_along*dim_across*dim_pol_wave);
            // Slice out one viewport
            std::vector<size_t> l1b_start = {0,ifov_start[ivp],0};
            std::vector<size_t> l1b_cnt = {dim_along,dim_across,dim_pol_wave};
            var_l1b_i_polsample.getVar(l1b_start,l1b_cnt,l1b_i_polsample.data());
            var_l1b_i_polsample_noise.getVar(l1b_start,l1b_cnt,l1b_i_polsample_noise.data());
            var_l1b_q.getVar(l1b_start,l1b_cnt,l1b_q.data());
            var_l1b_q_noise.getVar(l1b_start,l1b_cnt,l1b_q_noise.data());
            var_l1b_u.getVar(l1b_start,l1b_cnt,l1b_u.data());
            var_l1b_u_noise.getVar(l1b_start,l1b_cnt,l1b_u_noise.data());
            var_l1b_dolp_noise.getVar(l1b_start,l1b_cnt,l1b_dolp_noise.data());
            // Acquire noise covariance from all noise terms
            for (size_t imat=0 ; imat<dim_along*dim_across*dim_pol_wave ; imat++) {
                l1b_qu_noise_covar[imat] = (pow(l1b_dolp_noise[imat],2.0) * (pow(l1b_q[imat],2.0) + pow(l1b_u[imat],2.0)) - pow(l1b_q[imat]*l1b_q_noise[imat],2.0) - pow(l1b_u[imat]*l1b_u_noise[imat],2.0)) / (l1b_q[imat]*l1b_u[imat]);
            }
            factormat_cur = factormat.data();
            for (size_t iobs=0 ; iobs<dim_l1c_obs ; iobs++) {

                // Where to write
                std::vector<size_t> l1c_start = {iobs,0,ivp,0};
                std::vector<size_t> l1c_cnt = {1,1,1,dim_pol_wave};

                if (fail[iobs]) {

                    // Write fill values
                    std::vector<double> fillvalues(dim_pol_wave,NC_FILL_DOUBLE);
                    var_l1c_i_polsample.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_i_polsample_noise.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_q.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_q_noise.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_u.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_u_noise.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_dolp.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_dolp_noise.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_aolp.putVar(l1c_start,l1c_cnt,fillvalues.data());
                    var_l1c_aolp_noise.putVar(l1c_start,l1c_cnt,fillvalues.data());

                } else {

                    std::vector<double> l1c_i_polsample(dim_pol_wave,0.0);
                    std::vector<double> l1c_i_polsample_noise(dim_pol_wave,0.0);
                    std::vector<double> l1c_q(dim_pol_wave,0.0);
                    std::vector<double> l1c_q_noise(dim_pol_wave,0.0);
                    std::vector<double> l1c_u(dim_pol_wave,0.0);
                    std::vector<double> l1c_u_noise(dim_pol_wave,0.0);
                    std::vector<double> l1c_qu_noise_covar(dim_pol_wave,0.0);
                    std::vector<double> l1c_dolp(dim_pol_wave,0.0);
                    std::vector<double> l1c_dolp_noise(dim_pol_wave,0.0);
                    std::vector<double> l1c_aolp(dim_pol_wave,0.0);
                    std::vector<double> l1c_aolp_noise(dim_pol_wave,0.0);
                    double* l1b_i_polsample_cur = l1b_i_polsample.data();
                    double* l1b_i_polsample_noise_cur = l1b_i_polsample_noise.data();
                    double* l1b_q_cur = l1b_q.data();
                    double* l1b_q_noise_cur = l1b_q_noise.data();
                    double* l1b_u_cur = l1b_u.data();
                    double* l1b_u_noise_cur = l1b_u_noise.data();
                    double* l1b_qu_noise_covar_cur = l1b_qu_noise_covar.data();
                    for (size_t ialong=0 ; ialong<dim_along ; ialong++) {
                        for (size_t iacross=0 ; iacross<dim_across ; iacross++) {
                            for (size_t iwave=0 ; iwave<dim_pol_wave ; iwave++) {
                                l1c_i_polsample[iwave] += factormat_cur[ialong*dim_across+iacross] * l1b_i_polsample_cur[iwave];
                                // Turned into variances
                                l1c_i_polsample_noise[iwave] += pow(factormat_cur[ialong*dim_across+iacross] * l1b_i_polsample_noise_cur[iwave],2.0);
                                l1c_q[iwave] += factormat_cur[ialong*dim_across+iacross] * l1b_q_cur[iwave];
                                // Turned into variances
                                l1c_q_noise[iwave] += pow(factormat_cur[ialong*dim_across+iacross] * l1b_q_noise_cur[iwave],2.0);
                                l1c_u[iwave] += factormat_cur[ialong*dim_across+iacross] * l1b_u_cur[iwave];
                                // Turned into variances
                                l1c_u_noise[iwave] += pow(factormat_cur[ialong*dim_across+iacross] * l1b_u_noise_cur[iwave],2.0);
                                // This is already a variance. Still
                                // averages with averaging out noise
                                // assuming there is no correlation
                                // between one Q and the other U.
                                l1c_qu_noise_covar[iwave] +=
                                  pow(factormat_cur[ialong*dim_across+iacross],2.0) * l1b_qu_noise_covar_cur[iwave];
                            }
                            l1b_i_polsample_cur += dim_pol_wave;
                            l1b_i_polsample_noise_cur += dim_pol_wave;
                            l1b_q_cur += dim_pol_wave;
                            l1b_q_noise_cur += dim_pol_wave;
                            l1b_u_cur += dim_pol_wave;
                            l1b_u_noise_cur += dim_pol_wave;
                            l1b_qu_noise_covar_cur += dim_pol_wave;
                        }
                    }
                    for (size_t iwave=0 ; iwave<dim_pol_wave ; iwave++) {
                        l1c_dolp[iwave] = sqrt(pow(l1c_q[iwave],2.0) + pow(l1c_u[iwave],2.0));
                        l1c_aolp[iwave] = 0.5*atan2(l1c_u[iwave],l1c_q[iwave]) / math::deg_to_rad;
                        l1c_dolp_noise[iwave] = sqrt(
                          // Last term is still a variance
                          pow(l1c_q[iwave] / l1c_dolp[iwave],2.0) * l1c_q_noise[iwave] +
                          // Last term is still a variance
                          pow(l1c_u[iwave] / l1c_dolp[iwave],2.0) * l1c_u_noise[iwave] +
                          l1c_q[iwave]*l1c_u[iwave] / pow(l1c_dolp[iwave],2.0) * l1c_qu_noise_covar[iwave]
                        );
                        l1c_aolp_noise[iwave] = 0.5 * sqrt(
                          // Last term is still a variance
                          pow(l1c_u[iwave] / pow(l1c_dolp[iwave],2.0),2.0) * l1c_q_noise[iwave] +
                          // Last term is still a variance
                          pow(l1c_q[iwave] / pow(l1c_dolp[iwave],2.0),2.0) * l1c_u_noise[iwave] -
                          l1c_q[iwave]*l1c_u[iwave] / pow(l1c_dolp[iwave],4.0) * l1c_qu_noise_covar[iwave]
                        ) / math::deg_to_rad;
                        // Now turn the variances on I, q and u into noise
                        l1c_i_polsample_noise[iwave] = sqrt(l1c_i_polsample_noise[iwave]);
                        l1c_q_noise[iwave] = sqrt(l1c_q_noise[iwave]);
                        l1c_u_noise[iwave] = sqrt(l1c_u_noise[iwave]);
                    }
                    // Write the results
                    var_l1c_i_polsample.putVar(l1c_start,l1c_cnt,l1c_i_polsample.data());
                    var_l1c_i_polsample_noise.putVar(l1c_start,l1c_cnt,l1c_i_polsample_noise.data());
                    var_l1c_q.putVar(l1c_start,l1c_cnt,l1c_q.data());
                    var_l1c_q_noise.putVar(l1c_start,l1c_cnt,l1c_q_noise.data());
                    var_l1c_u.putVar(l1c_start,l1c_cnt,l1c_u.data());
                    var_l1c_u_noise.putVar(l1c_start,l1c_cnt,l1c_u_noise.data());
                    var_l1c_dolp.putVar(l1c_start,l1c_cnt,l1c_dolp.data());
                    var_l1c_dolp_noise.putVar(l1c_start,l1c_cnt,l1c_dolp_noise.data());
                    var_l1c_aolp.putVar(l1c_start,l1c_cnt,l1c_aolp.data());
                    var_l1c_aolp_noise.putVar(l1c_start,l1c_cnt,l1c_aolp_noise.data());
                }
                factormat_cur += dim_along*dim_across;
            }
        }

        spdlog::info("Interpolating solar and viewing geometry");
        // Geometry
        {
            std::vector<double> l1b_sza(dim_along*dim_across);
            std::vector<double> l1b_saa(dim_along*dim_across);
            std::vector<double> l1b_vza(dim_along*dim_across);
            std::vector<double> l1b_vaa(dim_along*dim_across);
            std::vector<Vector> l1b_sun(dim_along*dim_across);
            std::vector<Vector> l1b_view(dim_along*dim_across);

            // Slice out one viewport
            std::vector<size_t> l1b_start = {0,ifov_start[ivp]};
            std::vector<size_t> l1b_cnt = {dim_along,dim_across};
            var_l1b_sza.getVar(l1b_start,l1b_cnt,l1b_sza.data());
            var_l1b_saa.getVar(l1b_start,l1b_cnt,l1b_saa.data());
            var_l1b_vza.getVar(l1b_start,l1b_cnt,l1b_vza.data());
            var_l1b_vaa.getVar(l1b_start,l1b_cnt,l1b_vaa.data());
            // Turn angles into something interpolatable
            for (size_t imat=0 ; imat<dim_along*dim_across ; imat++) {
                // Go to the correct units
                l1b_sza[imat] *= math::deg_to_rad;
                l1b_saa[imat] *= math::deg_to_rad;
                l1b_vza[imat] *= math::deg_to_rad;
                l1b_vaa[imat] *= math::deg_to_rad;
                // Now, turn them into vectors.
                // We have to interpret zenith as altitude and azimuth
                // as longitude. But for the zenith, we need to flip
                // sine and cosine, because it is a zero (or 180 degrees)
                // angle that is in the singularity where the azimuth
                // does not matter.
                l1b_sun[imat] = Vector (
                    sin(l1b_sza[imat]) * cos(l1b_saa[imat]) ,
                    sin(l1b_sza[imat]) * sin(l1b_saa[imat]) ,
                    cos(l1b_sza[imat])
                );
                l1b_view[imat] = Vector (
                    sin(l1b_vza[imat]) * cos(l1b_vaa[imat]) ,
                    sin(l1b_vza[imat]) * sin(l1b_vaa[imat]) ,
                    cos(l1b_vza[imat])
                );
            }
            factormat_cur = factormat.data();
            for (size_t iobs=0 ; iobs<dim_l1c_obs ; iobs++) {

                // Where to write
                std::vector<size_t> l1c_start = {iobs,0,ivp};
                std::vector<size_t> l1c_cnt = {1,1,1};

                if (fail[iobs]) {

                    // Write fill values
                    double fillvalue = NC_FILL_DOUBLE;
                    var_l1c_sza.putVar(l1c_start,l1c_cnt,&fillvalue);
                    var_l1c_saa.putVar(l1c_start,l1c_cnt,&fillvalue);
                    var_l1c_vza.putVar(l1c_start,l1c_cnt,&fillvalue);
                    var_l1c_vaa.putVar(l1c_start,l1c_cnt,&fillvalue);

                } else {

                    Vector l1c_sun(0.0,0.0,0.0);
                    Vector l1c_view(0.0,0.0,0.0);
                    Vector* l1b_sun_cur = l1b_sun.data();
                    Vector* l1b_view_cur = l1b_view.data();
                    for (size_t ialong=0 ; ialong<dim_along ; ialong++) {
                        for (size_t iacross=0 ; iacross<dim_across ; iacross++) {
                            l1c_sun += factormat_cur[ialong*dim_across+iacross] * *l1b_sun_cur;
                            l1c_view += factormat_cur[ialong*dim_across+iacross] * *l1b_view_cur;
                            l1b_sun_cur++;
                            l1b_view_cur++;
                        }
                    }
                    // Convert vectors to angles
                    // Note that the averged vector is not normalized
                    l1c_sun.normalize();
                    l1c_view.normalize();
                    // Now they are
                    double l1c_sza = acos(l1c_sun[2]) / math::deg_to_rad;
                    double l1c_saa = atan2(l1c_sun[1],l1c_sun[0]) / math::deg_to_rad;
                    double l1c_vza = acos(l1c_view[2]) / math::deg_to_rad;
                    double l1c_vaa = atan2(l1c_view[1],l1c_view[0]) / math::deg_to_rad;

                    // Write the results
                    var_l1c_sza.putVar(l1c_start,l1c_cnt,&l1c_sza);
                    var_l1c_saa.putVar(l1c_start,l1c_cnt,&l1c_saa);
                    var_l1c_vza.putVar(l1c_start,l1c_cnt,&l1c_vza);
                    var_l1c_vaa.putVar(l1c_start,l1c_cnt,&l1c_vaa);
                }
                factormat_cur += dim_along*dim_across;
            }
        }
    } // Loop over viewports
}

} // namespace spex
