// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1x.h"

#include "ckd.h"
#include "l1a.h"
#include "spectra.h"

#include <algorithm>

namespace spex {

L1X::L1X(const gadfit::MPIVars& mpi,
         const CKD& ckd,
         const std::vector<L1A>& l1a_products,
         const ProcLevel level,
         const std::string& filename)
  : mpi { mpi }, ckd { ckd }, level { level }, filename { filename }
{
    if (mpi.rank != 0) {
        return;
    }
    NcObject nc { netCDF::NcFile::replace, filename };

    // Global attributes

    nc.putAtt("processing_level", "L1X");
    nc.putAtt("l1x_level", netCDF::ncInt, static_cast<int>(level));

    // Dimensions

    const auto number_of_images {
        nc.addDim("number_of_images", l1a_products.size()),
    };
    netCDF::NcDim samples_per_image {};
    netCDF::NcDim spatial_detector_pixels {};
    netCDF::NcDim spatial_samples_per_image {};
    netCDF::NcDim polarization_directions {};
    if (level <= ProcLevel::stray) {
        samples_per_image = nc.addDim(
          "samples_per_image", l1a_products.front().binning_table->nBins());
        spatial_detector_pixels =
          nc.addDim("spatial_detector_pixels", ckd.dim_detector_spat);
    } else if (level < ProcLevel::l1b) {
        spatial_samples_per_image =
          nc.addDim("spatial_samples_per_image", ckd.fov.n_act);
        polarization_directions =
          nc.addDim("polarization_directions", dims::pol);
    }
    // Spectral dimension is required for both the detector image and
    // when saving spectra. The spectra will have different sizes
    // depending on the viewport and ACT angle (FOV index) but to keep
    // storage simple we use the original number of columns (i.e. the
    // spectral dimension) when saving them.
    const auto spectral_detector_pixels {
        nc.addDim("spectral_detector_pixels", ckd.dim_detector_spec),
    };
    const auto hk_packets {
        nc.addDim("hk_packets", l1a_products.front().meta->hk_packets),
    };
    const auto sc_records {
        nc.addDim("SC_records", l1a_products.front().meta->sc_records),
    };
    const auto vector_elements { nc.addDim("vector_elements", dims::vec) };
    const auto quaternion_elements {
        nc.addDim("quaternion_elements", dims::quat),
    };
    constexpr std::array<uint8_t, 2> mask_values { 0, 1 };

    // Image attributes

    auto image_attributes { nc.addGroup("image_attributes") };
    auto image_time {
        image_attributes.addVar(
          "image_time", netCDF::ncDouble, number_of_images),
    };
    image_time.putAtt("long_name", "image time (seconds of day)");
    image_time.putAtt("valid_min", netCDF::ncDouble, 0.0);
    image_time.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
    image_time.putAtt("units", "seconds");

    auto image_CCSDS_sec {
        image_attributes.addVar(
          "image_CCSDS_sec", netCDF::ncUint, number_of_images),
    };
    image_CCSDS_sec.putAtt("long_name",
                           "image CCSDS time (seconds since 1970)");
    image_CCSDS_sec.putAtt("valid_min", netCDF::ncUint, 1900000000);
    image_CCSDS_sec.putAtt("valid_max", netCDF::ncUint, 2400000000);
    image_CCSDS_sec.putAtt("units", "seconds");

    auto image_CCSDS_subsec {
        image_attributes.addVar(
          "image_CCSDS_subsec", netCDF::ncUshort, number_of_images),
    };
    image_CCSDS_subsec.putAtt("long_name", "image CCSDS time (sub-seconds)");
    image_CCSDS_subsec.putAtt("valid_min", netCDF::ncUshort, 0);
    image_CCSDS_subsec.putAtt("valid_max", netCDF::ncUshort, 65535);
    image_CCSDS_subsec.putAtt("units", "1/65536 s");

    auto binning_table {
        image_attributes.addVar(
          "binning_table", netCDF::ncUbyte, number_of_images),
    };
    binning_table.putAtt("long_name", "binning-table ID");
    binning_table.putAtt("valid_min", netCDF::ncUbyte, 0);
    binning_table.putAtt("valid_max", netCDF::ncUbyte, 255);

    auto digital_offset {
        image_attributes.addVar(
          "digital_offset", netCDF::ncShort, number_of_images),
    };
    digital_offset.putAtt("long_name", "digital offset");
    digital_offset.putAtt("units", "1");

    auto nr_coadditions {
        image_attributes.addVar(
          "nr_coadditions", netCDF::ncUshort, number_of_images),
    };
    nr_coadditions.putAtt("_FillValue", netCDF::ncUshort, NC_FILL_USHORT);
    nr_coadditions.putAtt("long_name", "number of coadditions");
    nr_coadditions.putAtt("units", "1");

    auto exposure_time {
        image_attributes.addVar(
          "exposure_time", netCDF::ncDouble, number_of_images),
    };
    exposure_time.putAtt("_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
    exposure_time.putAtt("long_name", "exposure time");
    exposure_time.putAtt("units", "s");

    // Now that the variables have been defined fill in the values in
    // a single loop.
    for (size_t i_l1a {}; i_l1a < l1a_products.size(); ++i_l1a) {
        const std::vector<size_t> start { i_l1a };
        const std::vector<size_t> count { 1 };
        // Image_attributes
        image_time.putVar(start, count, &l1a_products.at(i_l1a).image_time);
        image_CCSDS_sec.putVar(start, count, &l1a_products.at(i_l1a).seconds);
        nr_coadditions.putVar(
          start, count, &l1a_products.at(i_l1a).nr_coadditions);
        const uint16_t subseconds {
            static_cast<uint16_t>(
              std::round(65535 * l1a_products.at(i_l1a).second_fraction)),
        };
        image_CCSDS_subsec.putVar(start, count, &subseconds);
        binning_table.putVar(
          start, count, &l1a_products.at(i_l1a).binning_table_id);
        digital_offset.putVar(
          start, count, &l1a_products.at(i_l1a).digital_offset);
        exposure_time.putVar(
          start, count, &l1a_products.at(i_l1a).exposure_time);
    }

    // Science data

    auto science_data { nc.addGroup("science_data") };
    if (level <= ProcLevel::stray) {
        std::vector<std::size_t> chunksize_detector {
            1, static_cast<size_t>(l1a_products.front().binning_table->nBins())
        };
        std::vector<std::size_t> chunksize_detector_unbinned {
            1,
            static_cast<size_t>(ckd.dim_detector_spat),
            static_cast<size_t>(ckd.dim_detector_spec)
        };
        // Binned and unbinned detector images
        auto detector_images {
            science_data.addVar("detector_images",
                                netCDF::ncDouble,
                                { number_of_images, samples_per_image }),
        };
        detector_images.setChunking(netCDF::NcVar::nc_CHUNKED,
                                    chunksize_detector);
        detector_images.putAtt("_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        detector_images.putAtt("long_name",
                               "(binned) detector images at current stage.");
        detector_images.putAtt("units", "counts");
        auto detector_images_unbinned {
            science_data.addVar("detector_images_unbinned",
                                netCDF::ncDouble,
                                { number_of_images,
                                  spatial_detector_pixels,
                                  spectral_detector_pixels }),
        };
        detector_images_unbinned.setChunking(netCDF::NcVar::nc_CHUNKED,
                                             chunksize_detector_unbinned);
        detector_images_unbinned.putAtt(
          "_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        detector_images_unbinned.putAtt(
          "long_name", "unbinned detector images at current stage.");
        detector_images_unbinned.putAtt("units", "counts");
        // Binned and unbinned detector image masks
        auto pixel_mask {
            science_data.addVar("pixel_mask",
                                netCDF::ncUbyte,
                                { number_of_images, samples_per_image }),
        };
        pixel_mask.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize_detector);
        pixel_mask.putAtt("_FillValue", netCDF::ncUbyte, NC_FILL_UBYTE);
        pixel_mask.putAtt("long_name", "(binned) detector pixel mask.");
        pixel_mask.putAtt("flag_values",
                          netCDF::ncUbyte,
                          mask_values.size(),
                          mask_values.data());
        pixel_mask.putAtt("flag_meanings", "good bad");
        auto pixel_mask_unbinned {
            science_data.addVar("pixel_mask_unbinned",
                                netCDF::ncUbyte,
                                { number_of_images,
                                  spatial_detector_pixels,
                                  spectral_detector_pixels }),
        };
        pixel_mask_unbinned.setChunking(netCDF::NcVar::nc_CHUNKED,
                                        chunksize_detector_unbinned);
        pixel_mask_unbinned.putAtt(
          "_FillValue", netCDF::ncUbyte, NC_FILL_UBYTE);
        pixel_mask_unbinned.putAtt("long_name",
                                   "unbinned detector pixel mask.");
        pixel_mask_unbinned.putAtt("flag_values",
                                   netCDF::ncUbyte,
                                   mask_values.size(),
                                   mask_values.data());
        pixel_mask_unbinned.putAtt("flag_meanings", "good bad");
        // Binned and unbinned detector image noise
        auto detector_noise {
            science_data.addVar("detector_noise",
                                netCDF::ncDouble,
                                { number_of_images, samples_per_image }),
        };
        detector_noise.setChunking(netCDF::NcVar::nc_CHUNKED,
                                   chunksize_detector);
        detector_noise.putAtt("_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        detector_noise.putAtt("long_name",
                              "(binned) noise of the detector images");
        detector_noise.putAtt("units", "counts");
        auto detector_noise_unbinned {
            science_data.addVar("detector_noise_unbinned",
                                netCDF::ncDouble,
                                { number_of_images,
                                  spatial_detector_pixels,
                                  spectral_detector_pixels }),
        };
        detector_noise_unbinned.setChunking(netCDF::NcVar::nc_CHUNKED,
                                            chunksize_detector_unbinned);
        detector_noise_unbinned.putAtt(
          "_FillValue", netCDF::ncDouble, NC_FILL_DOUBLE);
        detector_noise_unbinned.putAtt("long_name",
                                       "unbinned noise of the detector images");
        detector_noise_unbinned.putAtt("units", "counts");
    } else if (level < ProcLevel::l1b) {
        std::vector<std::size_t> chunksize {
            1,
            static_cast<size_t>(ckd.fov.n_act),
            static_cast<size_t>(dims::pol),
            static_cast<size_t>(ckd.dim_detector_spec),
        };
        const std::string unit { level == ProcLevel::fov ? "counts"
                                                         : "W.m-2.sr-1.um-1" };
        // Spectra can have different sizes which need to be recorded separately
        auto spectra_sizes {
            science_data.addVar(
              "spectra_sizes", netCDF::ncInt, spatial_samples_per_image),
        };
        spectra_sizes.putAtt("long_name",
                             "size (number of wavelengths) of each spectrum");
        spectra_sizes.putAtt("units", "1");
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            const std::vector<size_t> start { static_cast<size_t>(i_fov) };
            const std::vector<size_t> count { 1 };
            const auto size { ckd.wave_full.front().at(i_fov).size() };
            spectra_sizes.putVar(start, count, &size);
        }
        // Need to save wavelengths of all spectra because they differ
        // from those in the CKD which are unbinned.
        auto wavelengths {
            science_data.addVar("wavelengths",
                                netCDF::ncDouble,
                                { spatial_samples_per_image,
                                  polarization_directions,
                                  spectral_detector_pixels }),
        };
        wavelengths.putAtt("long_name",
                           "wavelengths of extracted spectra at current stage");
        wavelengths.putAtt("units", "nm");
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
                const std::vector<size_t> start { static_cast<size_t>(i_fov),
                                                  static_cast<size_t>(i_pol),
                                                  0 };
                const std::vector<size_t> count {
                    1, 1, ckd.wave_full.at(i_pol).at(i_fov).size()
                };
                if (level < ProcLevel::rad) {
                    wavelengths.putVar(
                      start, count, ckd.wave_full.at(i_pol).at(i_fov).data());
                } else {
                    wavelengths.putVar(
                      start, count, ckd.wave_common.at(i_fov).data());
                }
            }
        }
        // All spectra
        auto spectra {
            science_data.addVar("spectra",
                                netCDF::ncDouble,
                                { number_of_images,
                                  spatial_samples_per_image,
                                  polarization_directions,
                                  spectral_detector_pixels }),
        };
        spectra.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize);
        spectra.putAtt("long_name", "extracted spectra at current stage");
        spectra.putAtt("units", unit);
        // All spectral noise
        auto spectra_noise { science_data.addVar(
          "spectra_noise",
          netCDF::ncDouble,
          { number_of_images,
            spatial_samples_per_image,
            polarization_directions,
            spectral_detector_pixels }) };
        spectra_noise.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize);
        spectra_noise.putAtt(
          "long_name", "noise estimate on extracted spectra at current stage");
        spectra_noise.putAtt("units", unit);
        // All spectral masks
        auto spectra_mask {
            science_data.addVar("spectra_mask",
                                netCDF::ncUbyte,
                                { number_of_images,
                                  spatial_samples_per_image,
                                  polarization_directions,
                                  spectral_detector_pixels }),
        };
        spectra_mask.setChunking(netCDF::NcVar::nc_CHUNKED, chunksize);
        spectra_mask.putAtt("_FillValue", netCDF::ncUbyte, NC_FILL_UBYTE);
        spectra_mask.putAtt("long_name", "spectral mask at current stage");
        spectra_mask.putAtt("flag_values",
                            netCDF::ncUbyte,
                            mask_values.size(),
                            mask_values.data());
        spectra_mask.putAtt("flag_meanings", "good bad");
    }

    // Engineering data

    auto engineering_data { nc.addGroup("engineering_data") };
    auto hk_tlm_time {
        engineering_data.addVar("HK_tlm_time", netCDF::ncDouble, hk_packets),
    };
    hk_tlm_time.putAtt("long_name",
                       "HK telemetry packet time (seconds of day)");
    hk_tlm_time.putAtt("valid_min", netCDF::ncDouble, 0.0);
    hk_tlm_time.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
    hk_tlm_time.putAtt("units", "seconds");
    hk_tlm_time.putVar(l1a_products.front().meta->HK_tlm_time.data());

    auto temp_detector {
        engineering_data.addVar("temp_detector", netCDF::ncDouble, hk_packets),
    };
    temp_detector.putAtt("long_name", "detector temperature");
    temp_detector.putAtt("comment", "TS1 DEM Temperature (nominal)");
    temp_detector.putAtt("valid_min", netCDF::ncDouble, 260.0);
    temp_detector.putAtt("valid_max", netCDF::ncDouble, 300.0);
    temp_detector.putAtt("units", "K");
    temp_detector.putVar(l1a_products.front().meta->temp_detector.data());

    // Navigation data

    if (!l1a_products.front().meta->nav_att_time.empty()) {
        auto navigation_data { nc.addGroup("navigation_data") };
        auto att_time {
            navigation_data.addVar("att_time", netCDF::ncDouble, sc_records),
        };
        att_time.putAtt("long_name", "Attitude sample time (seconds of day)");
        att_time.putAtt("valid_min", netCDF::ncDouble, 0.0);
        att_time.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
        att_time.putAtt("units", "seconds");
        att_time.putVar(l1a_products.front().meta->nav_att_time.data());

        auto att_quat {
            navigation_data.addVar("att_quat",
                                   netCDF::ncDouble,
                                   { sc_records, quaternion_elements }),
        };
        att_quat.putAtt("long_name",
                        "attitude quaternions (J2000 to spacecraft)");
        att_quat.putAtt("valid_min", netCDF::ncDouble, -1.0);
        att_quat.putAtt("valid_max", netCDF::ncDouble, 1.0);
        att_quat.putAtt("units", "1");
        // Values will be inserted later along with velocities

        auto orb_time {
            navigation_data.addVar("orb_time", netCDF::ncDouble, sc_records),
        };
        orb_time.putAtt("long_name", "orbit vector time (seconds of day)");
        orb_time.putAtt("valid_min", netCDF::ncDouble, 0.0);
        orb_time.putAtt("valid_max", netCDF::ncDouble, 86400.999999);
        orb_time.putAtt("units", "seconds");
        orb_time.putVar(l1a_products.front().meta->nav_orb_time.data());

        auto orb_pos {
            navigation_data.addVar(
              "orb_pos", netCDF::ncDouble, { sc_records, vector_elements }),
        };
        orb_pos.putAtt("long_name", "orbit positions vectors (ECR)");
        orb_pos.putAtt("valid_min", netCDF::ncDouble, -7200000.0);
        orb_pos.putAtt("valid_max", netCDF::ncDouble, 7200000.0);
        orb_pos.putAtt("units", "m");
        // Values will be inserted later along with velocities

        auto orb_vel {
            navigation_data.addVar(
              "orb_vel", netCDF::ncDouble, { sc_records, vector_elements }),
        };
        orb_vel.putAtt("long_name", "orbit velocity vectors (ECR)");
        orb_vel.putAtt("valid_min", netCDF::ncDouble, -7600.0);
        orb_vel.putAtt("valid_max", netCDF::ncDouble, 7600.0);
        orb_vel.putAtt("units", "meters/seconds");
        orb_vel.putVar(l1a_products.front().meta->nav_orb_vel.data());

        std::vector<double> att_quat_values(
          l1a_products.front().meta->sc_records * dims::quat);
        std::vector<double> orb_pos_values(l1a_products.front().meta->sc_records
                                           * dims::vec);
        std::vector<double> orb_vel_values(orb_pos_values.size());
        for (int i {}; i < l1a_products.front().meta->sc_records; ++i) {
            for (int j {}; j < dims::vec; ++j) {
                att_quat_values.at(i * dims::quat + j) =
                  l1a_products.front().meta->nav_att_quat.at(i).v.v.at(j);
                orb_pos_values.at(i * dims::vec + j) =
                  l1a_products.front().meta->nav_orb_pos.at(i).v.at(j);
                orb_vel_values.at(i * dims::vec + j) =
                  l1a_products.front().meta->nav_orb_vel.at(i).v.at(j);
            }
            att_quat_values.at(i * dims::quat + dims::vec) =
              l1a_products.front().meta->nav_att_quat.at(i).r;
        }
        orb_pos.putVar(orb_pos_values.data());
        orb_vel.putVar(orb_vel_values.data());
        att_quat.putVar(att_quat_values.data());
    }
}

auto L1X::writeImage(const L1A& l1a) -> void
{
    std::unique_ptr<NcObject> nc {};
    // Make sure only one process is writing to the file at a time
    while (true) {
        try {
            nc = std::make_unique<NcObject>(netCDF::NcFile::write, filename);
            break;
        } catch (const netCDF::exceptions::NcHdfErr&) {
        }
    }
    auto science_data { nc->getGroup("science_data") };
    const std::vector<size_t> start { l1a.l1x_image_offset, 0 };
    const std::vector<size_t> count { 1, l1a.image.size() };
    const std::vector<size_t> start_unbinned { l1a.l1x_image_offset, 0, 0 };
    const std::vector<size_t> count_unbinned {
        1,
        static_cast<size_t>(ckd.dim_detector_spat),
        static_cast<size_t>(ckd.dim_detector_spec)
    };
    // Write current detector image in binned and unbinned formats
    science_data.getVar("detector_images")
      .putVar(start, count, l1a.image.data());
    std::vector<double> image_unbinned(ckd.npix);
    l1a.binning_table->unbin(l1a.image, image_unbinned);
    science_data.getVar("detector_images_unbinned")
      .putVar(start_unbinned, count_unbinned, image_unbinned.data());
    // Write current detector image mask
    std::vector<uint8_t> pixel_mask_u8(l1a.pixel_mask.size());
    std::transform(l1a.pixel_mask.cbegin(),
                   l1a.pixel_mask.cend(),
                   pixel_mask_u8.begin(),
                   [](const auto x) { return x ? 1 : 0; });
    science_data.getVar("pixel_mask")
      .putVar(start, count, pixel_mask_u8.data());
    std::vector<uint8_t> pixel_mask_unbinned_u8(ckd.npix);
    l1a.binning_table->unbin(pixel_mask_u8, pixel_mask_unbinned_u8);
    science_data.getVar("pixel_mask_unbinned")
      .putVar(start_unbinned, count_unbinned, pixel_mask_unbinned_u8.data());
    science_data.getVar("detector_noise")
      .putVar(start, count, l1a.noise.data());
    std::vector<double> noise_unbinned(ckd.npix);
    l1a.binning_table->unbin(l1a.noise, noise_unbinned);
    science_data.getVar("detector_noise_unbinned")
      .putVar(start_unbinned, count_unbinned, noise_unbinned.data());
}

auto L1X::writeSpectrum(const Spectra& spectra,
                        const size_t l1x_image_offset,
                        const int i_fov) -> void
{
    std::unique_ptr<NcObject> nc {};
    // Make sure only one process is writing to the file at a time
    while (true) {
        try {
            nc = std::make_unique<NcObject>(netCDF::NcFile::write, filename);
            break;
        } catch (const netCDF::exceptions::NcHdfErr&) {
        }
    }
    auto science_data { nc->getGroup("science_data") };
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        const std::vector<size_t> start { static_cast<size_t>(l1x_image_offset),
                                          static_cast<size_t>(i_fov),
                                          static_cast<size_t>(i_pol),
                                          0 };
        const std::vector<size_t> count {
            1, 1, 1, spectra.signal.front().size()
        };
        science_data.getVar("spectra").putVar(
          start, count, spectra.signal.at(i_pol).data());
        science_data.getVar("spectra_noise")
          .putVar(start, count, spectra.noise.at(i_pol).data());
        std::vector<uint8_t> mask_u8(spectra.signal.front().size());
        std::transform(spectra.mask.at(i_pol).cbegin(),
                       spectra.mask.at(i_pol).cend(),
                       mask_u8.begin(),
                       [](const auto x) { return x ? 1 : 0; });
        science_data.getVar("spectra_mask")
          .putVar(start, count, mask_u8.data());
    }
}

} // namespace spex
