// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for recording and saving a partially calibrated image. The
// data levels are L1A for the raw data and L1B for the calibrated
// radiance, DoLP, and AoLP spectra. Anything in between we call level
// L1X. For instance, we may calibrate an image up to and including
// stray light and write it to an output file. That file can then be
// read in during the next run and the calculation may proceed from
// where we left off.

#pragma once

#include "ckd.h"
#include "constants.h"
#include "l1a_metadata.h"

namespace spex {

class CKD;
class L1A;
class Spectra;

class L1X
{
private:
    const gadfit::MPIVars& mpi;
    const CKD& ckd;
    ProcLevel level {};
    // The output file, in NetCDF format, is opened and closed every
    // time we need to write something into it. For that we need to
    // store the file name.
    const std::string filename {};

public:
    L1X(const gadfit::MPIVars& mpi,
        const CKD& ckd,
        const std::vector<L1A>& l1a_products,
        const ProcLevel level,
        const std::string& filename);
    // Write the current state of an L1A instance to an output file if
    // the current state is calibrated for no more than stray light.
    auto writeImage(const L1A& l1a) -> void;
    // If we are past stray light calibration, i.e. we are extracting
    // spectra and perfoming actions on them, then call this function
    // instead. It writes data associated with a spectrum instead of a
    // partially calibrated detector image to file.
    //
    // Arguments:
    //   spectra - +/- spectra, noise, and the spectral mask
    //             corresponding to one ACT angle
    //   l1x_image_offset - which image in the L1X instance is owned
    //                      by this L1A instance
    //   i_fov - ACT angle index (field of view or FOV) for which the
    //           read function is called
    auto writeSpectrum(const Spectra& spectra,
                       const size_t l1x_image_offset,
                       const int i_fov) -> void;
    ~L1X() = default;
};

} // namespace spex
