// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Linear algebra routines

#pragma once

#include <vector>

extern "C"
{
    auto dgttrf_(const int*, double*, double*, double*, double*, int*, int*)
      -> void;
    auto dgttrs_(const char*,
                 const int*,
                 const int*,
                 const double*,
                 const double*,
                 const double*,
                 const double*,
                 const int*,
                 double*,
                 const int*,
                 int*) -> void;
    auto dppsv_(const char*,
                const int*,
                const int*,
                double*,
                double*,
                const int*,
                int*) -> void;
    auto ddot_(const int*, const double*, const int*, const double*, const int*)
      -> double;
}

// Compute the solution to a real system of linear equations A * X =
// B, where A is an N-by-N symmetric positive definite matrix stored
// in packed format and X and B are N-by-NRHS matrices.
auto dppsv(const auto& data,
           const auto& std_devs,
           std::vector<double>& B,
           std::vector<double>& state_vector) -> void
{
    // Number of state parameters
    const int n_states { static_cast<int>(B.size() / data.size()) };
    // Scale each row of data by noise (standard deviations)
    for (int i {}; i < static_cast<int>(data.size()); ++i) {
        for (int j {}; j < n_states; ++j) {
            B[i * n_states + j] /= std_devs[i];
        }
    }
    // B^T x B
    std::vector<double> BTB(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int j {}; j < n_states; ++j) {
            for (int k {}; k < static_cast<int>(data.size()); ++k) {
                BTB[i * n_states + j] +=
                  B[k * n_states + i] * B[k * n_states + j];
            }
        }
    }
    // B^T x Y where Y is the data vector
    std::vector<double> BTY(static_cast<int>(data.size()), 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < static_cast<int>(data.size()); ++k) {
            BTY[i] += B[k * n_states + i] * data[k] / std_devs[k];
        }
    }
    // Upper triangle of B^T x B
    std::vector<double> BTB_tri(n_states * (n_states + 1) / 2);
    for (int j {}; j < n_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j + 1) * j / 2] = BTB[i * n_states + j];
        }
    }
    // Covariance matrix of linear inversion initialized to unity
    std::vector<double> Sx(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        Sx[i * n_states + i] = 1.0;
    }
    constexpr char uplo { 'u' };
    int info {};
    dppsv_(
      &uplo, &n_states, &n_states, BTB_tri.data(), Sx.data(), &n_states, &info);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < n_states; ++k) {
            state_vector[i] += Sx[i * n_states + k] * BTY[k];
        }
    }
}
