// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "nonlincal.h"

#include "ckd.h"
#include "functions.h"
#include "l1a.h"
#include "l1a_metadata.h"
#include "lininv.h"
#include "matrix.h"

#include <cassert>
#include <spdlog/spdlog.h>

namespace spex {

auto Nonlincal::processInit() -> void
{
    // Create the fitting B-spline, meanwhile leaving all essenntial information
    // on the 'nonlin'-prefixed variables in the CKD structure.
    ckd.dim_nonlin_knot = settings.nonlin.knots.size();
    ckd.nonlin_knots.resize(ckd.dim_nonlin_knot);
    memcpy(ckd.nonlin_knots.data(),settings.nonlin.knots.data(),ckd.dim_nonlin_knot*sizeof(double));
    ckd.nonlin_order = settings.nonlin.order;
    b = std::make_unique<Bspline>(ckd.nonlin_order,ckd.dim_nonlin_knot,ckd.nonlin_knots.data());
    ckd.dim_nonlin_spline = b->nspline;

    // Recycled element: The abscissa
    abscissa.resize(nL1A());
    for (size_t il1a=0 ; il1a<nL1A() ; il1a++) {
        abscissa[il1a] = l1a_manager->l1a_products[il1a].exposure_time*l1a_manager->l1a_products[il1a].meta->illumination_level;
    }

    // Find out how many exposure times are used
    ckd.dim_nonlin_exptime = settings.nonlin.exposure_times.size();
    ckd.nonlin_exptimes.resize(ckd.dim_nonlin_exptime);
    memcpy(ckd.nonlin_exptimes.data(),settings.nonlin.exposure_times.data(),ckd.dim_nonlin_exptime*sizeof(double));
    // Verify that no exposure time is inside the tolerance limit of
    // two different selected exposure times.
    double tol2 = pow(settings.nonlin.exptime_tol,2.0);
    for (size_t iexptime=0 ; iexptime<ckd.dim_nonlin_exptime ; iexptime++) {
        double& exptime1 = ckd.nonlin_exptimes[iexptime];
        for (size_t iexptime_check=0 ; iexptime_check<iexptime ; iexptime_check++) {
            double& exptime0 = ckd.nonlin_exptimes[iexptime_check];
        }
    }

    // First, sort out which files belong to a chosen exposure time (iexptime). If it not not
    // associated to an exposure time, mark dim_nonlin_exptime.
    // Meanwhile, count the number of files per exposure time.
    std::vector<size_t> iexptime_l1a(nL1A());
    std::vector<size_t> nl1a_exptime(ckd.dim_nonlin_exptime,0);
    for (size_t il1a=0 ; il1a<nL1A() ; il1a++) {
        double& exptime_l1a = l1a_manager->l1a_products[il1a].exposure_time;
        bool found = false;
        for (size_t iexptime=0 ; iexptime<ckd.dim_nonlin_exptime ; iexptime++) {
            double& exptime_select = ckd.nonlin_exptimes[iexptime];
            if (exptime_l1a / exptime_select <= settings.nonlin.exptime_tol && exptime_select / exptime_l1a <= settings.nonlin.exptime_tol) {
                found = true;
                iexptime_l1a[il1a] = iexptime;
                nl1a_exptime[iexptime]++;
                break;
            }
        }
        if (!found) iexptime_l1a[il1a] = ckd.dim_nonlin_exptime;
    }

    // Verify that every exposure time has emough images
    for (size_t iexptime=0 ; iexptime<ckd.dim_nonlin_exptime ; iexptime++) {
        assert(!(nl1a_exptime[iexptime] < ckd.dim_nonlin_spline) && "Unable to perform non-linearity fit for exposure time");
    }

    // The following batches will be formed
    // - Highest illumination levels of chosen exposure times
    // --- dim_nonlin_exptime batches, not chunked
    // - All files
    // --- One batch, chunked
    // - All files for chosen exposure time (dim_nonlin_exptime batches of multiple files)
    // --- dim_nonlin_exptime batches, chunked

    // Add administration to the batches: The phase that belongs to the batch and the exposure time
    // that belongs to the batch (or dim_nonlin_exptime if it is not specific to an exposure time).

    // Total number of batches: dim_nonlin_exptime + nfrac + dim_nonlin_exptime*nfrac
    const int nbatch { ckd.dim_nonlin_exptime + 1 + ckd.dim_nonlin_exptime };
    batches.resize(nbatch);
    phase_batches.resize(nbatch);
    iexptime_batches.resize(nbatch);
    ifrac_batches.resize(nbatch);
    // Reverse this administration for here
    std::vector<size_t> ibatch_medianphase(ckd.dim_nonlin_exptime);
    std::vector<size_t> ibatch_linphase(1);
    std::vector<size_t> ibatch_nonlinphase(ckd.dim_nonlin_exptime); // Slow dimension is fraction. Quick dimension is exposure time.

    // First dim_nonlin_exptime batches are the whole images for scaling
    size_t ibatch = 0; // Counter
    // Of course, here ibatch is iexptime, but it is better to keep it clear which index is meant
    for (size_t iexptime=0 ; iexptime<ckd.dim_nonlin_exptime ; iexptime++) {
        ibatch_medianphase[iexptime] = ibatch;
        phase_batches[ibatch] = phase_median;
        iexptime_batches[ibatch] = iexptime;
        ifrac_batches[ibatch] = 1; // Not used
        ibatch++;
    }
    // Next there are nfrac series of lin, nonlin, nonlin ..., where number of nonlins is dim_nonlin_exptime.
    for (size_t ifrac {}; ifrac < 1; ++ifrac) {
        ibatch_linphase[ifrac] = ibatch;
        phase_batches[ibatch] = phase_lin;
        iexptime_batches[ibatch] = ckd.dim_nonlin_exptime; // Not used
        ifrac_batches[ibatch] = ifrac;
        ibatch++;
        for (size_t iexptime=0 ; iexptime<ckd.dim_nonlin_exptime ; iexptime++) {
            ibatch_nonlinphase[ifrac*ckd.dim_nonlin_exptime+iexptime] = ibatch;
            phase_batches[ibatch] = phase_nonlin;
            iexptime_batches[ibatch] = iexptime;
            ifrac_batches[ibatch] = ifrac;
            ibatch++;
        }
    }

    // Set the sizes of the batches
    for (size_t ibatch=0 ; ibatch<nbatch ; ibatch++) {
        L1ABatch& bat = batches[ibatch];
        if (phase_batches[ibatch] == phase_median) {
        } else if (phase_batches[ibatch] == phase_lin) {
            bat.l1a_products = { l1a_manager->l1a_products.begin(),
                                 l1a_manager->l1a_products.end() };
            for (size_t il1a=0 ; il1a<nL1A() ; il1a++) {
            }
        } else if (phase_batches[ibatch] == phase_nonlin) {
            size_t& iexptime = iexptime_batches[ibatch];
            size_t& nl1a_cur = nl1a_exptime[iexptime];
        }
    }

    // Now, store the L1A pointers in the nonlin and the median phases
    std::vector<double> illum_max(ckd.dim_nonlin_exptime,0.0);
    std::vector<size_t> cnt_exptime(ckd.dim_nonlin_exptime,0); // Batch element counter
    for (size_t il1a=0 ; il1a<nL1A() ; il1a++) {
        L1A* l1a_cur = &l1a_manager->l1a_products[il1a];
        size_t& iexptime = iexptime_l1a[il1a];
        if (iexptime != ckd.dim_nonlin_exptime) {
            // Append to nonlin list
            size_t& cnt = cnt_exptime[iexptime];
            for (size_t ifrac {}; ifrac < 1; ++ifrac) {
                size_t& ibatch = ibatch_nonlinphase[ifrac*ckd.dim_nonlin_exptime+iexptime];
                L1ABatch& bat = batches[ibatch];
                bat.l1a_products = { l1a_manager->l1a_products.begin(),
                                     l1a_manager->l1a_products.end() };
            }
            nonlin_il1a_tracker.push_back(il1a);
            cnt++;
            // If this is the largest illumination level in the exposure time, overwrite the L1A* in
            // the median phase.
            if (l1a_cur->meta->illumination_level > illum_max[iexptime]) {
                illum_max[iexptime] = l1a_cur->meta->illumination_level;
                // Overwrite the L1A* (or set it if this is the first time
                size_t& ibatch = ibatch_medianphase[iexptime];
                L1ABatch& bat = batches[ibatch];
                bat.l1a_products =
                  { l1a_manager->l1a_products.begin() + il1a,
                    l1a_manager->l1a_products.begin() + il1a + 1 };
            }
        }
    }

    // Shape the rest of the CKD
    ckd.nonlin_signal_scale.resize(ckd.npix*ckd.dim_nonlin_exptime);
    nonlin_lin_slope.resize(ckd.npix);
    nonlin_lin_chi2.resize(ckd.npix);
    ckd.nonlin_fit.resize(ckd.npix*ckd.dim_nonlin_exptime*ckd.dim_nonlin_spline);
    nonlin_chi2.resize(ckd.npix*ckd.dim_nonlin_exptime);
    // Set pointers to diagnostic CKD (owned by processor)
    ckd.diag_nonlin_lin_slope = nonlin_lin_slope;
    ckd.diag_nonlin_lin_chi2 = nonlin_lin_chi2;
    ckd.diag_nonlin_chi2 = nonlin_chi2;

    for (auto& batch : batches) {
        batch.genWorkRanges(mpi);
    }
}

// Non-linearity correction calibration software
auto Nonlincal::processBatch(const int ibatch) -> void
{
    if (phase_batches[ibatch] == phase_median) {
        process_median(batches[ibatch], iexptime_batches[ibatch]); // No fraction
    } else if (phase_batches[ibatch] == phase_lin) {
        process_lin(batches[ibatch], 0, ckd.npix); // No exptime
    } else if (phase_batches[ibatch] == phase_nonlin) {
        process_nonlin(batches[ibatch], 0, ckd.npix, iexptime_batches[ibatch]);
    }
}

auto Nonlincal::process_median(L1ABatch& batch, size_t iexptime) -> void
{
    if (settings.nonlin.scale_median) {
        double med {};
        median(
          ckd.npix, batch.l1a_products.front().get().image.data(), med);
        for (size_t ipix=0 ; ipix<ckd.npix ; ipix++) ckd.nonlin_signal_scale[ipix*ckd.dim_nonlin_exptime+iexptime] = med;
    } else {
        for (size_t ipix=0 ; ipix<ckd.npix ; ipix++) {
            ckd.nonlin_signal_scale[ipix*ckd.dim_nonlin_exptime+iexptime] =
              batch.l1a_products.front().get().image[ipix];
        }
    }
}

auto Nonlincal::process_lin(L1ABatch& batch, size_t ipix_start, size_t ipix_end) -> void
{
    // Linear result pointers
    double* linslopeptr = &nonlin_lin_slope[ipix_start];
    double* linchi2ptr = &nonlin_lin_chi2[ipix_start];

    spdlog::info("Calculating linear fit per pixel");
    for (size_t ipix=ipix_start ; ipix<ipix_end ; ipix++) {
        // Put fillvalues everywhere
        *linslopeptr = NC_FILL_DOUBLE;
        *linchi2ptr = NC_FILL_DOUBLE;

        // Breakable do-once loop
        for (bool ctd=true ; ctd ; ctd=false) {

            // Die if you are already dead
            if (ckd.pixel_mask[ipix]) break;
            ckd.pixel_mask[ipix] = true; // Unless you survive until the end of the loop

            // Gather the measurements at this pixel index
            std::vector<double> signal(batch.l1a_products.size());
            std::vector<double> noise(batch.l1a_products.size());
            std::vector<bool> pixel_mask(batch.l1a_products.size());
            for (size_t il1a=0 ; il1a<batch.l1a_products.size() ; il1a++) {
                signal[il1a] = batch.l1a_products[il1a].get().image[ipix];
                noise[il1a] = batch.l1a_products[il1a].get().noise[ipix];
                pixel_mask[il1a] = batch.l1a_products[il1a].get().pixel_mask[ipix];
            }
            // Filter for signals between the chosen range
            size_t nl1a_lin = 0;
            size_t nl1a_lin_healthy = 0;
            std::vector<double> abscissa_lin(batch.l1a_products.size()); // Maximum allocation size
            std::vector<double> signal_lin(batch.l1a_products.size()); // Maximum allocation size
            std::vector<double> s_y_lin(batch.l1a_products.size()); // Maximum allocation size
            std::vector<bool> pixel_mask_lin(batch.l1a_products.size()); // Maximum allocation size
            for (size_t il1a=0 ; il1a<batch.l1a_products.size() ; il1a++) {
                double& sig = signal[il1a];
                if (sig > settings.nonlin.lmin && sig < settings.nonlin.lmax) {
                    abscissa_lin[nl1a_lin] = abscissa[il1a];
                    signal_lin[nl1a_lin] = sig;
                    s_y_lin[nl1a_lin] = pow(noise[il1a],2.0);
                    pixel_mask_lin[nl1a_lin] = pixel_mask[il1a];
                    if (!pixel_mask_lin[nl1a_lin]) nl1a_lin_healthy++;
                    nl1a_lin++;
                }
            }

            // Encapsulate vectors to their intended sizes
            abscissa_lin.resize(nl1a_lin);
            signal_lin.resize(nl1a_lin);
            s_y_lin.resize(nl1a_lin);
            pixel_mask_lin.resize(nl1a_lin);

            // Nothing is possible if there are no reliable measurments
            if (nl1a_lin_healthy < 1) break;

            // Perform the linear fit. Note that if it fails, no mess is created in the result before
            // returning.
            if (linear_invert(1,nl1a_lin,abscissa_lin.data(),OPT_DIAG,s_y_lin.data(),&pixel_mask_lin,signal_lin.data(),linslopeptr) != 0) break;
            *linchi2ptr = 0.0;
            for (size_t il1a=0 ; il1a<nl1a_lin ; il1a++) {
                if (pixel_mask_lin[il1a]) continue;
                *linchi2ptr += pow((abscissa_lin[il1a] * *linslopeptr - signal_lin[il1a]),2.0)/s_y_lin[il1a] / (nl1a_lin_healthy-1);
            }

            // Apply mask criterion: Linear chi squared
            if (settings.nonlin.mask_lin_chi2_max != NC_FILL_DOUBLE && *linchi2ptr > settings.nonlin.mask_lin_chi2_max) break;

            // Survived to the end of the loop
            ckd.pixel_mask[ipix] = false; // Unless you survive until the end of the loop

        }

        // Progress pointers
        linslopeptr++;
        linchi2ptr++;

    }
}

auto Nonlincal::process_nonlin(L1ABatch& batch, size_t ipix_start, size_t ipix_end, size_t iexptime) -> void
{
    // Pointers to earlier calculated results
    double* linslopeptr = &nonlin_lin_slope[ipix_start];
    double* scaleptr = &ckd.nonlin_signal_scale[ipix_start*ckd.dim_nonlin_exptime+iexptime];
    // Pointers to results
    double* resptr = &ckd.nonlin_fit[ipix_start*ckd.dim_nonlin_exptime*ckd.dim_nonlin_spline+iexptime*ckd.dim_nonlin_spline];
    double* chi2ptr = &nonlin_chi2[ipix_start*ckd.dim_nonlin_exptime+iexptime];

    spdlog::info("Calculating nonlinearity fit per pixel");
    for (size_t ipix=ipix_start ; ipix<ipix_end ; ipix++) {
        // Fill fillvalues everywhere
        for (size_t ispline=0 ; ispline<ckd.dim_nonlin_spline ; ispline++) resptr[ispline] = NC_FILL_DOUBLE;
        *chi2ptr = NC_FILL_DOUBLE;

        // Breakable do-once loop
        for (bool ctd=true ; ctd ; ctd=false) {

            // Die if you are already dead
            if (ckd.pixel_mask[ipix]) {
                break;
            }
            ckd.pixel_mask[ipix] = true; // Unless you survive until the end of the loop

            std::vector<double> signal_corrected_scaled(batch.l1a_products.size());
            std::vector<double> signal_measured(batch.l1a_products.size());
            std::vector<double> s_y_measured(batch.l1a_products.size());
            std::vector<bool> pixel_mask_measured(batch.l1a_products.size());
            size_t nl1a_healthy = 0;
            for (size_t il1a=0 ; il1a<batch.l1a_products.size() ; il1a++) {
                L1A* l1a = &l1a_manager->l1a_products[il1a];
                signal_corrected_scaled[il1a] = *linslopeptr * l1a->exposure_time * l1a->meta->illumination_level / (*scaleptr);
                signal_measured[il1a] = l1a->image[ipix];
                s_y_measured[il1a] = pow(l1a->noise[ipix],2.0);
                pixel_mask_measured[il1a] = l1a->pixel_mask[ipix];
                if (!pixel_mask_measured[il1a]) nl1a_healthy++;
            }

            // Perform the fit
            std::vector<double> jac(ckd.dim_nonlin_spline*batch.l1a_products.size(), 0.0);
            b->jaccalc(batch.l1a_products.size(),signal_corrected_scaled.data(),jac.data());
            // Add hard constraint at zero. Zero signal should mean zero mismatch.
            double zero = 0.0; // Used as abscissa value and as measurement
            std::vector<double> jac_zero(ckd.dim_nonlin_spline);
            b->jaccalc(1,&zero,jac_zero.data());
            if (constrained_linear_invert(ckd.dim_nonlin_spline,batch.l1a_products.size(),1,jac.data(),jac_zero.data(),OPT_DIAG,s_y_measured.data(),&pixel_mask_measured,signal_measured.data(),&zero,resptr) != 0) break;
            *chi2ptr = 0.0;
            std::vector<double> mod(batch.l1a_products.size());
            Matrix::matmul_rl_fold_slow(ckd.dim_nonlin_spline,batch.l1a_products.size(),jac.data(),resptr,mod.data());
            for (size_t il1a=0 ; il1a<batch.l1a_products.size() ; il1a++) {
                if (pixel_mask_measured[il1a]) continue;
                *chi2ptr += pow(mod[il1a]-signal_measured[il1a],2.0)/s_y_measured[il1a] / (nl1a_healthy-ckd.dim_nonlin_spline+1); // The plus one is the hard constraint, reducing the DFS by one (or adding a measurement with no bias).
            }

            // Apply mask criterion: Nonlinear chi squared
            if (settings.nonlin.mask_chi2_max && *chi2ptr > settings.nonlin.mask_chi2_max) break;

            // Survived to the end of the loop
            ckd.pixel_mask[ipix] = false;

        }

        // Progress pointers
        linslopeptr++;
        scaleptr += ckd.dim_nonlin_exptime;
        resptr += ckd.dim_nonlin_exptime*ckd.dim_nonlin_spline;
        chi2ptr += ckd.dim_nonlin_exptime;

    }
}

} // namespace spex
