// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include "bspline.h"
#include "processor.h"

#include <cstddef>
#include <vector>

namespace spex {

// Detailed output of one series
struct DetailedOutputNonlin
{
    size_t nl1a; // Number of measurements in the exposure time series
    std::vector<double> signal_meas; // Measured signal, for all pixels
    std::vector<double> signal_corr; // Desired corrected signal, for all pixels
};

enum phase_t
{
    phase_median = 0,
    phase_lin = 1,
    phase_nonlin = 2
};

class Nonlincal : public Processor
{
public:
    Nonlincal(
      const gadfit::MPIVars& mpi, const SPEXSettings& settings, CKD& ckd)
      : Processor { mpi, settings, ProcLevel::nonlin, ckd }
    {}
    ~Nonlincal() = default;

private:
    auto processInit() -> void override;
    auto prepareBatches() -> void override {}
    // ibatch is exposure time number or illumination level number
    // (ascending)
    auto processBatch(const int ibatch) -> void override;

    auto process_median(L1ABatch& batch, size_t iexptime) -> void;
    auto process_lin(L1ABatch& batch, size_t ipix_start, size_t ipix_end) -> void;
    auto process_nonlin(L1ABatch& batch, size_t ipix_start, size_t ipix_end, size_t iexptime)
      -> void;

private:
    // NetCDF contents that is written, but not read
    std::vector<double> nonlin_lin_slope {}; // Slope fit during linear fit
    std::vector<double> nonlin_lin_chi2 {}; // Residual chi squared on the linear fit
    std::vector<double> nonlin_chi2 {}; // Residual chi squared on non-linearity correction

    // Batch administration
    std::vector<phase_t> phase_batches {}; // Phase (0=get median, 1=find linear relationship, 2=fit non-linearity)
    std::vector<size_t> iexptime_batches {}; // Exposure time index that belongs to the batch
    std::vector<size_t> ifrac_batches {}; // Fraction indices of batches

    // Abscissa (exposure time times illumination level, for all L1A instances)
    std::vector<double> abscissa {};

    // Fitting spline
    std::unique_ptr<Bspline> b {};

    std::vector<int> nonlin_il1a_tracker {};
};

} // namespace spex
