// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include "nc_object.h"
#include "vector.h"

namespace spex {

class Planet
{
private:
    double a {}; // Semi-major axis
    double b {}; // Semi-minor axis
    double tol {}; // Tolerance for searching latitude [radians]
    double ecc2 {}; // Eccentricity squared
    // For DEM
    NcObject nc_dem { netCDF::NcFile::read };
    size_t dim_dem_lat {}; // Number of latitudes of detailed elevation map
    size_t dim_dem_lon {}; // Number of longitudes of detailed elevation map
    std::vector<double> dem_lat {}; // Latitude axis of detailed elevation map
    std::vector<double> dem_lon {}; // Longitude axis of detailed elevation map
    // An issue is that the elevation map shows the surface height or
    // the sea bottom (negative) height. For sea pixels, we need the
    // water surface height. The water surface height is fillvalue for
    // the land pixels, so a mask is needed.
    netCDF::NcVar var_watermask {}; // Variable identifier for land/sea mask
    netCDF::NcVar var_landheight {}; // Variable identifier for land surface height
    netCDF::NcVar var_waterheight {}; // Variable identifier for water surface height

public:
    Planet(double semi_major_axis, // Semi-major axis [space unit]
           double semi_minor_axis, // Semi-minor aixs [space unit]
           double latitude_tolerance); // Tolerance for searching latitude [radians]
    ~Planet() = default;

    auto xyz(double lat, // Latitude
             double lon, // Longitude
             double alt, // Altitude
             Vector* pos) -> void; // Position (output)
    auto lla(Vector pos, // Position
             double* lat, // Latitude (output)
             double* lon, // Longitude (output)
             double* alt) -> void; // Altitude (output)
    // Filename for detailed elevation map
    auto include_dem(const std::string& filename) -> void;

    // Acquisition of DEM, from file or using placeholder
    auto dem(double lat, // Latitude in radians
             double lon, // Longitude in radians
             double& res) -> void; // Resulting elevation

private:
    auto northbound(double lat, // Attempted latitude
                    double xy, // Distance from pole axis
                    double z) -> bool; // Distance from equatorial plane
};

} // namespace spex
