// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "prnu_gen.h"

#include "bspline.h"
#include "ckd.h"
#include "l1a.h"
#include "l1a_metadata.h"
#include "matrix.h"
#include "spex_math.h"

#include <fstream>
#include <gadfit/lm_solver.h>
#include <spdlog/spdlog.h>
#include <string>

namespace spex {

auto PrnuGen::processBatch(const int /*i_batch*/) -> void
{
    const auto l1a_products { batches.front().l1a_products };
    // Light exposure (illumination strength multiplied by exposure
    // time)
    std::vector<double> photons(l1a_products.size(), 0.0);
    for (int i_l1a {}; i_l1a < static_cast<int>(l1a_products.size()); ++i_l1a) {
        photons[i_l1a] = l1a_products[i_l1a].get().meta->illumination_level
                         * l1a_products[i_l1a].get().exposure_time;
    }
    // Sensitivity is signal divided by the number of photons averaged
    // over the calibration measurements.
    std::vector<double> sensitivity(ckd.npix);
    std::vector<double> sensitivity_noise(ckd.npix);
    const double n_l1a_sqrt { std::sqrt(
      static_cast<double>(l1a_products.size())) };
    for (int i_l1a {}; i_l1a < static_cast<int>(l1a_products.size()); ++i_l1a) {
        for (int i {}; i < ckd.npix; ++i) {
            sensitivity[i] += l1a_products[i_l1a].get().image[i]
                              / photons[i_l1a] / l1a_products.size();
            if (ckd.pixel_mask[i]) {
                sensitivity_noise[i] = NC_FILL_DOUBLE;
            } else {
                sensitivity_noise[i] += l1a_products[i_l1a].get().noise[i]
                                        / (photons[i_l1a] * n_l1a_sqrt);
            }
        }
    }
    // Fit sensitivity with a low-order polynomial that is of order
    // [prnu][order_spat] in the spatial dimension and
    // [prnu][order_spec] in the spectral dimension. Also included are
    // all cross-terms whose order is not higher than max(order_spat,
    // order_spec).
    const int order_max { std::max(settings.prnu.order_spat,
                                   settings.prnu.order_spec) };
    gadfit::LMsolver solver {
        [this, order_max](const std::vector<gadfit::AdVar>& parameters,
                          const double idx_full) {
            // GADfit interface requires the second argument to be a
            // double precision number. Resolve it into an integer row
            // and column pixel index.
            // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
            const int idx_i_int { *(reinterpret_cast<const int*>(&idx_full)) };
            const int idx_j_int { *(
              // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
              reinterpret_cast<const int*>(&idx_full)
              // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-pointer-arithmetic)
              + 1) };
            // Before computing anything, construct all orders of the
            // column index: j^0, j^1, ..., j^order_spec.
            std::vector<double> idx_j(settings.prnu.order_spec + 1);
            idx_j[0] = 1.0;
            for (int j { 1 }; j <= settings.prnu.order_spec; ++j) {
                idx_j[j] = idx_j[j - 1] * idx_j_int;
            }
            // Orders of the row index are computed on the fly. Only
            // initialize the first one here: i^0 = 1.
            double idx_i { 1.0 };
            // Counter for iterating through the fitting parameter
            // array
            int i_par { -1 };
            gadfit::AdVar result { 0.0, gadfit::passive_idx };
            // Outer loop is over all orders of the row index
            for (int i {}; i <= settings.prnu.order_spat; ++i) {
                // Cross-term orders are less or equal to
                // order_max. Hence the j order has to be order_max-i
                // or less. Also, the j order cannot be higher than
                // order_spec.
                for (int j {}; j <= std::min(order_max - i,
                                             settings.prnu.order_spec.value);
                     ++j) {
                    result += parameters[++i_par] * idx_i * idx_j[j];
                }
                if (i != settings.prnu.order_spat) {
                    idx_i *= idx_i_int;
                }
            }
            return result;
        },
        mpi.comm
    };
    // Array with pixel row and column indices. As mentioned above,
    // these needs to be of type double whereas the two pixel indices
    // are of type int. Use reinterpret_cast to store the integer
    // indices.
    std::vector<double> pixel_indices(ckd.npix);
    static_assert(
      sizeof(double) / sizeof(int) >= 2
      && "size of double needs to be at least twice the size of integer");
    std::array<int, 2> index_pair {};
    for (int i {}; i < ckd.dim_detector_spat; ++i) {
        index_pair[0] = i;
        for (int j {}; j < ckd.dim_detector_spec; ++j) {
            index_pair[1] = j;
            pixel_indices[i * ckd.dim_detector_spec + j] =
              // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
              *(reinterpret_cast<double*>(index_pair.data()));
        }
    }
    solver.addDataset(pixel_indices, sensitivity, sensitivity_noise);
    spdlog::info("Sensitivity fit using the polynomial terms:");
    int i_par { -1 };
    for (int i {}; i <= settings.prnu.order_spat; ++i) {
        for (int j {};
             j <= std::min(order_max - i, settings.prnu.order_spec.value);
             ++j) {
            solver.setPar(++i_par, 0.0, true);
            spdlog::info("  i^{} x j^{}", i, j);
        }
    }
    solver.settings.iteration_limit = settings.prnu.fitting.iteration_limit;
    solver.settings.verbosity = gadfit::io::timings;
    solver.settings.min_n_blocks = settings.prnu.fitting.min_n_blocks;
    solver.fit();
    // Store the result in CKD
    ckd.prnu.prnu = { true, mpi, ckd.npix };
    std::vector<int> mpi_ranges {};
    genWorkRanges(ckd.dim_detector_spat, mpi.n_procs, mpi_ranges);
    for (int i { mpi_ranges.at(mpi.rank) }; i < mpi_ranges.at(mpi.rank + 1);
         ++i) {
        index_pair[0] = i;
        for (int j {}; j < ckd.dim_detector_spec; ++j) {
            index_pair[1] = j;
            // The PRNU is defined as the sensitivity divided by the fitted
            // (low-order) sensitivity.
            const int idx_full { i * ckd.dim_detector_spec + j };
            ckd.prnu.prnu[idx_full] =
              sensitivity[idx_full]
              / solver.getValue(
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
                *(reinterpret_cast<double*>(index_pair.data())));
        }
    }
#ifdef USE_MPI
    MPI_Barrier(mpi.comm);
#endif
    // Update pixel mask
    if (settings.prnu.mask_prnu_min) {
        updateMaskLt(ckd.prnu.prnu,
                     settings.prnu.mask_prnu_min.value(),
                     "pixels masked bad because their PRNU is less than",
                     ckd.pixel_mask);
    }
    if (settings.prnu.mask_prnu_max) {
        updateMaskGt(ckd.prnu.prnu,
                     settings.prnu.mask_prnu_max.value(),
                     "pixels masked bad because their PRNU is greater than",
                     ckd.pixel_mask);
    }
    if (settings.debug.at(static_cast<int>(ProcLevel::prnu)).value()) {
        ckd.prnu.sensitivity_fit = SharedArray<double> { true, mpi, ckd.npix };
        for (int i { mpi_ranges.at(mpi.rank) }; i < mpi_ranges.at(mpi.rank + 1);
             ++i) {
            index_pair[0] = i;
            for (int j {}; j < ckd.dim_detector_spec; ++j) {
                index_pair[1] = j;
                const int idx_full { i * ckd.dim_detector_spec + j };
                ckd.prnu.sensitivity_fit[idx_full] = solver.getValue(
                  *(reinterpret_cast<double*>(index_pair.data())));
            }
        }
    }
}

} // namespace spex
