// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "processor.h"

#include "io.h"
#include "l1a_flight.h"
#include "l1a_metadata.h"

#include <spdlog/spdlog.h>

namespace spex {

Processor::Processor(const gadfit::MPIVars& mpi,
                     const SPEXSettings& settings,
                     const ProcLevel proc_level,
                     CKD& ckd)
  : mpi { mpi }, settings { settings }, proc_level { proc_level }, ckd { ckd }
{
    // Do not change the CKD level if this is L1B or higher. If the
    // CKD is incomplete, e.g. the polarimetric section is missing,
    // then its level should stay at ProcLevel::pol. Increasing it
    // here will mislead L1B and higher processes thinking the CKD is
    // more mature than it actually is.
    if (proc_level < ProcLevel::l1b) {
        ckd.level = static_cast<ProcLevel>(static_cast<int>(proc_level) - 1);
    }
}

auto Processor::execute() -> void
{
    // STEP 1 - Welcome message

    switch (proc_level) {
    case ProcLevel::dark:
        printHeading("Dark CKD generation");
        break;
    case ProcLevel::noise:
        printHeading("Noise CKD generation");
        break;
    case ProcLevel::nonlin:
        printHeading("Nonlinearity CKD generation");
        break;
    case ProcLevel::prnu:
        printHeading("PRNU CKD generation");
        break;
    case ProcLevel::stray:
        printHeading("Stray light CKD generation");
        break;
    case ProcLevel::fov:
        printHeading("Field of view CKD generation");
        break;
    case ProcLevel::swath:
        printHeading("Swath vector CKD generation");
        break;
    case ProcLevel::wave:
        printHeading("Spectral CKD generation");
        break;
    case ProcLevel::rad:
        printHeading("Radiometric CKD generation");
        break;
    case ProcLevel::pol:
        printHeading("Polarimetric CKD generation");
        break;
    case ProcLevel::l1b:
        printHeading("L1B process");
        break;
    case ProcLevel::l1c:
        printHeading("L1C process");
        break;
    default:
        printHeading("Unknown process");
    }
    setLoggerLevel(mpi,
                   settings.debug.at(static_cast<int>(proc_level)).value());

    // STEP 2 - Return or continue

    if (settings.skip.at(static_cast<int>(proc_level))) {
        spdlog::info("Skipping process");
        return;
    }

    // STEP 3 - Timers

#ifdef USE_MPI
    MPI_Barrier(MPI_COMM_WORLD); // For timings
#endif
    // Start main timer for this process
    gadfit::Timer timer {};
    timer.start();
    // Set up the rest of the timers
    std::array<std::shared_ptr<gadfit::Timer>,
               static_cast<int>(ProcLevel::n_levels)>
      timers {};
    for (auto& timer : timers) {
        timer = std::make_shared<gadfit::Timer>();
        timer->reset();
    }

    // STEP 4 - Organize L1A files

    // Construct an L1A manager which is a container for all L1A
    // products. After this, the L1A products are available as
    // l1a_manager.l1a_products but only their metadata is read. The
    // image contents will be read later during calibration.
    switch (settings.l1a_type.at(static_cast<int>(proc_level))) {
    case L1AType::flight:
        l1a_manager = std::make_unique<L1AFlight>(mpi, false);
        break;
    case L1AType::calibration:
        l1a_manager = std::make_unique<L1ACalibration>(mpi, false);
        break;
    default:
        throw std::runtime_error { "unknown L1A type" };
    }
    l1a_manager->init(settings, proc_level, ckd);
    for (auto& l1a_product : l1a_manager->l1a_products) {
        l1a_product.timers = timers;
    }

    // STEP 5 - Binning table

    // Initialize the binning table, associate it with each L1A
    // object, and bin the CKD if running the L1B process.
    std::shared_ptr<BinningTable> binning_table {};
    std::shared_ptr<BinningTable> binning_table_stray {};
    if (!l1a_manager->l1a_products.empty()) {
        binning_table = std::make_shared<BinningTable>(
          mpi,
          settings.main.detector_spat,
          settings.main.detector_spec,
          settings.main.binningtable,
          l1a_manager->l1a_products.front().binning_table_id,
          settings.main.spectral_binning);
        // For stray light correction, the binning table dimensions
        // are either the same as the detector dimension or reduced by
        // a factor of two.
        if (proc_level == ProcLevel::l1b) {
            const int reduction { ckd.stray.reduced_kernels ? 2 : 1 };
            binning_table_stray = std::make_shared<BinningTable>(
              mpi,
              settings.main.detector_spat / reduction,
              settings.main.detector_spec / reduction,
              settings.main.binningtable,
              l1a_manager->l1a_products.front().binning_table_id,
              settings.main.spectral_binning);
        }
    }
    for (auto& l1a_product : l1a_manager->l1a_products) {
        l1a_product.binning_table = binning_table;
        l1a_product.binning_table_stray = binning_table_stray;
    }
    if (proc_level == ProcLevel::l1b) {
        binning_table->binCKD(ckd);
    }

    // STEP 6 - Initialize L1X instances (if any)

    l1a_manager->initL1X(settings, proc_level, ckd);

    // STEP 7 - Perform initialization, if any, of the current process

    processInit();
    prepareBatches();

    // STEP 8 - Background images

    // Separate L1A manager for background images
    L1ACalibration l1a_manager_bgr { mpi, true };
    l1a_manager_bgr.init(settings, proc_level, ckd);
    // Reduce the CKD level temporarily to dark, i.e. nonlinearity
    // or higher should not be applied to background images.
    const ProcLevel ckd_level_save { ckd.level };
    ckd.level = std::min(ckd_level_save, ProcLevel::dark);
    for (auto& l1a_product : l1a_manager_bgr.l1a_products) {
        l1a_product.timers = timers;
        l1a_product.binning_table = binning_table;
    }
    prepareBackground(l1a_manager_bgr);
    ckd.level = ckd_level_save;

    // STEP 9 - Print a summary of what will be done in this process

    if (nL1A() > 0) {
        spdlog::info("Total number of L1A files: {}", nL1A());
        spdlog::info("Processing steps:");
        // Flag indicating if any calibration steps are performed for
        // this process
        bool no_steps { true };
        const auto& l1x_level {
            l1a_manager->l1a_products.front().meta->l1x_level
        };
        for (int i_proc {}; i_proc < static_cast<int>(ProcLevel::n_levels);
             ++i_proc) {
            const auto proc { static_cast<ProcLevel>(i_proc) };
            if (l1a_manager->l1a_products.front().meta->l1x_level < proc
                && ckd.level >= proc && !settings.skip.at(i_proc)) {
                switch (proc) {
                case ProcLevel::dark:
                    spdlog::info("  Dark correction");
                    break;
                case ProcLevel::noise:
                    spdlog::info("  Noise estimation");
                    break;
                case ProcLevel::nonlin:
                    spdlog::info("  Nonlinearity correction");
                    break;
                case ProcLevel::prnu:
                    spdlog::info("  PRNU correction");
                    break;
                case ProcLevel::stray:
                    spdlog::info("  Stray light correction");
                    break;
                case ProcLevel::fov:
                    spdlog::info("  Spectra extraction (FOV)");
                    break;
                case ProcLevel::rad:
                    spdlog::info("  Radiometric correction");
                    break;
                case ProcLevel::pol:
                    spdlog::info("  Demodulation");
                    break;
                default:
                    continue;
                }
                no_steps = false;
            }
        }
        if (!l1a_manager_bgr.l1a_products.empty()) {
            spdlog::info("  Subtract background image");
            no_steps = false;
        }
        if (settings.dark_cols_left.at(static_cast<int>(proc_level)) > 0
            || settings.dark_cols_right.at(static_cast<int>(proc_level)) > 0) {
            spdlog::info("  Row-dependent offset");
            no_steps = false;
        }
        if (no_steps) {
            spdlog::info("  None");
        }
    }

    // STEP 10 - Perform the main loop over L1A batches

    // This is the core of the process. In a parallel run the range of
    // batches is given by mpi_ranges. Otherwise the range is [0,
    // nbatch).
    for (int i_batch { mpi_ranges.empty() ? 0 : mpi_ranges.at(mpi.rank) };
         i_batch < (mpi_ranges.empty() ? static_cast<int>(batches.size())
                                       : mpi_ranges.at(mpi.rank + 1));
         ++i_batch) {
        // Read L1A images (large data load) and calibrate them
        batches.at(i_batch).calibrate(
          settings, proc_level, proc_level != ProcLevel::l1b);
#ifdef USE_MPI
        if (mpi_ranges.empty()) {
            // If mpi_ranges is not empty then some processes will
            // never reach this point because i_batch runs over
            // non-overlapping ranges.
            MPI_Barrier(mpi.comm);
        }
#endif
        // Execute the processor per batch. Batch number may indicate
        // the phase of the process.
        processBatch(i_batch);
        // Discard used images
        batches.at(i_batch).destroy();
    }

    // STEP 11 - Do postprocessing, if any, for this process

    processFinalize();

    // STEP 12 - Print timings for processing step applied in this process

    timer.stop();
    spdlog::info("Timings:");
    spdlog::info(
      "         Dark correction: {:8.3f} s ({} calls)",
      timers.at(static_cast<int>(ProcLevel::dark))->totalWallTime(),
      timers.at(static_cast<int>(ProcLevel::dark))->totalNumberOfCalls());
    spdlog::info(
      "        Noise estimation: {:8.3f} s ({} calls)",
      timers.at(static_cast<int>(ProcLevel::noise))->totalWallTime(),
      timers.at(static_cast<int>(ProcLevel::noise))->totalNumberOfCalls());
    spdlog::info(
      " Nonlinearity correction: {:8.3f} s ({} calls)",
      timers.at(static_cast<int>(ProcLevel::nonlin))->totalWallTime(),
      timers.at(static_cast<int>(ProcLevel::nonlin))->totalNumberOfCalls());
    spdlog::info(
      "         PRNU correction: {:8.3f} s ({} calls)",
      timers.at(static_cast<int>(ProcLevel::prnu))->totalWallTime(),
      timers.at(static_cast<int>(ProcLevel::prnu))->totalNumberOfCalls());
    spdlog::info(
      "  Stray light correction: {:8.3f} s ({} calls)",
      timers.at(static_cast<int>(ProcLevel::stray))->totalWallTime(),
      timers.at(static_cast<int>(ProcLevel::stray))->totalNumberOfCalls());
    spdlog::info(
      "Spectra extraction (FOV): {:8.3f} s ({} calls)",
      timers.at(static_cast<int>(ProcLevel::fov))->totalWallTime(),
      timers.at(static_cast<int>(ProcLevel::fov))->totalNumberOfCalls());
    spdlog::info(
      " Radiometric calibration: {:8.3f} s ({} calls)",
      timers.at(static_cast<int>(ProcLevel::rad))->totalWallTime(),
      timers.at(static_cast<int>(ProcLevel::rad))->totalNumberOfCalls());
    spdlog::info(
      "            Demodulation: {:8.3f} s ({} calls)",
      timers.at(static_cast<int>(ProcLevel::pol))->totalWallTime(),
      timers.at(static_cast<int>(ProcLevel::pol))->totalNumberOfCalls());
    // Print detailed timings for each MPI process
    if (mpi.n_procs > 1) {
        std::vector<gadfit::Times> times(mpi.n_procs);
        gatherTimes(mpi, timer, times);
        for (int i_rank {}; i_rank < mpi.n_procs; ++i_rank) {
            spdlog::info("Wall time for MPI process {:3}: {:8.3f} s",
                         i_rank,
                         times.at(i_rank).wall);
        }
        for (int i_rank {}; i_rank < mpi.n_procs; ++i_rank) {
            spdlog::info(" CPU time for MPI process {:3}: {:8.3f} s",
                         i_rank,
                         times.at(i_rank).cpu);
        }
    } else {
        spdlog::info("   Wall time for process: {:8.3f} s",
                     timer.totalWallTime());
        spdlog::info("    CPU time for process: {:8.3f} s",
                     timer.totalCPUTime());
    }
    // To be on the safe side because parts of the processor include
    // asynchronous MPI.
#ifdef USE_MPI
    MPI_Barrier(MPI_COMM_WORLD);
#endif
}

auto Processor::prepareBatches() -> void
{
    batches.resize(1);
    batches.front().l1a_products = { l1a_manager->l1a_products.begin(),
                                     l1a_manager->l1a_products.end() };
    batches.front().genWorkRanges(mpi);
}

auto Processor::prepareBackground(L1ACalibration& l1a_manager_bgr) -> void
{
    // Put all background images into one batch and calibrate them in
    // parallel. This is mostly to save memory.
    L1ABatch batch_bgr {};
    batch_bgr.l1a_products = { l1a_manager_bgr.l1a_products.begin(),
                               l1a_manager_bgr.l1a_products.end() };
    batch_bgr.genWorkRanges(mpi);
    batch_bgr.calibrate(settings, proc_level);
    for (auto& l1a : l1a_manager->l1a_products) {
        for (auto& l1a_bgr : l1a_manager_bgr.l1a_products) {
            if (l1a_bgr.meta->viewport == l1a.meta->viewport
                || l1a_bgr.meta->viewport == 0) {
                l1a.bgr_image = &l1a_bgr.image;
                break;
            }
        }
    }
}

auto Processor::batchViewport() -> void
{
    // Create a temporary wrapper array that we use to sort L1A
    // products by viewport
    std::vector<std::reference_wrapper<L1A>> l1a_refs {
        l1a_manager->l1a_products.begin(), l1a_manager->l1a_products.end()
    };
    std::ranges::sort(l1a_refs, [](const auto& a, const auto& b) {
        return a.get().meta->viewport < b.get().meta->viewport;
    });
    batches.resize(ckd.dim_vp);
    for (auto batch { batches.begin() };
         const auto& [vp_idx, angle] : viewport_idx_to_angle) {
        // Find the first L1A product that matches the given viewport
        auto l1a_beg { std::ranges::find_if(
          l1a_refs, [vp_idx = vp_idx](const auto& x) {
              return x.get().meta->viewport == vp_idx;
          }) };
        // Find last L1A product with the same viewport as l1a_beg
        auto l1a_end { std::ranges::find_if(
          l1a_beg, l1a_refs.cend(), [vp_idx = vp_idx](const auto& x) {
              return x.get().meta->viewport != vp_idx;
          }) };
        // Assign this range of L1A products to the current batch
        batch->l1a_products = { l1a_beg, l1a_end };
        batch->genWorkRanges(mpi);
        ++batch;
    }
}

auto Processor::batchOne() -> void
{
    batches.resize(nL1A());
    for (int i {}; i < nL1A(); ++i) {
        batches.at(i).l1a_products = { l1a_manager->l1a_products.begin() + i,
                                       l1a_manager->l1a_products.begin() + i
                                         + 1 };
        batches.at(i).genWorkRanges();
    }
}

Processor::~Processor()
{
    // Same comment as in the constructor. The CKD level should only
    // be increased here if we are generating the CKD.
    if (proc_level < ProcLevel::l1b) {
        ckd.level = static_cast<ProcLevel>(static_cast<int>(ckd.level) + 1);
    }
}

} // namespace spex
