// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Base class for all the steps of generating the CKD, L1B process,
// and L1C process. Each class inheriting from Processor performs one
// task, e.g. generating the nonlinearity CKD or running the L1B
// process (applying the CKD to one of more L1A files).

#pragma once

#include "constants.h"
#include "l1a_batch.h"
#include "l1a_calibration.h"
#include "nc_object.h"
#include "spex_settings.h"

namespace gadfit {

class MPIVars;

} // namespace gadfit

namespace spex {

class CKD;

class Processor
{
protected:
    const gadfit::MPIVars& mpi;
    // Each instance of Processor has a binding to the set of all user
    // editable configuration variables.
    const SPEXSettings& settings {};
    // Level or type of the current process
    const ProcLevel proc_level {};
    // Like L1ABatch::mpi_ranges but stores the range of batches that
    // each MPI process should process. The starting batch index for a
    // given process is mpi_ranges[mpi.rank] and the end index is
    // mpi_ranges[mpi.rank+1]-1.
    std::vector<int> mpi_ranges {};
    // Object to manage all L1A products and their metadata
    std::unique_ptr<L1AManager> l1a_manager {};
    // A non-constant reference to the CKD, meaning each processor can
    // edit its contents, even though for L1B or L1C there is no need
    // to do so.
    CKD& ckd;
    // List of L1A instances of the current batch, to be used during
    // process_batch. Here, the image data is read and calibrated up
    // to the desired level.
    std::vector<L1ABatch> batches {};
    // Total number of L1A products
    auto nL1A() const -> int
    {
        return static_cast<int>(l1a_manager->l1a_products.size());
    }
    // Sort all L1A products by viewport and assign to L1A
    // batches. Each process may define its own way of organizing L1A
    // products in prepareBatches but since sorting by viewport is
    // common we include the default viewport batching implementation
    // here. It may be called from prepareBatches.
    auto batchViewport() -> void;
    // Another commonly use batching scheme. Each L1A product is in
    // its own batch. The number of batches is the number of L1A
    // products.
    auto batchOne() -> void;
    // Set the CKD level (maturity) to the current process
    Processor(const gadfit::MPIVars& mpi,
              const SPEXSettings& settings,
              const ProcLevel proc_level,
              CKD& ckd);

public:
    // Main protocol of executing the processor. Executes one process
    // which can be any of the CKD generation steps, L1B process, or
    // L1C process.
    auto execute() -> void;
    // Increase the CKD level (maturity). If this is followed by
    // another instance of Processor, the CKD level will be set again
    // in the constructor so there is a little bit of redundancy here.
    virtual ~Processor();

private:
    // Each process may define its own way for organizing the L1A
    // products into batches. The default behavior is to put all of
    // them into one batch.
    virtual auto prepareBatches() -> void;
    // Read in and calibrate background images and associate them with
    // L1A products. The default implementation is fine for all
    // processes except stray light.
    virtual auto prepareBackground(L1ACalibration& l1a_manager_bgr) -> void;
    // Steps in the execute function that may be optionally redefined
    // by sub-classes. By default these functions do nothing.
    virtual auto processInit() -> void {}
    // Generate a CKD or run an L1B/C process for all L1A images in a
    // given batch indexed by ibatch. If a process (e.g. PRNU
    // generation) processes all images in a single step the argument
    // is ignored.
    virtual auto processBatch(const int) -> void {}
    // Any post-processing of data once the batch has been processed
    virtual auto processFinalize() -> void {}
};

} // namespace spex
