// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class that implements a shared array using the MPI RMA model

#pragma once

#include <gadfit/mpi.h>
#include <span>

namespace spex {

template <typename T>
class SharedArray
{
private:
    MPI_Win win {};
    // If initialized with shared=false, each process gets a copy of
    // data stored in data_no_shared. We have to use a raw pointer
    // here because std::vector<bool> does not work properly with
    // std::span.
    T* data_no_shared {};
    // Regardless of whether the data is shared or not, it is
    // accessible from data_shared.
    std::span<T> data_shared {};

public:
    SharedArray() = default;
    SharedArray(const SharedArray&) = default;
    // If shared=true, allocates shared memory on the root
    // process. The allocation size is n_values values of type T on
    // root and 0 on other processes.
    SharedArray(
      const bool shared, const gadfit::MPIVars& mpi, const int n_values)
    {
        if (shared && mpi.comm != MPI_COMM_NULL) {
#ifdef USE_MPI
            T* data_ptr {};
            MPI_Win_allocate_shared(
              static_cast<MPI_Aint>(sizeof(T) * (mpi.rank == 0 ? n_values : 0)),
              sizeof(T),
              MPI_INFO_NULL,
              mpi.comm,
              &data_ptr,
              &win);
            MPI_Aint size {};
            int disp_unit {};
            MPI_Win_shared_query(win, 0, &size, &disp_unit, &data_ptr);
            data_shared = std::span { data_ptr, static_cast<size_t>(n_values) };
#endif
        } else {
            storeNoShared(n_values);
        }
    }

    // Shorthand for
    //   SharedArray(false, gadfit::MPIVars { MPI_COMM_NULL }, n_values)
    SharedArray(const int n_values)
    {
        storeNoShared(n_values);
    }

    // Create a local buffer and do not use shared memory
    auto storeNoShared(const int n_values) -> void
    {
        data_no_shared = new T[n_values] {};
        data_shared =
          std::span { data_no_shared, static_cast<size_t>(n_values) };
    }

    auto operator=(SharedArray&& shared_array) noexcept -> SharedArray&
    {
        std::swap(win, shared_array.win);
        std::swap(data_no_shared, shared_array.data_no_shared);
        data_shared = shared_array.data_shared;
        return *this;
    }

    [[nodiscard]] auto operator[](const int i) -> T&
    {
        return data_shared[i];
    }

    [[nodiscard]] auto operator[](const int i) const -> T&
    {
        return data_shared[i];
    }

    [[nodiscard]] auto data() const -> T*
    {
        return data_shared.data();
    }

    [[nodiscard]] auto size() const -> size_t
    {
        return data_shared.size();
    }

    [[nodiscard]] auto empty() const -> bool
    {
        return data_shared.empty();
    }

    [[nodiscard]] auto cbegin() const -> auto
    {
        return data_shared.begin();
    }

    [[nodiscard]] auto cend() const -> auto
    {
        return data_shared.end();
    }

    [[nodiscard]] auto begin() -> auto
    {
        return data_shared.begin();
    }

    [[nodiscard]] auto end() -> auto
    {
        return data_shared.end();
    }

    ~SharedArray()
    {
#ifdef USE_MPI
        if (static_cast<bool>(win)) {
            MPI_Win_free(&win);
        }
#endif
        delete[] data_no_shared;
    }
};

// Resize and initialize an array of shared arrays. Implementing a
// SharedArray copy constructor would allow us to call
// std::vector::resize directly. However, construction of a new shared
// array from an existing one (copy constructor) is not possible in
// the current implementation because it requires knowledge of the MPI
// environment (an external resource).
template <typename T>
auto resize(std::vector<SharedArray<T>>& shared_arrays,
            const gadfit::MPIVars& mpi,
            const int vector_size,
            const int shared_array_size) -> void
{
    shared_arrays.clear();
    shared_arrays.resize(vector_size);
    for (auto& shared_array : shared_arrays) {
        shared_array = { true, mpi, shared_array_size };
    }
}

} // namespace spex
