// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "spectra.h"

#include "cubic_spline.h"

#include <numeric>

namespace spex {

auto Spectra::extractFOV(const CKD& ckd,
                         const SharedArray<double>& image,
                         const SharedArray<double>& image_noise,
                         const SharedArray<bool>& pixel_mask,
                         const int i_fov) -> void
{
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        signal[i_pol].resize(ckd.dim_detector_spec);
        noise[i_pol].resize(ckd.dim_detector_spec);
        mask[i_pol].resize(ckd.dim_detector_spec);
    }
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int i {}; i < ckd.dim_detector_spec; ++i) {
            const int& ipix_dn { ckd.fov.pix1[i_pol][i_fov][i] };
            const int& ipix_up { ckd.fov.pix2[i_pol][i_fov][i] };
            // Weight of the "down" pixel
            const double& weight { ckd.fov.weight[i_pol][i_fov][i] };
            const bool bad_dn { ipix_dn == math::idx_fill_value
                                || pixel_mask[ipix_dn] };
            const bool bad_up { ipix_up == math::idx_fill_value
                                || pixel_mask[ipix_up] };
            if (bad_dn && bad_up) {
                signal[i_pol][i] = NC_FILL_DOUBLE;
                noise[i_pol][i] = NC_FILL_DOUBLE;
                mask[i_pol][i] = true;
            } else if (bad_dn) {
                signal[i_pol][i] = image[ipix_up];
                noise[i_pol][i] = image_noise[ipix_up];
            } else if (bad_up) {
                signal[i_pol][i] = image[ipix_dn];
                noise[i_pol][i] = image_noise[ipix_dn];
            } else {
                signal[i_pol][i] =
                  weight * image[ipix_dn] + (1.0 - weight) * image[ipix_up];
                const double&& dn { weight * image_noise[ipix_dn] };
                const double&& up { (1.0 - weight) * image_noise[ipix_up] };
                noise[i_pol][i] = std::sqrt(dn * dn + up * up);
            }
        }
    }
}

auto Spectra::calibrate(const CKD& ckd,
                        const double exposure_time,
                        const int i_fov) -> void
{
    const double exposure_time_inv { 1.0 / exposure_time };
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int i {}; i < static_cast<int>(signal.front().size()); ++i) {
            if (mask[i_pol][i]) {
                continue;
            }
            if (ckd.rad_spectra[i_pol][i_fov][i] == NC_FILL_DOUBLE) {
                mask[i_pol][i] = true;
            } else {
                signal[i_pol][i] *=
                  ckd.rad_spectra[i_pol][i_fov][i] * exposure_time_inv;
                noise[i_pol][i] *=
                  ckd.rad_spectra[i_pol][i_fov][i] * exposure_time_inv;
            }
        }
    }
}

auto Spectra::interpolateToCommon(const CKD& ckd, const int i_fov) -> void
{
    std::array<std::vector<double>, dims::pol> wavelengths {};
    std::array<std::vector<double>, dims::pol> signal_in {};
    std::array<std::vector<double>, dims::pol> noise_in {};
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        wavelengths[i_pol].reserve(signal.front().size());
        signal_in[i_pol].reserve(signal.front().size());
        noise_in[i_pol].reserve(signal.front().size());
        for (int i {}; i < static_cast<int>(signal.front().size()); ++i) {
            if (!mask[i_pol][i]) {
                wavelengths[i_pol].push_back(ckd.wave_full[i_pol][i_fov][i]);
                signal_in[i_pol].push_back(signal[i_pol][i]);
                noise_in[i_pol].push_back(noise[i_pol][i]);
            }
        }
    }
    std::array<CubicSpline, dims::pol> signal_spline {
        CubicSpline { wavelengths[pol_minus], signal_in[pol_minus] },
        CubicSpline { wavelengths[pol_plus], signal_in[pol_plus] }
    };
    std::array<CubicSpline, dims::pol> noise_spline {
        CubicSpline { wavelengths[pol_minus], noise_in[pol_minus] },
        CubicSpline { wavelengths[pol_plus], noise_in[pol_plus] }
    };
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        for (int i {}; i < static_cast<int>(signal.front().size()); ++i) {
            signal[i_pol][i] =
              signal_spline[i_pol].eval(ckd.wave_common[i_fov][i]);
            noise[i_pol][i] =
              noise_spline[i_pol].eval(ckd.wave_common[i_fov][i]);
        }
    }
}

} // namespace spex
