// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "spex_math.h"

#include "constants.h"

namespace spex {

auto binaryFindIdx(const std::vector<double>& list, const double x) -> int
{
    int i_begin {};
    int i_end { static_cast<int>(list.size() - 1) };
    int i_mid {};
    if (x < list.front() || x > list.back()) {
        return math::idx_fill_value;
    } else if (x == list.front()) {
        return i_begin;
    } else if (x == list.back()) {
        return i_end;
    }
    while (true) {
        i_mid = (i_begin + i_end) / 2;
        if (x < list[i_mid]) {
            i_end = i_mid - 1;
        } else if (x < list[i_mid + 1]) {
            return i_mid;
        } else {
            i_begin = i_mid + 1;
        }
    }
}

auto genWorkRanges(const int work_size,
                   const int n_work_units,
                   std::vector<int>& work_ranges,
                   std::vector<int>& proc_map) -> void
{
    // First attempt to distribute all work equally among the work units
    std::vector<int> workloads(n_work_units, work_size / n_work_units);
    // If the amount of work (e.g. number of L1A batches) is not an
    // exact multiple of work units (e.g. MPI processes), assign a bit
    // more work to a subset of work units.
    for (int i {}; i < work_size % n_work_units; ++i) {
        ++workloads.at(i);
    }
    // Work ranges are defined such that the ith work unit processes
    // data from work_ranges[i] to work_ranges[i + 1].
    work_ranges.resize(n_work_units + 1);
    work_ranges.front() = 0;
    for (int i { 1 }; i < n_work_units + 1; ++i) {
        work_ranges.at(i) = work_ranges.at(i - 1) + workloads.at(i - 1);
    }
    // Map size is the number of work elements which is given by the
    // last element of work_ranges.
    proc_map.resize(work_ranges.back());
    // First element of work is assigned to process 0
    proc_map.front() = 0;
    // Index of the work element that signals the beginning of the
    // work range of the next process
    auto next_work_start { work_ranges.cbegin() + 1 };
    for (int i_wrk { 1 }; i_wrk < static_cast<int>(proc_map.size()); ++i_wrk) {
        // If index of the current work element is equal or higher
        // than the start of the next work range, assign this work
        // element to the next process (map value increased by one
        // compared to the previous value).
        if (i_wrk >= *next_work_start) {
            proc_map[i_wrk] = proc_map[i_wrk - 1] + 1;
            // Shift the boundary of work range
            ++next_work_start;
        } else {
            // Assign this work element to the same process as before
            proc_map[i_wrk] = proc_map[i_wrk - 1];
        }
    }
}

auto genWorkRanges(const int work_size,
                   const int n_work_units,
                   std::vector<int>& work_ranges) -> void
{
    std::vector<int> _ {};
    genWorkRanges(work_size, n_work_units, work_ranges, _);
}

auto updateMaskLt(const SharedArray<double>& data,
                  const double threshold,
                  const std::string& info_str,
                  std::vector<bool>& mask) -> void
{
    updateMask(data, threshold, info_str, mask);
}

auto updateMaskGt(const SharedArray<double>& data,
                  const double threshold,
                  const std::string& info_str,
                  std::vector<bool>& mask) -> void
{
    updateMask<std::greater<double>>(data, threshold, info_str, mask);
}

} // namespace spex
