// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "spex_settings.h"

#include "io.h"

namespace spex {

const std::string SPEXSettings::l1x_raw_text {
    "Name of L1X output files after just co-adding and converting to floats"
};
const std::string SPEXSettings::l1x_dark_text {
    "Name of L1X output files after dark correction"
};
const std::string SPEXSettings::l1x_noise_text {
    "Name of L1X output files after noise characterization"
};
const std::string SPEXSettings::l1x_nonlin_text {
    "Name of L1X output files after non-linearity correction"
};
const std::string SPEXSettings::l1x_prnu_text {
    "Name of L1X output files after PRNU correction"
};
const std::string SPEXSettings::l1x_stray_text {
    "Name of L1X output files after stray light correction"
};
const std::string SPEXSettings::l1x_fov_text {
    "Name of L1X output files for extraced spectra in detector counts"
};
const std::string SPEXSettings::l1x_rad_text {
    "Name of L1X output files for extracted spectra in radiance units"
};
const std::string SPEXSettings::dark_cols_left_text {
    "Number of columns corresponding to the left side of an /unbinned/\n"
    "detector whose median is subtracted from each row. If dark_cols_right is\n"
    "also given, the median is taken over left and right columns. This\n"
    "correction is applied\nfor "
};
const std::string SPEXSettings::dark_cols_right_text {
    "Number of columns corresponding to the right side of an /unbinned/\n"
    "detector whose median is subtracted from each row. If dark_cols_right is\n"
    "also given, the median is taken over left and right columns. This\n"
    "correction is applied\nfor"
};
const std::string SPEXSettings::discard_frames_front_text {
    "Number of frames to discard from the beginning of L1A files to mitigate\n"
    "memory effects"
};
const std::string SPEXSettings::discard_frames_back_text {
    "Number of frames to discard from the end of L1A files to mitigate memory\n"
    "effects"
};

auto SPEXSettings::scanKeys() -> void
{
    // Main
    scan(main.processes);
    scan(main.ckd_file_in);
    scan(main.ckd_file_out);
    scan(main.detector_spec);
    scan(main.detector_spat);
    scan(main.detector_spec_full);
    scan(main.detector_spat_full);
    scan(main.n_viewports);
    scan(main.binningtable);
    scan(main.row_dep_bgr);
    scan(main.spectral_binning);
    scan(main.debug);

    // Dark offset and current calibration
    scan(dark.order);
    scan(dark.nominal_temperature);
    scan(dark.saturation_fraction_start);
    scan(dark.chi2_increment_threshold);
    scan(dark.mask_chi2_max);
    scan(dark.mask_offset_min);
    scan(dark.mask_offset_max);
    scan(dark.mask_current_min);
    scan(dark.mask_current_max);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::dark)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::dark)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::dark)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::dark)));
    scan(debug.at(static_cast<int>(ProcLevel::dark)));

    // Noise calibration
    scan(noise.mask_g_min);
    scan(noise.mask_g_max);
    scan(noise.mask_n_min);
    scan(noise.mask_n_max);
    scan(noise.intensity_threshold);
    scan(noise.fitting.iteration_limit);
    scan(noise.fitting.initial_g);
    scan(noise.fitting.initial_n);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::noise)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::noise)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::noise)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::noise)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::noise)));
    scan(discard_frames_front.at(static_cast<int>(ProcLevel::noise)));
    scan(discard_frames_back.at(static_cast<int>(ProcLevel::noise)));
    scan(debug.at(static_cast<int>(ProcLevel::noise)));

    // Nonlinearity calibration
    scan(nonlin.lmin);
    scan(nonlin.lmax);
    scan(nonlin.exposure_times);
    scan(nonlin.exptime_tol);
    scan(nonlin.scale_median);
    scan(nonlin.knots);
    scan(nonlin.order);
    scan(nonlin.mask_lin_chi2_max);
    scan(nonlin.mask_chi2_max);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::nonlin)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::nonlin)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::nonlin)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::nonlin)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::nonlin)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::nonlin)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::nonlin)));
    scan(discard_frames_front.at(static_cast<int>(ProcLevel::nonlin)));
    scan(discard_frames_back.at(static_cast<int>(ProcLevel::nonlin)));
    scan(debug.at(static_cast<int>(ProcLevel::nonlin)));

    // PRNU calibration
    scan(prnu.order_spat);
    scan(prnu.order_spec);
    scan(prnu.mask_prnu_min);
    scan(prnu.mask_prnu_max);
    scan(prnu.fitting.iteration_limit);
    scan(prnu.fitting.min_n_blocks);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::prnu)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::prnu)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::prnu)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::prnu)));
    scan(l1x_nonlin.at(static_cast<int>(ProcLevel::prnu)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::prnu)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::prnu)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::prnu)));
    scan(debug.at(static_cast<int>(ProcLevel::prnu)));

    // Stray light calibration
    scan(stray.exposure_intensity_thres);
    scan(stray.kernel_hole_radius);
    scan(stray.kernel_extent);
    scan(stray.reduced_kernels);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::stray)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::stray)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::stray)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::stray)));
    scan(l1x_nonlin.at(static_cast<int>(ProcLevel::stray)));
    scan(l1x_prnu.at(static_cast<int>(ProcLevel::stray)));
    scan(dark_cols_left.at(static_cast<int>(ProcLevel::stray)));
    scan(dark_cols_right.at(static_cast<int>(ProcLevel::stray)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::stray)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::stray)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::stray)));
    scan(debug.at(static_cast<int>(ProcLevel::stray)));

    // FOV calibration
    scan(fov.n_knots);
    scan(fov.bspline_order);
    scan(fov.exclude_columns_left);
    scan(fov.exclude_columns_right);
    scan(fov.act_sampling);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::fov)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::fov)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::fov)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::fov)));
    scan(l1x_nonlin.at(static_cast<int>(ProcLevel::fov)));
    scan(l1x_prnu.at(static_cast<int>(ProcLevel::fov)));
    scan(l1x_stray.at(static_cast<int>(ProcLevel::fov)));
    scan(dark_cols_left.at(static_cast<int>(ProcLevel::fov)));
    scan(dark_cols_right.at(static_cast<int>(ProcLevel::fov)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::fov)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::fov)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::fov)));
    scan(debug.at(static_cast<int>(ProcLevel::fov)));

    // Swath calibration
    scan(swath.theodolite_reference_filename);
    scan(swath.theodolite_spotcheck_filename);
    scan(swath.viewport_alt_angles);
    scan(swath.theodolite_alt_angle_tol);
    scan(swath.actscan_sampling);
    scan(swath.act_linesearch_tol);
    scan(swath.act_angle_tol);
    scan(swath.insignificant_island);
    scan(swath.maxreso);
    scan(swath.backgroundorder);
    scan(swath.gaussfit);
    scan(swath.domain_range);
    scan(swath.maxiter);
    scan(swath.magtol);
    scan(swath.postol);
    scan(swath.widthtol);
    scan(swath.backgroundtol);
    scan(swath.steptolerance);
    scan(swath.initial_step_reducer);
    scan(swath.reducerfactor_fail);
    scan(swath.reducerfactor_success);
    scan(swath.reducer_limit);
    scan(swath.order);
    scan(swath.min_fit);
    scan(swath.rotmat);
    scan(swath.rot_warningthreshold);
    scan(swath.theo_altmodel_maxiter);
    scan(swath.theo_altmodel_tol);
    scan(swath.theo_altmodel_steptolerance);
    scan(swath.theo_altmodel_initial_step_reducer);
    scan(swath.theo_altmodel_reducerfactor_fail);
    scan(swath.theo_altmodel_reducerfactor_success);
    scan(swath.theo_altmodel_reducer_limit);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::swath)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::swath)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::swath)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::swath)));
    scan(l1x_nonlin.at(static_cast<int>(ProcLevel::swath)));
    scan(l1x_prnu.at(static_cast<int>(ProcLevel::swath)));
    scan(l1x_stray.at(static_cast<int>(ProcLevel::swath)));
    scan(l1x_fov.at(static_cast<int>(ProcLevel::swath)));
    scan(dark_cols_left.at(static_cast<int>(ProcLevel::swath)));
    scan(dark_cols_right.at(static_cast<int>(ProcLevel::swath)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::swath)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::swath)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::swath)));
    scan(debug.at(static_cast<int>(ProcLevel::swath)));

    // Spectral calibration
    scan(wave.peak_insignificant_island);
    scan(wave.peak_maxreso);
    scan(wave.peak_backgroundorder);
    scan(wave.peak_gaussfit);
    scan(wave.peak_domain_range);
    scan(wave.peak_maxiter);
    scan(wave.peak_magtol);
    scan(wave.peak_postol);
    scan(wave.peak_widthtol);
    scan(wave.peak_backgroundtol);
    scan(wave.peak_steptolerance);
    scan(wave.peak_initial_step_reducer);
    scan(wave.peak_reducerfactor_fail);
    scan(wave.peak_reducerfactor_success);
    scan(wave.peak_reducer_limit);
    scan(wave.refspec_insignificant_island);
    scan(wave.refspec_maxreso);
    scan(wave.refspec_backgroundorder);
    scan(wave.refspec_gaussfit);
    scan(wave.refspec_domain_range);
    scan(wave.refspec_maxiter);
    scan(wave.refspec_magtol);
    scan(wave.refspec_postol);
    scan(wave.refspec_widthtol);
    scan(wave.refspec_backgroundtol);
    scan(wave.refspec_steptolerance);
    scan(wave.refspec_initial_step_reducer);
    scan(wave.refspec_reducerfactor_fail);
    scan(wave.refspec_reducerfactor_success);
    scan(wave.refspec_reducer_limit);
    scan(wave.order);
    scan(wave.npeak);
    scan(wave.peak_offset);
    scan(wave.reference_wavelengths);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::wave)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::wave)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::wave)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::wave)));
    scan(l1x_nonlin.at(static_cast<int>(ProcLevel::wave)));
    scan(l1x_prnu.at(static_cast<int>(ProcLevel::wave)));
    scan(l1x_stray.at(static_cast<int>(ProcLevel::wave)));
    scan(l1x_fov.at(static_cast<int>(ProcLevel::wave)));
    scan(dark_cols_left.at(static_cast<int>(ProcLevel::wave)));
    scan(dark_cols_right.at(static_cast<int>(ProcLevel::wave)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::wave)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::wave)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::wave)));
    scan(debug.at(static_cast<int>(ProcLevel::wave)));

    // Radiometric calibration
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::rad)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::rad)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::rad)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::rad)));
    scan(l1x_nonlin.at(static_cast<int>(ProcLevel::rad)));
    scan(l1x_prnu.at(static_cast<int>(ProcLevel::rad)));
    scan(l1x_stray.at(static_cast<int>(ProcLevel::rad)));
    scan(l1x_fov.at(static_cast<int>(ProcLevel::rad)));
    scan(dark_cols_left.at(static_cast<int>(ProcLevel::rad)));
    scan(dark_cols_right.at(static_cast<int>(ProcLevel::rad)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::rad)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::rad)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::rad)));
    scan(debug.at(static_cast<int>(ProcLevel::rad)));

    // Polarimetric calibration
    scan(pol.delta_prior_file);
    scan(pol.mt_prior_file);
    scan(pol.aolp_from);
    scan(pol.aolp_offset);
    scan(pol.aolp_sign_flip);
    scan(pol.fit_window_min);
    scan(pol.fit_window_max);
    scan(pol.fitting.delta_order);
    scan(pol.fitting.eps_order);
    scan(pol.fitting.mt_order);
    scan(pol.print_debug.enabled);
    scan(pol.print_debug.viewport);
    scan(pol.print_debug.fov);
    scan(pol.print_debug.output_folder);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::pol)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::pol)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::pol)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::pol)));
    scan(l1x_nonlin.at(static_cast<int>(ProcLevel::pol)));
    scan(l1x_prnu.at(static_cast<int>(ProcLevel::pol)));
    scan(l1x_stray.at(static_cast<int>(ProcLevel::pol)));
    scan(l1x_fov.at(static_cast<int>(ProcLevel::pol)));
    scan(l1x_rad.at(static_cast<int>(ProcLevel::pol)));
    scan(dark_cols_left.at(static_cast<int>(ProcLevel::pol)));
    scan(dark_cols_right.at(static_cast<int>(ProcLevel::pol)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::pol)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::pol)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::pol)));
    scan(debug.at(static_cast<int>(ProcLevel::pol)));

    // L1B process
    scan(l1b.outputfile);
    scan(l1b.order);
    scan(l1b.knots);
    scan(l1b.wave_min);
    scan(l1b.wave_max);
    scan(l1b.l1b_wavelength);
    scan(l1b.intensity_wavelength);
    scan(l1b.resolving_power);
    scan(l1b.gauss_range_tol);
    scan(l1b.telescope_errors);
    scan(l1b.geolocation);
    scan(l1b.utcfile);
    scan(l1b.demfile);
    scan(l1b.semi_major_axis);
    scan(l1b.semi_minor_axis);
    scan(l1b.latitude_tol);
    scan(l1b.mountain);
    scan(l1b.movedistance);
    scan(l1b.extremeweight);
    scan(l1b.geolocation_tol);
    scan(l1b.l1b_output_level);
    scan(l1b.first_proc_rel_workload);
    scan(l1b.mpi_send_size);
    scan(l1b.last_calibration_step);
    scan(l1b.image_start);
    scan(l1b.image_end);
    scan(l1b.disable_io);
    scan(l1b.disable_mpi_comm);
    scan(l1b.noise_propagation);
    scan(l1a_file_list.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1a_files.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1x_raw.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1x_dark.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1x_nonlin.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1x_prnu.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1x_stray.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1x_fov.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1x_rad.at(static_cast<int>(ProcLevel::l1b)));
    scan(dark_cols_left.at(static_cast<int>(ProcLevel::l1b)));
    scan(dark_cols_right.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1a_bgr_files.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1a_bgr_file_list.at(static_cast<int>(ProcLevel::l1b)));
    scan(l1a_type.at(static_cast<int>(ProcLevel::l1b)));
    scan(debug.at(static_cast<int>(ProcLevel::l1b)));

    // L1C process
    scan(l1c.demfile);
    scan(l1c.semi_major_axis);
    scan(l1c.semi_minor_axis);
    scan(l1c.latitude_tol);
    scan(l1c.l1bfile);
    scan(l1c.l1cfile);
    scan(l1c.lat);
    scan(l1c.lon);
    scan(debug.at(static_cast<int>(ProcLevel::l1c)));
}

auto SPEXSettings::getReducedConfig() const -> std::string
{
    auto reduced_config { YAML::Clone(config) };
    for (const auto& it : proc_level_str_to_enum) {
        if (std::ranges::find(main.processes, it.second)
            == main.processes.end()) {
            reduced_config.remove(it.first);
        }
    }
    YAML::Emitter out {};
    out.SetBoolFormat(YAML::YesNoBool);
    out.SetNullFormat(YAML::LowerNull);
    out << reduced_config;
    return out.c_str();
}

auto SPEXSettings::checkParameters() -> void
{
    // Main
    if (main.processes.empty()) {
        throw std::runtime_error { "[main][processes] is empty, must select at "
                                   "least one process" };
    }
    std::ranges::sort(main.processes);
    // If a process is not listed in main.processes, set its skip
    // value to true. Then CKD::write will not attempt to write for
    // that process.
    for (int i_proc {}; i_proc < static_cast<int>(skip.size()); ++i_proc) {
        skip.at(i_proc) =
          std::ranges::find(main.processes, static_cast<ProcLevel>(i_proc))
          == main.processes.end();
    }
    // CKD file must be present for L1B but is otherwise optional
    checkPresenceOfFile(main.ckd_file_in,
                        !skip.at(static_cast<int>(ProcLevel::l1b)));
    // CKD output file name must be given for CKD generation steps
    if (const auto* const first_proc { std::ranges::find(skip, false) };
        first_proc - skip.cbegin() < static_cast<int>(ProcLevel::l1b)
        && main.ckd_file_out.empty()) {
        throw std::runtime_error { main.ckd_file_out.keyToStr()
                                   + " is required when generating a CKD" };
    }
    // Make sure CKD output is writable
    checkFileWritable(main.ckd_file_out);
    // Set default full detector dimensions and swap the values
    if (!main.detector_spec_full) {
        main.detector_spec_full = main.detector_spec;
    }
    if (!main.detector_spat_full) {
        main.detector_spat_full = main.detector_spat;
    }
    if (main.detector_spec % main.detector_spec_full.value() != 0) {
        throw std::runtime_error { "[main][detector_spec] must be a multiple "
                                   "of [main][detector_spec_reduced]" };
    }
    if (main.detector_spat % main.detector_spat_full.value() != 0) {
        throw std::runtime_error { "[main][detector_spat] must be a multiple "
                                   "of [main][detector_spat_reduced]" };
    }
    int tmp { main.detector_spec };
    main.detector_spec = main.detector_spec_full.value();
    main.detector_spec_full = tmp;
    tmp = main.detector_spat;
    main.detector_spat = main.detector_spat_full.value();
    main.detector_spat_full = tmp;
    // If data needs to be prebinned (detector_spat_reduced and/or
    // detector_spec_reduced) were given then adjust some input
    // parameters using these binning factors.
    const int pre_bin_spat { main.detector_spat_full.value()
                             / main.detector_spat };
    const int pre_bin_spec { main.detector_spec_full.value()
                             / main.detector_spec };
    const int pre_bin { pre_bin_spat * pre_bin_spec };

    // Common to some or all calibration steps
    if (pre_bin) {
        for (auto& dark_cols : dark_cols_left) {
            dark_cols /= pre_bin;
        }
        for (auto& dark_cols : dark_cols_right) {
            dark_cols /= pre_bin;
        }
    }

    // Noise calibration
    if (pre_bin) {
        if (noise.mask_n_min) {
            noise.mask_n_min *=
              static_cast<int>(std::sqrt(static_cast<double>(pre_bin)));
        }
        if (noise.mask_n_max) {
            noise.mask_n_max *=
              static_cast<int>(std::sqrt(static_cast<double>(pre_bin)));
        }
    }

    // FOV calibration
    for (auto& act_sampling : fov.act_sampling) {
        act_sampling *= math::deg_to_rad;
    }
    if (pre_bin) {
        fov.exclude_columns_left /= pre_bin;
        fov.exclude_columns_right /= pre_bin;
    }

    // Swath calibration
    swath.act_angle_tol *= math::deg_to_rad;
    swath.act_linesearch_tol *= math::deg_to_rad;
    if (swath.actscan_sampling) {
        swath.actscan_sampling.value() *= math::deg_to_rad;
    }
    for (auto& viewport_alt_angles : swath.viewport_alt_angles) {
        viewport_alt_angles *= math::deg_to_rad;
    }
    swath.theodolite_alt_angle_tol *= math::deg_to_rad;

    // Polarimetric calibration
    checkPresenceOfFile(pol.delta_prior_file,
                        !skip.at(static_cast<int>(ProcLevel::pol)));
    checkPresenceOfFile(pol.mt_prior_file,
                        !skip.at(static_cast<int>(ProcLevel::pol)));
    if (pol.fitting.mt_order
        && pol.fitting.mt_order
             > std::max(pol.fitting.delta_order, pol.fitting.eps_order)) {
        throw std::runtime_error { "[pol][fitting][mt_order] should not be "
                                   "higher than delta_order/eps_order" };
    }
    if (!skip.at(static_cast<int>(ProcLevel::pol))) {
        if (!pol.aolp_from) {
            throw std::runtime_error {
                "mandatory keyword missing: [pol][aolp_from]"
            };
        }
        if (!pol.aolp_offset) {
            throw std::runtime_error {
                "mandatory keyword missing: [pol][aolp_offset]"
            };
        }
        if (!pol.aolp_sign_flip) {
            throw std::runtime_error {
                "mandatory keyword missing: [pol][aolp_sign_flip]"
            };
        }
        pol.aolp_offset.value() *= math::deg_to_rad;
    }

    // L1B
    l1b.first_proc_rel_workload.value =
      std::min(l1b.first_proc_rel_workload.value, 1.0);
    if (!skip.at(static_cast<int>(ProcLevel::l1b)) && l1b.outputfile.empty()) {
        throw std::runtime_error { l1b.outputfile.keyToStr()
                                   + " is required for L1B" };
    }
    checkFileWritable(l1b.outputfile);

    // L1C
    checkPresenceOfFile(l1c.l1bfile,
                        !skip.at(static_cast<int>(ProcLevel::l1c)));
    checkFileWritable(l1c.l1cfile);

    // Common
    for (int i_lev {}; i_lev < static_cast<int>(ProcLevel::n_levels); ++i_lev) {
        // Open file lists such as l1a_file_list for a given process
        // and copy the file names into corresponding variables such
        // as l1a_files.
        extractFilesFromList(
          l1a_file_list.at(i_lev), l1a_files.at(i_lev), false);
        extractFilesFromList(
          l1a_bgr_file_list.at(i_lev), l1a_bgr_files.at(i_lev), false);
        // Check if all paths point to existing files
        for (const auto& file : l1a_files.at(i_lev)) {
            checkPresenceOfFile(file, true);
        }
        for (const auto& file : l1a_bgr_files.at(i_lev)) {
            checkPresenceOfFile(file, true);
        }
        // Unless the user set the debug level of a given process, set
        // its value according to the one from the main section.
        if (!debug.at(i_lev)) {
            debug.at(i_lev) = main.debug;
        }
    }
}

} // namespace spex
