// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for storing all user defined configuration parameters for
// SPEXone. A single instance is used throughout the code.
//
// When adding a new parameter two places need to be updated: i)
// define the parameter by editing one of structs below or by defining
// a new struct and ii) call the scan function on the parameter in
// scanKeys so it can be read from the YAML file and/or printed on the
// screen. You may optionally add checks in Settings::checkParameters
// for things like correct file paths.

#pragma once

#include "settings.h"

#include <netcdf.h>

namespace spex {

class SPEXSettings : public Settings
{
private:
    // Info fields for some of the parameters below. These are used
    // for multiple settings.
    static const std::string l1x_raw_text;
    static const std::string l1x_dark_text;
    static const std::string l1x_noise_text;
    static const std::string l1x_nonlin_text;
    static const std::string l1x_prnu_text;
    static const std::string l1x_stray_text;
    static const std::string l1x_fov_text;
    static const std::string l1x_rad_text;
    static const std::string dark_cols_left_text;
    static const std::string dark_cols_right_text;
    static const std::string discard_frames_front_text;
    static const std::string discard_frames_back_text;

    auto checkParameters() -> void override;

public:
    // Main (driver) section of configuration
    struct
    {
        Setting<std::vector<ProcLevel>> processes {
            { "main", "processes" },
            {},
            "Processes to be executed. Allowed values: dark, noise, nonlin,\n"
            "prnu, stray, fov, swath, wave, rad, pol, l1b, l1c."
        };
        Setting<std::string> ckd_file_in {
            { "main", "ckd_file_in" },
            {},
            "Input CKD file name. For steps not listed in 'processes', if the\n"
            "corresponding section is present in the input CKD file, the data\n"
            "is taken from that file and copied to the output CKD file.\n"
            "Needed for L1B and optionally CKD generation."
        };
        Setting<std::string> ckd_file_out {
            { "main", "ckd_file_out" },
            {},
            "Output CKD file name (mandatory for CKD generation)"
        };
        Setting<int> detector_spat {
            { "main", "detector_spat" },
            2048,
            "Number of detector pixels in the spatial dimension"
        };
        Setting<int> detector_spec {
            { "main", "detector_spec" },
            2048,
            "Number of detector pixels in the spectral dimension"
        };
        // The user defines the reduced dimension but it is easier to
        // save the original full dimension here. detector_spec is
        // then swapped by detector_spec_full, i.e. detector_spec
        // becomes the reduced dimension.
        Setting<std::optional<int>> detector_spat_full {
            { "main", "detector_spat_reduced" },
            "If set, produce a CKD with this spatial dimension. Calibration\n"
            "measurement are then immediately binned with the factor\n"
            "detector_spec / detector_spec_reduce in the spatial dimension\n"
            "and the CKD is generated using the binned (reduced) data."
        };
        Setting<std::optional<int>> detector_spec_full {
            { "main", "detector_spec_reduced" },
            "Like detector_spat_reduced but in the spectral dimension"
        };
        Setting<int> n_viewports { { "main", "n_viewports" },
                                   5,
                                   "Number of viewports" };
        Setting<std::string> binningtable {
            { "main", "binningtable" },
            {},
            "File with all the used binning tables"
        };
        Setting<int> stray_van_cittert_steps {
            { "main", "stray_van_cittert_steps" },
            3,
            "Number of Van Cittert iterations for stray light correction"
        };
        Setting<bool> row_dep_bgr {
            { "main", "row_dep_bgr" },
            false,
            "Whether row-dependent offset, if used, is applied to both the\n"
            "calibration image and the background image. Default is false\n"
            "meaning that the background image is subtracted first and then\n"
            "row-dependent offset, if used, is only applied to the background\n"
            "corrected image."
        };
        Setting<bool> spectral_binning {
            { "main", "spectral_binning" },
            true,
            "Whether to spectrally bin the FOV and higher CKDs",
            SettingLevel::debug
        };
        Setting<bool> debug {
            { "main", "debug" },
            false,
            "Run all processes in debug mode, which saves additional\n"
            "information in the generated product (CKD or an L1B/L1C product)\n"
            "and produces more verbose screen output. This setting can be\n"
            "overwritten by the debug setting of individual sections",
            SettingLevel::debug
        };
    } main;

    // Dark offset and current calibration
    struct
    {
        Setting<int> order {
            { "dark", "order" },
            0,
            "B-spline order for temperature fit (one higher than polynomial\n"
            "order)"
        };
        Setting<double> nominal_temperature { { "dark", "nominal_temperature" },
                                              NC_FILL_DOUBLE,
                                              "Standard temperature" };
        Setting<double> saturation_fraction_start {
            { "dark", "saturation_fraction_start" },
            NC_FILL_DOUBLE,
            "Fraction of maximum signal to include in initial fit"
        };
        Setting<double> chi2_increment_threshold {
            { "dark", "chi2_increment_threshold" },
            NC_FILL_DOUBLE,
            "Threshold of chi square increment where to stop"
        };
        Setting<std::optional<double>> mask_chi2_max {
            { "dark", "mask_chi2_max" },
            "Maximum fit residual chi squared for dark fit"
        };
        Setting<std::optional<double>> mask_offset_min {
            { "dark", "mask_offset_min" },
            "Minimum fixed dark signal (independent of integration time)"
        };
        Setting<std::optional<double>> mask_offset_max {
            { "dark", "mask_offset_max" },
            "Maximum fixed dark signal (independent of integration time)"
        };
        Setting<std::optional<double>> mask_current_min {
            { "dark", "mask_current_min" },
            "Minimum dark signal added per second of integration time"
        };
        Setting<std::optional<double>> mask_current_max {
            { "dark", "mask_current_max" },
            "Maximum dark signal added per second of integration time"
        };
    } dark;

    // Noise calibration
    struct
    {
        Setting<std::optional<double>> mask_g_min {
            { "noise", "mask_g_min" },
            "Pixels whose inverse noise gain (g) is lower than this value are\n"
            "masked bad"
        };
        Setting<std::optional<double>> mask_g_max {
            { "noise", "mask_g_max" },
            "Pixels whose inverse noise gain (g) is higher than this value\n"
            "are masked bad"
        };
        Setting<std::optional<double>> mask_n_min {
            { "noise", "mask_n_min" },
            "Pixels whose dark noise (n) is lower than this value are masked\n"
            "bad"
        };
        Setting<std::optional<double>> mask_n_max {
            { "noise", "mask_n_max" },
            "Pixels whose dark noise (n) is higher than this value are masked\n"
            "bad"
        };
        Setting<double> intensity_threshold {
            { "noise", "intensity_threshold" },
            730.0,
            "Signal intensities higher than this value are excluded from the\n"
            "noise model fit"
        };
        struct
        {
            Setting<int> iteration_limit {
                { "noise", "fitting", "iteration_limit" },
                100,
                "Iteration limit",
                SettingLevel::debug
            };
            Setting<double> initial_g {
                { "noise", "fitting", "initial_g" },
                0.08,
                "Initial value for inverse noise gain (g)",
                SettingLevel::debug
            };
            Setting<double> initial_n { { "noise", "fitting", "initial_n" },
                                        3.0,
                                        "Initial value for dark noise (n)",
                                        SettingLevel::debug };
        } fitting;
    } noise;

    // Nonlinearity calibration
    struct
    {
        Setting<double> lmin {
            { "nonlin", "lmin" },
            NC_FILL_DOUBLE,
            "Lower border (in measured signal) of linear domain"
        };
        Setting<double> lmax {
            { "nonlin", "lmax" },
            NC_FILL_DOUBLE,
            "Upper border (in measured signal) of linear domain"
        };
        Setting<std::vector<double>> exposure_times {
            { "nonlin", "exposure_times" },
            {},
            "Exposure times for which the non-linearity is evaluated"
        };
        Setting<double> exptime_tol {
            { "nonlin", "exptime_tol" },
            {},
            "Relative tolerance for exposure time grouping"
        };
        Setting<bool> scale_median {
            { "nonlin", "scale_median" },
            true,
            "Flag to scale the spline knots with the median of the saturated\n"
            "image. Disable to use individual-pixel saturation levels"
        };
        Setting<std::vector<double>> knots {
            { "nonlin", "knots" },
            {},
            "B-spline knots in signal domain"
        };
        Setting<int> order { { "nonlin", "order" }, 0, "B-spline order" };
        Setting<double> mask_lin_chi2_max {
            { "nonlin", "mask_lin_chi2_max" },
            NC_FILL_DOUBLE,
            "Maximum residual chi squared on the linear fit"
        };
        Setting<std::optional<double>> mask_chi2_max {
            { "nonlin", "mask_chi2_max" },
            "Maximum residual chi squared on non-linearity correction"
        };
    } nonlin;

    // PRNU calibration
    struct
    {
        Setting<int> order_spat {
            { "prnu", "order_spat" },
            0,
            "Polynomial order for the spatial dimension"
        };
        Setting<int> order_spec {
            { "prnu", "order_spec" },
            0,
            "Polynomial order for the spectral dimension"
        };
        Setting<std::optional<double>> mask_prnu_min {
            { "prnu", "mask_prnu_min" },
            "Minimum pixel response non-uniformity"
        };
        Setting<std::optional<double>> mask_prnu_max {
            { "prnu", "mask_prnu_max" },
            "Maximum pixel response non-uniformity"
        };
        struct
        {
            Setting<int> iteration_limit {
                { "prnu", "fitting", "iteration_limit" },
                100,
                "Iteration limit",
                SettingLevel::debug
            };
            Setting<int> min_n_blocks {
                { "prnu", "fitting", "min_n_blocks" },
                100,
                "Minimum number of blocks of a 2D block-cyclic array held by\n"
                "each process. For minimal MPI communication set this values \n"
                "to 1 but for better load balancing set it to a larger value.",
                SettingLevel::debug
            };
        } fitting;
    } prnu;

    // Stray light calibration
    struct
    {
        Setting<double> exposure_intensity_thres {
            { "stray", "exposure_intensity_thres" },
            730.0,
            "For pixel light levels higher than this threshold, switch to an\n"
            "image with a longer exposure time instead. This parameter\n"
            "affects how composite images are built up from multiple exposure\n"
            "times."
        };
        Setting<double> kernel_hole_radius {
            { "stray", "kernel_hole_radius" },
            10.0,
            "Radius of the center hole when constructing the kernels"
        };
        Setting<int> kernel_extent {
            { "stray", "kernel_extent" },
            512,
            "Kernels are confined to a square box with the side\n"
            "2*kernel_extent. Beyond that the kernel is taken to be zero."
        };
        Setting<bool> reduced_kernels {
            { "stray", "reduced_kernels" },
            true,
            "If true, kernels are reduced in size to\n"
            "detector_spec/2 * detector_spat/2, e.g. 1024 * 1024. Otherwise\n"
            "the kernels have the same dimensions as the detector.\n"
            "Implication is that for deconvolution the images don't have to\n"
            "be fully unbinned which reduces computational cost."
        };
    } stray;

    // FOV calibration
    struct
    {
        Setting<int> n_knots {
            { "fov", "n_knots" },
            2,
            "Number of B-spline knots when smoothing the FOV curves"
        };
        Setting<int> bspline_order {
            { "fov", "bspline_order" },
            3,
            "B-spline (polynomial) order when smoothing the FOV curves"
        };
        Setting<int> exclude_columns_left {
            { "fov", "exclude_columns_left" },
            {},
            "Number of columns from the left side of the detector to be\n"
            "excluded when determining peaks. The edge columns have no useful\n"
            "signal which can lead to instabilities when trying to fit with a\n"
            "Gaussian."
        };
        Setting<int> exclude_columns_right {
            { "fov", "exclude_columns_right" },
            {},
            "Same as exclude_columns_left but applies to the right edge (long\n"
            "wavelengths) of the detector."
        };
        Setting<std::vector<double>> act_sampling {
            { "fov", "act_sampling" },
            {},
            "ACT angle sampling per viewport in degrees"
        };
    } fov;

    // Swath calibration
    struct
    {
        Setting<std::string> theodolite_reference_filename {
            { "swath", "theodolite_reference_filename" },
            {},
            "NetCDF file name of the theodolite reference file (a lot of\n"
            "measurement on the wrong day)"
        };
        Setting<std::string> theodolite_spotcheck_filename {
            { "swath", "theodolite_spotcheck_filename" },
            {},
            "NetCDF file name of the theodolite spot-check file (one\n"
            "measurement on correct day)"
        };
        Setting<std::vector<double>> viewport_alt_angles {
            { "swath", "viewport_alt_angles" },
            {},
            "Expected ALT angles of each viewport (also used to put the\n"
            "viewports in the correct order)"
        };
        Setting<double> theodolite_alt_angle_tol {
            { "swath", "theodolite_alt_angle_tol" },
            1.0,
            "Maximum allowed different between ALT angle of theodolite\n"
            "measurement and expected ALT angle of a viewport"
        };
        Setting<std::optional<double>> actscan_sampling {
            { "swath", "actscan_sampling" },
            "Sampling for initial scan over swath to find the correct FOVcal\n"
            "ACT angle. Default is to use swathcal stage ACT angle."
        };
        Setting<double> act_linesearch_tol {
            { "swath", "act_linesearch_tol" },
            1e-7,
            "Convergence tolerance in line search over ACT angles"
        };
        Setting<double> act_angle_tol {
            { "swath", "act_angle_tol" },
            {},
            "Difference between across-track angles that are considered to be\n"
            "the same"
        };
        Setting<int> insignificant_island {
            { "swath", "insignificant_island" },
            0,
            "Setting for peak detection: Size of peaks or non-peaks to be\n"
            "discarded"
        };
        Setting<int> maxreso {
            { "swath", "maxreso" },
            128,
            "Highest denominator in setting threshold to detect peaks"
        };
        Setting<int> backgroundorder {
            { "swath", "backgroundorder" },
            1,
            "Degrees of freedom of background polynomial for peak detection"
        };
        Setting<bool> gaussfit { { "swath", "gaussfit" },
                                 true,
                                 "Flag for performing gauss fit" };
        Setting<double> domain_range {
            { "swath", "domain_range" },
            5.0,
            "Range in a-priori FWHM that a Gauss contains in split\n"
            "splitgaussfit"
        };
        Setting<int> maxiter {
            { "swath", "maxiter" },
            0,
            "Number of iterations for fitting Gauss peak shape"
        };
        Setting<double> magtol {
            { "swath", "magtol" },
            NC_FILL_DOUBLE,
            "Tolerance of peak magnitude (in units of a-priori magnitude)"
        };
        Setting<double> postol { { "swath", "postol" },
                                 NC_FILL_DOUBLE,
                                 "Tolerance of peak positions (in indices)" };
        Setting<double> widthtol { { "swath", "widthtol" },
                                   NC_FILL_DOUBLE,
                                   "Tolerance of peak FWHM (in indices)" };
        Setting<double> backgroundtol {
            { "swath", "backgroundtol" },
            NC_FILL_DOUBLE,
            "Tolerance of background B-spline coefficients (in median signals)"
        };
        Setting<double> steptolerance {
            { "swath", "steptolerance" },
            1.1,
            "Inversion step acceptance criterion for fitting Gauss peak shape"
        };
        Setting<double> initial_step_reducer {
            { "swath", "initial_step_reducer" },
            4.0,
            "Starting value of the Levenberg-Marquardt step control parameter"
        };
        Setting<double> reducerfactor_fail {
            { "swath", "reducerfactor_fail" },
            2.5,
            "Multiplication of step reducer for a failed step"
        };
        Setting<double> reducerfactor_success {
            { "swath", "reducerfactor_success" },
            0.5,
            "Multiplication of step reducer for a successful step"
        };
        Setting<double> reducer_limit {
            { "swath", "reducer_limit" },
            NC_FILL_DOUBLE,
            "Maximum step control reducer for convergence"
        };
        Setting<std::vector<int>> order {
            { "swath", "order" },
            {},
            "Degrees of freedom for polynomial fit of ALT against ACT angles\n"
            "per viewport, each with a minimum number of successful\n"
            "associations. Start with highest."
        };
        Setting<std::vector<int>> min_fit {
            { "swath", "min_fit" },
            {},
            "Minimum required associations to perform the fit with the\n"
            "respective order. Start with highest."
        };
        Setting<std::vector<double>> rotmat {
            { "swath", "rotmat" },
            {},
            "Rotation matrix from spacecraft to instrument coordinates. Nine\n"
            "elements. Source direction is quick. Destination dimension is\n"
            "slow."
        };
        Setting<double> rot_warningthreshold {
            { "swath", "rot_warningthreshold" },
            1e-8,
            "Allowed error in check if rotmat is a proper rotation matrix"
        };
        Setting<int> theo_altmodel_maxiter {
            { "swath", "theo_altmodel_maxiter" },
            100,
            "Maximum number of iterations in non-linear inversion for\n"
            "ALT-model in the theodolite."
        };
        Setting<double> theo_altmodel_tol {
            { "swath", "theo_altmodel_tol" },
            1e-7,
            "Convergence tolerance on vector elements in non-linear inversion\n"
            "for ALT-model in the theodolite"
        };
        Setting<double> theo_altmodel_steptolerance {
            { "swath", "theo_altmodel_steptolerance" },
            1.1,
            "Inversion step acceptance criterion for non-linear inversion for\n"
            "ALT-model in the theodolite"
        };
        Setting<double> theo_altmodel_initial_step_reducer {
            { "swath", "theo_altmodel_initial_step_reducer" },
            4.0,
            "Starting value of the Levenberg-Marquardt step control\n"
            "parameter for non-linear inversion for ALT-model in the theodolite"
        };
        Setting<double> theo_altmodel_reducerfactor_fail {
            { "swath", "theo_altmodel_reducerfactor_fail" },
            2.5,
            "Multiplication of step reducer for a failed step for non-linear\n"
            "inversion for ALT-model in the theodolite"
        };
        Setting<double> theo_altmodel_reducerfactor_success {
            { "swath", "theo_altmodel_reducerfactor_success" },
            0.5,
            "Multiplication of step reducer for a successful step for\n"
            "non-linear inversion for ALT-model in the theodolite"
        };
        Setting<double> theo_altmodel_reducer_limit {
            { "swath", "theo_altmodel_reducer_limit" },
            1e-7,
            "Maximum step control reducer for convergence for non-linear\n"
            "inversion for ALT-model in the theodolite"
        };
    } swath;

    // Spectral calibration
    struct
    {
        Setting<int> peak_insignificant_island {
            { "wave", "peak_insignificant_island" },
            1,
            "Setting for peak detection: Size of peaks or non-peaks to be\n"
            "discarded"
        };
        Setting<int> peak_maxreso {
            { "wave", "peak_maxreso" },
            128,
            "Highest denominator in setting threshold to detect peaks"
        };
        Setting<double> peak_backgroundorder {
            { "wave", "peak_backgroundorder" },
            1,
            "Degrees of freedom of background polynomial for peak detection"
        };
        Setting<bool> peak_gaussfit { { "wave", "peak_gaussfit" },
                                      true,
                                      "Flag for performing gauss fit" };
        Setting<double> peak_domain_range {
            { "wave", "peak_domain_range" },
            5.0,
            "Range in a-priori FWHM that a Gauss contains in split\n"
            "splitgaussfit"
        };
        Setting<int> peak_maxiter {
            { "wave", "peak_maxiter" },
            0,
            "Number of iterations for fitting Gauss peak shape"
        };
        Setting<double> peak_magtol {
            { "wave", "peak_magtol" },
            NC_FILL_DOUBLE,
            "Tolerance of peak magnitude (in units of a-priori magnitude)"
        };
        Setting<double> peak_postol {
            { "wave", "peak_postol" },
            NC_FILL_DOUBLE,
            "Tolerance of peak positions (in indices)"
        };
        Setting<double> peak_widthtol { { "wave", "peak_widthtol" },
                                        NC_FILL_DOUBLE,
                                        "Tolerance of peak FWHM (in indices)" };
        Setting<double> peak_backgroundtol {
            { "wave", "peak_backgroundtol" },
            NC_FILL_DOUBLE,
            "Tolerance of background B-spline coefficients (in median signals)"
        };
        Setting<double> peak_steptolerance {
            { "wave", "peak_steptolerance" },
            1.1,
            "Inversion step acceptance criterion for fitting Gauss peak shape"
        };
        Setting<double> peak_initial_step_reducer {
            { "wave", "peak_initial_step_reducer" },
            4.0,
            "Starting value of the Levenberg-Marquardt step control parameter"
        };
        Setting<double> peak_reducerfactor_fail {
            { "wave", "peak_reducerfactor_fail" },
            2.5,
            "Multiplication of step reducer for a failed step"
        };
        Setting<double> peak_reducerfactor_success {
            { "wave", "peak_reducerfactor_success" },
            0.5,
            "Multiplication of step reducer for a successful step"
        };
        Setting<double> peak_reducer_limit {
            { "wave", "peak_reducer_limit" },
            100,
            "Maximum step control reducer for convergence"
        };
        Setting<int> refspec_insignificant_island {
            { "wave", "refspec_insignificant_island" },
            {},
            "In reference spectrum: Setting for peak detection: Size of peaks\n"
            "or non-peaks to be discarded"
        };
        Setting<int> refspec_maxreso {
            { "wave", "refspec_maxreso" },
            128,
            "In reference spectrum: Highest denominator in setting threshold\n"
            "to detect peaks"
        };
        Setting<int> refspec_backgroundorder {
            { "wave", "refspec_backgroundorder" },
            1,
            "In reference spectrum: Degrees of freedom of background\n"
            "polynomial for peak detection"
        };
        Setting<bool> refspec_gaussfit {
            { "wave", "refspec_gaussfit" },
            true,
            "In reference spectrum: Flag for performing gauss fit"
        };
        Setting<double> refspec_domain_range {
            { "wave", "refspec_domain_range" },
            5.0,
            "In reference spectrum: Range in a-priori FWHM that a Gauss\n"
            "contains in split splitgaussfit"
        };
        Setting<int> refspec_maxiter {
            { "wave", "refspec_maxiter" },
            0,
            "In reference spectrum: Number of iterations for fitting Gauss\n"
            "peak shape"
        };
        Setting<double> refspec_magtol {
            { "wave", "refspec_magtol" },
            NC_FILL_DOUBLE,
            "In reference spectrum: Tolerance of peak magnitude (in units of\n"
            "a-priori magnitude)"
        };
        Setting<double> refspec_postol {
            { "wave", "refspec_postol" },
            NC_FILL_DOUBLE,
            "In reference spectrum: Tolerance of peak positions (in indices)"
        };
        Setting<double> refspec_widthtol {
            { "wave", "refspec_widthtol" },
            NC_FILL_DOUBLE,
            "In reference spectrum: tolerance of peak FWHM (in indices)"
        };
        Setting<double> refspec_backgroundtol {
            { "wave", "refspec_backgroundtol" },
            NC_FILL_DOUBLE,
            "In reference spectrum: Tolerance of background B-spline\n"
            "coefficients (in median signals)"
        };
        Setting<double> refspec_steptolerance {
            { "wave", "refspec_steptolerance" },
            NC_FILL_DOUBLE,
            "In reference spectrum: Inversion step acceptance criterion for\n"
            "fitting Gauss peak shape"
        };
        Setting<double> refspec_initial_step_reducer {
            { "wave", "refspec_initial_step_reducer" },
            NC_FILL_DOUBLE,
            "In reference spectrum: Starting value of the\n"
            "Levenberg-Marquardt step control parameter"
        };
        Setting<double> refspec_reducerfactor_fail {
            { "wave", "refspec_reducerfactor_fail" },
            NC_FILL_DOUBLE,
            "In reference spectrum: Multiplication of step reducer for a\n"
            "failed step"
        };
        Setting<double> refspec_reducerfactor_success {
            { "wave", "refspec_reducerfactor_success" },
            NC_FILL_DOUBLE,
            "In reference spectrum: Multiplication of step reducer for a\n"
            "successful step"
        };
        Setting<double> refspec_reducer_limit {
            { "wave", "refspec_reducer_limit" },
            NC_FILL_DOUBLE,
            "In reference spectrum: Maximum step control reducer for\n"
            "convergence"
        };
        Setting<int> order {
            { "wave", "order" },
            0,
            "Degrees of freedom in fit of wavelength against pixel index"
        };
        Setting<int> npeak { { "wave", "npeak" },
                             0,
                             "Number of peak wavelengths" };
        Setting<double> peak_offset {
            { "wave", "peak_offset" },
            {},
            "Offset added to all fitted peaks found in the reference spectrum",
        };
        Setting<std::vector<double>> reference_wavelengths {
            { "wave", "reference_wavelengths" },
            {},
            "Values to replace the peaks found in the references spectrum.\n"
            "Run the wavelength generation in debug mode to be sure which\n"
            "peaks are used."
        };
    } wave;

    // Polarimetric calibration
    struct
    {
        Setting<std::string> delta_prior_file {
            { "pol", "delta_prior_file" },
            {},
            "File containing prior data for retardance"
        };
        Setting<std::string> mt_prior_file {
            { "pol", "mt_prior_file" },
            {},
            "File containing prior data for telescope polarization parameter mt"
        };
        Setting<std::optional<EGSEMember>> aolp_from {
            { "pol", "aolp_from" },
            "Name of the EGSE member from which to read the AoLP values.\n"
            "Allowed values: POLARIZER, ACT_ANGLE"
        };
        Setting<std::optional<double>> aolp_offset {
            { "pol", "aolp_offset" },
            "Value in degrees added to each "
            "polarizer angle read from the egse\n"
            "variable"
        };
        Setting<std::optional<bool>> aolp_sign_flip {
            { "pol", "aolp_sign_flip" },
            "Whether the polarizer angles read from the egse variable are\n"
            "interpreted as having the opposite sign"
        };
        Setting<double> fit_window_min {
            { "pol", "fit_window_min" },
            {},
            "Shortest wavelength in nm included in the fitting procedure.\n"
            "This parameter could be useful in excluding bad data points near\n"
            "the edge of the detector that could have an adverse effect on\n"
            "the fitting procedure.",
            SettingLevel::debug
        };
        Setting<double> fit_window_max {
            { "pol", "fit_window_max" },
            NC_FILL_DOUBLE,
            "Longest wavelength in nm included in the fitting procedure. This\n"
            "parameter could be useful in excluding bad data points near the\n"
            "edge of the detector that could have an adverse effect on the\n"
            "fitting procedure.",
            SettingLevel::debug
        };
        struct
        {
            Setting<int> delta_order {
                { "pol", "fitting", "delta_order" },
                4,
                "Polynomial order of retardance used in the model",
            };
            Setting<int> eps_order {
                { "pol", "fitting", "eps_order" },
                4,
                "Polynomial order of efficiency used in the model",
            };
            // Internally, the default value is -1
            Setting<std::optional<int>> mt_order {
                { "pol", "fitting", "mt_order" },
                "Polynomial order of telescope polarization Mueller element\n"
                "used in the model. Default is to not fit for mt.",
            };
        } fitting;
        struct
        {
            Setting<bool> enabled {
                { "pol", "print_debug", "enabled" },
                false,
                "Whether to print the modulation data and the corresponding\n"
                "fits for each value of AoLP of a given viewport and field of\n"
                "view (FOV). Also printed are values of the fitting\n"
                "parameters.",
                SettingLevel::debug
            };
            Setting<int> viewport {
                { "pol", "print_debug", "viewport" },
                {},
                "Viewport for which to print the fitting results. Counting\n"
                "starts from 0. Ignored if debug_fov < 0.",
                SettingLevel::debug
            };
            Setting<int> fov {
                { "pol", "print_debug", "fov" },
                {},
                "Field of view for which to print the fitting results.\n"
                "Counting starts from 0 and is relative to the edge of the\n"
                "given viewport. Ignored if debug_vp < 0.",
                SettingLevel::debug
            };
            Setting<std::string> output_folder {
                { "pol", "print_debug", "output_folder" },
                {},
                "Full or relative path to the folder that will contain the\n"
                "fitting results. Only used with debug_vp and debug_fov.\n"
                "Default is to use the same directory from which the\n"
                "executable is run.",
                SettingLevel::debug
            };
        } print_debug;
    } pol;

    struct
    {
        Setting<std::string> outputfile { { "l1b", "outputfile" },
                                          {},
                                          "Output L1B file" };
        Setting<int> order {
            { "l1b", "order" },
            {},
            "B-spline order of wavelength dependence of q and u"
        };
        Setting<std::vector<double>> knots {
            { "l1b", "knots" },
            {},
            "Wavelengths of knot positions in B-spline"
        };
        Setting<std::optional<double>> wave_min {
            { "l1b", "wave_min" },
            "Wavelength (nm) to start the demodulation window"
        };
        Setting<std::optional<double>> wave_max {
            { "l1b", "wave_max" },
            "Wavelength (nm) to end the demodulation window"
        };
        Setting<std::vector<double>> l1b_wavelength {
            { "l1b", "l1b_wavelength" },
            {},
            "Central wavelengths of polarimetric product"
        };
        Setting<std::vector<double>> intensity_wavelength {
            { "l1b", "intensity_wavelength" },
            {},
            "Wavelengths for intensity-only grid"
        };
        Setting<double> resolving_power {
            { "l1b", "resolving_power" },
            NC_FILL_DOUBLE,
            "Full-width half-maximum of Gauss convolution of polarimetric L1B\n"
            "product"
        };
        Setting<double> gauss_range_tol {
            { "l1b", "gauss_range_tol" },
            1e-5,
            "Values higher than this are considered non-zero in the Gaussian\n"
            "convolution kernel. Zero values are not present in the packed\n"
            "matrix format which speeds up the calculation. This affects the\n"
            "accuracy of the convolved I, Q, U and related quantities."
        };
        Setting<bool> telescope_errors {
            { "l1b", "telescope_errors" },
            true,
            "Whether to include telescope correction (mt) in error\n"
            "propagation. Disabling this saves time but introduces a small\n"
            "error in the noise estimate. The values of I, q, u and derived\n"
            "quantities are unaffected."
        };
        Setting<bool> geolocation {
            { "l1b", "geolocation" },
            true,
            "Wether to perform geolocation (disable when using OCAL\n"
            "measurements as test)",
        };
        Setting<std::string> utcfile { { "l1b", "utcfile" },
                                       {},
                                       "UTC Time-difference file" };
        Setting<std::string> demfile {
            { "l1b", "demfile" },
            {},
            "Detailed elevation map. Leave empty for placeholder DEM function."
        };
        Setting<double> semi_major_axis { { "l1b", "semi_major_axis" },
                                          NC_FILL_DOUBLE,
                                          "Semi-major axis of the earth" };
        Setting<double> semi_minor_axis { { "l1b", "semi_minor_axis" },
                                          NC_FILL_DOUBLE,
                                          "Semi-minor axis of the earth" };
        Setting<double> latitude_tol {
            { "l1b", "latitude_tol" },
            NC_FILL_DOUBLE,
            "Tolerance for iterative calculation of latitude during\n"
            "geolocation (radians)"
        };
        Setting<double> mountain { { "l1b", "mountain" },
                                   NC_FILL_DOUBLE,
                                   "Highest mountain to expect" };
        Setting<double> movedistance {
            { "l1b", "movedistance" },
            NC_FILL_DOUBLE,
            "Distance to move in one iteration avoiding to skip mountains"
        };
        Setting<double> extremeweight {
            { "l1b", "extremeweight" },
            NC_FILL_DOUBLE,
            "Clip on weight factor for guess during geolocation convergence"
        };
        Setting<double> geolocation_tol {
            { "l1b", "geolocation_tol" },
            NC_FILL_DOUBLE,
            "Tolerance for convergence along line of sight (length units)"
        };
        Setting<int> l1b_output_level {
            { "l1b", "l1b_output_level" },
            {},
            "Level of detailed output. Own implementation, because output\n"
            "file is not the CKD file."
        };
        Setting<double> first_proc_rel_workload {
            { "l1b", "first_proc_rel_workload" },
            0.9,
            "Relative workload on the first MPI process compared to the other\n"
            "processes. Example: if this value is 0.9 (i.e. 90%) then with 10\n"
            "MPI processes and 1000 images process 0 is assigned the first 90\n"
            "images. The reason process 0 should be dealt less images is that\n"
            "it also needs to perform I/O operations."
        };
        Setting<int> mpi_send_size {
            { "l1b", "mpi_send_size" },
            10,
            "Number of images each MPI task processes before sending them to\n"
            "process 0. Default is 10, i.e. an asynchronous MPI send is\n"
            "initiated after every 10th image. Increasing this will reduce\n"
            "the amount of communication but increase memory usage on all MPI\n"
            "processes."
        };
        Setting<std::optional<ProcLevel>> last_calibration_step {
            { "l1b", "last_calibration_step" },
            "If given, the last calibration step to be executed. Allowed\n"
            "values: dark, noise, nonlin, prnu, stray, fov, swath, wave, rad.",
            SettingLevel::debug
        };
        Setting<int> image_start {
            { "l1b", "image_start" },
            {},
            "First image to be processed (counting starts at 0)",
            SettingLevel::debug
        };
        Setting<std::optional<int>> image_end {
            { "l1b", "image_end" },
            "Last image to be processed (up to the number of images minus 1)",
            SettingLevel::debug
        };
        Setting<bool> disable_io {
            { "l1b", "disable_io" },
            false,
            "If true, MPI process 0 will not write data to an output file.\n"
            "This allows to test the IO overhead.",
            SettingLevel::debug
        };
        Setting<bool> disable_mpi_comm {
            { "l1b", "disable_mpi_comm" },
            false,
            "If true, MPI processes calibrate images but do not send the\n"
            "results to MPI process 0. This allows to test the overhead of\n"
            "MPI communication.",
            SettingLevel::debug
        };
        Setting<bool> noise_propagation {
            { "l1b", "noise_propagation" },
            true,
            "Whether to propagate noise through demodulation or turn it off\n"
            "to save time. Note that the noise is still included in the L1B\n"
            "product but with incorrect values.",
            SettingLevel::debug
        };
    } l1b;

    struct
    {
        Setting<std::string> demfile {
            { "l1c", "demfile" },
            {},
            "Detailed elevation map. Leave empty for placeholder DEM function."
        };
        Setting<double> semi_major_axis { { "l1c", "semi_major_axis" },
                                          NC_FILL_DOUBLE,
                                          "Earth property (move to database)" };
        Setting<double> semi_minor_axis { { "l1c", "semi_minor_axis" },
                                          NC_FILL_DOUBLE,
                                          "Earth property (move to database)" };
        Setting<double> latitude_tol {
            { "l1c", "latitude_tol" },
            NC_FILL_DOUBLE,
            "Tolerance for iterative calculation of latitude during\n"
            "geolocation (radians)"
        };
        Setting<std::string> l1bfile {
            { "l1c", "l1bfile" },
            {},
            "Input L1B file (no L1A files are used here)"
        };
        Setting<std::string> l1cfile { { "l1c", "l1cfile" },
                                       {},
                                       "Output L1C file" };
        Setting<std::vector<double>> lat { { "l1c", "lat" },
                                           {},
                                           "Latitudes of target positions" };
        Setting<std::vector<double>> lon { { "l1c", "lon" },
                                           {},
                                           "Longitudes of target positions" };
    } l1c;

    // Configuration parameters used by most processes. For simplicity
    // and consistency, all parameter lists have the same length which
    // is equal to the number of process levels (see ProcLevel). Some
    // parameters are thus unused (or undefined for now). For
    // instance, even though l1a_files spans over all processes, no
    // L1A files are currently used as input for the L1C process.
    std::array<Setting<std::vector<std::string>>,
               static_cast<int>(ProcLevel::n_levels)>
      l1a_files { {
        {}, // l1a
        {}, // raw
        { { "dark", "l1a_files" },
          {},
          "L1A files for generating the dark CKD" },
        { { "noise", "l1a_files" },
          {},
          "L1A files for generating the noise CKD" },
        { { "nonlin", "l1a_files" },
          {},
          "L1A files for generating the nonlinearity CKD" },
        { { "prnu", "l1a_files" },
          {},
          "L1A files for generating the PRNU CKD" },
        { { "stray", "l1a_files" },
          {},
          "L1A files for generating the stray light CKD" },
        { { "fov", "l1a_files" },
          {},
          "L1A files for generating the field of view CKD" },
        { { "swath", "l1a_files" },
          {},
          "L1A files for generating the swath vector CKD" },
        { { "wave", "l1a_files" },
          {},
          "L1A files for generating the spectral CKD" },
        { { "rad", "l1a_files" },
          {},
          "L1A files for generating the radiometric CKD" },
        { { "pol", "l1a_files" },
          {},
          "L1A files for generating the polarimetric CKD" },
        { { "l1b", "l1a_files" }, {}, "L1A file for the L1B process" },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1a_file_list { {
        {}, // l1a
        {}, // raw
        { { "dark", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "dark CKD" },
        { { "noise", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "noise CKD" },
        { { "nonlin", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "nonlinearity CKD" },
        { { "prnu", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "PRNU CKD" },
        { { "stray", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "stray light CKD" },
        { { "fov", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "field of view CKD" },
        { { "swath", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "swath vector CKD" },
        { { "wave", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "spectral CKD" },
        { { "rad", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "radiometric CKD" },
        { { "pol", "l1a_file_list" },
          {},
          "File containing a list of L1A files for generating the "
          "polarimetric CKD" },
        { { "l1b", "l1a_file_list" },
          {},
          "File containing the L1A file for the L1B process" },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_raw { {
        {}, // l1a
        {}, // raw
        { { "dark", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "noise", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "nonlin", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "prnu", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "stray", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "fov", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "swath", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "wave", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "rad", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "pol", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "l1b", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_dark { {
        {}, // l1a
        {}, // raw
        {}, // dark
        { { "noise", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "nonlin", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "prnu", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "stray", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "fov", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "swath", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "wave", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "rad", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "pol", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "l1b", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_nonlin { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        { { "prnu", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "stray", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "fov", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "swath", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "wave", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "rad", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "pol", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "l1b", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_prnu { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        { { "stray", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "fov", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "swath", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "wave", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "rad", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "pol", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "l1b", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_stray { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        {}, // stray
        { { "fov", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "swath", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "wave", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "rad", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "pol", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "l1b", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_fov { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        {}, // stray
        {}, // fov
        { { "swath", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        { { "wave", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        { { "rad", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        { { "pol", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        { { "l1b", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_rad { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        {}, // stray
        {}, // fov
        {}, // swath
        {}, // wave
        {}, // rad
        { { "pol", "l1x_rad" }, {}, l1x_rad_text, SettingLevel::debug },
        { { "l1b", "l1x_rad" }, {}, l1x_rad_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      dark_cols_left { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        { { "stray", "dark_cols_left" },
          {},
          dark_cols_left_text + "stray light images only." },
        { { "fov", "dark_cols_left" },
          {},
          dark_cols_left_text + "FOV images only." },
        { { "swath", "dark_cols_left" },
          {},
          dark_cols_left_text + "swath images only." },
        { { "wave", "dark_cols_left" },
          {},
          dark_cols_left_text + "wavelength images only." },
        { { "rad", "dark_cols_left" },
          {},
          dark_cols_left_text + "radiometric images only." },
        { { "pol", "dark_cols_left" },
          {},
          dark_cols_left_text + "polarimetric images only." },
        { { "l1b", "dark_cols_left" },
          {},
          dark_cols_left_text + "the l1b process only." },
        {} // l1c
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      dark_cols_right { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        { { "stray", "dark_cols_right" },
          {},
          dark_cols_right_text + "stray light images only." },
        { { "fov", "dark_cols_right" },
          {},
          dark_cols_right_text + "FOV images only." },
        { { "swath", "dark_cols_right" },
          {},
          dark_cols_right_text + "swath images only." },
        { { "wave", "dark_cols_right" },
          {},
          dark_cols_right_text + "wavelength images only." },
        { { "rad", "dark_cols_right" },
          {},
          dark_cols_right_text + "radiometric images only." },
        { { "pol", "dark_cols_right" },
          {},
          dark_cols_right_text + "polarimetric images only." },
        { { "l1b", "dark_cols_right" },
          {},
          dark_cols_right_text + "the l1b process only." },
        {} // l1c
      } };
    std::array<Setting<std::vector<std::string>>,
               static_cast<int>(ProcLevel::n_levels)>
      l1a_bgr_files { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        { { "nonlin", "l1a_bgr_files" },
          {},
          "Nonlinearity background measurements" },
        { { "prnu", "l1a_bgr_files" }, {}, "PRNU background measurements" },
        { { "stray", "l1a_bgr_files" },
          {},
          "Stray light background measurements" },
        { { "fov", "l1a_bgr_files" },
          {},
          "Field of view background measurements" },
        { { "swath", "l1a_bgr_files" },
          {},
          "Swath calibration background measurements" },
        { { "wave", "l1a_bgr_files" },
          {},
          "Spectral calibration background measurements" },
        { { "rad", "l1a_bgr_files" },
          {},
          "Radiometric background measurements" },
        { { "pol", "l1a_bgr_files" },
          {},
          "Polarimetric background measurements" },
        { { "l1b", "l1a_bgr_files" },
          {},
          "Background measurements for the L1B process" },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1a_bgr_file_list { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        { { "nonlin", "l1a_bgr_file_list" },
          {},
          "File containing a list of nonlinearity "
          "background measurements" },
        { { "prnu", "l1a_bgr_file_list" },
          {},
          "File containing a list of PRNU "
          "background measurements" },
        { { "stray", "l1a_bgr_file_list" },
          {},
          "File containing a list of stray light "
          "background measurements" },
        { { "fov", "l1a_bgr_file_list" },
          {},
          "File containing a list of field of view "
          "background measurements" },
        { { "swath", "l1a_bgr_file_list" },
          {},
          "File containing a list of swath calibration "
          "background measurements" },
        { { "wave", "l1a_bgr_file_list" },
          {},
          "File containing a list of spectral "
          "calibration background measurements" },
        { { "rad", "l1a_bgr_file_list" },
          {},
          "File containing a list of radiometric "
          "background measurements" },
        { { "pol", "l1a_bgr_file_list" },
          {},
          "File containing a list of polarimetric "
          "background measurements" },
        { { "l1b", "l1a_bgr_file_list" },
          {},
          "File containing a list of L1B  background "
          "measurements" },
        {} // l1c
      } };
    std::array<Setting<L1AType>, static_cast<int>(ProcLevel::n_levels)>
      l1a_type { {
        {}, // l1a
        {}, // raw
        { { "dark", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "noise", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "nonlin", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "prnu", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "stray", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "fov", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "swath", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "wave", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "rad", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "pol", "l1a_type" },
          L1AType::calibration,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        { { "l1b", "l1a_type" },
          L1AType::flight,
          "L1A file type. Allowed values: calibration, flight",
          SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      discard_frames_front { {
        {}, // l1a
        {}, // raw
        {}, // dark
        { { "noise", "discard_frames_front" }, {}, discard_frames_front_text },
        { { "nonlin", "discard_frames_front" }, {}, discard_frames_front_text },
        {}, // prnu
        {}, // stray
        {}, // fov
        {}, // swath
        {}, // wave
        {}, // pol
        {}, // l1b
        {}, // l1c
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      discard_frames_back { {
        {}, // l1a
        {}, // raw
        {}, // dark
        { { "noise", "discard_frames_back" }, {}, discard_frames_back_text },
        { { "nonlin", "discard_frames_back" }, {}, discard_frames_back_text },
        {}, // prnu
        {}, // stray
        {}, // fov
        {}, // swath
        {}, // wave
        {}, // pol
        {}, // l1b
        {}, // l1c
      } };
    // Make sure to call value() when evaluating these! Otherwise
    // you're evaluating std::optional itself which always returns
    // true because it is set in SPEXSettings::checkParameters.
    std::array<Setting<std::optional<bool>>,
               static_cast<int>(ProcLevel::n_levels)>
      debug { {
        {}, // l1a
        {}, // raw
        { { "dark", "debug" },
          "Run dark CKD generation in debug mode",
          SettingLevel::debug },
        { { "noise", "debug" },
          "Run noise CKD generation in debug mode",
          SettingLevel::debug },
        { { "nonlin", "debug" },
          "Run nonlinearity CKD generation in debug mode",
          SettingLevel::debug },
        { { "prnu", "debug" },
          "Run PRNU CKD generation in debug mode",
          SettingLevel::debug },
        { { "stray", "debug" },
          "Run stray light CKD generation in debug mode",
          SettingLevel::debug },
        { { "fov", "debug" },
          "Run field of view CKD generation in debug mode",
          SettingLevel::debug },
        { { "swath", "debug" },
          "Run swath vector CKD generation in debug mode",
          SettingLevel::debug },
        { { "wave", "debug" },
          "Run spectral CKD generation in debug mode",
          SettingLevel::debug },
        { { "rad", "debug" },
          "Run radiometric CKD generation in debug mode",
          SettingLevel::debug },
        { { "pol", "debug" },
          "Run polarimetric CKD generation in debug mode",
          SettingLevel::debug },
        { { "l1b", "debug" },
          "Run L1B process in debug mode",
          SettingLevel::debug },
        { { "l1c", "debug" },
          "Run L1C process in debug mode",
          SettingLevel::debug },
      } };

    // Variables that are inferred from the input configuration files
    // but are not directly editable by user
    std::array<bool, static_cast<int>(ProcLevel::n_levels)> skip {};

    SPEXSettings() = default;
    SPEXSettings(const std::string& yaml_file) : Settings { yaml_file } {}
    auto scanKeys() -> void override;
    // Return a clone of the configuration with nodes removed that are
    // not present in [main][processes]. For instance, if swath is not
    // in [main][processes] then the swath section is removed even if
    // specified in the input configuration file.
    auto getReducedConfig() const -> std::string;
    ~SPEXSettings() = default;
};

} // namespace spex
