// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "swathcal.h"

#include "array.h"
#include "bspline.h"
#include "ckd.h"
#include "functions.h"
#include "l1a.h"
#include "l1a_metadata.h"
#include "lininv.h"
#include "matrix.h"
#include "splitgaussfit.h"

#include <cassert>
#include <spdlog/spdlog.h>

namespace spex {

// Swath calculation
// This is a placeholder. It should calculate the swath vectors from calibration
// measurements, but now, it just makes up swath vectors.
auto Swathcal::processInit() -> void
{

    // Check legal settings
    // Check if any tolerance domain overlaps between different viewports
    for (size_t ivp1=0 ; ivp1<ckd.dim_vp-1 ; ivp1++) {
        for (size_t ivp2=ivp1+1 ; ivp2<ckd.dim_vp ; ivp2++) {
            double diff = abs(settings.swath.viewport_alt_angles[ivp2]-settings.swath.viewport_alt_angles[ivp1]);
        }
    }
    nattempt = settings.swath.min_fit.size();
    size_t dimcheck = settings.swath.order.size();
    // Rotation matrix
    std::vector<double> rtr(6, 0.0);
    auto rotmat { settings.swath.rotmat };
    Matrix::matmul_rr_fold_quick_sym(3,3,rotmat.data(),rotmat.data(),rtr.data());
    if (abs(rtr[0]-1.0) > settings.swath.rot_warningthreshold) {
        spdlog::warn("Rotation matrix check times its transpose has {:.12f} "
                     "instead of one at location (X,X)",
                     rtr[0]);
    }
    if (abs(rtr[1]) > settings.swath.rot_warningthreshold) {
        spdlog::warn("Warning: Rotation matrix check times its transpose has "
                     "{:.12f} instead of zero at location (X,Y) and (Y,X)",
                     rtr[1]);
    }
    if (abs(rtr[2]-1.0) > settings.swath.rot_warningthreshold) {
        spdlog::warn("Warning: Rotation matrix check times its transpose has "
                     "{:.12f} instead of one at location (Y,Y)",
                     rtr[2]);
    }
    if (abs(rtr[3]) > settings.swath.rot_warningthreshold) {
        spdlog::warn("Warning: Rotation matrix check times its transpose has "
                     "{:.12f} instead of zero at location (X,Z) and (Z,X)",
                     rtr[3]);
    }
    if (abs(rtr[4]) > settings.swath.rot_warningthreshold) {
        spdlog::warn("Warning: Rotation matrix check times its transpose has "
                     "{:.12f} instead of zero at location (Y,Z) and (Z,Y)",
                     rtr[4]);
    }
    if (abs(rtr[5]-1.0) > settings.swath.rot_warningthreshold) {
        spdlog::warn("Warning: Rotation matrix check times its transpose has "
                     "{:.12f} instead of one at location (Z,Z)",
                     rtr[5]);
    }
    double det =
        settings.swath.rotmat[0]*settings.swath.rotmat[4]*settings.swath.rotmat[8] +
        settings.swath.rotmat[1]*settings.swath.rotmat[5]*settings.swath.rotmat[6] +
        settings.swath.rotmat[2]*settings.swath.rotmat[3]*settings.swath.rotmat[7] -
        settings.swath.rotmat[0]*settings.swath.rotmat[5]*settings.swath.rotmat[7] -
        settings.swath.rotmat[1]*settings.swath.rotmat[3]*settings.swath.rotmat[8] -
        settings.swath.rotmat[2]*settings.swath.rotmat[4]*settings.swath.rotmat[6]
    ;
    // Transform to quaternion so that rotation can be applied in one line
    double realpart = 0.5*sqrt(settings.swath.rotmat[0]+settings.swath.rotmat[4]+settings.swath.rotmat[8]+1.0);
    q_inst = Quaternion(
        (settings.swath.rotmat[7]-settings.swath.rotmat[5]) / (4.0*realpart),
        (settings.swath.rotmat[2]-settings.swath.rotmat[6]) / (4.0*realpart),
        (settings.swath.rotmat[3]-settings.swath.rotmat[1]) / (4.0*realpart),
        realpart
    ).conjug();
    // Input is conjugated. The input is the spacecraft to SPEX matrix.
    // We use a quaternion to rotate SPEX to the spacecraft.
    q_inst.normalize(); // Getting rid of eventual rounding errors possibly within the warning threshold

    // Log the instrument rotation
    if (q_inst[3] == 1.0) {
        spdlog::debug("No instrument rotation");
    } else {
        double ang = 2.0*acos(q_inst[3]);
        Vector axis(q_inst[0],q_inst[1],q_inst[2]);
        axis.normalize();
        spdlog::debug("Instrument rotation {:.7f} degrees along axis "
                      "({:.7f},{:.7f},{:.7f})",
                      ang / math::deg_to_rad,
                      axis[0],
                      axis[1],
                      axis[2]);
    }

    // Make batches per viewport per ACT angle

    // The L1A format allows illumination of multiple viewport, but these
    // should all have a star stimulus, so that should not be possible.
    // Therefore, we assume and assert that all files only have one
    // illuminated viepowrt (and also not zero).

    // Minimum number of ACT angles per viewport to make a fit
    min_fit_poorest = settings.swath.min_fit[nattempt-1];

    // Allocate the batches array with more than enough batches
    batches.resize(nL1A());
    int nbatch {};
    // Running batch pointer
    L1ABatch* bat_cur = batches.data();

    // Administration
    ibatch_start.resize(ckd.dim_vp);
    nbatch_vp.resize(ckd.dim_vp);
    nfail_vp.resize(ckd.dim_vp,0);

    // Sort L1A products by viewports and ACT angles
    std::vector<std::reference_wrapper<L1A>> l1a_refs {
        l1a_manager->l1a_products.begin(), l1a_manager->l1a_products.end()
    };
    std::ranges::sort(
      l1a_refs,
      [](const auto& l1a_1, const auto& l1a_2) {
          if (l1a_1.get().meta->viewport < l1a_2.get().meta->viewport) {
              return true;
          }
          if (l1a_1.get().meta->viewport > l1a_2.get().meta->viewport) {
              return false;
          }
          return l1a_1.get().meta->act_angle < l1a_2.get().meta->act_angle;
      });
    std::vector<double> act_angles_unsorted(nL1A(), 0.0);
    std::vector<double> act_angles_sorted(nL1A(), 0.0);
    for (size_t il1a=0 ; il1a<nL1A() ; il1a++) {
        act_angles_unsorted[il1a] = l1a_refs[il1a].get().meta->act_angle;
        act_angles_sorted[il1a] = l1a_refs[il1a].get().meta->act_angle;
    }
    uint8_t l1a_viewport_cur = 1;
    for (size_t ivp=0 ; ivp<ckd.dim_vp ; ivp++) {
        ibatch_start[ivp] = nbatch;
        double* act_prv = NULL; // Previous ACT angle
        // Open first empty batch in the viewport
        size_t nl1a_batch = 0;
        int i_prev {};
        for (size_t il1a=0 ; il1a<nL1A() ; il1a++) {
            L1A& l1a = l1a_refs[il1a];
            if (l1a.meta->viewport != l1a_viewport_cur) {
                i_prev = il1a + 1;
                continue; // Only consider own viewport
            }
            double* act_cur = &act_angles_sorted[il1a];
            if (il1a + 1 == nL1A()
                || std::abs(l1a_refs[i_prev].get().meta->act_angle
                            - l1a_refs[il1a + 1].get().meta->act_angle) > 1e-10
                || l1a_refs[i_prev].get().meta->viewport
                != l1a_refs[il1a + 1].get().meta->viewport){
                bat_cur->l1a_products = { l1a_refs.begin() + i_prev,
                                          l1a_refs.begin() + il1a + 1 };
                i_prev = il1a + 1;
            }
            if (act_prv == NULL || (*act_cur - *act_prv <= settings.swath.act_angle_tol)) {
                // Write L1A instance in current batch
                nl1a_batch++;
            } else {
                // Close off previous batch
                bat_cur++;
                nbatch++;
                nl1a_batch = 0;
                // Write file in new batch
                nl1a_batch++;
            }
            act_prv = act_cur;
        }
        // Close off last batch in the viewport
        assert(nl1a_batch != 0 && "There seems to be no L1A file for viewport");
        bat_cur++;
        nbatch++;
        // Administration
        nbatch_vp[ivp] = nbatch-ibatch_start[ivp];
        assert(!(nbatch_vp[ivp] < min_fit_poorest));
        l1a_viewport_cur *= 2;
    }

    // Encapsulate batch array
    batches.resize(nbatch);
    for (auto& batch : batches) {
        batch.genWorkRanges(mpi);
    }

    // Check that all L1A files are from one viewport only
    // If there are skipped viewports, not all L1A files are in a batch. That is okay
    size_t cnt = 0;
    l1a_viewport_cur = 1;
    for (size_t ivp=0 ; ivp<ckd.dim_vp ; ivp++) {
        for (size_t il1a=0 ; il1a<nL1A() ; il1a++) {
            L1A& l1a = l1a_manager->l1a_products[il1a];
            if (l1a.meta->viewport == l1a_viewport_cur) cnt++;
        }
        l1a_viewport_cur *= 2;
    }
    assert(cnt == nL1A()
           && "Not all L1A files are from exactly one viewport");

    // Shape arrays
    act_fovcal.resize(nbatch); // Used to express polynomials in and sample at L1B FOVs (based on detector positions)
    act_swathcal.resize(nbatch); // Stage ACT angles during these measurements
    alt_swathcal.resize(nbatch); // Peak ALT angles
    // Both act_swathcal and alt_swathcal are expressed as polynomials of act_fovcal.
    // Here, act_fovcal represents the detector positioning.
    // They are sampled on the act_fovcals from the L1B (are in CKD).
    // And then, the swathcal stage angles are fed to the theodolite
    // system and turned into swath vectors.
    mask.resize(nbatch); // Mask, for if an ACT angle seems not to
    mag.resize(nbatch); // Magnitude of fit, for weighing or filtering
    // During finalize, the fit will be performed

    theo = std::make_unique<Theodolite>();
    auto viewport_alt_angles { settings.swath.viewport_alt_angles };
    theo->calculate_reference(
        ckd.dim_vp,
        settings.swath.theodolite_reference_filename,
        viewport_alt_angles.data(),
        settings.swath.theodolite_alt_angle_tol,
        settings.swath.theo_altmodel_maxiter,
        settings.swath.theo_altmodel_tol,
        settings.swath.theo_altmodel_steptolerance,
        settings.swath.theo_altmodel_initial_step_reducer,
        settings.swath.theo_altmodel_reducerfactor_fail,
        settings.swath.theo_altmodel_reducerfactor_success,
        settings.swath.theo_altmodel_reducer_limit
    );

    theo->calculate_spotcheck(
        settings.swath.theodolite_spotcheck_filename
    );
}

auto Swathcal::processBatch(const int i_vp) -> void
{
    // Figure out which viewport it is
    size_t ivp = ckd.dim_vp; // Will be overwritten
    for (size_t attempt=0 ; attempt<ckd.dim_vp ; attempt++) {
        if (i_vp >= ibatch_start[attempt]) ivp = attempt;
        // Do not break. The last viewport in which this condition is true
        // is the right one.
    }
    // If anything in the process fails, we will throw away the entire
    // batch. That means that eventual results in the middle will also
    // be discarded.
    bool fail = true;
    // Breakable do-once loop
    for (bool ctd=true ; ctd ; ctd=false) {

        // Write swathcal ACT angles. They are all the same inside the batch
        // besides eventual rounding differences covered by a tolerance.
        act_swathcal[i_vp] =
          0.5 * (batches.at(i_vp).l1a_products.front().get().meta->act_angle
                 + batches.at(i_vp).l1a_products.back().get().meta->act_angle);
        // In principle, they are all equal
        spdlog::info(
          "Viewport {}, ACT angle {} deg",
          viewport_idx_to_angle.at(
            batches.at(i_vp).l1a_products.front().get().meta->viewport),
          act_swathcal[i_vp] / math::deg_to_rad);

        // Figure out where the spectrum is on the detector. By doing so,
        // the corresponding FOVcal ACT angle is derived, which because
        // of renewal of the laboratory setup may differ from the ACT angle
        // at the measruements themselves (swathcal). To ensure that there
        // is a spectrum, we add up all the images as they should have their
        // either-or-not visible spectrum on the same detector pixels.
        std::vector<double> image_agg(ckd.npix,0.0);
        std::vector<bool> mask_agg(ckd.npix,false);

        // Add L1As to the image and mask, but ignore the noise
        for (size_t il1a=0 ; il1a<batches[i_vp].l1a_products.size() ; il1a++) {
            // This only works for non-binned L1A. It can be extended so
            // that it works with L1As with all the same binning table.
            for (size_t ipix=0 ; ipix<ckd.npix ; ipix++) {
                if (batches.at(i_vp).l1a_products[il1a].get().pixel_mask[ipix]) mask_agg[ipix] = true;
                image_agg[ipix] += batches.at(i_vp).l1a_products[il1a].get().image[ipix];
            }
        }

        std::vector<double> linesearch_act(3,NC_FILL_DOUBLE);
        double linesearch_strength = NC_FILL_DOUBLE;
        double& act_swath_start = ckd.fov.act_angles[ckd.fov.start_vp[ivp]];
        double& act_swath_end = ckd.fov.act_angles[ckd.fov.start_vp[ivp]+ckd.fov.n_act_vp[ivp]-1];
        act_fovcal[i_vp] = act_swathcal[i_vp];
        spdlog::debug("Effective (FOVcal) ACT angle: {:.7f}",
                      act_fovcal[i_vp] / math::deg_to_rad);

        // Now, the contents of the algorithm. Get the along-track angle.

        // Sort the L1A instances by ALT angle
        std::vector<double> alt_angles_unsorted(batches[i_vp].l1a_products.size(), 0.0);
        for (size_t il1a=0 ; il1a<batches[i_vp].l1a_products.size() ; il1a++) {
            alt_angles_unsorted[il1a] =
              batches.at(i_vp).l1a_products[il1a].get().meta->alt_angle;
        }
        std::vector<double> alt_angles_sorted(batches[i_vp].l1a_products.size(), 0.0);
        std::vector<size_t> l1a_indexarray(batches[i_vp].l1a_products.size(), 0.0);
        Array<double>::sort(batches[i_vp].l1a_products.size(),alt_angles_unsorted.data(),alt_angles_sorted.data(),l1a_indexarray.data());

        // Signal converted to a scalar. This is a measure of how strong the
        // signal with that ALT angle is.
        std::vector<double> signalmedians(batches[i_vp].l1a_products.size(), 0.0);

        for (size_t il1a=0 ; il1a<batches[i_vp].l1a_products.size() ; il1a++) {
            act_extract(
              ivp,
              batches.at(i_vp).l1a_products[l1a_indexarray[il1a]].get().image.data(),
              batches.at(i_vp).l1a_products[l1a_indexarray[il1a]].get().pixel_mask,
              act_fovcal[i_vp],
              signalmedians[il1a]);
        }

        // Now we have the median signals as function of the ALT angle.
        // Abscissa: ALT-angle is alt_angles_sorted.
        // Ordinate: Median signals is signalmedians.

        // Fit a Gauss through these
        // Peak fitter
        Splitgaussfit spl(1,settings.swath.backgroundorder,batches[i_vp].l1a_products.size()); // One peak
        spl.setGaussfit(settings.swath.gaussfit); // Flag for performing gauss fit
        spl.setMaxiter(settings.swath.maxiter); // Number of iterations for fitting Gauss peak shape
        spl.setMagtol(settings.swath.magtol); // Tolerance of peak magnitude (in units of a-priori magnitude)
        spl.setPostol(settings.swath.postol); // Tolerance of peak positions (in indices)
        spl.setWidthtol(settings.swath.widthtol); // Tolerance of peak FWHM (in indices)
        spl.setBackgroundtol(settings.swath.backgroundtol); // Tolerance of background B-spline coefficients (in median signals)
        spl.setSteptolerance(settings.swath.steptolerance); // Inversion step acceptance criterion for fitting Gauss peak shape
        spl.setInitialStepReducer(settings.swath.initial_step_reducer); // Starting value of the Levenberg-Marquardt step control parameter
        spl.setReducerfactorFail(settings.swath.reducerfactor_fail); // Multiplication of step reducer for a failed step
        spl.setReducerfactorSuccess(settings.swath.reducerfactor_success); // Multiplication of step reducer for a successful step
        spl.setReducerLimit(settings.swath.reducer_limit); // Maximum step control reducer for convergence

        // Set abscissa and ordinate
        spl.setX(alt_angles_sorted.data());
        spl.setMeas(signalmedians.data());
        // Too complicated to take noise into account
        // Pixel mask should not be there

        std::vector<double> res(3, 0.0);
        if (spl.solve(settings.swath.insignificant_island,settings.swath.domain_range,settings.swath.maxreso,res.data()) != 0) break; // Fail. No proper peak fitted
        alt_swathcal[i_vp] = res[1];
        mask[i_vp] = false;
        mag[i_vp] = res[0];
        spdlog::debug("Associated ALT angle {:.7f}",
                      alt_swathcal[i_vp] / math::deg_to_rad,ivp);
        fail = false;
    }
    if (fail) {
        act_fovcal[i_vp] = NC_FILL_DOUBLE;
        alt_swathcal[i_vp] = NC_FILL_DOUBLE;
        mask[i_vp] = true;
        mag[i_vp] = NC_FILL_DOUBLE;
        spdlog::debug("Fit failed");
        nfail_vp[ivp]++;
        // Check if this viewport lost its chance to have enough successful fits
        assert(!(nbatch_vp[ivp]-nfail_vp[ivp] < min_fit_poorest) && "Too many failed fit in viewport");
    }
}

// Private function for spectrum extraction, an often-repeated action
// during processBatch. This extraction does not work for binned images.
auto Swathcal::act_extract(size_t ivp,
                           double* image,
                           const SharedArray<bool>& pixel_mask,
                           double act,
                           double& strength) -> void
{
    // Interpolate over FOV ACT angles
    // TODO: Maybe we should do this with a spline, also in the FOV
    double* ckd_act_cur = &ckd.fov.act_angles[ckd.fov.start_vp[ivp]];
    size_t ifov_left = NC_FILL_UINT64;
    if (act <= ckd_act_cur[0]) {
        ifov_left = 0;
    } else if (act >= ckd_act_cur[ckd.fov.n_act_vp[ivp]-1]) {
        ifov_left = ckd.fov.n_act_vp[ivp]-2;
    } else {
        for (size_t ifov_right = 1 ; ifov_right < ckd.fov.n_act_vp[ivp] ; ifov_right++) {
            if (ckd_act_cur[ifov_right] >= act) {
                ifov_left = ifov_right - 1;
                break;
            }
        }
    }
    assert(ifov_left != NC_FILL_UINT64 && "Variable ifov_left not properly initialization during search loop");
    // Get the weight factor. This implies linear extrapolation if needed
    double weightleft = (ckd_act_cur[ifov_left+1]-act) / (ckd_act_cur[ifov_left+1]-ckd_act_cur[ifov_left]);

    // Interpolate the FOV CKD on this ACT angle
    // Extract the spectra to get the median
    std::vector<double> spectra(dims::pol*ckd.dim_detector_spec);
    for (size_t ipol=0 ; ipol<dims::pol ; ipol++) {
        for (size_t ispec=0 ; ispec<ckd.dim_detector_spec ; ispec++) {
            double fov_ispat_interpol =
              weightleft
                * ckd.fov.spat_indices[ipol][ckd.fov.start_vp[ivp] + ifov_left]
                                      [ispec]
              + (1.0 - weightleft)
                  * ckd.fov.spat_indices[ipol][ckd.fov.start_vp[ivp] + ifov_left
                                               + 1][ispec];
            int ispat_left = static_cast<int>(fov_ispat_interpol); // Floor
            double weightleft = ispat_left+1-fov_ispat_interpol;
            size_t ipix_left = ispat_left*ckd.dim_detector_spec + ispec;
            size_t ipix_right = (ispat_left+1)*ckd.dim_detector_spec + ispec;
            if (ispat_left < 0 || ispat_left > (int)ckd.dim_detector_spat-2) {
                spectra[ipol*ckd.dim_detector_spec+ispec] = 0.0;
            } else {
                if (pixel_mask[ipix_left]) {
                    if (pixel_mask[ipix_right]) {
                        spectra[ipol*ckd.dim_detector_spec+ispec] = 0.0;
                    } else {
                        spectra[ipol*ckd.dim_detector_spec+ispec] = image[ipix_right];
                    }
                } else {
                    if (pixel_mask[ipix_right]) {
                        spectra[ipol*ckd.dim_detector_spec+ispec] = image[ipix_left];
                    } else {
                        spectra[ipol*ckd.dim_detector_spec+ispec] = weightleft * image[ipix_left] + (1.0-weightleft) * image[ipix_right];
                    }
                }
            }
        }
    }

    // Convert the two spectra to one median
    median(dims::pol*ckd.dim_detector_spec,spectra.data(),strength);
}

auto Swathcal::processFinalize() -> void
{
    spdlog::info("Evaluating ACT and ALT angles on L1B FOVs");

    // These intermediate results are saved as members, purely as
    // optional detailed output.
    l1b_act_angles.resize(ckd.fov.n_act); // Intermediate result
    l1b_alt_angles.resize(ckd.fov.n_act); // Intermediate result
    // Running pointers for abscissa and actual angles sampled
    // on L1B FOVs.
    double* l1b_abscissa_cur = ckd.fov.act_angles.data();
    double* l1b_act_angles_cur = l1b_act_angles.data();
    double* l1b_alt_angles_cur = l1b_alt_angles.data();

    // Construct CKD
    ckd.swath_swathvectors.resize(ckd.fov.n_act*dims::vec,NC_FILL_DOUBLE);
    double* swathvector_cur = ckd.swath_swathvectors.data(); // Running pointer of the end result
    ckd.swath_vectorplane_normals.resize(ckd.dim_vp*dims::vec,NC_FILL_DOUBLE);
    double* vectorplane_normal_cur = ckd.swath_vectorplane_normals.data(); // Running pointer of fitted plane

    for (size_t ivp=0 ; ivp<ckd.dim_vp ; ivp++) {
            // Number of data points, for all three of FOVcal ACT angles (abscissa),
            // swathcal ACT angles (ordinate) and swathcal ALT angles (ordinate).
            size_t& ndat = nbatch_vp[ivp];

            // Set mask and set larger weight for fits with higher magnitude
            std::vector<double> local_s_y(ndat,1.0);
            std::vector<bool> local_mask(ndat); // Because you cannot offset vectors<bool>
            size_t nfit_success = 0; // Meanwhile, count the number of successful fits
            for (size_t idat=0 ; idat<ndat ; idat++) {
                local_mask[idat] = mask[ibatch_start[ivp]+idat];
                if (!local_mask[idat]) {
                    local_s_y[idat] = 1.0/pow(mag[ibatch_start[ivp]+idat],2.0);
                    nfit_success++;
                }
            }
            size_t order = 0; // Must be overwritten
            for (size_t iattempt=0 ; iattempt<nattempt ; iattempt++) {
                if (nfit_success >= settings.swath.min_fit[iattempt]) {
                    order = settings.swath.order[iattempt];
                    break;
                }
            }
            spdlog::debug(
              "Fitting ACT and ALT with {} degrees of freedom for viewport {}",
              order,
              ivp);

            // Construct polynomial fitter
            std::vector<double> knots = {-90.0 * math::deg_to_rad, 90 * math::deg_to_rad}; // Knot values do not matter
            Bspline b(order,2,knots.data());

            // Jacobian, common for ACT and ALT
            std::vector<double> mat(b.nspline*ndat, 0.0);
            b.jaccalc(ndat,&act_fovcal[ibatch_start[ivp]],mat.data());
            // Fit relationship for ACT angles
            std::vector<double> res_act(b.nspline, 0.0);
            linear_invert(b.nspline,ndat,mat.data(),OPT_DIAG,local_s_y.data(),&local_mask,&act_swathcal[ibatch_start[ivp]],res_act.data());
            // Evaluate the fit on the L1B FOVs
            b.jacapply(ckd.fov.n_act_vp[ivp],l1b_abscissa_cur,res_act.data(),l1b_act_angles_cur);

            // Repeat for ALT angles
            std::vector<double> res_alt(b.nspline, 0.0);
            linear_invert(b.nspline,ndat,mat.data(),OPT_DIAG,local_s_y.data(),&local_mask,&alt_swathcal[ibatch_start[ivp]],res_alt.data());
            // Evaluate the fit on the L1B FOVs
            b.jacapply(ckd.fov.n_act_vp[ivp],l1b_abscissa_cur,res_alt.data(),l1b_alt_angles_cur);

            // Turn these angles into swath vectors
            for (size_t ifov=0 ; ifov<ckd.fov.n_act_vp[ivp] ; ifov++) {
                Vector res_irf {}; // Instrument reference frame
                theo->calculate_swath_vector(l1b_act_angles_cur[ifov],l1b_alt_angles_cur[ifov],res_irf);
                Vector res_srf = q_inst.quaternion_rotate(res_irf); // Satellite reference frame
                res_srf.get(&swathvector_cur[ifov*dims::vec]); // Copy result to CKD
            }

            // Fit a plane through all the swath vectors.
            // It is a plane that goes through (0,0,0).
            // In the ideal case, the equation is ax + cz = 0 and
            // y does not matter. In reality, y may matter a bit.
            // Let us use the model x + by + cz = 0, fixing the x
            // contribution to one so that the solution becomes unique.
            // This reasoning depends on the fact that the swath vectors
            // are already in satellite coordinates and that x is the
            // flight direction, z is nadir and y is to the right.
            // With just the instrument reference frame, this had not
            // been safe.

            // There are two state parameters and nfov_vp measurements
            std::vector<double> jac(2*ckd.fov.n_act_vp[ivp], 0.0);
            std::vector<double> meas(ckd.fov.n_act_vp[ivp], 0.0);
            std::vector<double> res_bc(2, 0.0);
            // The linear model will be: by + cz = -x with state vector
            // [b,c] and known parameters x,y,z.
            for (size_t ifov=0 ; ifov<ckd.fov.n_act_vp[ivp] ; ifov++) {
                double* xyz = &swathvector_cur[ifov*dims::vec]; // Swath vector as double pointer
                jac[ifov] = xyz[1]; // d/db = y
                jac[ckd.fov.n_act_vp[ivp]+ifov] = xyz[2]; // d/dc = z
                meas[ifov] = -xyz[0]; // Measurement is -x
            }
            linear_invert(2,ckd.fov.n_act_vp[ivp],jac.data(),OPT_NONE,NULL,NULL,meas.data(),res_bc.data());

            // Normalize, using the Vector class
            Vector normal_cur(1.0,res_bc[0],res_bc[1]);
            normal_cur.normalize();

            // Write down results
            normal_cur.get(vectorplane_normal_cur);
        // Progress pointers
        l1b_abscissa_cur += ckd.fov.n_act_vp[ivp];
        l1b_act_angles_cur += ckd.fov.n_act_vp[ivp];
        l1b_alt_angles_cur += ckd.fov.n_act_vp[ivp];
        swathvector_cur += ckd.fov.n_act_vp[ivp]*dims::vec;
        vectorplane_normal_cur += dims::vec;

    }
}

} // namespace spex
