// Test the packed matrix format

#include "testing.h"

#include "bspline_data.h"

#include <array>
#include <cmath>
#include <spexone_cal/bspline.h>
#include <spexone_cal/linalg.h>

TEST_CASE("Least squares")
{
    // This is the same least squares test as for B-splines
    // (tests/bspline.cpp) but using packed instead of normal
    // matrices.
    constexpr int n_knots { 15 };
    constexpr double d_beg { 454.0 };
    constexpr double d_end { 465.0 };
    constexpr int order { 3 };
    std::vector<double> knots(n_knots, 0.0);
    for (int i {}; i < n_knots; ++i) {
        knots[i] = d_beg + i * (d_end - d_beg) / (knots.size() - 1);
    }
    spex::BSpline b_spline { order, knots };
    const int n_points { static_cast<int>(x_data.size()) };
    const int n_states { b_spline.nStates() };
    spex::PackedMatrix B { n_points, n_states };
    // Construct the B-spline matrix row by row in packed format
    for (int i {}; i < n_points; ++i) {
        b_spline.evalBasis(x_data[i], i, B);
    }
    for (int i {}; i < n_points; ++i) {
        for (int k {}; k < B.row_sizes[i]; ++k) {
            const int idx { B.row_locations[i] + k };
            B.data[idx] /= std::sqrt(noise[i]);
        }
    }
    // Unlike with normal matrices in tests/bspline.cpp, here we also
    // need to explicitly construct B^T.
    spex::PackedMatrix BT { n_states, n_points };
    BT.transpose(B);
    std::vector<double> BTB(n_states * n_states, 0.0);
    BT.multRightTransposed(BT, BTB);
    std::vector<double> BTY(n_points, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < BT.row_sizes[i]; ++k) {
            const int packed_idx { BT.row_locations[i] + k };
            const int col_idx { BT.global_col_indices[i] + k };
            BTY[i] +=
              BT.data[packed_idx] * y_data[col_idx] / std::sqrt(noise[col_idx]);
        }
    }
    // From here on, everything is the same as in tests/bspline.cpp
    // because we are dealing with relatively small matrices that do
    // not need to be packed.
    std::vector<double> BTB_tri(n_states * (n_states + 1) / 2);
    for (int j {}; j < n_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j+1)*j/2] = BTB[i * n_states + j];
        }
    }
    std::vector<double> Sx(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        Sx[i * n_states + i] = 1.0;
    }
    const char uplo { 'u' };
    int info {};
    dppsv_(&uplo,
           &n_states,
           &n_states,
           BTB_tri.data(),
           Sx.data(),
           &n_states,
           &info);
    std::vector<double> x_result(n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < n_states; ++k) {
            x_result[i] += Sx[i * n_states + k] * BTY[k];
        }
    }
    constexpr std::array ref {
        97.3564075280751,
        99.3872290797305,
        98.1036228704817,
        107.3865396455,
        110.358715544159,
        117.060445583357,
        118.3129794002,
        117.893750905433,
        120.711696780604,
        122.159064328912,
        122.272854054485,
        120.610347238412,
        118.534057858669,
        112.036272414326,
        107.469363015539,
        103.322570875858,
        99.5583898586142
    };
    for (int i {}; i < n_states; ++i) {
        REQUIRE(x_result.at(i) == approx(ref.at(i)));
    }
}
