#pragma once

#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>
#include <cmath>
#include <numeric>

// Tolerance in units of epsilon for all floating point comparisons
constexpr double defeault_tolerance_factor { 1e2 };

// Small wrapper around the Catch2 Approx class. The optional argument
// can be used to increase the tolerance for the cases where different
// configurations lead to slightly different results, for instance MKL
// vs the fallback linear algebra routines.
auto approx(const auto x,
            const double tolerance_factor = defeault_tolerance_factor)
  -> Catch::Approx
{
    Catch::Approx app { x };
    return app
      .epsilon(tolerance_factor * std::numeric_limits<decltype(x)>::epsilon())
      .margin(1e-300);
}

// Mean average difference (MAE) of two arrays divided by the sum of
// all elements in one of the arrays. This is the same as MAE per
// element normalized by the average value of the array.
template <typename T>
static auto mae(T& arr1, T& arr2) -> double
{
    const double sum {
        std::transform_reduce(
          arr1.begin(),
          arr1.end(),
          arr2.begin(),
          0.0,
          std::plus<>(),
          [](const auto a, const auto b) { return std::abs(a - b); }),
    };
    return sum / std::accumulate(arr1.cbegin(), arr1.cend(), 0.0);
}
